//******************************************************************************
// Copyright (c) 2002-2017 by Friedemann Seebass, Germany.
// Dieses Programm ist freie Software. Siehe die GNU GPL Notiz in Datei main.c.
// This program is free software. See the GNU GPL notice in file main.c.
// Projekt: Almetare - Alle meine Taschenrechner
//******************************************************************************
/// \file calc_dlg_base.cpp
///       Implementierung der Basisfunktionalitaet des Taschenrechner-Dialogs 
///       aller Taschenrechner (Modul base).
//******************************************************************************
// Wann       Wer      Was
// ---------- -------- ---------------------------------------------------------
// 05.10.2005 fse      execKeyEvent(): neue Tastenkombinationen Ctrl-Insert und
//                     Shift-Insert
// 05.10.2005 fse      pasteText(): neuerlicher Bugfix
// 28.01.2005 fse      boot(): Sounds nur laden, wenn in MORE_REALISTIC-Modus
// 23.01.2005 fse      pasteText(): Bugfix
// 15.11.2003 fse      boot(): evtl. "always on top" setzen
// 11.11.2003 fse      boot(): Application-Icon setzen
// 10.08.2003 fse      extractSym(): Formatierung des Loggings
// 08.08.2003 fse      neue Funktion pasteText() und Handling von STRG-v
// 18.05.2003 fse      execKeyEvent() um Backspace-Taste erweitert
// 30 03 2003 fse      Aufruf neuer virtueller Funktion adjustRealisticSkin()
// 28.03.2003 fse      Aufruf von preExtractSym() in extractSym();
// 16.03.2003 fse      Flags in updateDisplayAfterDigit()
// 12.05.2002 fse      erzeugt
//******************************************************************************

#include <iomanip>        // setw()                              
#include <qapplication.h> // setPalette()
#include <qobjectlist.h>  // QObjectList
#include <qpushbutton.h>
#include <qdir.h>
#include <qpainter.h>
#include <qmessagebox.h>
#include <qtooltip.h>
#include <qclipboard.h>
#include <qnamespace.h>   // LeftButton, NoButton

#include "cfg_dlg.h"
#include "calc_dlg_base.h"
#include "util.h"         // toLower()

//******************************************************************************
// Klasse CalcDlgBase:
//******************************************************************************

// Eine Initialisierung (Setzen der Members, die Daten aus
// CalculatorBase benoetigen), ist hier nicht moeglich, da CalculatorBase
// zu diesem Zeitpunkt noch nicht konstruiert wurde. Dies muss daher spaeter
// vom Dialog angestossen werden.
// Eigene Members koennen jedoch initialisiert werden.
// fse, 30.05.02
// fse, 29.04.03: mMaxDigitNum mit 8 initialisieren (Wert passt fuer TI-30 und
//                fx-85v, nicht fuer fx-3600P, daher jetzt keine
//                Funktionskonstante mehr)

CalcDlgBase::CalcDlgBase()
{
	mMaxDigitNum = 8;
}

//******************************************************************************

/// Initialisiert einen Taschenrechner.
/// Laedt die passende Pixmap und die Sounds, Basisinitialisierung.
/// \param cbP    Zeiger auf die Calculator-Basisklasse des Rechners
/// \param dlgP   Zeiger auf die Qt-Dialogbox
/// \param wflags diverse Qt-Dialog-Flags
//  fse, 23.02.02
//  fse, 11.11.03: Setzen des Application-Icons
//  fse, 15.11.03: evtl. "always on top" setzen
//  fse, 28.01.05: Sounds nur laden, wenn in MORE_REALISTIC-Modus

void CalcDlgBase::boot(CalculatorBase* cbP, QDialog* dlgP, Qt::WFlags wflags)
{
	Cfgs*  cfgsP         = cbP->getCfgsP();
	string accPath       = cbP->getAccPath();
	string applIconPic   = accPath + "/almetare.png";
	string backgroundPic = accPath + "/calc.png";
	string digitsPic     = accPath + "/digits.png";
	string maskPic       = accPath + "/mask.png";
	string prsPath       = accPath + "/wav/prs";
	string rlsPath       = accPath + "/wav/rls";

	mCalcP         = cbP;
	mCalcDlgP      = dlgP;
	mIsOff         = false;
	mIsSoundLoaded = false;

	// Laden der Grafik- und Sound-Dateien:
	mIsPixLoaded = mBackgroundPixmap.load(backgroundPic.c_str());
	if (mIsPixLoaded) mIsPixLoaded = mDigitsPixmap.load(digitsPic.c_str());
	if (mApplIconPixmap.load(applIconPic.c_str()))
		mCalcDlgP->setIcon(mApplIconPixmap); // kleines Icon links oben im Fenster setzen
	if (cfgsP->mDisplayMode == MORE_REALISTIC)
		mIsSoundLoaded = loadWavs(prsPath, mPrsSounds);
	if (mIsSoundLoaded) mIsSoundLoaded = loadWavs(rlsPath, mRlsSounds);
	if (!mIsPixLoaded && cfgsP->mDisplayMode <= REALISTIC)
		cfgsP->mDisplayMode = SIMPLE;

	// Originalgroessen aller Widgets zwecks Skalierung sichern:
	makeWgtMap();

	// grafische Darstellung erzeugen:
	switch(cfgsP->mDisplayMode)
	{
		case MORE_REALISTIC:
		case REALISTIC     : setRealisticSkin(); break;
		case SIMPLE        : setSimpleSkin()   ; break;
		case DEBG          : setDebugSkin ()   ; break;
		default            : setSimpleSkin();
	}
	if (cfgsP->mWidth  == 0) cfgsP->mWidth  = mDefaultWidth;
	if (cfgsP->mHeight == 0) cfgsP->mHeight = mDefaultHeight;
	mCalcDlgP->setGeometry(cfgsP->mXpos, cfgsP->mYpos, cfgsP->mWidth, cfgsP->mHeight);
	updateDisplay();

	// Shortcut-Tabelle und Tool-Tips erzeugen:
	makeShortCutTab();
	if (cfgsP->mShowToolTips)
		addToolTips();

	// Dialog evtl. auf "always on top" setzen:
	setResetAlwaysOnTop(cfgsP->mAlwaysOnTop, wflags, false);
}

//************************************** ***************************************

/// Laedt saemtliche in den Verzeichnissen wav/prs und wav/rls vorhandenen
/// Dateien in Listen von QSound-Objekten.
/// \param path   Pfad zum Verzeichnis mit den Wav-Dateien
/// \param qsPvec Vektor mit den Zeigern auf die QSound-Objekte
/// \return       true (anscheinend war hier mal geplant, bei Fehlern false
///               zurueckzugeben...)
//  fse, 16.12.01

bool CalcDlgBase::loadWavs(string path, vector<QSound*>& qsPvec)
{
	QDir d(path.c_str());
	QStringList files = d.entryList();
	QSound* qsP;
	QStringList::Iterator it = files.begin();

	it++; it++; // die ersten beiden Eintraege sind die Verzeichnisse
	for (; it != files.end(); ++it)
	{
		string fullFileName((*it).latin1());
		fullFileName = path + "/" + fullFileName;
		qsP = new QSound(fullFileName.c_str());
		//cout << fullFileName << ", " << qsP << "\n";
		qsPvec.push_back(qsP);
	}
	return true;
}

//************************************** ***************************************

/// Speichert bei der Konstruktion des Taschenrechner-Hauptdialogs die Groessen
/// und Positionen aller Widgets in der Map mOrgWgtMap. Die Daten werden spaeter
/// zur Skalierung gebraucht.
/// Falls das Widget ein Button ist, werden auch die Texte gespeichert, da diese
/// durch Setzen einer Pixmap geloescht werden.
//  fse, 12.05.02

void CalcDlgBase::makeWgtMap()
{
	const QObjectList* qolP = mCalcDlgP->children();
	QListIterator<QObject> it(*qolP);
	WgtData  wgtData;
	QObject* qoP;
	QWidget* qwP;
	QButton* qbP;

	// Originalgroesse des Dialogs merken:
	if (mIsPixLoaded)
		mOrgDlgSize = mBackgroundPixmap.size();
	else
		mOrgDlgSize = QSize(mDefaultWidth*2, mDefaultHeight*2);
	
	// Originalgroessen aller Widgets merken:
	qoP = it.current();
	while (qoP != NULL)
	{
		qwP = static_cast<QWidget*>(qoP);
		wgtData.rect = qwP->geometry();
		qbP = dynamic_cast<QButton*>(qwP);  // falls Button auch Text merken
		if (qbP) wgtData.text = qbP->text();
		mOrgWgtMap[qwP->name()] = wgtData;
		++it;
		qoP = it.current();
	}
	/*	
	// Ausgabe zur Kontrolle:
	WgtRectMapIT mapIt;
	QRect qr;
	for (mapIt = mOrgWgtMap.begin(); mapIt != mOrgWgtMap.end(); ++mapIt)
	{
		cout << (*mapIt).first << ": " << endl;
		qr = (*mapIt).second;
		cout << qr.x() << ", " << qr.y() << ", " << qr.width() << ", " << qr.height() << endl;
	}
	*/	
}

//************************************** ***************************************

/// Versteckt diverse Felder und stellt die Pixmaps auf dem Hintergrund und auf
/// den Buttons dar.
//  fse, 04.06.02
//  fse, 30.03.03

void CalcDlgBase::setRealisticSkin()
{
	resizeAllWidgets(mCalcDlgP); // um alle Pixmaps zu setzen
	hideOrShowAllLabels(true);
	mLcdStrBaseP ->hide();
	mModesBaseP  ->hide();
	mDisplayBaseP->hide();
//	mSeebassBaseP->hide();
	adjustRealisticSkin();  // virtuell fuer fx-3600P
}

//************************* ************************* **************************

/// Geht alle Widgets durch und versteckt oder zeigt alle Labels.
/// \param isToHide wenn true, werden die Labels versteckt, sonst gezeigt
//  fse, 18.07.02

void CalcDlgBase::hideOrShowAllLabels(bool isToHide)
{
	const QObjectList*     qolP = mCalcDlgP->children();
	QListIterator<QObject> it(*qolP);
	QObject*               qoP;
	QLabel*                qlP;

	// alle Widgets durchgehen:
	qoP = it.current();
	while (qoP != NULL)
	{
		qlP = dynamic_cast<QLabel*>(qoP);
		if (qlP)
		{
			if (isToHide) qlP->hide(); else qlP->show();
		}
		++it;
		qoP = it.current();
	}
}

//************************************** ***************************************

/// Entfernt alle Pixmaps und restauriert die Beschriftung der Buttons.
//  fse, 04.06.02

void CalcDlgBase::setSimpleSkin()
{
	QPalette               p = QApplication::palette();
	const QObjectList*     qolP = mCalcDlgP->children();
	QListIterator<QObject> it(*qolP);
	QObject*               qoP;
	QButton*               bP;
	WgtData                wgtData;
	
	mCalcDlgP->setBackgroundPixmap(NULL);
	hideOrShowAllLabels(false);
	mLcdStrBaseP ->hide();
	mModesBaseP  ->show();
	mDisplayBaseP->show();
//	mSeebassBaseP->show();
	// Beschriftung der Buttons wiederherstellen:
	qoP = it.current();
	while (qoP != NULL)
	{
		bP = dynamic_cast<QButton*>(qoP);
		if (bP)
		{
			wgtData = mOrgWgtMap[bP->name()];
			bP->setText(wgtData.text);
			bP->setPalette(p);
		}
		++it;

		qoP = it.current();
	}
	resizeAllWidgets(mCalcDlgP); // um evtl. Schriftgroesse anzupassen
	adjustSimpleSkin();          // virtuell, setzt z. B. Farben fuer bestimmte Buttons
}

//************************************** ***************************************

/// Wie RealisticSkin, setzt aber die Beschriftung der Buttons und macht alle
/// Felder sichtbar.
//  fse, 15.06.02

void CalcDlgBase::setDebugSkin()
{
	setSimpleSkin();             // restauriert Button-Texte
//	resizeAllWidgets(mCalcDlgP); // setzt Pixmaps wird schon in setSimpleSkin() gemacht
	mLcdStrBaseP->show();
	adjustDebugSkin(); // virtuell, setzt z. B. Farben fuer bestimmte Buttons
}

//************************************** ***************************************

/// Geht die Short-Cut-Map durch und versieht jeden Button darin mit dem
/// Shortcut als Tooltip.
//  fse, 18.08.02

void CalcDlgBase::addToolTips()
{
	ShortCutTabIT it;
	string shortCut;
	QButton* bP;
	char c;

	for (it = mShortCutTab.begin(); it != mShortCutTab.end(); ++it)
	{
		c = (*it).first;
		shortCut = "Shortcut key: '";
		if (c == '\r') shortCut += "Enter"; else shortCut += c;
		shortCut += "' ";
		bP = (*it).second;
		QToolTip::add(bP, shortCut.c_str());
	}
}

//************************************** ***************************************

/// Setzt das Dialog-Flag Qt::WStyle_StaysOnTop des Rechnerdialogs, wenn
/// alwaysOnTop true; andernfalls wird es zurueckgesetzt.
/// Das Setzen erfolgt mit reparent(). Die Qt-Funktion setWFlags() ist protected
/// und kann daher hier nicht ueber den Zeiger mCalcDlgP aufgerufen werden.
/// Ausserdem scheint die Funktion nicht zu funktionieren.
/// \param alwaysOnTop wenn true, wird Flag gesetzt, sonst wird's zurueckgesetzt
/// \param wflags      Flags zur Gestaltung der Dialog-Titelzeilenelemente
/// \param showIt      fuer reparent(): Dialog neu zeichnen
//  fse, 15.11.03

void CalcDlgBase::setResetAlwaysOnTop(bool alwaysOnTop, Qt::WFlags wflags, bool showIt)
{
	unsigned int mask;
	bool oldAlwaysOnTop = (wflags & Qt::WStyle_StaysOnTop) != 0;

	// auf Aenderung pruefen (weil es bei reparent() immer stoerend flackert):
	if (alwaysOnTop == oldAlwaysOnTop)
		return;

	// Dialog evtl. auf "always on top" setzen:
	if (alwaysOnTop)
	{
		wflags = Qt::WStyle_Customize | wflags | Qt::WStyle_StaysOnTop;
	}
	else
	{
		mask    = UINT_MAX;              // alle Bits setzen
		mask   ^= Qt::WStyle_StaysOnTop; // exkl. verodern => alle Bits gesetzt ausser WStyle_StaysOnTop
		wflags &= mask;                  // verunden => wflags wie zuvor, aber ohne WStyle_StaysOnTop
		wflags |= Qt::WStyle_Customize;  // das braucht anscheinend Qt
	}
	#ifdef _WIN32
	QPoint p = QPoint(mCalcDlgP->geometry().x(), mCalcDlgP->geometry().y());
	mCalcDlgP->reparent(NULL, wflags, p, showIt); // sonst huepft der Rechner!?
	#else
	mCalcDlgP->reparent(NULL, wflags, mCalcDlgP->pos(), showIt);
	#endif
	if (!mApplIconPixmap.isNull())
		mCalcDlgP->setIcon(mApplIconPixmap); // kleines Icon links oben im Fenster erneut setzen
}

//******************************************************************************

/// Geht die Short-Cut-Map durch und loescht alle Tooltips der Buttons.
//  fse, 25.08.02

void CalcDlgBase::removeToolTips()
{
	ShortCutTabIT it;
	QButton* bP;

	for (it = mShortCutTab.begin(); it != mShortCutTab.end(); ++it)
	{
		bP = (*it).second;
		QToolTip::remove(bP);
	}

}

//******************************************************************************

/// Initialisierung, die nach "Einschalten" (mButOnP) des Rechners durchgefuehrt
/// wird.
//  fse, 15.06.02

void CalcDlgBase::initCalcDlgBase()
{
	setLastType(B_ON);
	mDisplayStr = "0";
	mIsOff = false;
	mCalcP->process("init");
	setDisplayFields();
}

//******************************************************************************

/// Loescht das LCD. Das Rechteck mLcdRect wird wieder mit dem Original aus der
/// urspruenglichen Pixmap ueberschrieben.
//  fse, 30.06.02

void CalcDlgBase::clearLcd()
{
	QWMatrix m;
	QPixmap* pmP  = const_cast<QPixmap*>(mCalcDlgP->backgroundPixmap());
	QPixmap  clearedLcdPm (mOrgLcdRect.width(), mOrgLcdRect.height());
	QPoint   qp(mLcdRect.topLeft());
	QRect    qr(QPoint(0,0), mLcdRect.size()); // Groesse Ziel

	if (!mIsPixLoaded) return;

	// leeres LCD aus Original-Pixmap in neue Pixmap kopieren:
	bitBlt(&clearedLcdPm, QPoint(0, 0), &mBackgroundPixmap, mOrgLcdRect);

	// umskalieren und ins aktuelle LCD-Feld kopieren:
	m.scale(mXfact, mYfact);
	bitBlt(pmP, qp, &clearedLcdPm.xForm(m), qr);
}

//******************************************************************************

/// Bringt Message-Box auf den Bildschirm, wenn die Maus zum ersten Mal den
/// Rechnerdialog betritt.

void CalcDlgBase::showHint()
{
	static bool isFirstCall = true;
	int rc;

	if (isFirstCall && mCalcP->getCfgsP()->mShowHint)
	{
		isFirstCall = false;
		rc = QMessageBox::information(mCalcDlgP, mCalcDlgP->caption(),
		     "<H3>Hint: Press 't' for help and configuration.</H3><BR>Do you want to see this message again on each program start?",
		     "&No", "&Yes");
		if (rc == 0)
			mCalcP->getCfgsP()->mShowHint = false;
	}
}

//******************************************************************************

/// Ermittelt aus der aktuellen Groesse und der Originalgroesse des Hauptdialogs
/// mainDlgP Streckfaktoren in x- und y-Richtung.
/// Mit diesen werden alle Widgets des Dialogs bzgl. Position und Groesse
/// angepasst.
/// \param mainDlgP Zeiger auf die Rechner-Dialogbox
//  fse, 12.05.02

void CalcDlgBase::resizeAllWidgets(QDialog* mainDlgP)
{
	const QObjectList*     qolP = mainDlgP->children();
	QListIterator<QObject> it(*qolP);
	QObject*               qoP;
	QWidget*               qwP;
	QRect                  newDlgRect = mCalcDlgP->geometry();
	QWMatrix               m;
	DisplayMode            dm = mCalcP->getCfgsP()->mDisplayMode;

	// Streckfaktoren:
	mXfact = (double)newDlgRect.width () / (double)mOrgDlgSize.width ();
	mYfact = (double)newDlgRect.height() / (double)mOrgDlgSize.height();
	// Hintergrundbild, Ziffern-Pixmap und LCD-Groesse skalieren:
	if (dm != SIMPLE && mIsPixLoaded)
	{
		int x, y, w, h;
		mOrgLcdRect.coords(&x, &y, &w, &h);
		mLcdRect.setCoords(int(x*mXfact), int(y*mYfact), int(w*mXfact), int(h*mYfact));
		m.scale(mXfact, mYfact);
		mCalcDlgP->setBackgroundPixmap(mBackgroundPixmap.xForm(m));
		mScaledDigitsPixmap = mDigitsPixmap.xForm(m);
	}
	// alle Widgets durchgehen:
	qoP = it.current();
	while (qoP != NULL)
	{
		string name = qoP->name();
		qwP = static_cast<QWidget*>(qoP);
		setNewSize(qwP);
		++it;
		qoP = it.current();
	}
	setDisplayFields();
}

//************************************** ***************************************

/// Berechnet fuer das Widget qwP eine neue Position und Groesse abhaengig von
/// der aktuellen Groesse des Hauptdialogs und der Originalgroesse.
/// \param qwP Zeiger auf das Widget, dessen Groesse berechnet werden soll
//  fse, 12.05.02

void CalcDlgBase::setNewSize(QWidget* qwP)
{
//	QRect       newDlgRect = mCalcDlgP->geometry();
	WgtData     wgtData    = mOrgWgtMap[qwP->name()];
	QRect       orgWgtRect = wgtData.rect;
	int         x = (int)(orgWgtRect.x     () * mXfact + .5);
	int         y = (int)(orgWgtRect.y     () * mYfact + .5);
	int         w = (int)(orgWgtRect.width () * mXfact + .5);
	int         h = (int)(orgWgtRect.height() * mYfact + .5);
	int         height, heightCor;
	QRect       newWgtRect(x, y, w, h);
	QButton*    bP  = dynamic_cast<QButton*  >(qwP);
	QLabel*     lbP = dynamic_cast<QLabel*   >(qwP);
	QLineEdit*  leP = dynamic_cast<QLineEdit*>(qwP);
	string      name = qwP->name();
	DisplayMode dm = mCalcP->getCfgsP()->mDisplayMode;

	// neue Groesse setzen:
	qwP->setGeometry(newWgtRect);
	// falls Button oder Label und nicht realistisch, Textgroesse anpassen:
	if ((bP || lbP) && (dm == SIMPLE || dm == DEBG))
	{
		QFont f("Helvetica");
		//QFont f = bP->font();
		height = qwP->height();
		heightCor = height <= 10 ? 1 : 2;
		if (bP) f.setPixelSize(height/2); else f.setPixelSize(height-heightCor);
		qwP->setFont(f);
	}
	// falls Button und realistisch: Pixmap anpassen:
	if (bP && (dm == REALISTIC || dm == MORE_REALISTIC)) 
		rescalePixmap(bP, orgWgtRect);

	// falls Edit-Feld: Schriftgroesse anpassen:
	if (leP)
	{
		QFont f = leP->font();
		string name = leP->name();
		height = leP->height();
		if (name == "mModesP")
			f.setPixelSize(height*5/12);
		else
			f.setPixelSize(height/2);
		qwP->setFont(f);
	}
}

//************************* ************************* **************************

/// Kopiert die Hintergrund-Pixmap in einen Button und setzt Palette mit grauem
/// Rand.
/// \param bP         Zeiger auf den zu bearbeitenden Button
/// \param orgWgtRect Groesse des originalen, unskalierten Buttons
//  fse, 15.06.02

void CalcDlgBase::rescalePixmap(QButton* bP, const QRect& orgWgtRect)
{
	QWMatrix m;
	QPixmap orgPm(orgWgtRect.width(), orgWgtRect.height());

	bitBlt(&orgPm, QPoint(0, 0), &mBackgroundPixmap, orgWgtRect);
	m.scale(mXfact, mYfact);
	QPalette p = QApplication::palette();
	bP->setPixmap(orgPm.xForm(m));
	p.setColor(QColorGroup::Light, Qt::darkGray);
	bP->setPalette(p);
}

//******************************************************************************
// Button-Slots:
//******************************************************************************

/// Extrahiert das Symbol aus dem Button-Namen, indem das fuehrende "m" und
/// abschliessende "P" entfernt wird.
/// Ausserdem wird das Symbol gleich auf ein anderes gemapped, wenn eine Shift-
/// Ebene aktiv ist.
/// \param senderP Zeiger auf den Button, der gedrueckt wurde
/// \return        String mit dem Symbol des Buttons
//  fse, 06.07.02
//  fse, 28.03.03: Aufruf preExtractSym()
//  fse, 10.08.03: Formatierung des Loggings mit setw(4)

string CalcDlgBase::extractSym(QPushButton* senderP)
{
	string buttonName = senderP->name();
	string in = buttonName.substr(1, buttonName.size()-2);
	string sym = in;

	preExtractSym(sym); // nur fuer Casio fx3600P noetig
	sym = mapDlgBase(sym);
	mCalcP->getComputerP()->map(sym);
	ONL_LOG(gnl) << "button " << setw(4) << in << " mapped to symbol " << sym << NL;
	return sym;
}

//************************************** ***************************************

/// Fuehrt grundlegendes Mapping durch von den Symbolen, die zwar in der
/// Konsolenumgebung verwendet werden, aber nicht in Widget-Namen verwendet
/// werden koennen.
/// \param in String mit dem Symbol, wie es der Button-Name liefert
/// \return   String mit dem Symbol, wie es der Konsolenrechner versteht

string CalcDlgBase::mapDlgBase(string& in) const
{
	in = Util::toLower(in);
	if      ("equ"   == in) in = "=";
	else if ("add"   == in) in = "+";
	else if ("sub"   == in) in = "-";
	else if ("mul"   == in) in = "*";
	else if ("div"   == in) in = "/";
	else if ("prc"   == in) in = "%";
	else if ("bra"   == in) in = "(";
	else if ("ket"   == in) in = ")";
	else if ("point" == in) in = ".";
	else if ("exp"   == in) in = "pi"; // eigentlich nur fuer Casio-Rechner
	return in;
}

	
//******************************************************************************

/// Holt die Display-Felder aus der zum Taschenrechner gehoerenden
/// Display-Klasse.
//  fse, 20.05.02

void CalcDlgBase::copyBaseDisplayFields()
{
	mDisplayStr = mCalcP->getDisplayP()->getValue();
	mDisplayBaseP->setText(mCalcP->getDisplayP()->getValue().c_str());
	mModesBaseP  ->setText(mCalcP->getDisplayP()->getModes().c_str());
	mLcdStrBaseP ->setText(mCalcP->getDisplayP()->getSeven().c_str());
}

//******************************************************************************

/// Slot-Funktionen fuer die meisten Buttons.
/// Der Name des gedrueckten Buttons wird ermittelt; bis auf das "m" und "P"
/// (z.B. im Namen "mSignP") ist der Name i.d.R. identisch mit dem String, den
/// der Calculator in seiner Funktion process() erwartet. Dieser String wird
/// daher an ihn geschickt.
/// Wenn als letztes eine Zahl (sie befindet sich in mDisplayStr) eingegeben
/// wurde, muss zunaechst diese  der process()-Funktion uebergeben werden.
/// \param  sym      zu verarbeitendes Symbol
/// \retval lastSymb das Symbol, das beim letzten Aufruf uebergeben worden war
//  fse, 27.10.01

void CalcDlgBase::butActionBase(string& sym, string& lastSymb)
{
	static string lastSym = "";
	if (mLastType == B_DIGIT || mLastType == B_ENG)
		mCalcP->process(mDisplayStr);
	mCalcP->process(sym);
	copyBaseDisplayFields();
	setDisplayFields(true);
	if (sym == "pi")
		setLastType(B_PI);  // sauberer waere: eigene Funktionen
	else
		setLastType(B_ACTION);
	lastSymb = lastSym;
	lastSym = sym;
}

//******************************************************************************

/// Funktion, die fuer die Bearbeitung aller Modus-Tasten aufgerufen wird.
/// \param sym das Symbol mit dem Modus ("inv" etc.)
//  fse, 16.03.03: neues Flag "m", damit die Modes (insbesondere INV) nicht
//                 "ENG" abschalten

void CalcDlgBase::butAnyModeBase(string& sym)
{
//	DisplayBase* dsplP = mCalcP->getDisplayP();
//	DisplayMode dm = mCalcP->getCfgsP()->mDisplayMode;
	string dummy;

	mCalcP->process(sym);
	updateDisplayAfterDigit("m" + mDisplayStr); // "m" ist Flag
//	setLastType(B_ACTION);
//	mModesBaseP->setText(mCalcP->getDisplayP()->getModes().c_str());
//	setLastType(B_ANY_MODE);
}

//******************************************************************************

/// Wird von allen Zifferntasten aufgerufen.
/// Entweder ist in sym tatsaechlich eine Ziffer, dann wird sie i.d.R. an die
/// Anzeige angehaengt.
/// Oder in sym ist keine Ziffer (die Ziffer wurde wegen Shift oder Mode
/// umgemappt), dann muss die entsprechende Aktion ausgefuehrt werden.
/// \param sym das Symbol mit der Ziffer
//  fse, 10.10.01

void CalcDlgBase::butDigitBase(string& sym)
{
	char c = sym[0];
	
	if (isdigit(c) || c == '.')
	{
		butDigitDigit(sym);
		setLastType(B_DIGIT);
	}
	else
	{
		butDigitAction(sym);
//		setLastType(mLastButOneType);
	}
//	updateDisplayAfterDigit();
}

//************************************** ***************************************

/// Die Ziffer ist wirklich eine Ziffer (nicht etwa auf ein anderes Symbol
/// umgemappt). Die Ziffer ersetzt daher die Anzeige komplett oder wird an die
/// Anzeige angehaengt.
/// Wenn die Zahl gerade in Exponentialschreibweise dargestellt wird, darf die
/// Zahl nicht einfach angehaengt werden, sondern die letzten beiden Ziffern des
/// Exponenten werden nach links geshifted.
/// \param sym das Symbol mit der Ziffer
//  fse, 06.07.02
//  fse, 29.04.02: statt const int maxLen = 8 wird jetzt mMaxDigitNum benutzt,
//                 da der fx-3600P maximal 11 Ziffern in der Anzeige darstellt

void CalcDlgBase::butDigitDigit(string& sym)
{
	int len;
	bool isPoint = mDisplayStr.find('.') != string::npos;
	bool isFrac  = mDisplayStr.find('/') != string::npos;

	if (mDisplayStr == "0" || mLastType == B_ACTION || mLastType == B_ON
	                       || mLastType == B_PI     || mLastType == B_ENG)
	{ // Anzeige komplett ersetzen:
		mDisplayStr = sym;
	}
	else
	{ // Zeichen anhaengen:
		if (sym=="." && (isPoint || isFrac)) return; // Punkt schon da => tu nix
		if (mDisplayStr.find('e') != string::npos)
		{ // Zahl ist in Exponentialschreibweise:
			int last = mDisplayStr.size() - 1;
			mDisplayStr[last-1] = mDisplayStr[last];
			mDisplayStr.erase(last);
			mDisplayStr += sym;
		}
		else
		{
			len = mCalcP->getNumStrLen(mDisplayStr);
			if (len < mMaxDigitNum) mDisplayStr += sym;
		}
	}
	updateDisplay(mDisplayStr);
}

//************************************** ***************************************

/// Die urspruengliche Ziffer wurde wegen einer Shift-Operation auf ein anderes
/// Symbol umgemappt. Dieses neue Symbol wird jetzt stattdessen prozessiert.
/// \param sym das neue Symbol
//  fse, 07.06.02

void CalcDlgBase::butDigitAction(string& sym)
{
	string dummy, displayStr = mDisplayStr;

	if (sym == "deg" || sym == "rad" || sym == "gra")
	{
		mCalcP->process(sym);
		updateDisplayAfterDigit(mDisplayStr);
	}
	else
	{
		butActionBase(sym, dummy);
		setLastType(B_ACTION);
	}
//	if (mLastButOneType == B_DIGIT)
//	{
//		mDisplayStr = displayStr;
//		mDisplayBaseP->setText(mDisplayStr.c_str());
//	}
}

//************************************** ***************************************

/// Erneuert die Members mModes und mSeven der Display-Klasse und das LCD
/// entsprechend.
/// \param inputStr wenn "", wird der String in mDisplayStr verwendet, um die 
///                 Anzeige zu erstellen, sonst inputStr
//  fse, 06.07.02
//  fse, 16.03.03: Flags "e" und "m", damit Mode-Tasten (insbesondere INV)
//                 nicht sofort die ENG-Modi abschalten.
//                 Falls gerade ein ENG-Mode aktiv (isEngOld ist true) und
//                 die aktuelle Taste eine Mode-Taste ist (isMode ist true),
//                 wird nur die Mode-Zeile aktualisiert, nicht die
//                 Ziffern-Ausgabe.

void CalcDlgBase::updateDisplayAfterDigit(string inputStr)
{
	DisplayBase* dsplP = mCalcP->getDisplayP();
	static bool isEngOld  = false;
	static bool isModeOld = false;
	bool isEng  = inputStr[0] == 'e'; // wurde von butEngBoth() aufgerufen
	bool isMode = inputStr[0] == 'm'; // wurde von butAnyModeBase() aufgerufen

	if (isEng || isMode) inputStr.replace(0, 1, ""); // String wieder reparieren
	dsplP->setModes();
	if (isMode && isEngOld)
	{
		mModesBaseP  ->setText(dsplP->getModes().c_str());
	}
	else
	{
		if (inputStr == "")
			dsplP->makeSeven(mDisplayStr);
		else
			dsplP->makeSeven(inputStr);
		mLcdStrBaseP ->setText(dsplP->getSeven().c_str());
		mModesBaseP  ->setText(dsplP->getModes().c_str());
		mDisplayBaseP->setText(mDisplayStr.c_str());
	}
	setDisplayFields();

	isModeOld = isMode;
	if (!isMode)
		isEngOld = isEng; // Altes Eng-Flag nur ueberschreiben, wenn aktuell
		                  // keine Mode-Taste; sonst wuerde der zweite Druck
		                  // auf die Mode-Taste den ENG-Modus doch wieder abschalten.
}

//************************************** ***************************************

/// Erneuert die Members der Display-Klasse, die drei Edit-Felder des Dialogs und
/// das LCD.
/// Wenn inputStr leer ist, werden die Members der Display-Klasse allenfalls
/// umformatiert, der Wert der Anzeige wird nicht geaendert. Sinn: Z.B. der
/// Button "abc" muss eine Umformatierung durchfuehren, ohne Werte zu aendern.
/// Wenn inputStr gefuellt ist, wird dieser String als mSeven in die
/// Display-Klasse und als Eingabestring in das Dialog-Edit-Feld *mDisplayP
/// uebertragen, mValue und mModes aber ebenfalls nicht geaendert. Sinn: Bei der
/// Eingabe von Ziffern muessen diese sofort angezeigt werden, ohne dass sich der
/// Wert mVal in der Display-Klasse gleich aendern darf, da die Eingabe ja noch
/// nicht abgeschlossen ist.
/// \param inputStr wenn "", wird der String in mDisplayStr verwendet, um die 
///                 Anzeige zu erstellen, sonst inputStr
//  fse, 12.07.02

void CalcDlgBase::updateDisplay(string inputStr)
{
	DisplayBase* dsplP = mCalcP->getDisplayP();
//	DisplayMode dm = mCalcP->getCfgsP()->mDisplayMode;

	if (inputStr == "")
	{
		dsplP->makeDisplay();
		mDisplayStr = dsplP->getValue();
		mDisplayBaseP->setText(dsplP->getValue().c_str());
	}
	else
	{
		mDisplayBaseP->setText(inputStr.c_str());
		dsplP->makeSeven(inputStr);
		dsplP->setModes();
	}
	mLcdStrBaseP->setText(dsplP->getSeven().c_str());
	mModesBaseP ->setText(dsplP->getModes().c_str());
	setDisplayFields();
}

//******************************************************************************

/// Waehrend der Eingabe einer Zahl muss das Vorzeichen gewechselt werden, ohne
/// die Eingabe zu unterbrechen; sonst muss im Calculator die Sign-Routine
/// aufgerufen werden.
//  fse, 15.10.01

void CalcDlgBase::butSignBase()
{
//	DisplayBase* dsplP = mCalcP->getDisplayP();
	
	ONL_LOG(gnl) << "button 'sign'" << NL;
	if (isError() || mIsOff) return;
	if (mLastType == B_DIGIT || mLastType == B_EE)
	{
		string::size_type ePos = mDisplayStr.find('e');
		if (ePos == string::npos)
		{ // keine Exponentialdarstellung:
			if (mDisplayStr[0] == '-')
				mDisplayStr = mDisplayStr.substr(1);
			else
				mDisplayStr = "-" + mDisplayStr;
		}
		else
		{ // Exponentialdarstellung:
			string::size_type signPos = ePos + 1;
			if (mDisplayStr[signPos] == '-')
				mDisplayStr[signPos] = '+';
			else
				mDisplayStr[signPos] = '-';
		}
		updateDisplay(mDisplayStr);
		//setLastType(B_DIGIT);
	}
	else
	{
		string sym = "sign", dummy;
		butActionBase(sym, dummy);
		//mCalcP->process("sign");
		//copyBaseDisplayFields();
		//setLastType(B_ACTION);
	}
//	updateDisplayAfterDigit();
//	dsplP->setVal(atof(mDisplayStr.c_str()));
//	dsplP->makeSeven();
//	mLcdStrBaseP->setText(dsplP->getSeven().c_str());
//	mDisplayBaseP->setText(mDisplayStr.c_str());
//	setDisplayFields();
}

//******************************************************************************

/// Entfernt das letzte Zeichen aus der Anzeige.
/// Wenn die Anzeige in Exponentialdarstellung ist, werden die Ziffern des
/// Exponenten nach rechts geshifted bis er nur noch Nullen enthaelt, dann wird
/// er ganz entfernt.
//  fse, 18.05.03: Funktion von Fx85vDlg nach hier verschoben

void CalcDlgBase::butBack()
{
	int               last        = mDisplayStr.size() - 1;
	bool              isLastPoint = mDisplayStr[last] == '.';
	string::size_type ePos        = mDisplayStr.find('e');
	bool              isSci       = ePos != string::npos;
	string            sym, numStr, mnt, exp;

	ONL_LOG(gnl) << "button 'back'" << NL;
	if (isError() || mLastType != B_DIGIT) return;
	sym = "back";
//	resetAllModes();
	if (isSci)
	{ // Zahl ist in Exponentialschreibweise:
		exp = mDisplayStr.substr(ePos+2);
		if (atoi(exp.c_str()) == 0)
		{ // Exponent enthaelt nur noch Nullen => entfernen:
			mDisplayStr = mDisplayStr.erase(ePos-1);
		}
		else
		{ // Ziffern des Exponenten nach recht schieben, "0" rueckt nach:
			Util::shiftStr(exp, "0", Util::RIGHT);
			mDisplayStr = mDisplayStr.substr(0, ePos+2) + exp;
		}
	}
	else
	{ // Zahl ist nicht in Exponentialschreibweise:
		if (last == 0)
		{
			mDisplayStr = "0";
		}
		else
		{
			mDisplayStr.erase(last);
			if (isLastPoint) butBack(); // so keine neue Kodierung der Ueberpruefung auf last == 0 noetig
			if (mDisplayStr == "-") butBack();
		}
	}
	updateDisplay(mDisplayStr);
}

//******************************************************************************

/// Fuehrt den Konfigurationsdialog aus.
/// \param dlgP   Zeiger auf den Vater-Dialog
/// \param wflags die Flag-Parameter des Rechnerdialogs
//  fse, 30.11.01
//  fse, 15.11.03: evtl. "always on top" setzen

void CalcDlgBase::butTypeBase(QWidget* dlgP, Qt::WFlags wflags)
{
	CfgDlg cfgDlg(this, mApplIconPixmap, dlgP, 0, true);
	Cfgs* cfgsP = mCalcP->getCfgsP();
	int ret;

	ONL_LOG(gnl) << "button: CalcType" << NL;
	ret = cfgDlg.exec();
	if (ret == 1)
	{
		cfgsP->updtCfgFile();
		if (cfgsP->mZoom != 0)
		{
			QRect rect = mCalcDlgP->geometry();
			int x = rect.x();
			int y = rect.y();

			int w = mIsPixLoaded ? mBackgroundPixmap.width () : mDefaultWidth  * 2;
			int h = mIsPixLoaded ? mBackgroundPixmap.height() : mDefaultHeight * 2;
			w = w * cfgsP->mZoom / 100;
			h = h * cfgsP->mZoom / 100;
			mCalcDlgP->setGeometry(x, y, w, h);
		}
		switch(cfgsP->mDisplayMode)
		{
			case DEBG          : setDebugSkin()    ; break;
			case SIMPLE        : setSimpleSkin()   ; break;
			case REALISTIC     :
			case MORE_REALISTIC: setRealisticSkin(); break;
			default: setSimpleSkin();
		}
		if (cfgsP->mShowToolTips)
			addToolTips();
		else
			removeToolTips();
		cfgsP->startStopLogging();
		mIsOff = false;
		setDisplayFields(); // um ggf. die Timer v. TI-30 zu starten
		setResetAlwaysOnTop(cfgsP->mAlwaysOnTop, wflags, true);
	}
}

//******************************************************************************

/// Spielt einen der Sounds im Vektor mPrsSounds beim Druecken der Taste.
//  fse, 22.12.01

void CalcDlgBase::playPressBase()
{
	DisplayMode dm = mCalcP->getCfgsP()->mDisplayMode;
	int rnd, ind;

	if (dm == MORE_REALISTIC || dm == DEBG)
	{
		if (mPrsSounds.size() == 0) return;
		rnd = rand();
		ind = rnd * mPrsSounds.size() / (RAND_MAX+1);
		mPressedClock = clock();
		mPrsSounds[ind]->play();
	}
}

//******************************************************************************

/// Spielt einen der Sounds im Vektor mRlsSounds beim Loslassen der Taste.
/// Der Sound darf nur dann gespielt werden, wenn der vom Druecken der Taste
/// schon abgespielt ist, da man sonst nur das Loslassen hoert. 
/// Nach 0.15 Sek. sollte der Druecken-Sound vorbei sein.
//  fse, 22.12.01

void CalcDlgBase::playReleaseBase()
{
	DisplayMode dm = mCalcP->getCfgsP()->mDisplayMode;
	int rnd, ind;

	if (dm == MORE_REALISTIC || dm == DEBG)
	{
		if (mRlsSounds.size() == 0) return;
		rnd = rand();
		ind = rnd * mRlsSounds.size() / (RAND_MAX+1);
		clock_t releasedClock = clock();
 		if ((float)(releasedClock - mPressedClock)/(float)CLOCKS_PER_SEC >= .15)
			mRlsSounds[ind]->play();
	}
}

//******************************************************************************

/// Wandelt Tastatur-Event in Button-Pressed-Event um.
/// Dazu wird die Liste der Shortcuts durchgesehen und das Event an den dort
/// zur Taste vermerkten Button geschickt.
/// \param evP Tastatur-Event
/// \param me  entsprechendes Maus-Event
//  fse, 08.08.03: Neuer Case-Zweig Qt::Key_V zum Einfuegen von Zahlen in die
//                 Anzeige; Umbau von isPressed-Logik.
//  fse, 05.10.05: Neue Tastenkombinationen Ctrl-Insert (=Copy) und Shift-Insert
//                 (=Paste)

void CalcDlgBase::execKeyEvent(QKeyEvent* evP, QMouseEvent& me)
{
	int ascii = evP->ascii();
	int key   = evP->key();
	ShortCutTabIT it = mShortCutTab.find(ascii);

	evP->accept();
	switch(key)
	{
		case Qt::Key_Backspace:
		{
			static bool isPressed = false;
			isPressed = !isPressed;
			if (isPressed) // nur bei jedem zweiten Aufruf Zeichen loeschen, da
				butBack();  // Event sowohl beim Druecken wie Loslassen kommt
			break;
		}
		case Qt::Key_F1:
		{
			static bool isPressed = false;
			isPressed = !isPressed;
			if (isPressed)
			{
				CfgDlg cfgDlg(this, mApplIconPixmap, mCalcDlgP, 0, true);
				cfgDlg.butHelp();
			}
			break;
		}
		case Qt::Key_Insert:
		{
			static bool isPressed = false;
			isPressed = !isPressed;
			if (isPressed && evP->state() == Qt::ControlButton)
				QApplication::clipboard()->setText(mDisplayStr.c_str());
			else
				if (isPressed && evP->state() == Qt::ShiftButton)
					pasteText();
			break;
		}
		case Qt::Key_C:
		{
			static bool isPressed = false;
			isPressed = !isPressed;
			if (isPressed && evP->state() == Qt::ControlButton)
			{
				QApplication::clipboard()->setText(mDisplayStr.c_str());
			}
			else
			{
				QApplication::sendEvent(mShortCutTab[ascii], &me);
			}
			break;
		}
		case Qt::Key_V:
		{
			static bool isPressed = false;
			isPressed = !isPressed;
			if (isPressed && evP->state() == Qt::ControlButton)
			{
				pasteText();
			}
			break;
		}
		case Qt::Key_Escape:
		{
			QApplication::sendEvent(mShortCutTab['a'], &me);
			break;
		}
		default:
		{
			if (it == mShortCutTab.end())
				return;
			QApplication::sendEvent(mShortCutTab[ascii], &me);
		}
	}
}

//************************************** ***************************************

/// Kopiert den Textinhalt des Clipboards zeichenweise ins Display. Dabei wird
/// fuer jedes Zeichen ein Tastendruck simuliert, das Zeichen also als Shortcut
/// gemaess der Tabelle mShortCutTab[] interpretiert.
/// Auf diese Weise ist also ebenfalls eine Art Programmierung des Rechners
/// moeglich.
/// Allerdings sollen normalerweise mit dieser Funktion Zahlen und keine
/// "Programme" in die Anzeige gebracht werden, und diese sehen z.B so aus:
///    -2.3e-025.
/// mShortCutTab[] interpretiert nun aber "-" nicht als Vorzeichen, sondern als
/// Subtraktion; auch ist die Reihenfolge der Zeichen falsch, wenn in die
/// Anzeige diese Zahl eingegeben wird. Die waere naemlich so:
///    "2.3_e025_".
/// Also muss der String entsprechend umgestellt werden.
/// Um trotzdem auch die "Programmierung" zu erlauben, wird nun zunaechst
/// geprueft, ob im Clipboard eine einzelne Zahl steht. Ist dies der Fall, wird
/// eine entsprechende Umstellung der Zeichen vorgenommen. Ist dies nicht der
/// Fall, werden alle Zeichen so prozessiert, wie sie im Clipboard stehen.
//  fse, 08.08.03
//  fse, 23.01.05: Fehler beim Adressieren von exp[0] behoben
//  fse, 05.10.05: Fehler beim Exponenten behoben

void CalcDlgBase::pasteText()
{
	QMouseEvent       mePressed (QEvent::MouseButtonPress  , QPoint(5,5), Qt::LeftButton, Qt::LeftButton);
	QMouseEvent       meReleased(QEvent::MouseButtonRelease, QPoint(5,5), Qt::LeftButton, Qt::NoButton);
	QString           qtext = QApplication::clipboard()->text();
//	string            text = QApplication::clipboard()->text().latin1();
	string            text;
	string::size_type pos;
	string            mnt, exp;
	string            e = "";
	bool              isSci;
	unsigned int      i;
	int               ascii;
	ShortCutTabIT it;

	if (qtext.isEmpty())
		return;
	text = qtext.latin1();
	// pruefen, ob einzelne Zahl im Clipboard:
	pos = text.find_first_not_of("0123456789.e+-");
	if (pos == string::npos)
	{
		// keine anderen Zeichen => vermutlich Zahl im Clipboard => umstellen:
		isSci = Util::getMantExp(text, mnt, exp);
		if (mnt[0] == '-')
			mnt = mnt.substr(1) + '_';
		if (isSci)
		{
			if (exp[0] == '-')
				exp = exp.substr(1) + '_';
			else
				exp = "0";
			e = "e";
		}
		text = mnt + e + exp;
	}

	// Text durchlaufen:
	for (i = 0; i < text.size(); ++i)
	{
		ascii = text[i];
		it = mShortCutTab.find(ascii);
		if (it != mShortCutTab.end())
		{
			QApplication::sendEvent(mShortCutTab[ascii], &mePressed);
			QApplication::sendEvent(mShortCutTab[ascii], &meReleased);
		}
	}
}

//******************************************************************************
//************************************** ***************************************
//************************* ************************* **************************
//****************** ******************* ******************* *******************
//************** *************** *************** *************** ***************
//************ ************ ************ ************ ************ *************
