//******************************************************************************
// Copyright (c) 2002-2004 by Friedemann Seebass, Germany.
// Dieses Programm ist freie Software. Siehe GNU GPL Notiz in Datei main.cpp.
// This program is free software. See the GNU GPL notice in file main.cpp.
// Projekt: Almetare - Alle meine Taschenrechner
//******************************************************************************
/// \file calc_sci.cpp
///       Klassen fuer alle wissenschaftlichen Taschenrechner (Modul base).
///       In dieser Datei befinden sich die Implementierungen der Klassen, die
///       alle wissenschaftlichen Taschenrechner benoetigen.
//******************************************************************************
// Wann       Wer      Was
// ---------- -------- ---------------------------------------------------------
// 11.06.2004 fse      DisplaySci::DisplaySci(): Initialisierung einiger
//                     vergessener Members
// 17.11.2003 fse      strstream durch sstream ersetzt
// 11.11.2003 fse      SymTrigFunction::process(): lastInP auf != NULL pruefen
// 17.03.2003 fse      Trennzeichen '#' bei Parametern durch '@' ersetzt. Siehe
//                     Kommentar in calculator.cpp
// 24.03.2002 fse      erzeugt
//******************************************************************************

//******************************************************************************

#include <sstream>
#include "util.h"     // Util::shiftPoint()
#include "calc_sci.h"

//******************************************************************************

int AcShiftPoint::mShiftNumMax = 4;

//******************************************************************************
// Klasse ComputerSci:
//******************************************************************************

/// Der Konstruktor fuegt der Symboltabelle die zusaetzlichen Symbole fuer
/// wissenschaftliche Taschenrechner hinzu.

ComputerSci::ComputerSci()
	: mIsInv(false), mIsHyp(false),
	  mIsFixInput(false), mIsSciInput(false), mAngMode(DEG)
{
	addSciSymsToSymTab();
}

//************************************** ***************************************

/// Funktion fuegt der Symboltabelle die zusaetzlichen Symbole fuer
/// wissenschaftliche Taschenrechner hinzu.

void ComputerSci::addSciSymsToSymTab()
{	
	string sym;
	Symbol* symP;

	symP = new NumPi    ; sym = symP->getSym(); mSymTab[sym] = symP;	
	symP = new OpPow    ; sym = symP->getSym(); mSymTab[sym] = symP;	
	symP = new OpRoot   ; sym = symP->getSym(); mSymTab[sym] = symP;	
	symP = new OpMod    ; sym = symP->getSym(); mSymTab[sym] = symP;	
	symP = new OpGcd    ; sym = symP->getSym(); mSymTab[sym] = symP;	
	symP = new OpLcm    ; sym = symP->getSym(); mSymTab[sym] = symP;	
	symP = new FnRcp    ; sym = symP->getSym(); mSymTab[sym] = symP;	
	symP = new FnSqr    ; sym = symP->getSym(); mSymTab[sym] = symP;	
	symP = new FnEx     ; sym = symP->getSym(); mSymTab[sym] = symP;	
	symP = new FnPow10  ; sym = symP->getSym(); mSymTab[sym] = symP;	
	symP = new FnLn     ; sym = symP->getSym(); mSymTab[sym] = symP;	
	symP = new FnLog    ; sym = symP->getSym(); mSymTab[sym] = symP;	
	symP = new FnSin    ; sym = symP->getSym(); mSymTab[sym] = symP;	
	symP = new FnAsin   ; sym = symP->getSym(); mSymTab[sym] = symP;	
	symP = new FnSinh   ; sym = symP->getSym(); mSymTab[sym] = symP;	
	symP = new FnAsinh  ; sym = symP->getSym(); mSymTab[sym] = symP;	
	symP = new FnCos    ; sym = symP->getSym(); mSymTab[sym] = symP;	
	symP = new FnAcos   ; sym = symP->getSym(); mSymTab[sym] = symP;	
	symP = new FnCosh   ; sym = symP->getSym(); mSymTab[sym] = symP;	
	symP = new FnAcosh  ; sym = symP->getSym(); mSymTab[sym] = symP;	
	symP = new FnTan    ; sym = symP->getSym(); mSymTab[sym] = symP;	
	symP = new FnAtan   ; sym = symP->getSym(); mSymTab[sym] = symP;	
	symP = new FnTanh   ; sym = symP->getSym(); mSymTab[sym] = symP;	
	symP = new FnAtanh  ; sym = symP->getSym(); mSymTab[sym] = symP;	
	symP = new AcInv    ; sym = symP->getSym(); mSymTab[sym] = symP;	
	symP = new AcConst  ; sym = symP->getSym(); mSymTab[sym] = symP;	
	symP = new AcDeg    ; sym = symP->getSym(); mSymTab[sym] = symP;	
	symP = new AcRad    ; sym = symP->getSym(); mSymTab[sym] = symP;	
	symP = new AcGra    ; sym = symP->getSym(); mSymTab[sym] = symP;	
	symP = new AcNrm    ; sym = symP->getSym(); mSymTab[sym] = symP;	
	symP = new AcSci    ; sym = symP->getSym(); mSymTab[sym] = symP;	
	symP = new AcFix    ; sym = symP->getSym(); mSymTab[sym] = symP;	
	symP = new AcFrc    ; sym = symP->getSym(); mSymTab[sym] = symP;	
	symP = new AcMix    ; sym = symP->getSym(); mSymTab[sym] = symP;	
	symP = new AcStd    ; sym = symP->getSym(); mSymTab[sym] = symP;	
	symP = new AcEe     ; sym = symP->getSym(); mSymTab[sym] = symP;	
	symP = new AcEng    ; sym = symP->getSym(); mSymTab[sym] = symP;	
	symP = new AcEngBack; sym = symP->getSym(); mSymTab[sym] = symP;	
}

//******************************************************************************

/// Funktion erlaubt es, ein Symbol vor der eigentlichen Verarbeitung zu
/// veraendern. Insbesondere wird ein Mapping durchgefuehrt, wenn ein Modus
/// (Invers, Hyperbolicus, ...) aktiv ist. Z.B. wird im Invers-Modus das
/// Symbol "sin" zu "asin".
/// \retval in Eingabesymbol und das gemappte Symbol
/// \return    true, wenn das Symbol vollstaendig prozessiert worden ist

bool ComputerSci::preProcess(string& in)
{
	bool isMapped    = map(in);
	bool isProcessed = false;
	
	isMapped = isMapped;
	return isProcessed;
}

//******************************************************************************

/// Prueft verschiedene Shift-Ebenen und versucht, das Symbol sym entsprechend
/// anzuupassen.
/// \retval sym Eingabesymbol und das gemappte Symbol
/// \return     true, wenn das Symbol veraendert worden ist
//  fse, 06.07.2

bool ComputerSci::mapSci(string& sym)
{
	bool isMapped = false;
	
	if (mIsInv ) isMapped = mapInvSci(sym);
	if (mIsHyp ) isMapped = mapHypSci(sym);
	if (mIsFixInput || mIsSciInput) isMapped = mapFixSciInputSci(sym);
	return isMapped;
}

//************************************** ***************************************

/// Ordnet einem Symbol dasjenige Symbol zu, das ihm im Invers-Modus entspricht.
/// \retval sym Eingabesymbol und das gemappte Symbol
/// \return     true, wenn das Symbol veraendert worden ist
//  fse, 26.05.02

bool ComputerSci::mapInvSci(string& sym)
{
	if ("sin" == sym) { sym = "asin" ; return true; }
	if ("cos" == sym) { sym = "acos" ; return true; }
	if ("tan" == sym) { sym = "atan" ; return true; }
	if ("log" == sym) { sym = "pow10"; return true; }
	if ("ln"  == sym) { sym = "ex"   ; return true; }
	if ("pow" == sym) { sym = "root" ; return true; }
	return false;
}

//************************************** ***************************************

/// Ordnet einem Symbol dasjenige Symbol zu, das ihm im Hyperbolicus-Modus
/// entspricht (sin -> sinh, cos -> cosh etc.).
/// \retval sym Eingabesymbol und das gemappte Symbol
/// \return     true, wenn das Symbol veraendert worden ist
//  fse, 26.05.02

bool ComputerSci::mapHypSci(string& sym)
{
	if (sym ==  "sin" || sym ==  "cos" || sym ==  "tan"
	 || sym == "asin" || sym == "acos" || sym == "atan")
	{
		sym += "h";
		return true;
	}
	return false;
}

//************************************** ***************************************

/// Macht aus einer Ziffer ein "fix"- oder "sci"-Symbol, wenn zuvor "fix" oder
/// "sci" eingegeben wurde.
/// "fix" oder "sci" wurden zuvor ohne Parameter eingegeben. Die naechste Ziffer
/// muss nun als Parameter interpretiert werden.
/// \retval sym Eingabesymbol und das gemappte Symbol
/// \return     true, wenn das Symbol veraendert worden ist
//  fse, 06.07.02
//  fse, 17.03.03: Trennzeichen '#' durch '@' ersetzt

bool ComputerSci::mapFixSciInputSci(string& sym)
{
	char c = sym[0];
	bool ret = false;

	if (sym.size() != 1) goto lbl;
	if (c >= '0' && c <= '9')
	{
		sym = "";
		if (mIsFixInput) sym = "fix@";
		if (mIsSciInput) sym = "sci@";
		sym += c;
		ret = true;
	}
lbl:
	mIsFixInput = mIsSciInput = false;
	return ret;
}

//******************************************************************************

/// Erzeugt String mit den wichtigen Member-Variablen.
/// \retval String mit den wichtigen Member-Variablen
//  fse, 13.07.02

string ComputerSci::makeCompSciInfoStr()
{
	ostringstream str;
	string tmp;

	str << makeCompBaseInfoStr() << "\n";
	str << "       mIsInv: " << Util::trimStr(mIsInv  ) << "  mIsFixInput: " << mIsFixInput<< "\n";
	str << "       mIsHyp: " << Util::trimStr(mIsHyp  ) << "    mPrcWhole: " << mIsSciInput<< "\n";
	str << "     mAngMode: " << Util::trimStr(mAngMode);
//	str << ends;
	return str.str();
}

//******************************************************************************
// Klasse DisplaySci:
//******************************************************************************

/// Initialisiert das Display.
//  fse, 11.06.04: Initialisierung einiger vergessener Members
//  fse, 19.04.14: mForceSci wird mit 0.0000001 statt mit 0,001 initialisiert,
//                 da der TI-30 sonst bei z. B. 0.25/100 als Ergebnis 2.5e-3
//                 statt 0.0025 geliefert hat. Die anderen beiden Rechner
//                 initialisieren den Wert in ihren Konstruktoren mittels
//                 restoreNvRam() selbst, so dass fuer sie der Wert hier egal.

DisplaySci::DisplaySci(ComputerBase* cbP)
	: DisplayBase(cbP), mDspMode(NRM), mFrcMode(STD), mInv("..."), mHyp("..."),
	  mDeg("..."), mRad("..."), mGra("..."),
	  mSci("..."), mFix("..."), mIsSxgMode(false),
	  mDigits(-1), mShiftNum(0), mEng(0), mForceSci(0.0000001)
{
	mCompP = static_cast<ComputerSci*>(DisplayBase::mCompP);
	setModesSci();
}

//******************************************************************************

/// Schaut im Computer, welche Modes gesetzt sind, und baut den Mode-String des
/// Displays entsprechend zusammen.
//  fse, 12.04.02

void DisplaySci::setModes()
{
	setModesSci();
	mModes =  mInv + " " + mMem + " " + mK + " " + mHyp + " "
	        + mDeg + " " + mRad + " " + mGra + " " + mSci + " " + mFix;
}

//************************************** ***************************************

/// Setzt die Modus-Strings auf die eingestellten Werte.
//  Aus setModes() ausgelagert, weil auch von den abgeleiteten Klassen augerufen.
//  fse, 13.04.02

void DisplaySci::setModesSci()
{
	bool         isInv = mCompP->getIsInv();
	bool         ishyp = mCompP->getIsHyp();
	AngMode      am    = mCompP->getAngMode();

	setModesBase();
	setInv(isInv);
	setHyp(ishyp);
	setAngMode(am);
	setDspMode(mDspMode);
}

//************************* ************************* **************************

/// Setzt die Winkelmodus-Strings.
/// \param am eingestellter Winkelmodus

void DisplaySci::setAngMode(AngMode am)
{
	switch(am)
	{
		case DEG: setDeg(); break;
		case RAD: setRad(); break;
		case GRA: setGra(); break;
		default: setDeg();
	}
}

//************************* ************************* **************************

/// Setzt die Display-Modus-Strings.
/// \param dm eingestellter Display-Modus

void DisplaySci::setDspMode(DspMode dm)
{
	mDspMode = dm;
	switch(dm)
	{
		case NRM: setNrm(); break;
		case SCI: setSci(); break;
		case FIX: setFix(); break;
		default: setNrm();
	}
}

//******************************************************************************

/// Wandelt den Wert in mVal in einen String, aber nur, wenn irgendwelche
/// Aenderungen am Wert oder der Darstellung erfolgt sind. Dies ist wichtig,
/// da makeDisplay() so auch aufgerufen werden kann, wenn in der Oberflaeche
/// Eingaben gemacht wurden, diese aber noch nicht im Computer gelandet sind.
//  fse, 07.07.02

void DisplaySci::makeValue()
{
	static Number  val = 0;
	static int     digits    = mDigits;
	static DspMode dspMode   = mDspMode;
	static FrcMode frcMode   = mFrcMode;
	static int     shiftNum  = mShiftNum;
	static int     eng       = mEng;
	static bool    isSxgMode = mIsSxgMode;
	static double  forceSci  = mForceSci;
	int    format, size;
	Symbol* lastInP = mCompP->getLastInP();

	if (!mVal.isFrc()) mFrcMode = STD;
	if (!(dynamic_cast<AcEng*>(lastInP) || dynamic_cast<AcEngBack*>(lastInP))) 
		mEng = 0;
	if (!(dynamic_cast<AcEe*>(lastInP) || dynamic_cast<AcInv*>(lastInP)))
		mShiftNum = 0;
	//bool uneq = mVal != val; // zum Testen in MSVC, da Probleme mit -1.#IND
	if (    mVal != val       || mDigits != digits   || dspMode   != mDspMode
	 || shiftNum != mShiftNum || frcMode != mFrcMode || isSxgMode != mIsSxgMode
	 || eng      != mEng      || forceSci != mForceSci)
	{
		format = mDspMode;
		if (mFrcMode != STD) format = 2 + mFrcMode;
		if (mIsSxgMode     ) format = 5;
		if (mEng     != 0  ) format = 1;
		mValue = mVal.makeStr(format, mDigits, mShiftNum, mForceSci);
		size = mValue.size();
		if (mValue[0] == '-') --size;
		if (mValue.find('/') != string::npos && size > 8)
		{ // wenn Bruchdarstellung zu gross, in Dezimalzahl aendern:
			mVal.frcToDbl();
			mFrcMode = STD;
			format = mDspMode;
			mValue = mVal.makeStr(format, mDigits, mShiftNum);
		}
  		val        = mVal;
		digits     = mDigits;
		dspMode    = mDspMode;
		eng        = mEng;
		shiftNum   = mShiftNum;
		frcMode    = mFrcMode;
		isSxgMode  = mIsSxgMode;
		mIsSxgMode = false;
		forceSci   = mForceSci;
	}
}

//******************************************************************************

/// Erzeugt den String fuer die Sieben-Segment-Anzeige so wie sie der
/// Taschenrechner darstellen wuerde. Das Ergebnis wird in der Klassenvariablen
/// mSeven abgelegt.
/// Wenn mShiftNum != 0, wird die Zahl in Exponentialschreibweise dargestellt
/// und das Komma der Mantisse um diese Zahl nach rechts (>0) oder links (<0)
/// verschoben. mShiftNum wird von AcEe::process() gesetzt, um der Display-
/// Klasse diesen Wunsch mitzuteilen. Da das nur temporaer geschehen soll, wird
/// mShiftNum dann wieder zurueckgesetzt.
/// \param displayStr normalerweise wird der Wert in mVal in einen String
///                   umgewandelt; wenn dieser Parameter uebergeben wird, wird
///                   stattdessen mSeven so gesetzt, wie dieser String als
///                   Sieben-Segment-Anzeige aussehen wuerde.
//  fse, 09.05.02

void DisplaySci::makeSeven(string displayStr)
{
	if (displayStr == "")
		mSeven = mVal.makeStr(mDspMode, mDigits, mShiftNum);
	else
		mSeven = displayStr;
	mSevenRaw = mSeven;
	mShiftNum = 0;
}

//******************************************************************************
// Weitere Symbole in wissenschaftlichen Rechnern:
//******************************************************************************

/// Berechnet den Reziprokwert der Zahl a.
/// \param a Zahl, deren Reziprokwert ermittelt werden soll
/// \return  Reziprokwert von a

Number FnRcp::calcVal(const Number& a)
{
	if (a.isFrc())
	{
		try
		{
			Number n(a.frc().rcp());
			return n;
		}
		catch (ZeroDivide)
		{
			Number b = a; // weil a const ist, geht a.frcToDbl() nicht
			b.frcToDbl();
			return Number(1.0/b);
		}
	}
	else
	{
		return Number(1.0/a);
	}
}

//******************************************************************************

/// Prozessiert das pi-Symbol. Pi wird berechnet und auf den Stack geworfen.

void NumPi::process()
{
	double val = asin(1.0)*2.0;
	
	processCommon();
	mCompP->pushNum(val);
}

//******************************************************************************

/// Prozessiert alle Winkelfunktions-Symbole.
//  fse, 02.08.03: Wenn unmittelbar vor der Funktion ein Operator eingegeben
//                 wurde, darf die Zahl vom Stack nicht gepopped, sondern nur
//                 getopped werden. Siehe Kommentar bei
//                 SymOtherFunction::process().
//  fse, 04.08.03: Korrektur, falls STO oder INV das letzte Symbol.
//  fse, 11.11.03: lastInP auf != NULL pruefen

void SymTrigFunction::process()
{
//	Number arg = mCompP->popNum(); // fse, 02.08.03
	Number arg;                    //
	Number res;
	bool isArea = getSym()[0] == 'a';
	Symbol* lastInP       = getLastInP();

	if (lastInP && lastInP->getSym() == "inv") // Abfrage mit dynamic_cast<AcInv> geht
		lastInP = mCompP->getLastButOneInP();   // hier nicht, da AcInv erst in calc_sci.h
	if (dynamic_cast<SymOperator*>(lastInP) || // fse, 02./04.08.03
	    dynamic_cast<AcSto*      >(lastInP))   //
		arg = mCompP->topNum();                 //
	else                                       //
		arg = mCompP->popNum();                 //
	
	if (isArea) // Area-Funktionen
	{
		res = calcVal(arg.getVal());
		res = radToDegOrGra(res.getVal());
	}
	else
	{
		arg = DegOrGraToRad(arg.getVal());
		res = calcVal(arg);
	}
	mCompP->pushNum(res);
}

//************************************** ***************************************

/// Wandelt einen Winkel von Neugrad oder Altgrad ins Bogenmass um.
/// Noetig, da die Compiler-Funktionen grundsaetzlich im Bogenmass rechnen.
/// \param ang Winkel in Neugrad oder Altgrad
/// \return   Winkel im Bogenmass
//  fse, 07.10.01

double SymTrigFunction::DegOrGraToRad(double ang) const
{
	static double piHalf = asin(1.0);
	ComputerSci* compSciP = static_cast<ComputerSci*>(mCompP);

	switch(compSciP->getAngMode())
	{
		case RAD: return ang;
		case DEG: return ang * piHalf/ 90.0;
		case GRA: return ang * piHalf/100.0;
		default: return ang;
	}
}

//************************************** ***************************************

/// Wandelt einen Winkel vom Bogenmass in Altgrad oder Neugrad um.
/// Noetig, da die Compiler-Funktionen grundsaetzlich im Bogenmass rechnen.
/// \param ang Winkel im Bogenmass
/// \return   Winkel in Neugrad oder Altgrad
//  fse, 07.10.01

double SymTrigFunction::radToDegOrGra(double ang) const
{
	static double piHalf = asin(1.0);
	ComputerSci* compSciP = static_cast<ComputerSci*>(mCompP);

	switch(compSciP->getAngMode())
	{
		case RAD: return ang;
		case DEG: return ang *  90.0/piHalf;
		case GRA: return ang * 100.0/piHalf;
		default: return ang;
	}
}

//******************************************************************************

/// Setzt Invers-Modus.
//  fse, 08.10.01

void AcInv::process()
{
	ComputerSci* compP = static_cast<ComputerSci*>(mCompP);
	bool invMode = compP->getIsInv();
	
	compP->setIsInv(!invMode);
}

//******************************************************************************

/// Setzt Hyperbolikus-Modus.
//  fse, 01.07.02

void AcHyp::process()
{
	ComputerSci* compP = static_cast<ComputerSci*>(mCompP);
	bool isHyp = compP->getIsHyp();
	
	compP->setIsHyp(!isHyp);
}

//******************************************************************************

/// Setzt Konstanten-Modus.
//  fse, 08.10.01

void AcConst::process()
{
	mCompP->setConstVal(mCompP->topNum());
	mCompP->setConstOpP(mCompP->topSym()); // Konstanten-Operator speichern
}

//******************************************************************************

/// Setzt Display-Modus auf NRM.
//  fse, 14.04.02

void AcNrm::process()
{
	DisplaySci*  dsplP = static_cast<DisplaySci* >(mDsplP);

	dsplP->setDigits(-1);
	dsplP->setShiftNum(0);
	dsplP->setDspMode(NRM);
}

//******************************************************************************

/// Setzt Display-Modus auf SCI. Der Parameter digits wurde zuvor durch
/// mCompP->extractParam() extrahiert und gespeichert.
//  fse, 14.04.02

void AcSci::process()
{
	DisplaySci*  dsplP = static_cast<DisplaySci* >(mDsplP);
	ComputerSci* compP = static_cast<ComputerSci*>(mCompP);
	int digits = mCompP->getParam();

	if (digits == NO_PARAM)
	{
		compP->setIsSciInput(true);
		compP->setLastInPisNotToSet(true);
	}
	else
	{
		dsplP->setDspMode(SCI);
		if (digits == 0 || digits > 8) digits = 8;
		dsplP->setDigits(digits-1); // Casio: digits sind alle Stellen, nicht nur Nachkomma wie in C++
	}
}

//******************************************************************************

/// Setzt Display-Modus auf FIX. Der Parameter digits wurde zuvor durch
/// mCompP->extractParam() extrahiert und gespeichert.
//  fse, 14.04.02

void AcFix::process()
{
	DisplaySci*  dsplP = static_cast<DisplaySci* >(mDsplP);
	ComputerSci* compP = static_cast<ComputerSci*>(mCompP);
	int digits = compP->getParam();

	if (digits == NO_PARAM)
	{
		compP->setIsFixInput(true);
		compP->setLastInPisNotToSet(true);
	}
	else
	{
		dsplP->setDigits(digits);
		dsplP->setDspMode(FIX);
	}
}

//******************************************************************************

/// Verschiebt Dezimalpunkt der Anzeige um die Anzahl von Stellen, die im
/// Parameter uebergeben wird.
/// Falls der Parameter 9999, Verschiebung um +-1 vom vorherigen Wert. 9999 wird
/// von Calculator::process() eingetragen, wenn der Parameter weggelassen wird.
//  fse, 09.05.02

void AcEe::process()
{
	ComputerSci* compP = static_cast<ComputerSci*>(mCompP);
	DisplaySci*  dsplP = static_cast<DisplaySci*>(mDsplP);
	int          actShiftNum = dsplP->getShiftNum();
	int          shiftNum = mCompP->getParam();
	bool         isSci = dsplP->getDspMode() == SCI;
	bool         isInv = compP->getIsInv();
//	Symbol*      lastInP = getLastInP();
	AcSci sci;
		
//	if (!dynamic_cast<AcEe*>(lastInP) && !dynamic_cast<AcInv*>(lastInP))
//		mShiftNum = 0;
//	if (!isSci)
//		if (compP->getIsInv()) mShiftNum = 1; else mShiftNum = -1;
	if (!isSci)
		if (compP->getIsInv()) actShiftNum = 1; else actShiftNum = -1;
	mCompP->setParam(dsplP->getDigits());
	sci.process();
	if (shiftNum == 9999)
	{
//		if (isInv) --mShiftNum; else ++mShiftNum;
		if (isInv) --actShiftNum; else ++actShiftNum;
//		if (mShiftNum > mShiftNumMax) --mShiftNum;
		if (actShiftNum > mShiftNumMax) --actShiftNum;
		//cout << mShiftNum << "<<<<<\n";
	}
	else
	{
//		mShiftNum = shiftNum;
		actShiftNum = shiftNum;
	}
//	dsplP->setShiftNum(mShiftNum);
	dsplP->setShiftNum(actShiftNum);
}

//******************************************************************************

/// Verschiebt Dezimalpunkt der Anzeige so nach links, dass der Exponent ein
/// Vielfaches von 3 ist. Wiederholter Aufruf verschiebt ihn um weitere drei
/// Stellen.
//  fse, 14.07.02

void AcEng::process()
{
	shiftEng(Util::RIGHT);
}

//******************************************************************************

/// Verschiebt Dezimalpunkt der Anzeige so nach links, dass der Exponent ein
/// Vielfaches von 3 ist. Wiederholter Aufruf verschiebt ihn um weitere drei
/// Stellen.
//  fse, 14.07.02

void AcEngBack::process()
{
	shiftEng(Util::LEFT);
}

//************************************** ***************************************

/// Verschiebt Dezimalpunkt der Anzeige so nach rechts oder links, dass der
/// Exponent ein Vielfaches von 3 ist.
/// Wiederholter Aufruf verschiebt ihn um weitere drei Stellen.
/// \param dir Richtung der Verschiebung (Util::RIGHT, Util::LEFT)

void AcShiftPoint::shiftEng(Util::Direction dir)
{
	DisplaySci* dsplP = static_cast<DisplaySci*>(mDsplP);
	int         eng   = dsplP->getEng();
	int         delta = (dir == Util::LEFT) ? -1 : 1;
	string      value;
	double      m;
	int         e, shiftNum;

	dsplP->setShiftNum(0);
	dsplP->setEng(0);
	dsplP->makeValue();        // Originalzustand herstellen
	value = dsplP->getValue(); // Original-String holen
	Util::getMantExp(value, m, e);
	eng +=delta;
	// evtl. Zaehlerkorrektur bei Ueberschreiten der 0:
	if (eng == 0)
	{
		if (e == 0)
			eng += 2*delta; // damit beim Richtungswechsel nicht zweimal e+000
		else
			eng += delta;   // hier kommt e+000 nicht vor
	}
	shiftNum = Util::getEngShiftNum(e, eng);
	dsplP->setEng(eng);
	dsplP->setShiftNum(shiftNum);
}

//******************************************************************************
//******************************************************************************
