//******************************************************************************
// Copyright (c) 2002-2019 by Friedemann Seebass, Germany.
// Dieses Programm ist freie Software. Siehe GNU GPL Notiz in Datei main.cpp.
// This program is free software. See the GNU GPL notice in file main.cpp.
// Projekt: Almetare - Alle meine Taschenrechner
//******************************************************************************
/// \file calculator.cpp
///       Basisklassen fuer alle Taschenrechner (Modul base)
//******************************************************************************
// Wann       Wer      Was
// ---------- -------- ---------------------------------------------------------
// 29.05.2005 fse      CalculatorBase::CalculatorBase(): mHomeDir wird unter
//                     Windows aus der Registry gelesen
// 12.06.2004 fse      ComputerBase::initMembers(): zusaetzl. Initialisierungen
// 21.02.2004 fse      getTitle() modifiziert
// 17.11.2003 fse      strstream durch sstream ersetzt
// 02.08.2003 fse      letzte Aenderung in ComputerBase::initStacks() wieder
//                     rueckgaengig gemacht und nach
//                     SymOperator::processCommon() verlegt
// 14.05.2003 fse      initStacks(): Zahlenstack mit 0 initialisieren
// 31.03.2003 fse      neue Funktion getAllSyms()
// 25.03.2003 fse      neue virtuelle Funktion Computer::aftProcess2()
// 17.03.2003 fse      Trennzeichen '#' der Parameter in extractParam() durch
//                     '@' ersetzt, da ersteres Kommentarzeichen in der
//                     Ini-Datei ist und der fx-3600P diese Zeichen in den
//                     Programmstrings ablegt
// 16.03.2003 fse      Konstruktor Calculator mit Argument argc
// 01.03.2003 fse      Bugfix in CalculatorBase()
// 01.09.2001 fse      erzeugt
//******************************************************************************

#include <iostream>
#include <sstream>
#ifdef _WIN32
#include <windows.h>      // fuer Auslesen Home_dir. aus Registry
#else
#include <unistd.h>       // getuid()
#include <pwd.h>          // struct passwd
#endif
#include <iomanip>        // setw()
#include "calculator.h"
#include "logging.h"
#include "util.h"
#include "calc_sci.h"     // AcInv

using namespace std;

//******************************************************************************

ComputerBase* Symbol::mCompP;
DisplayBase*  Symbol::mDsplP;
// extern ComputerBase* newComputer(); 
// extern DisplayBase* newDisplay();

//template class DrawableList<Point>;        // explizite Instanziierung
//template Log& (operator<<) <ComputerBase>(Log& log, const ComputeBase& out);
//template ostream& operator<< <NumStack>(ostream& os, list<NumStack>& lst);

//******************************************************************************
// Klasse Error:
//******************************************************************************

/// Ausgabeoperator, wandelt enum ErrId in Strings.
/// \param os Ausgabe-Stream
/// \param e  Error-ID
/// \return   Ostream-Referenz

ostream& operator<<(ostream& os, const Error& e)
{
	switch (e.mErrId)
	{
		case OK         : os << "OK"                     ; break;
		case SYM_EXP    : os << "Symbol expected!"       ; break;
		case OP_EXP     : os << "Operator expected!"     ; break;
		case TWO_OPS_EXP: os << "Two operators expected!"; break;
		default         : os << "Unknown error code!"    ;
	}
	return os;
}

//******************************************************************************
// Klasse DisplayBase:
//******************************************************************************

/// Konstruktor. Setzt den Zeiger auf den Computer und die anderen Members.

DisplayBase::DisplayBase(ComputerBase* cbP) : mCompP(cbP), mValue("0"), mMem(".")
{
	setModes();
	makeSeven();
}

//******************************************************************************

/// Setzt den Wert der Anzeige auf val und fuellt die Strings mModes, mValue
/// und mSeven entsprechend.
/// \param val auszugebende Zahl
//  fse, 19.07.02

void DisplayBase::makeDisplay(const Number& val)
{
	Symbol* lastInP = mCompP->getLastInP();

	mVal = val;
	setModes();
	if (dynamic_cast<AcInv*>(lastInP)) return;
	makeValue();
	makeSeven();
}

//******************************************************************************

/// Wie makeDisplay(const Number& val), aber der Wert wird nicht neu gesetzt.
/// Moeglicherweise sind aber andere Parameter der Darstellungg
/// (z.B. DisplayMode) geaendert worden, daher diese Funktion.
//  fse, 19.07.02

void DisplayBase::makeDisplay()
{
	Symbol* lastInP = mCompP->getLastInP();

	setModes();
	if (dynamic_cast<AcInv*>(lastInP)) return;
	makeValue();
	makeSeven();
}

//************************************** ***************************************

/// Erzeugt einen String des Wertes, wie ihn der Taschenrechner darstellen
/// wuerde.
/// \param displayStr falls nicht "", wird aus diesem String der Ausgabestring
///                   erzeugt (nur relevant fuer die ueberladenden Funktionen)

void DisplayBase::makeSeven(string displayStr)
{
	mSevenRaw = mSeven = mVal.makeStr(0, -1);
	displayStr = displayStr;
}

//******************************************************************************

/// Setzt den String fuer die Rechner-Modi (DEG, SCI etc) zusammen.

void DisplayBase::setModes()
{
	setModesBase();
	mModes = mMem + " " + mK;
}

//************************************** ***************************************

/// Setzt den String fuer die Rechner-Modi (DEG, SCI etc) zusammen.
/// Aus setModes() ausgelagert, weil auch von den abgeleiteten Klassen augerufen.
//  fse, 13.04.02

void DisplayBase::setModesBase()
{
	bool isMem   = mCompP->getMemory() != 0.0;
	bool isConst = mCompP->getConstOpP() != NULL;

	setMem(isMem);
	setK(isConst);
}

//******************************************************************************

/// Prueft, ob bei einer Berechnung ein Fehler aufgetreten ist. Dabei wird
/// ausgenutzt, dass (zumindest bei Float-Arithmetik) bei der Umwandlung eines
/// fehlerhaften Double-Wertes in einen String ein Fehlerstring erzeugt wird.
/// Die Alternative waere eine Exception-Behandlung; das waere im Hinblick
/// darauf, dass dies bei Integer-Rechnung von den Compilern standardmaessig
/// so gehandhabt wird, wahrscheinlich die sauberere Loesung; das hat sich
/// allerdings erst spaeter herausgestellt.

bool DisplayBase::isError()
{
	//cout << mValue << endl;
	if (mValue=="1.#IND" || mValue=="-1.#IND" || mValue=="1.#INF" || mValue=="-1.#INF" // MSVC
	 || mValue=="inf"    || mValue=="-inf"    || mValue=="nan")                        // GNU
	{
		return true;
	}
	return false;
}

//******************************************************************************
/*
// Uebersetzt den numerischen Wert der Anzeige in einen String.
// fse, 14.04.02

void DisplayBase::makeValue()
{
	strstream str;

	str << mVal << ends;
	mValue = str.str();
}
*/
//******************************************************************************

/// Ausgabeoperator fuer Klasse DisplayBase. Wird fuer die Ausgabe auf die
/// Konsole benoetigt und gibt die Modes, die "Sieben-Segment"-Anzeige und den
/// rohen, numerischen Wert aus.

ostream& operator<<(ostream& os, const DisplayBase* dbP)
{
	os << "-------------------------\n";
	os << dbP->getModes() + "\n";
	os << dbP->getSeven() + "\n";
	os << dbP->getValue() + "\n";
	os << "-------------------------";
	return os;
}

//******************************************************************************
// Klasse ComputerBase:
//******************************************************************************

/// Konstruktor erzeugt Symboltabelle, initialisiert Symbol- und Nummern-Stack.
//  06.10.01

ComputerBase::ComputerBase() : mDsplP(0), mLastInPisNotToSet(false), mIsToInit(true)
{
	createSymTabBase();
	initStacks();
	initMembers();
}

//************************************** ***************************************

/// Erzeugt Symboltabelle mit allen Symbolen des "minimalen" Taschenrechners.
/// Da die Behandlung von Klammern integraler Bestandteil des Stack-Konzeptes
/// ist, gehoeren auch Klammern zum minimalen Taschenrechner, obwohl einfache 

/// Taschenrechner i.d.R. keine Klammern kennen.
//  fse, 08.04.02

void ComputerBase::createSymTabBase()
{	
	string sym;
	Symbol* symP;

	symP = new NumDmy     ; sym = symP->getSym(); mSymTab[sym] = symP;	
	symP = new OpAdd      ; sym = symP->getSym(); mSymTab[sym] = symP;	
	symP = new OpSub      ; sym = symP->getSym(); mSymTab[sym] = symP;	
	symP = new OpMul      ; sym = symP->getSym(); mSymTab[sym] = symP;	
	symP = new OpDiv      ; sym = symP->getSym(); mSymTab[sym] = symP;	
	symP = new OthBra     ; sym = symP->getSym(); mSymTab[sym] = symP;	
	symP = new OthKet     ; sym = symP->getSym(); mSymTab[sym] = symP;	
	symP = new FnSqrt     ; sym = symP->getSym(); mSymTab[sym] = symP;	
	symP = new AcHelp     ; sym = symP->getSym(); mSymTab[sym] = symP;	
	symP = new AcEqu      ; sym = symP->getSym(); mSymTab[sym] = symP;	
	symP = new AcPrc      ; sym = symP->getSym(); mSymTab[sym] = symP;	
	symP = new AcSto      ; sym = symP->getSym(); mSymTab[sym] = symP;	
	symP = new AcRcl      ; sym = symP->getSym(); mSymTab[sym] = symP;	
	symP = new AcSum      ; sym = symP->getSym(); mSymTab[sym] = symP;	
	symP = new AcSub      ; sym = symP->getSym(); mSymTab[sym] = symP;	
	symP = new AcExc      ; sym = symP->getSym(); mSymTab[sym] = symP;	
	symP = new AcSign     ; sym = symP->getSym(); mSymTab[sym] = symP;	
	symP = new AcInit     ; sym = symP->getSym(); mSymTab[sym] = symP;	
	symP = new AcInitWoMem; sym = symP->getSym(); mSymTab[sym] = symP;
	symP = new AcAc       ; sym = symP->getSym(); mSymTab[sym] = symP;
	symP = new AcC        ; sym = symP->getSym(); mSymTab[sym] = symP;
}

//******************************************************************************

/// Gibt alle Symbole der Symboltabelle auf die Konsole aus.
//  fse, 08.04.02

void ComputerBase::printSymHelp()
{
	SymTabIT it;
	int count = 0;
	string str;

	cout << "Known symbols:\n";
	for (it = mSymTab.begin(); it != mSymTab.end(); ++it)
	{
		// Manipulatoren funktionieren nicht mit GNU:
		//cout << setw(10) << ios::right << *(*it).second << ": "     // geht nicht
		//     << setw(25) << ios::left  << (*(*it).second).getHlp(); // mit GNU
		// ausserdem funktioniert Feldausrichtung nur mit C-Strings, nicht Strings:
		cout.width(10); cout.setf(ios::right, ios::adjustfield); cout << (*(*it).second).getSym().c_str() << ": ";
		cout.setf(ios::left , ios::adjustfield); cout.width(25); cout << (*(*it).second).getHlp().c_str();
		if (++count %  1 == 0) cout << endl; // falls mehr Eintr. pro Zeile
		if (  count % 23 == 0)
		{
			cout << "enter 'm' for more, 's' for stop: " << flush;
			cin >> str;
			if (str == "s" || str == "q") break;
		}
	}
	cout << endl;
}

//******************************************************************************

/// Gibt saemtliche Symbole und Infos im String zurueck.
//  fse, 31.03.03

string ComputerBase::getAllSyms()
{
	SymTabIT it;
	ostringstream str;

	for (it = mSymTab.begin(); it != mSymTab.end(); ++it)
	{
		// Manipulatoren funktionieren nicht mit GNU:
		//cout << setw(10) << ios::right << *(*it).second << ": "     // geht nicht
		//     << setw(25) << ios::left  << (*(*it).second).getHlp(); // mit GNU
		// ausserdem funktioniert Feldausrichtung nur mit C-Strings, nicht Strings:
		str.width(10); str.setf(ios::right, ios::adjustfield); str << (*(*it).second).getSym().c_str() << ": ";
		str.setf(ios::left , ios::adjustfield); str.width(25); str << (*(*it).second).getHlp().c_str() << endl;
	}
//	str << ends;
	return str.str();
}

//******************************************************************************

/// Schreibt alle Symbole in die Log-Datei.
//  fse, 08.04.02

void ComputerBase::showSymTab()
{
	SymTabIT it;

	for (it = mSymTab.begin(); it != mSymTab.end(); ++it)
	{
		cout << *(*it).second << NL;
		ONL_LOG(gnl) << *(*it).second << NL;

	}
}

//************************************** ***************************************

/// Initialisiert die beiden Stack-Listen.
/// Die Stacks koennen leider nicht bereits im Konstruktor initialisiert werden,
/// da mLogFileP zum Zeitpunkt des Aufrufs (Member von Calculator) noch nicht
/// gesetzt ist.
//  fse, 27.10.01
//  fse, 02.12.01: Da es jetzt die Log-Klasse gibt muss kein mLogFileP mehr
//                 angelegt werden und die Initialisierung kann wieder im 
//                 Konstruktor erfolgen.
//  fse, 14.05.03: Zur Initialisierung 0 auf den Zahlenstack legen, weil z.B.
//                 die Casio Rechner sonst nach dem Programmstart und Eingabe
//                 von "*6=" 36 statt 0 ausgeben.
//  fse, 02.08.03: Letzte Aenderung wieder rueckgaengig gemacht, da sonst
//                 "6*= 6*= 0" statt 36 ergibt, da "=" auch initStacks()
//                 aufruft. Dieser Fall wird jetzt direkt in
//                 SymOperator::processCommon() behandelt.

void ComputerBase::initStacks()
{
	mNumStack.clear();
	mSymStack.clear();
	mNumStack.setListP(&mNumStackList);
	mSymStack.setListP(&mSymStackList);
	mNumStackList.clear();
	mSymStackList.clear();
	mNumStackList.push_back(mNumStack); // ein numStack muss schon in der Liste sein
	mSymStackList.push_back(mSymStack); // ein symStack muss schon in der Liste sein
//	mNumStackList.back().push_back(0);  // Stack mit 0 initialisieren
}

//************************************** ***************************************

/// Initialisiert die Members von ComputerBase.
/// Aus Konstruktor ausgelagert, weil auch von AcInit::process() aufgerufen.
/// \param clrMem wenn true, wird der Memory-Speicher geloescht
//  fse, 08.04.02
//  fse, 12.06.04: mLastButOneInP und mParam initialisieren

void ComputerBase::initMembers(bool clrMem)
{
	mLastInP       = NULL;
	mLastButOneInP = NULL; 
	mLastOpP       = NULL;
	mConstOpP      = NULL;
	mConstVal      = 0.0;
	mParam         = 0;

	if (clrMem) mMemory   = 0.0;
	mBraLev   = 0;
	mPrcWhole = 0.0;
}

//******************************************************************************

/// Die mit new angelegten Symbole muessen wieder geloescht werden.
//  fse, 08.04.02

ComputerBase::~ComputerBase()

{
	for (SymTabIT it = mSymTab.begin(); it != mSymTab.end(); it++)
		delete (*it).second;
}

//******************************************************************************

/// Speichert den Zeiger auf das zuletzt eingegebene Symbol.
/// Alle Symbole mit Parametern benoetigen normalerweise zwei Durchlaeufe (z.B.
/// Kout: im zweiten Durchlauf wird daraus "Kout@2"). Beim ersten Durchlauf darf
/// mLastInP nicht gesetzt werden, da die Bearbeitung noch nicht abgeschlossen
/// ist. Daher koennen diese Symbole das durch Setzen von mLastInPisNotToSet
/// unterbinden.
/// \param sP Zeiger auf das Symbol, das in mLastInP gespeichert wird
//  fse, 28.07.02

void ComputerBase::setLastInP(Symbol* sP )
{
	if (mLastInPisNotToSet)
	{
		mLastInPisNotToSet = false;
		return;
	}
	mLastButOneInP = mLastInP;
	mLastInP = sP ;
}

//******************************************************************************

/// Speichert die Zeiger auf die beiden zuletzt eingegebene Operator-Symbole.
/// \param sP Zeiger auf das Symbol, das in mLastInP gespeichert wird
//  fse, 20.06.14

void ComputerBase::setLastOpP(SymOperator* sP )
{
	mLastButOneOpP = mLastOpP;
	mLastOpP = sP ;
}

//******************************************************************************

/// Wirft neue Stacks auf die Stack-Listen.
/// Wird benoetigt, wenn die Klammerebene erhoeht wird.
//  fse, 06.10.01

void ComputerBase::pushNewStacks()
{
	mNumStackList.push_back(mNumStack);
	mSymStackList.push_back(mSymStack);
}

//******************************************************************************

/// Holt die letzten Stacks von den Stack-Listen.
/// Wird benoetigt, wenn die Klammerebene erniedrigt wird.
//  fse, 06.10.01

void ComputerBase::popStacks()
{
	mNumStackList.pop_back();
	mSymStackList.pop_back();
}

//******************************************************************************

/// Holt die vorherige Eingabe vom Zahlenstack, wenn es kein Operator war.
/// Das ist z.B. noetig, wenn direkt nacheinander zwei Zahlen eingegeben werden:
/// Die zweite muss die erste auf dem Stack ersetzen.
/// Diese Funktion muss in anderen Taschenrechnern erweitert werden um weitere
/// Symbole; z.B. gilt das auch fuer die Tasten "Kout@1", ..., "Kout@6" der
/// Casio-Taschenrechner.
/// Daher muss eine virtuelle Funktion aufgerufen werden, die wiederum einige
/// Funktionen wie diese hier aufruft.
/// \return true, wenn wirklich was gepopped wurde
//  fse, 27.04.02

bool ComputerBase::popNumDepOnLastSymBase()
{
	Symbol* lastInP = getLastInP();
	
	if (isSymEmpty()
	   || dynamic_cast<NumDmy*     >(lastInP)
	   || dynamic_cast<SymFunction*>(lastInP)
	   || dynamic_cast<AcPrc*      >(lastInP)
	   || dynamic_cast<OthBra*     >(lastInP)
	   || dynamic_cast<AcEqu*      >(lastInP)
	   || dynamic_cast<AcRcl*      >(lastInP)
	   || dynamic_cast<AcC*        >(lastInP)
	   || dynamic_cast<AcExc*      >(lastInP))
	{
		(void)popNum();
		return true;
	}
	return false;
}

//******************************************************************************

/// Berechnet den letzten Operator auf dem Symbol-Stack.
/// Dazu werden die benoetigten Operanden vom Zahlen-Stack geholt, die
/// Operation ausgefuehrt und das Ergebnis wieder auf dem Zahlen-Stack abgelegt.
/// \return Element von Klasse Error
//  fse, 06.10.01

Error ComputerBase::compute()

{
	Error err;
	SymOperator* soP;
	Number a, b;
	
	if (isSymEmpty())
	{
		err.set(SYM_EXP);
		return err;
	}
	soP = popSym();
	if (isNumEmpty())
	{
		err.set(OP_EXP);
		return err;
	}
	else
	{
		b = popNum();
	}
	if (isNumEmpty())
	{
		err.set(TWO_OPS_EXP);
		handleEmptyNumStack(a, soP, b);
	}
	else
	{
		a = popNum();
	}
	pushNum(soP->calcVal(a, b));
	return err;
}

//************************************** ***************************************

/// Erfindet einen zweiten Operanden, falls ein Operator mit nur einem berechnet
/// werden soll. Muss von den abgeleiteten Klassen je nach Taschenrechner
/// ueberladen werden.
/// \param a   erster Operand
/// \param soP Zeiger auf Operatorsymbol
/// \retval b  erfundener zweiter Operator
//  fse, 08.04.02

void ComputerBase::handleEmptyNumStack(Number& a, SymOperator* soP, Number& b)
{
	a=a;soP=soP;b=b;
	// !!!
}

//******************************************************************************
// virtuelle Funktionen:
// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * 

/// Untersucht, ob in den Konstanten-Modus gewechselt werden oder dieser
/// abgeschaltet werden muss.
/// Bedingungen fuers Einschalten (bei Casio-Rechnern):
/// - Es muss zwei Mal unmittelbar hintereinander derselbe Operator eingegeben
///   worden sein.
/// - Es duerfen keine Klammern eingegeben worden sein.
/// Bedingung fuers Ausschalten:
/// - Irgendein Operator wurde eingegeben.
//  fse, 03.10.01 // !!!

SymOperator* ComputerBase::checkConstMode(SymOperator* opP)
{
	opP=opP;
	if (mConstOpP)
	{ // wenn der Konstanten-Mode eingeschaltet, fuehrt neuer Operator zum Abschalten
		mConstOpP = NULL;
	}
	return mConstOpP;
}

//******************************************************************************

/// Erzeugt passende Anzeige nach Eingabe von "(".
/// Die Rechner unterscheiden sich darin, welche Zahl im Display erscheint, wenn
/// "(" eingegeben wird. (TI: behaelt aktuelle Zahl bei, sie wird nur geloescht,
/// wenn dann vor dem Operator noch eine andere eingegeben wird; Casio: schreibt
/// "0" ins Display).
/// Als Default-Verhalten ist hier das Casio-Verhalten implementiert.
//  fse, 13.04.02

void ComputerBase::handleDisplayAfterBra()
{
	pushNewStacks();
	pushNum(0.0);
}

//******************************************************************************

/// Prozentrechnung des Solar-Rechners:
///
/// - 1: a* b%: a*b/100
/// - 2: a/ b%: a*100/b
/// - 3: a+ b%: a+a*b/100
/// - 4: a- b%: a-a*b/100
///
/// - 5: a%   : a
/// - 6: a*%  : a*a/100          : wie oben mit b = a
/// - 7: a/%  : 100/a
/// - 8: a+%  : 0
/// - 9: a-%  : 0

// fse, 09.04.02 // !!!

void ComputerBase::handlePercent()
{
	SymOperator* soP;
	Number a, b;
	Number val;
	string op;

	if (isNumEmpty()) return; // Prozentzeichen ohne alles wird nichts
	b = popNum();
	if (isSymEmpty())
	{ // Fall 5:


		pushNum(b);
		return;
	}
	soP = popSym();
	op = soP->getSym();
	if (isNumEmpty())
	{ // Faelle 6 bis 9:
		if      (op == "*") val = b * b / 100.0; // Fall 6
		else if (op == "/") val = 100.0 / b;     // Fall 7
		else if (op == "+") val = 0.0;           // Fall 8
		else if (op == "-") val = 0.0;           // Fall 9
		pushNum(val);
		return;
	}
	a = popNum();
	if      (op == "*") val = a * b / 100.0;     // Fall 1
	else if (op == "/") val = a * 100.0 / b;     // Fall 2
	else if (op == "+") val = a + a * b / 100.0; // Fall 3
	else if (op == "-") val = a - a * b / 100.0; // Fall 4
	pushNum(val);
}

//******************************************************************************

/// Erzeugt fuers Debugging String mit den wichtigen Member-Variablen.
/// \return String mit den wichtigen Member-Variablen
//  fse, 13.07.02

string ComputerBase::makeCompBaseInfoStr()
{
	ostringstream str;
	string tmp;

	if (mConstOpP) tmp = mConstOpP->getSym(); else tmp = "-";
	str << "mNumStackList: " << mNumStackList << "\n";
	str << "mSymStackList: " << mSymStackList << "\n";
	str << "    mConstVal: " << Util::trimStr(mConstVal.makeStr()) << "      mBraLev: " << mBraLev       << "\n";

	str << "    mConstOpP: " << Util::trimStr(tmp                ) << "    mPrcWhole: " << mPrcWhole     << "\n";
	str << "      mMemory: " << Util::trimStr(mMemory  .makeStr());
//	str << ends;
	return str.str();
}

//******************************************************************************

/// Ausgabeoperator gibt wichtige Member-Variablen aus.

ostream& operator<<(ostream& os, ComputerBase& c)
{
	os << "ConstVal = " << c.mConstVal << "\n";
	os << "ConstOp  = ";
	if (c.mConstOpP) os << c.mConstOpP->getSym(); else os << "NULL";
	os << "\n";
//	os << "NumStackList: " << c.mNumStackList << "\n";
//	os << "SymStackList: " << c.mSymStackList;
	return os;
}

//******************************************************************************
// Klasse Calculator:
//******************************************************************************

/// Konstruktor tut eigentlich nichts, da die Konstruktion von den abgeleiteten
/// Klassen uebernommen wird.
//  fse, 29.09.01
//  fse, 01.03.03: qApp darf hier nicht verwendet werden, da Programm ja auch
//                 als Konsolenanwendung gestartet werden kann.
//  fse, 29.05.05: unter Windows wird mHomeDir jetzt aus Registry gelesen

CalculatorBase::CalculatorBase(int argc, char** argv)
{
	string::size_type slashPos;
	#ifdef _WIN32
	char homeRegPath[] = "Software\\Microsoft\\Windows\\CurrentVersion\\Explorer\\Shell Folders";
	int ret;
	HKEY hkey;
	unsigned long type, len = 255;
	unsigned char path[256]; // = "D:\home\fs"
	char* cP = (char*)path;

	ret = RegOpenKeyEx(HKEY_CURRENT_USER, homeRegPath, 0L, KEY_QUERY_VALUE, &hkey);
	if (ret != ERROR_SUCCESS)
	{
		ERR_LOG(hgh) << "Error in opening registry key '" << homeRegPath << "' (Return code: " << ret << "!" << NL;
	}
	ret = RegQueryValueEx(hkey, "Personal", NULL, &type, path, &len);
	if (ret != ERROR_SUCCESS)
	{
		ERR_LOG(hgh) << "Error in reading registry key '" << homeRegPath << "' (Return code: " << ret << "!" << NL;
	}
	ret = RegCloseKey(hkey);
	if (ret != ERROR_SUCCESS)
	{
		ERR_LOG(hgh) << "Error in closing registry key '" << homeRegPath << "' (Return code: " << ret << "!" << NL;
	}
	mHomeDir = cP;
	#else
	pid_t uid = getuid();
	passwd* pwdstructP = getpwuid(uid);
	mHomeDir  = pwdstructP->pw_dir;
	#endif
	mProgName = "CalculatorBase";
	mProgVers = "V.0.5";
	mProgPath = argv[0];
	mIsDlgApp = !(argc > 1 && strchr(argv[1], 'c'));
	slashPos  = mProgPath.find_last_of("/\\"); // nach letztem "/" bzw. "\" suchen
	mProgPath = mProgPath.substr(0, slashPos+1);
	mAccPath     = ""; // wird in abgel. Klasse gesetzt
	mIniFileName = ""; // wird in abgel. Klasse gesetzt
}

//******************************************************************************

/// Destruktor raeumt die Objekte auf, die von den abgeleiteten Calculatoren
/// angelegt wurden, da dies immer dieselben sind.
//  fse, 11.05.02

CalculatorBase::~CalculatorBase()
{
	delete mDisplayP;
	delete mComputerP ;
	delete mCfgsP;
	mDisplayP  = NULL;
	mComputerP = NULL;
	mCfgsP     = NULL;
}

//******************************************************************************

/// Erzeugt fuer die Konsolenversion einen String mit einem Titelbild.
/// \return String mit Titelbild

string CalculatorBase::getTitle()
{
	ostringstream str;
	
	str << "---------------------------------------------------\n";
	str << "                       ___ _____         __    ___\n";
	str << "   /\\\\   |    |\\\\  /| |__    |    /\\\\   |  \\\\ |__ \n";
	str << "  /__\\\\  |    | \\\\/ | |      |   /__\\\\  |__/  |   \n";
	str << " /    \\\\ |___ |     | |___   |  /    \\\\ |  \\\\ |___\n";
	str << "\n";
	str << "            ALle MEine TAschenREchner\n";
	str << "            Virtual pocket calculators\n";
	str << "\n";
	str << "  Copyright (c) 2002-2019 by\n";
	str << "  Friedemann Seebass, Germany\n";
	str << "  Published under the GPL\n";
	str << "  E-Mail: fs@friedemann-seebass.de\n";
	str << "  Web: http://www.friedemann-seebass.de\n";
//	str << "\n";
	str << "  This is version " << getProgVers() << " of\n";
	str << "  Almetare's " << getProgName() << "\n";
//	str << "\n";
//	str << "----------------------------------------------------"; // << ends;
	return str.str();
}

//******************************************************************************

/// Uebernimmt bei der Konsolenversion die Steuerung der Eingabe und Berechnung
/// und Ausgabe der Werte.
/// Bei Tests ohne Eingabe ueber Konsole und stattdessen Einlesen aus String
/// "expr": Zeile mit cin einkommentieren und die mit getNextSym() aktivieren.

double CalculatorBase::calculate()
{
	string expr = "hyp 3 / 2 = q";
	string in;

//	ONL_LOG(med) << NL << *mComputerP << NL;
//	ONL_LOG(med) << NL << mDisplay << NL;
	mDisplayP->makeDisplay(0);
	cout << getTitle() << endl;
	cout << mDisplayP << endl;
	for (;;)
	{
		ostringstream str;
		cin >> in;
		//in = getNextSym(expr);
		//cout << in << ", >" << expr << "<\n";
		if (in == "quit" || in == "q") break;

		process(in);
//		ONL_LOG(med) << NL << mComputer << NL;
//		str.clear();
//		if (mComputerP->isNumEmpty())
//			str << "0"; // << ends;
//		else
//			str << mComputerP->topNum(); // << ends;
//		mDisplayP->setSeven(str.str());

//		ONL_LOG(med) << NL << mDisplay << NL;
		cout << mDisplayP << NL;
	}
	return 0.0;

}

//************************************** ***************************************

/// Holt aus dem String expr das erste Symbol.
/// \param expr String mit einer ganzen Tastenfolge (z.B. "3.4 * 3 sin ="
/// \return     String mit dem ersten Symbol

string CalculatorBase::getNextSym(string& expr) const
{
	string::size_type pos = expr.find(" ");
	string sym(expr, 0, pos);
	string newExpr(expr, pos+1, string::npos);

	expr = newExpr;
	if (pos == string::npos) expr = "";
	return sym;
}

//************************************** ***************************************

/// Verarbeitet ein einzelnes Symbol.
/// Dies ist die zentrale Funktion zur Bearbeitung von Eingaben. Im String in 
/// wird ein Symbol erwartet; das sollte entweder eine Zahl sein oder eine
/// Zeichenkette, die sich in der Symboltabelle befindet. Die Zeichenkette wird
/// in der Symboltabelle gesucht, damit ist dann ein Symbol-Pointer gefunden.
/// Das Symbol wird durch die process()-Funktion des Symbols verarbeitet.
/// Wenn das Symbol nicht gefunden wird, wird von der Funktion isSym() eine
/// Fehlermeldung ausgegeben.
/// \param in Eingabe-String, der eine Zahl oder ein Symbol enthalten sollte
//  fse, 02.09.01
//  fse, 22.06.14: Wir merken uns jetzt auch den letzten Operator, damit die
//                 Sonderbehandlung von "+" und "-" nach Prozentrechnung der
//                 Casio-Rechner nur nach "*" und "/" erfolgt. Siehe
//                 SymOperator::processAddSub().

void CalculatorBase::process(string in)
{
	bool isNumber;
	bool isProcessed;
	Number val;

	DBG_LOG(gnl) << "processing: " << in << NL;
	isProcessed = mComputerP->preProcess(in);
	if (isProcessed) return;
	if (mDisplayP->isError() && !(in == "ac" || in == "initwomem")) return;
	isNumber = in.size() >= 2 && in[0] == '-' && (isdigit(in[1]) || in[1] == '.');
	isNumber = isNumber || isdigit(in[0]) || in[0] == '.';
	if (isNumber)
	{
		Symbol* sP = mComputerP->getSymP("0");
		NumDmy* numP = static_cast<NumDmy*>(sP);  	
		numP->setSym(in);
		(void)numP->process();
		mComputerP->setLastInP(numP);
	}
	else
	{
		Symbol* inP;
		in = mComputerP->extractParam(in);
		if (!mComputerP->isSym(in)) return;
		inP = mComputerP->getSymP(in);
		(void)inP->process();
		mComputerP->setLastInP(inP);
		if (dynamic_cast<SymOperator*>(inP))           // fse, 22.06.14
			mComputerP->setLastOpP((SymOperator*)inP); //
	}
	if (!mComputerP->isNumEmpty())
		val = mComputerP->topNum();
	mComputerP->aftProcess(in);
//	cout << mComputerP->makeCompInfoStr() << "\n";
	mDisplayP->makeDisplay(val);
	mComputerP->aftProcess2(in);
}

//************************* ************************* **************************

/// Prueft, ob Symbol sym in der Symboltabelle ist, und gibt true oder false
/// zurueck.
/// \param sym zu untersuchende Zeichenkette
/// \return    true, wenn das Symbol in der Symboltabelle gefunden wurde
//  fse, 08.09.01

bool ComputerBase::isSym(const string& sym)
{
	SymTabIT it = mSymTab.find(sym);
	if (it != mSymTab.end()) return true;
	ERR_LOG(hgh) << "Symbol '" << sym << "' not in symbol table!" << NL;
	return false;
}

//************************* ************************* **************************

/// Extrahiert einen moeglicherweise vorhandenen Parameter des Symbols.
/// Einige Operatoren ("sci", "fix") benoetigen einen Parameter. Dieser wird bei
/// der Eingabe, durch "@" getrennt, an den Operator angehaengt und hier wieder
/// extrahiert.
/// Bei "ee" (Exponentialtaste des TI-30) Sonderbehandlung: Wenn hier der
/// Parameter weggelassen wird, wird er auf 9999 gesetzt; dies wird in Symbol
/// AcEe dann ausgewertet.
/// \param in zu untersuchende Zeichenkette
//  fse, 14.04.02
//  fse, 17.03.03: Trennzeichen '#' durch '@' ersetzt, da ersteres
//                 Kommentarzeichen in der Ini-Datei ist und der fx-3600P diese
//                 Zeichen in den Programmstrings ablegt

string ComputerBase::extractParam(string& in)
{
	string::size_type hashPos = in.find('@');
	string par;

	if (hashPos == string::npos)
	{
		mParam = NO_PARAM;
		return in;
	}
	if (hashPos == string::npos) return in;
	par = in.substr(hashPos+1);
	in = in.substr(0, hashPos);
	mParam = atoi(par.c_str());
	return in;
}

//******************************************************************************
//************************************** ***************************************
//************************* ************************* **************************
//****************** ******************* ******************* *******************
//************** *************** *************** *************** ***************
//************ ************ ************ ************ ************ *************
