//******************************************************************************
// Copyright (c) 2002-2004 by Friedemann Seebass, Germany.
// Dieses Programm ist freie Software. Siehe GNU GPL Notiz in Datei main.cpp.
// This program is free software. See the GNU GPL notice in file main.cpp.
// Projekt: Almetare - Alle meine Taschenrechner
//******************************************************************************
/// \file calculator.h
///       Basisklassen fuer alle Taschenrechner (Modul base)
//******************************************************************************
// Wann       Wer      Was
// ---------- -------- ---------------------------------------------------------
// 31.03.2003 fse      neue Funktion getAllSyms()
// 25.03.2003 fse      neue virtuelle Funktion Computer::aftProcess2()
// 16.03.2003 fse      Konstruktor Calculator mit Argument argc
// 01.03.2003 fse      Bugfix in CalculatorBase
// 01.09.2001 fse      erzeugt
//******************************************************************************

#ifndef CALCULATOR_H
#define CALCULATOR_H

#ifndef __cplusplus
#error calculator.h is only for C++!
#endif

//******************************************************************************

#include <string>
#include <map>
#include "stack.h"
#include "symbols.h"
#include "cfgs.h"

using namespace std;

//******************************************************************************

class Symbol;
class Calculator;
class SymOperator;
class Number;

typedef Stack<Number> NumStack;       ///< Stack fuer Zahlen
typedef Stack<SymOperator*> SymStack; ///< Stack fuer Symbole
typedef list<NumStack> NumStackList;  ///< Liste von Zahlen-Stacks
typedef list<SymStack> SymStackList;  ///< Liste von Symbol-Stacks
typedef map<string, Symbol*> SymTabT; ///< Symboltabelle
typedef SymTabT::iterator SymTabIT;   ///< Iterator fuer die Symboltabelle

//******************************************************************************
// Klasse Error:
//******************************************************************************

enum ErrId { OK, SYM_EXP, OP_EXP, TWO_OPS_EXP }; /// Art des Fehlers
enum { NO_PARAM = 9999 };                        /// hat Symbol Parameter?

/// Klasse zur Fehlerausgabe.
/// Die Klasse entspricht dem enum ErrId plus einer Ausgabe des Fehlers in
/// Ostream.

class Error
{
	ErrId mErrId; ///< Art des Fehlers
public:
	Error(ErrId id = OK) : mErrId(id) {}
	void set(ErrId id) { mErrId = id; }
	friend ostream& operator<<(ostream& os, const Error& e);	
};

//******************************************************************************
// Klasse Computer:
//******************************************************************************

/// Basisklasse aller Computer.
/// Die Klasse bildet zusammen mit den Symbolen das Rechenwerk des Calculators.
/// Sie enthaelt einen Zahlen-Stack, einen Symbolstack, verschiedene "Register"
/// (mMemory, mConstVal etc.) und Zustandsflags (RAD, DEG etc.).
/// Fuer jede Klammerebene wird ein eigener Zahlen- und Symbol-Stack angelegt;
/// daher gibt es in Wahrheit eine Liste von Zahlen-Stacks und eine Liste von
/// Symbol-Stacks.
/// Unterschiede zwischen den Rechnern werden in einem Satz virtueller
/// Funktionen nachgebildet.

class ComputerBase
{
protected:
	DisplayBase* mDsplP;             ///< Zeiger auf zugehoeriges Display
	SymTabT      mSymTab;            ///< Symboltabelle
	NumStackList mNumStackList;      ///< Zahlen-Stack-Liste zum Rechnen
	SymStackList mSymStackList;      ///< Symbol-Stack-Liste zum Rechnen
	NumStack     mNumStack;          ///< Default-Zahlen-Stack: ist immer leer
	SymStack     mSymStack;          ///< Default-Symbol-Stack: ist immer leer
	Symbol*      mLastInP;           ///< das zuletzt eingegebene Symbol
	Symbol*      mLastButOneInP;     ///< das vorletzte eingegebene Symbol
	SymOperator* mLastOpP;	         ///< der zuletzt eingegebene Operator
	SymOperator* mLastButOneOpP;	 ///< der vorletzte eingegebene Operator
	Number       mMemory;            ///< Speicherregister
	int          mBraLev;            ///< Klammerebene
	Number       mPrcWhole;          ///< bei Prozentrechnung: das "Ganze"
	SymOperator* mConstOpP;          ///< bei Konstantenrechnung: der Operator
	Number       mConstVal;          ///< bei Konstantenrechnung: der konstante Wert
	int          mParam;             ///< bei manchen Operatoren uebergebener Parameter
	bool         mLastInPisNotToSet; ///< damit kann Setzen von mLastInP verhindert werden
	bool         mIsToInit;          ///< wird von AcAc::process() auf true gesetzt und nur vom fx-3600P ausgewertet

	bool popNumDepOnLastSymBase();
	virtual void handleEmptyNumStack(Number& a, SymOperator* soP, Number& b);

public:
	// Initialisierung:
	ComputerBase();
	virtual ~ComputerBase();
	void         createSymTabBase();
	void         showSymTab();
	void         initStacks();
	void         initMembers(bool clrMem = true);
	void         clearNum()   { mNumStackList.back().clear(); }
	// Zugriff:
	Number       getConstVal     () const { return mConstVal     ; }
	Number       getPrcWhole     () const { return mPrcWhole     ; }
	Number       getMemory       () const { return mMemory       ; }
	int          getBraLev       () const { return mBraLev       ; }
	int          getParam        () const { return mParam        ; }
	SymOperator* getConstOpP     () const { return mConstOpP     ; }
	SymOperator* getLastOpP      () const { return mLastOpP      ; }
	SymOperator* getLastButOneOpP() const { return mLastButOneOpP; }
	Symbol*      getLastInP      () const { return mLastInP      ; }
	Symbol*      getLastButOneInP() const { return mLastButOneInP; }
	Symbol*      getSymP(const string& sym) { return mSymTab[sym]; }
	void         setDsplP  (DisplayBase* dsplP) { mDsplP  = dsplP; }
	void         setConstVal(const Number& val) { mConstVal = val; }
	void         setPrcWhole(const Number& val) { mPrcWhole = val; }
	void         setMemory  (const Number& val) { mMemory   = val; }
	void         setParam   (const int     val) { mParam    = val; }
	void         setLastOpP (SymOperator*  soP);
	void         setConstOpP(SymOperator*  soP) { mConstOpP = soP; }
	void         setLastInPisNotToSet(bool is ) { mLastInPisNotToSet = is; }
	void         setLastInP (Symbol*       sP );
	void         setIsToInit(bool          in ) { mIsToInit = in; }
	// Stack-Manipulation:
	void         pushNewStacks();
	void         popStacks();
	void         pushNum(const Number& n) { mNumStackList.back().push(n); }
	void         pushSym(SymOperator* sP) { mSymStackList.back().push(sP ); }
	Number       getTopButOneNum() { return mNumStackList.back().getTopButOne(); }
	void         setTopButOneNum(const Number& n) { mNumStackList.back().setTopButOne(n); }
	bool         isNumEmpty() { return mNumStackList.back().empty(); }
	bool         isSymEmpty() { return mSymStackList.back().empty(); }
	Number       popNum()     { return mNumStackList.back().pop  (); }
	Number       topNum()     { return mNumStackList.back().top  (); }
	SymOperator* popSym()     { return mSymStackList.back().pop  (); }
	SymOperator* topSym()     { return mSymStackList.back().top  (); }
	// Sonstiges:
	void         incBraLev() { ++mBraLev; }
	void         decBraLev() { --mBraLev; }
	void         printSymHelp();
	string       getAllSyms();
	bool         isSym(const string& sym);
	Error        compute();
	Error        swapNum();
	string       extractParam(string& in);
	string       makeCompBaseInfoStr();

	virtual bool         preProcess (string& in) { in = in; return false; }
	virtual void         aftProcess (string& in) { in = in; }
	virtual void         aftProcess2(string& in) { in = in; }
	virtual SymOperator* checkConstMode(SymOperator* opP);
	virtual void         handleDisplayAfterBra();
	virtual void         handlePercent();
	virtual void         popNumDepOnLastSym() { (void)popNumDepOnLastSymBase(); }
	virtual bool         map(string& sym) { sym = sym; return false; }
	virtual string       makeCompInfoStr() { return makeCompBaseInfoStr(); }

	friend ostream& operator<<(ostream& os, ComputerBase& c);
};

//******************************************************************************
// Klasse DisplayBase:
//******************************************************************************

/// Basisklasse aller Displays.
/// Die Klasse ist verantwortlich fuer die Darstellung aller Eingaben und 
/// Ergebnisse sowie der Anzeige der Betriebsmodi (DEG, RAD, SCI etc.).
/// Fuer die Ausgabe dieser Daten auf der Konsole besitzt sie einen
/// Ausgabeoperator, die Dialogboxen koennen die Daten selbst abfragen.
/// Unterschiede zwischen den Rechnern werden in einem Satz virtueller
/// Funktionen nachgebildet.

class DisplayBase
{
protected:
	ComputerBase* mCompP;    ///< Zeiger auf Computer mit den Modes
	string        mValue;    ///< Anzeige unformatiert
	string        mSeven;    ///< Sieben-Segment-Anzeige, formatiert wie Original
	string        mSevenRaw; ///< dito, aber noch mit Spaces statt '#' und evtl. e+002
	string        mMem;      ///< Memory-Anzeige
	string        mK;        ///< Konstante
	string        mModes;    ///< "INV MOD M K LRN ITG hyp SD LR DEG RAD GRA FIX SCI"
						     // Casio fx-3600P: "ON INV M K LRN Idx LR SD DEG RAD GRA"
						     // Casio fx-85v  : "   s m M K hyp     SD LR DEG RAD GRA FIX SCI"
	Number  mVal;            ///< numerischer Wert der Sieben-Segment-Anzeige

	void setModesBase();
public:
	DisplayBase(ComputerBase* cbP);
	virtual ~DisplayBase() {}
	// Zugriff:
	string getValue   () const { return mValue   ; }
	string getSeven   () const { return mSeven   ; }
	string getSevenRaw() const { return mSevenRaw; }
	string getModes   () const { return mModes   ; }
	Number getVal     () const { return mVal     ; }
	void setMem  (bool isOn) { if (isOn) mMem = "M"; else mMem = "."; }
	void setK    (bool isOn) { if (isOn) mK   = "K"; else mK   = "."  ; }
	void setCompP(ComputerBase* cP)    { mCompP = cP   ; }
	void setVal  (const Number& val  ) { mVal   = val  ; makeValue(); }
	void setValue(const string& value) { mValue = value; }
	void setSeven(const string& seven) { mSeven = seven; }
	// Sonstiges:
	void makeDisplay(const Number& val);
	void makeDisplay();
	bool isError();
	void setError() { mValue = "nan"; }
	virtual void setModes();
	virtual void makeValue() { mValue = mVal.makeStr(0, 15); }
	virtual void makeSeven(string displayStr = "");
	virtual int  getNumStrLen(string dsplStr) const { return dsplStr.size(); }
	friend ostream& operator<<(ostream& os, const DisplayBase* dbP);
};

//******************************************************************************
// Klasse CalculatorBase:
//******************************************************************************

/// Basisklasse aller Taschenrechner.
/// Ein Taschenrechner besteht aus einem Gehaeuse, einem Rechenwerk und einer
/// Anzeige sowie aus einem Algorithmus, der Zeichenketten entgegennimmt und an
/// den Computer weiterleitet. CalculatorBase (bzw. die hiervon abgeleiteten
/// Klassen bilden das Gehaeuse, das die anderen Komponenten als Members (bzw.
/// Zeiger darauf) enthaelt.
//  fse, 06.10.01
//  fse, 01.03.03: Konstruktor bekommt argv als Argument

class CalculatorBase
{
protected:
	string        mProgName;     ///< z.B. "TI-30", "Casio fx-85v", ...
	string        mProgPath;     ///< Pfad, unter dem das Binary liegt
	string        mProgVers;     ///< Programmversion zwecks Vrgl. in Ini-Datei
	string        mHomeDir;      ///< Home-Verzeichnis des Users (falls Linux)
	string        mAccPath;      ///< Zubehoer-Ordner
	string        mIniFileName;  ///< Name der Initialisierungsdatei (ohne Pfad)
	ComputerBase* mComputerP;    ///< Zeiger auf den Computer 
	DisplayBase*  mDisplayP;     ///< Zeiger auf die Anzeige
	Cfgs*         mCfgsP;        ///< Zeiger auf die Konfigurationsdaten
	bool          mIsDlgApp;     ///< als Konsolen- oder Dialogapplik. gestartet?
public:
	CalculatorBase(int argc, char** argv);
	~CalculatorBase();
	// Zugriff:
	ComputerBase* getComputerP() const { return mComputerP; }
	DisplayBase * getDisplayP () const { return mDisplayP ; }
	Cfgs*         getCfgsP    () const { return mCfgsP    ; }
	string        getProgName () const { return mProgName ; }
	string        getProgVers () const { return mProgVers ; }
	string        getAccPath  () const { return mAccPath  ; }
	bool          getIsDlgApp () const { return mIsDlgApp ; }
	// Sonstiges:
	double calculate();	
	void   process(string in);
	string getTitle();
	string getNextSym(string& expr) const;
	bool   isError() const { return mDisplayP->isError(); }
	int    getNumStrLen(string dsplStr)	{ return mDisplayP->getNumStrLen(dsplStr); }
};


//******************************************************************************
//************************************** ***************************************
//************************* ************************* **************************

#endif // !CALCULATOR_H

