//******************************************************************************
// Copyright (c) 2002-2005 by Friedemann Seebass, Germany.
// Dieses Programm ist freie Software. Siehe GNU GPL Notiz in Datei main.cpp.
// This program is free software. See the GNU GPL notice in file main.cpp.
// Projekt: Almetare - Alle meine Taschenrechner
//******************************************************************************
/// \file cfg_dlg.cpp
///       Implementierung der Konfigurations-Dialogbox (Modul base)
//******************************************************************************
// Wann       Wer      Was
// ---------- -------- ---------------------------------------------------------
// 24.01.2005 fse      butSearch(): Korrektur in Zuweisung
// 18.06.2004 fse      butHelp(): Berechnung von helpFile geaendert
// 15.11.2003 fse      accept(): Setzen von mAlwaysOnTop
// 11.11.2003 fse      CfgDlg(): Application-Icon setzen
// 12.05.2003 fse      getDefaultBrowserPath(): Korrektur fuer Windows XP
// 09.11.2002 fse      "About..."-Button hinzugefuegt
// 23.12.2001 fse      erzeugt
//******************************************************************************

#include <qradiobutton.h>
#include <qpushbutton.h>
#include <qcheckbox.h>
#include <qmessagebox.h>
#include <qlabel.h>
#include <qfiledialog.h>
#include <qregexp.h>

#ifdef _WIN32
#include <windows.h> // fuer Auslesen der Registry
#include <process.h> // spawn
#endif

#include "logging.h"
#include "cfg_dlg.h"
#include "about_dlg.h"

//******************************************************************************

/// Konstruktor.
/// \param calcBaseP      Zeiger auf den aufrufenden Taschenrechnerdialog.
/// \param applIconPixmap Pixmap mit dem Application-Icon links oben im Dialog
/// \param parent         Zeiger auf Vaterdialog
/// \param name           Name des Dialogs
/// \param modal          Flag, ob Dialog modal oder nicht
/// \param fl             diverse Qt-Dialog-Flags
//  fse, 02.10.02
//  fse, 11.11.03: Application-Icon des Dialogs setzen
//  fse, 15.11.03: neu: mCfgsP->mAlwaysOnTop

CfgDlg::CfgDlg(CalcDlgBase* calcBaseP,
               QPixmap&     applIconPixmap, // fse, 11.11.03
               QWidget*     parent,
               const char*  name,
               bool         modal,
               WFlags       fl)
    : CfgBaseDlg(parent, name, modal, fl),
      mApplIconPixmap(applIconPixmap)
{
	string browserPath;

	mParentP = calcBaseP;
	mCfgsP   = calcBaseP->getCalcP()->getCfgsP();
	if (!mParentP->getIsPixLoaded())
	{
		mRbMoreRealisticP->setEnabled(false);
		mRbRealisticP    ->setEnabled(false);
	}
	mCfgsP->mZoom = 0;
	switch(mCfgsP->mDisplayMode)
	{
		case MORE_REALISTIC: mRbMoreRealisticP->setChecked(true); break;
		case REALISTIC     : mRbRealisticP    ->setChecked(true); break;
		case SIMPLE        : mRbSimpleP       ->setChecked(true); break;
		case DEBG          : mRbDebugP        ->setChecked(true); break;
		default            : mRbRealisticP    ->setChecked(true);
	}	
	if (mCfgsP->mLogMode     ) mCbLogP        ->setChecked(true);
	if (mCfgsP->mShowToolTips) mCbToolTipP    ->setChecked(true);
	if (mCfgsP->mAlwaysOnTop ) mCbAlwaysOnTopP->setChecked(true);
	mBrowserPath = mCfgsP->mBrowserPath;
	if (mBrowserPath == "default")
		mCbDefaultP->setChecked(true);
	else
		mCbDefaultP->setChecked(false);
	if (!mApplIconPixmap.isNull())
		setIcon(mApplIconPixmap); // kleines Icon links oben im Fenster setzen
	cbEnDisable();
}

//******************************************************************************

/// Destruktor.

CfgDlg::~CfgDlg()
{
    // no need to delete child widgets, Qt does it all for us
}

//******************************************************************************

/// Enabled und disabled den Search-Button und das Browser-Pfad-Eingabefeld in
/// Abhaengigkeit von der Check-Box "default browser".
//  fse, 22.12.01

void CfgDlg::cbEnDisable()
{
	string browserPath;

	if (mCbDefaultP->isChecked())
	{
		mBtnSearchP->setEnabled(false);
		mLeBrowserPathP->setEnabled(false);
		getDefaultBrowserPath(browserPath);
		mLeBrowserPathP->setText(browserPath.c_str());
	}
	else
	{
		mBtnSearchP->setEnabled(true);
		mLeBrowserPathP->setEnabled(true);
		mLeBrowserPathP->setText(mBrowserPath.c_str());
	}
}

//************************************** ***************************************

/// Holt den Pfad zum Default-Browser.
/// Windows: Pfad wird aus der Registry geholt.
/// Linux:   Pfad ist einfach "mozilla". Netscape geht nicht, da man dem Browser
///          anscheinend keine Datei zum Anzeigen mitgeben kann.
//  fse, 23.12.01
//  fse, 12.05.03: Korrektur der Anfuehrungszeichen in Windows XP

bool CfgDlg::getDefaultBrowserPath(string& pathStr)
{
#ifdef _WIN32
	pathStr = "Retrieving default path error!";
	char httpRegPath[] = "http\\shell\\open\\command";
	int ret;
	HKEY hkey;
	unsigned long type, len = 255;
	unsigned char path[256]; // = "D:\PROGRAMME\KOMMUNIKATION\NETSCAPE\COMMUNICATOR\PROGRAM\NETSCAPE.EXE";
	string::size_type slashPos;

	ret = RegOpenKeyEx(HKEY_CLASSES_ROOT, httpRegPath, 0L, KEY_QUERY_VALUE, &hkey);
	if (ret != ERROR_SUCCESS)
	{
		ERR_LOG(hgh) << "Error in opening registry key '" << httpRegPath << "' (Return code: " << ret << "!" << NL;
		return false;
	}
	ret = RegQueryValueEx(hkey, NULL, NULL, &type, path, &len);
	if (ret != ERROR_SUCCESS)
	{
		ERR_LOG(hgh) << "Error in reading registry key '" << httpRegPath << "' (Return code: " << ret << "!" << NL;
		return false;
	}
	ret = RegCloseKey(hkey);
	if (ret != ERROR_SUCCESS)
	{
		ERR_LOG(hgh) << "Error in closing registry key '" << httpRegPath << "' (Return code: " << ret << "!" << NL;
		return false;
	}
	string str((char*)path);
	slashPos = str.find_last_of("/\\"); // falls im Pfad Leerzeichen sind, von hinten nach Slash suchen
	str = str.substr(0, str.find(" ", slashPos)); // evtl. Parameter abschneiden
	// Windows XP schliesst den String in Anfuehrungszeichen ein - die
	// muessen weg fuer das spaetere spawn():
	if (str[0] == '"')
		str = str.substr(1, str.size()-2);
	pathStr = str;
	return true;
#else
	pathStr = "mozilla";
	return true;
#endif
}

//******************************************************************************

/// Slot fuer den Button zur Auswahl des Browsers.
//  fse, 24.01.05: mBrowserPath = s.latin1(); statt mBrowserPath = s;

void CfgDlg::butSearch()
{
	QString s(QFileDialog::getOpenFileName());
	if (!s.isEmpty())
	{
		mBrowserPath = s.latin1();
		mLeBrowserPathP->setText(s);
	}
	return;
}

//******************************************************************************

/// Slot fuer den OK-Button.
//  fse, 15.11.03: Setzen von mCfgsP->mAlwaysOnTop.

void CfgDlg::accept()
{
	if      (mRb100P->isChecked()) mCfgsP->mZoom = 100;
	else if (mRb75P ->isChecked()) mCfgsP->mZoom =  75;
	else if (mRb50P ->isChecked()) mCfgsP->mZoom =  50;
	else if (mRb33P ->isChecked()) mCfgsP->mZoom =  33;

	if      (mRbMoreRealisticP->isChecked()) mCfgsP->mDisplayMode = MORE_REALISTIC;
	else if (mRbRealisticP    ->isChecked()) mCfgsP->mDisplayMode = REALISTIC     ;
	else if (mRbSimpleP       ->isChecked()) mCfgsP->mDisplayMode = SIMPLE        ;
	else if (mRbDebugP        ->isChecked()) mCfgsP->mDisplayMode = DEBG          ;
	
	if (mCbLogP->isChecked())
		mCfgsP->mLogMode = true;
	else
		mCfgsP->mLogMode = false;

	if (mCbToolTipP->isChecked())
		mCfgsP->mShowToolTips = true;
	else
		mCfgsP->mShowToolTips = false;

	if (mCbAlwaysOnTopP->isChecked())
		mCfgsP->mAlwaysOnTop = true;
	else
		mCfgsP->mAlwaysOnTop = false;

	if (mCbDefaultP->isChecked()) 
		mCfgsP->mBrowserPath = "default";
	else 
		mCfgsP->mBrowserPath = mLeBrowserPathP->text().latin1();
	
	done(1);
}

//******************************************************************************

/// Slot fuer den About-Button. Zeigt About-Dialogbox an.
//  fse, 01.11.02

void CfgDlg::butAbout()
{
	char vrs[100];
	char name[100];
	WFlags fl = WStyle_Customize | WStyle_NormalBorder | WStyle_Title | WStyle_SysMenu;

	strcpy(vrs , mParentP->getCalcP()->getProgVers().c_str());
	strcpy(name, mParentP->getCalcP()->getProgName().c_str());
	AboutDlg aboutDlg(name, mApplIconPixmap, vrs, this, 0, true, fl);
	aboutDlg.exec();
}

//******************************************************************************

/// Slot fuer den Hilfe-Button. Zeigt Hilfedatei im HTML-Browser an.
//  fse, 22.12.01
//  fse, 18.06.04: pwd helpFile nur voranstellen, wenn nicht ohnehin schon
//                 absoluter Pfad (sonst klappt Laden des Files nicht, wenn
//                 Aufruf des Rechners unter Linux per Desktop-Icon)

void CfgDlg::butHelp()
{
	string pwd = QDir::currentDirPath().latin1();
	string helpFile = mParentP->getCalcP()->getAccPath() + "/doc/start.html";
//	string helpFile = /* pwd + "/" + */ mParentP->getCalcP()->getAccPath() + "/doc/start.html";
	string browserPath, command;
	int ret = 0;

	if (helpFile[0] != '/' && helpFile[0] != '\\')
		helpFile = "\"" + pwd + "/" + helpFile + "\"";
	browserPath = mLeBrowserPathP->text().latin1();
	#ifdef _WIN32
		for (int i = 0; i < browserPath.size(); ++i)
			if (browserPath[i] == '/') browserPath[i] = '\\';
		for (i = 0; i < helpFile.size(); ++i)
			if (helpFile[i] == '/') helpFile[i] = '\\';
		/*
		browserPath = "\"" + browserPath + "\""; // in Anfz. einschliessen, da Namen Spaces enthalten koennen
		command = browserPath + " " + helpFile + " &";
		ret = system(command.c_str());
		if (ret != 0)
		{
			string msg = "Command<BR>'" + command + "'<BR>failed!" \
						 "<BR>Try to select another HTML browser with the 'Search' button.";
			ERR_LOG(med) << "Command '" << command << "' failed with return code "
			             << ret << " and errno " << errno << "!" << NL;
			ret = QMessageBox::warning(this, "Error starting help browser", msg.c_str(), "&OK");
		}
		*/
		ret = _spawnl(_P_NOWAIT, browserPath.c_str(), "dummy", helpFile.c_str(), NULL );
		if (ret < 0)
		{
			string msg = "Spawning<BR>'" + browserPath + "'<BR>" +
			             "with argument<BR>'" + helpFile + "'<BR>failed!" \
						 "<BR>Try to select another HTML browser with the 'Search' button.";
			ERR_LOG(med) << "Spawning '" << browserPath << " " << helpFile << "' failed!" << NL;
			ret = QMessageBox::warning(this, "Error starting help browser", msg.c_str(), "&OK");
		}
	#else
		command = browserPath + " file:" + helpFile + " &";
		ret = system(command.c_str());
		if (ret != 0)
		{
			string msg = "Command<BR>'" + command + "'<BR>failed!" \
						 "<BR>Try to select another HTML browser with the 'Search' button.";
			ERR_LOG(med) << "Command '" << command << "' failed!" << NL;
			ret = QMessageBox::warning(this, "Error starting help browser", msg.c_str(), "&OK");
		}
	#endif
}

//******************************************************************************
//************************************** ***************************************
//************************* ************************* **************************
//****************** ******************* ******************* *******************
//************** *************** *************** *************** ***************
//************ ************ ************ ************ ************ *************
