//******************************************************************************
// Copyright (c) 2002-2004 by Friedemann Seebass, Germany.
// Dieses Programm ist freie Software. Siehe GNU GPL Notiz in Datei main.cpp.
// This program is free software. See the GNU GPL notice in file main.cpp.
// Projekt: Almetare - Alle meine Taschenrechner
//******************************************************************************
/// \file cfgs.cpp
///       Klasse mit allen Parametern, die in der Konfigurationsdatei stehen
///       (Modul base)
//******************************************************************************
// Wann       Wer      Was
// ---------- -------- ---------------------------------------------------------
// 17.11.2003 fse      strstream durch sstream ersetzt
// 12.11.2003 fse      neuer Parameter alwaysOnTop
// 20.03.2003 fse      dbgLog->gDbgLog, errLog->gErrLog, onlLog->gOnlLog
// 17.03.2003 fse      setDefaults(): mNvRam changed from "|       |" to
//                                                        "|||||||||"
// 01.03.2003 fse      Datei wird auch dann neu geschrieben, wenn Version oder
//                     Programmname nicht stimmt
// 27.11.2001 fse      erzeugt
//******************************************************************************

#include <fstream>
#include <sstream>
#include <iomanip> // setw()
#include "cfgs.h"
#include "logging.h"

//******************************************************************************
// Klasse Cfgs:
//******************************************************************************

/// Konstruktor liest die Konfigurationsdatei ein oder belegt die Parameter mit
/// Default-Werten.
/// \param accPath     Pfad zum Zubehoer-Ordner (obsolet, da Datei jetzt im
///                    Wurzelverzeichnis liegt)
/// \param cfgFileName Name der einzulesenden Konfigurationsdatei
/// \param progName    Name des Programms, das die Datei einliest; dieser Name
///                    wird als Parameter in der Datei erwartet
/// \param progVers    Versionsnummer des Programms, das die Datei einliest;
///                    dieser String wird als Parameter in der Datei erwartet
//  fse, 01.10.02
//  fse, 01.03.03: schreibe Datei auch dann neu, wenn Version oder
//                 Programmname nicht stimmt

Cfgs::Cfgs(string accPath, string cfgFileName, string progName, string progVers)
{
	accPath = accPath; // entfernt Compiler-Warnung "unused parameter"
	                   // (vielleicht braucht man accPath nochmal...)
	setDefaults();
	mCfgFileName = cfgFileName;
	mProgName = progName;
	mProgVers = progVers;
	if (readCfgFile() < 0) (void)writeDefaultCfgFile();
}

//************************************** ***************************************

/// Setzt alle Werte, die in der Konfigurationsdatei geaendert werden koennen,
/// auf Default-Werte.
//  fse, 29.11.01
//  fse, 17.03.03: mNvRam changed from "|       |" to "|||||||||"
//  fse, 15.11.03: mAlwaysOnTop

void Cfgs::setDefaults()
{
	mXpos          = 50;
	mYpos          = 20;
	mWidth         = 0;
	mHeight        = 0;
	mDisplayMode   = MORE_REALISTIC;
	mShowHint      = true;
	mShowToolTips  = true;
	mAlwaysOnTop   = false;
	mLogMode       = false;
	mNvRam         = "|||||||||";
	mBrowserPath   = "default";
	mOnlLogFile    = "log.txt";
	mDbgLogFile    = "log.txt";
	mErrLogFile    = "log.txt";
	mMaxLogLen     = 50;
	mOnlChannels   = gnl | cns;
	mDbgChannels   = 0;
	mErrChannels   = all | cns;
	mExclFileToLog = "";
}

//************************************** ***************************************

/// Liest alle Werte aus der Konfigurationsdatei ein und schreibt sie in die
/// Members.
/// \return  0: OK;
///         -1: Lesen de Konfig.datei gescheitert
///         -2: Programmname stimmt nicht mit Namen in der Datei ueberein
///         -3: Programmname stimmt nicht mit Version in der Datei ueberein
//  fse, 27.11.00
//  fse, 12.11.03: neuer Parameter alwaysOnTop

int Cfgs::readCfgFile()
{
	Config::Err err;
	Config cfg(mCfgFileName, err);
	string par, name, version, displayMode, nvRam, width, height, alwaysOnTop, showHint;
	string showToolTips, logMode, browserPath, onlFile, dbgFile, errFile, fileToLog;
	string maxLogLen;
	string onlChan, dbgChan, errChan;
	
	if (err != Config::OK) return -1;
	(void)readParam(cfg, "name", name, "PROGRAM");
	if (name != mProgName)
	{
		ERR_LOG(hgh) << "INI file for '" << mProgName << "' not found!\n"
		             << "Found '" << name << "' instead. "
		             << "Using defaults." << NL;
		return -2;
	}
	(void)readParam(cfg, "version", version, "PROGRAM");
	if (version != mProgVers)
	{
		ERR_LOG(hgh) << "INI file for version '" << mProgVers
		             << "' expected but INI file for version '" << version << "' found! "
		             << "Deleting old one and using defaults." << NL;
		cfg.remove();
		return -3;
	}
	if (readParam(cfg, "displayMode" , displayMode , "PROGRAM")) mDisplayMode   = strToDisplayMode(displayMode);
	if (readParam(cfg, "nvRam"       , nvRam       , "PROGRAM")) mNvRam         = nvRam;
	if (readParam(cfg, "xPos"        , par         , "PROGRAM")) mXpos          = atoi(par.c_str());
	if (readParam(cfg, "yPos"        , par         , "PROGRAM")) mYpos          = atoi(par.c_str());
	if (readParam(cfg, "width"       , width       , "PROGRAM")) mWidth         = atoi(width.c_str());
	if (readParam(cfg, "height"      , height      , "PROGRAM")) mHeight        = atoi(height.c_str());
	if (readParam(cfg, "alwaysOnTop" , alwaysOnTop , "PROGRAM")) mAlwaysOnTop   = strToBool(alwaysOnTop);
	if (readParam(cfg, "showHint"    , showHint    , "PROGRAM")) mShowHint      = strToBool(showHint);
	if (readParam(cfg, "showToolTips", showToolTips, "PROGRAM")) mShowToolTips  = strToBool(showToolTips);
	if (readParam(cfg, "logMode"     , logMode     , "PROGRAM")) mLogMode       = strToBool(logMode);
	if (readParam(cfg, "browserPath" , browserPath , "PROGRAM")) mBrowserPath   = browserPath;
	if (readParam(cfg, "onlFile"     , onlFile     , "LOGGING")) mOnlLogFile    = onlFile;
	if (readParam(cfg, "dbgFile"     , dbgFile     , "LOGGING")) mDbgLogFile    = dbgFile;
	if (readParam(cfg, "errFile"     , errFile     , "LOGGING")) mErrLogFile    = errFile;
	if (readParam(cfg, "maxLogLen"   , maxLogLen   , "LOGGING")) mMaxLogLen     = atoi(maxLogLen.c_str());
	if (readParam(cfg, "fileToLog"   , fileToLog   , "LOGGING")) mExclFileToLog = fileToLog;
	if (readParam(cfg, "dbgChannels" , dbgChan     , "LOGGING")) mDbgChannels   = Log::strToChannels(dbgChan);
	if (readParam(cfg, "onlChannels" , onlChan     , "LOGGING")) mOnlChannels   = Log::strToChannels(onlChan);
	if (readParam(cfg, "errChannels" , errChan     , "LOGGING")) mErrChannels   = Log::strToChannels(errChan);
	if (mLogMode)
	{	// Achtung: Reihenfolge wichtig! (s. Log::init()):
		(void)gOnlLog.init(onlFile, mOnlChannels, mMaxLogLen, fileToLog, mProgName, mProgVers);
		(void)gErrLog.init(errFile, mErrChannels, mMaxLogLen, fileToLog, mProgName, mProgVers);
		(void)gDbgLog.init(dbgFile, mDbgChannels, mMaxLogLen, fileToLog, mProgName, mProgVers);
	}
	else
	{
		(void)gOnlLog.init("", 0, mMaxLogLen, "", mProgName, mProgVers);
		(void)gErrLog.init("", 0, mMaxLogLen, "", mProgName, mProgVers);
		(void)gDbgLog.init("", 0, mMaxLogLen, "", mProgName, mProgVers);
	}
	return 0;
}

//************************************** ***************************************

/// Lesen der Konfigurationsdatei ging schief!
/// Schreibt Konfigurationsdatei mit den aktuellen Werten (das sind bei der
/// Konstruktion der Klasse die Default-Werte) und Kommentaren.
/// \return  0: OK;
///         -1: Schreiben der Konfigurationsdatei fehlgeschlagen
//  fse, 18.11.01
//  fse, 12.11.03: neuer Parameter alwaysOnTop

int Cfgs::writeDefaultCfgFile()
{
	ofstream cfgFile(mCfgFileName.c_str());
	const int w = 14; // Breite des Wert-Feldes
	
	if (!cfgFile)
	{
		ERR_LOG(hgh) << "Writing default config file '" << mCfgFileName << "' failed!" << NL;
		return -1;
	}
	(void)cfgFile.setf(ios::left, ios::adjustfield);
	cfgFile << "# configuration file for Almetare's " << mProgName  << endl;
	cfgFile << ""                                                   << endl;
	cfgFile << "[PROGRAM]"                                          << endl;
	cfgFile << "	name         = " << setw(w) << mProgName.c_str()  << endl;
	cfgFile << "	version      = " << setw(w) << mProgVers.c_str()  << endl;
	cfgFile << "	nvRam        = " << setw(w) << mNvRam.c_str()                           << "# non volatile RAM - you should not edit this entry" << endl;
	cfgFile << "	displayMode  = " << setw(w) << displayModeToStr(mDisplayMode).c_str()   << "# allowed: 'debug', 'simple', 'realistic' or 'moreRealistic'" << endl;
	cfgFile << "	xPos         = " << setw(w) << mXpos                                    << "# x position of the dialog in pixels" << endl;
	cfgFile << "	yPos         = " << setw(w) << mYpos                                    << "# y position of the dialog in pixels" << endl;
	cfgFile << "	width        = " << setw(w) << mWidth                                   << "# width  of the dialog in pixels (0: default)" << endl;
	cfgFile << "	height       = " << setw(w) << mHeight                                  << "# height of the dialog in pixels (0: default)" << endl;
	cfgFile << "	alwaysOnTop  = " << setw(w) << boolToStr(mAlwaysOnTop).c_str()          << "# allowed: 'on' or 'off'" << endl;
	cfgFile << "	showHint     = " << setw(w) << boolToStr(mShowHint).c_str()             << "# allowed: 'on' or 'off'" << endl;
	cfgFile << "	showToolTips = " << setw(w) << boolToStr(mShowToolTips).c_str()         << "# allowed: 'on' or 'off'" << endl;
	cfgFile << "	logMode      = " << setw(w) << boolToStr(mLogMode).c_str()              << "# allowed: 'on' or 'off'" << endl;
	cfgFile << "	browserPath  = " << setw(w) << mBrowserPath.c_str()                     << "# allowed: 'default' or path to browser" << endl;
	cfgFile << endl;
	cfgFile << "[LOGGING]" << endl;
	cfgFile << "	onlFile      = " << setw(w) << mOnlLogFile.c_str()                      << "# name of file for normal logging" << endl;
	cfgFile << "	dbgFile      = " << setw(w) << mDbgLogFile.c_str()                      << "# name of file for debug logging" << endl;
	cfgFile << "	errFile      = " << setw(w) << mErrLogFile.c_str()                      << "# name of file for error logging" << endl;
	cfgFile << "	maxLogLen    = " << setw(w) << mMaxLogLen                               << "# max. length of files in kByte (if longer: -> *.bak)" << endl;
	cfgFile << "	onlChannels  = " << setw(w) << Log::channelsToStr(mOnlChannels).c_str() << "# allowed: 'ctl' 'mem' 'gnl' 'hgh' 'med' 'low' 'all' 'cns'" << endl;
	cfgFile << "	dbgChannels  = " << setw(w) << Log::channelsToStr(mDbgChannels).c_str() << "# allowed: 'ctl' 'mem' 'gnl' 'hgh' 'med' 'low' 'all' 'cns'" << endl;
	cfgFile << "	errChannels  = " << setw(w) << Log::channelsToStr(mErrChannels).c_str() << "# allowed: 'ctl' 'mem' 'gnl' 'hgh' 'med' 'low' 'all' 'cns'" << endl;
	cfgFile << "	fileToLog    = " << setw(w) << mExclFileToLog.c_str()                   << "# name of the exclusively logged file" << endl;
	cfgFile.close();
	return 0;
}

//************************************** ***************************************

/// Liest Parameter param ein und gibt Fehlermeldung auf ERR_LOG aus, falls es
/// nicht geklappt hat.
/// \param  cfg     Referenz auf Config-Objekt
/// \param  param   String mit Namen des einzulesenden Parmeters
/// \param  section String mit Namen der Sektion, in der der einzulesende
///                 Parameter steht
/// \retval val     String mit dem eingelesenen Wert des Parameters
/// \return         true, wenn das Einlesen geklappt hat, sonst false
// fse, 27.11.01

bool Cfgs::readParam(Config& cfg, const string& param, string& val, const string& section) const
{
	Config::Err err = cfg.readParameter(param, val, section);
	if (err != Config::OK)
	{
		ERR_LOG(hgh) << cfg.errToStr() << NL;
		return false;
	}
	val = val.substr(0, val.find_last_not_of(" ")+1); // trailing spaces abschneiden
	return true;
}

//******************************************************************************

/// Schreibt alle veraenderbaren Membervariablen in die Konfigurationsdatei.
//  fse, 28.11.01
//  fse, 12.11.03: neuer Parameter alwaysOnTop

void Cfgs::updtCfgFile()
{
	Config::Err err;
	Config cfg(mCfgFileName, err);
	string displayMode, logMode;
	
	if (err == Config::OK)
	{
		ostringstream xPos, yPos, width, height;
		xPos   << mXpos  ; // << ends;
		yPos   << mYpos  ; // << ends;
		width  << mWidth ; // << ends;
		height << mHeight; // << ends;
		(void)writeParam(cfg, "nvRam"       , mNvRam                        , "PROGRAM");
		(void)writeParam(cfg, "displayMode" , displayModeToStr(mDisplayMode), "PROGRAM");
		(void)writeParam(cfg, "xPos"        , xPos.str()                    , "PROGRAM");
		(void)writeParam(cfg, "yPos"        , yPos.str()                    , "PROGRAM");
		(void)writeParam(cfg, "width"       , width.str()                   , "PROGRAM");
		(void)writeParam(cfg, "height"      , height.str()                  , "PROGRAM");
		(void)writeParam(cfg, "alwaysOnTop" , boolToStr(mAlwaysOnTop)       , "PROGRAM");
		(void)writeParam(cfg, "showHint"    , boolToStr(mShowHint)          , "PROGRAM");
		(void)writeParam(cfg, "showToolTips", boolToStr(mShowToolTips)      , "PROGRAM");
		(void)writeParam(cfg, "logMode"     , boolToStr(mLogMode)           , "PROGRAM");
		(void)writeParam(cfg, "browserPath" , mBrowserPath                  , "PROGRAM");
	}
	else
	{
		ERR_LOG(hgh) << cfg.errToStr() << NL;
	}
}

//************************************** ***************************************

/// Schreibt Parameter param und gibt Fehlermeldung auf ERR_LOG aus, falls es
/// nicht geklappt hat.
/// \param  cfg     Referenz auf Config-Objekt
/// \param  param   String mit Namen des zu schreibenden Parmeters
/// \param  section String mit Namen der Sektion, in der der zu schreibende
///                 Parameter steht
/// \retval val     String mit dem zu schreibenden Wert des Parameters
/// \return         true, wenn das Schreiben geklappt hat, sonst false
//  fse, 28.11.01

bool Cfgs::writeParam(Config& cfg, const string& param, const string& val, const string& section) const
{
	Config::Err err = cfg.writeParameter(param, val, section);
	if (err != Config::OK)
	{
		ERR_LOG(hgh) << cfg.errToStr() << NL;
		return false;
	}
	return true;
}

//******************************************************************************
// Service-Routinen:
//******************************************************************************

/// Wandelt DisplayMode dm in String um. Wird zum Schreiben des Display-Modes
/// in die Datei benoetigt.
/// \param dm in String zu wandelnder Display-Mode
/// \return   Display-Mode als String
//  fse, 29.11.01

const string Cfgs::displayModeToStr(DisplayMode dm) const
{
	switch(dm)
	{
		case DEBG          : return "debug";
		case SIMPLE        : return "simple";
		case REALISTIC     : return "realistic";
		case MORE_REALISTIC: return "moreRealistic";
		default: return "moreRealistic";
	}
}

//******************************************************************************

/// Wandelt String str in DisplayMode um. Wird zum Lesen des Display-Modes
/// benoetigt.
/// \param str String mit dem Display-Mode
/// \return    DisplayMode, der dem String entspricht
//  fse, 29.11.01

DisplayMode Cfgs::strToDisplayMode(const string& str) const
{
  	if      (str.find("debug"        ) != string::npos) return DEBG;
  	else if (str.find("simple"       ) != string::npos) return SIMPLE;
  	else if (str.find("realistic"    ) != string::npos) return REALISTIC;
  	else if (str.find("moreRealistic") != string::npos) return MORE_REALISTIC;
  	return MORE_REALISTIC;
}

//******************************************************************************

/// Wandelt bool bl in "on" oder "off".
/// \param bl true oder false
/// \return   "on" oder "off"
//  fse, 29.11.01

const string Cfgs::boolToStr(bool bl) const
{
	if (bl) return "on"; else return "off";
}

//******************************************************************************

/// Wandelt String str in bool um.
/// \param str "on" oder "off"
/// \return    true oder false
//  fse, 29.11.01

bool Cfgs::strToBool(const string& str) const
{
	if      (str.find("on" ) != string::npos) return true;
	else if (str.find("off") != string::npos) return false;
	return false;
}

//******************************************************************************

/// Initialisiert die Log-Objekte bzw. stoppt das Logging je nach mLogMode.
//  02.12.01

void Cfgs::startStopLogging()
{
	if (mLogMode)
	{
		(void)gOnlLog.init(mOnlLogFile, mOnlChannels, mMaxLogLen, mExclFileToLog, mProgName, mProgVers);
		(void)gErrLog.init(mErrLogFile, mErrChannels, mMaxLogLen, mExclFileToLog, mProgName, mProgVers);
		(void)gDbgLog.init(mDbgLogFile, mDbgChannels, mMaxLogLen, mExclFileToLog, mProgName, mProgVers);
	}
	else
	{
		gOnlLog.stopLogging();
		gErrLog.stopLogging();
		gDbgLog.stopLogging();
	}
}

//******************************************************************************
// zum Testen:
//******************************************************************************
/*
void test()
{
	Cfgs cfgs;
	
	cfgs.mLogMode = true;
	cfgs.mDisplayMode = SIMPLE;
	cfgs.updtCfgFile();	
}
*/
//******************************************************************************
//************************************** ***************************************
//************************* ************************* **************************
