//******************************************************************************
// Copyright (c) 2002-2004 by Friedemann Seebass, Germany.
// Dieses Programm ist freie Software. Siehe GNU GPL Notiz in Datei main.cpp.
// This program is free software. See the GNU GPL notice in file main.cpp.
// Projekt: Almetare - Alle meine Taschenrechner
//******************************************************************************
/// \file number.cpp
///       Klasse der Zahlen (Kombination aus double und Bruechen) (Modul base)
//******************************************************************************
// Wann       Wer      Was
// ---------- -------- ---------------------------------------------------------
// 17.11.2003 fse      strstream durch sstream ersetzt
// 08.03.2003 fse      neue Operatoren <= und >=
// 21.04.2002 fse      erzeugt
//******************************************************************************

#include <sstream>   // str << ...
#include <math.h>    // fabs()
#include <limits.h>  // numeric_limits<double>epsilon()
#include "util.h"    // Util::doubleToStr()
#include "number.h"

//******************************************************************************
// Klasse Number:
//******************************************************************************

/// Konstruktor wandelt gegebenen String in eine Zahl.
/// Untersucht, ob der eingegebene String als Bruch interpretiert werden kann:
/// - es darf kein "." vorkommen
/// - es darf kein "e" vorkommen
/// - Zahl muss kleiner als 100000 sein, da sonst ein Overflow passieren koennte
/// - wenn "/" gefunden, muessen Zaehler und Nenner kleiner 100000 sein
/// \param sym Eingabe-String, der eine Zahl enthalten muss
///            (Bruch, double oder int)
//  fse, 17.07.02

Number::Number(const string& sym)
{
	const double maxNum = 100000.0; // numr oder dnom mehr als 5 Stellen => kein Bruch
	double val, dWhole, dNumr, dDnom;
	long   numr, dnom;
	string::size_type slashPos    = sym.find("/");
	string::size_type slashPos2nd = sym.find("/", slashPos+1);
	string::size_type pointPos    = sym.find(".");
	string::size_type ePos        = sym.find("e");

	if (slashPos != string::npos)
	{ // "/" enthalten => Bruch, falls Zaehler und Nenner klein genug
		if (slashPos2nd == string::npos)
		{ // normaler Bruch ohne ganzzahligen Anteil:
			dNumr = atof(sym.substr(0, slashPos).c_str());
			dDnom = atof(sym.substr(slashPos+1 ).c_str());
		}
		else
		{ // gemischter Bruch:
			dWhole = atof(sym.substr(0, slashPos).c_str());
			dNumr = atof(sym.substr(slashPos+1, slashPos2nd-slashPos).c_str());
			dDnom = atof(sym.substr(slashPos2nd+1 ).c_str());
			dNumr += dWhole*dDnom;
		}
		if (fabs(dNumr) >= maxNum || fabs(dDnom) >= maxNum || pointPos != string::npos)
		{ // doch kein Bruch:
			mVal = dNumr/dDnom;
			mFrc = 0;
			mIsFrc = false;
			return;
		}
		else
		{ // tatsaechlich Bruch:
			numr = static_cast<long>(dNumr);
			dnom = static_cast<long>(dDnom);
			mVal = 0.0;
			mFrc = Fraction(numr, dnom);
			mIsFrc = true;
			return;
		}
	}
	else
	{ // zwar kein "/", aber vielleicht doch als Bruch darstellbar:
		val = atof(sym.c_str());
		if (fabs(val) >= maxNum || pointPos != string::npos || ePos != string::npos)
		{ // zu gross oder "." enthalten => kein Bruch:
			mVal = val;
			mFrc = 0;
			mIsFrc = false;
			return;
		}
		else
		{ // klein genug => Bruch mit Nenner = 1:
			numr = static_cast<long>(val);
			dnom = 1;
			mVal = 0.0;
			mFrc = Fraction(numr, dnom);
			mIsFrc = true;
			return;
		}
	}
}

//******************************************************************************

/// Konvertiert die Zahl *this von einem Bruch in einen Double.
//  fse, 05.07.02

Number Number::frcToDbl()
{
	if (mIsFrc)
	{
		mVal = mFrc.getVal();
		mIsFrc = false;
	}
	return *this;
}

//******************************************************************************

/// Konvertiert die Zahl *this von einem Double in einen Bruch.
//  fse, 05.07.02

Number Number::dblToFrc()
{
	if (!mIsFrc && mVal == (int)mVal)
	{
		mFrc = (int)mVal;

		mIsFrc = true;
	}
	return *this;
}

//******************************************************************************

/// Wandelt Number in String um. Falls kein Bruch, wird gewuenschtes Format und
/// Stellenanzahl beruecksichtigt, inclusive Verschiebung des Dezimalpunkts der
/// Mantisse, falls Zahl im wissenschaftlichen Format.
/// \param format steuert Darstellungsart der Ausgabe, darf folgende Werte haben:
///               - format = 0: NRM - Standardformat
///               - format = 1: SCI - wissenschaftliches Format
///               - format = 2: FIX - fixes Format
///               - format = 3: FRC - Bruchdarstellung
///               - format = 4: MIX - Bruchdarstellung als gemischter Bruch
///               - format = 5: SXG - Sexagesimaldarstellung
/// \param digits   Anzahl der Nachkommastellen (fix) oder der Nachkommastellen der
///                 Mantisse (sci)
/// \param shiftNum Anzahl der Stellen, um die das Komma der Mantisse einer Zahl
///                 im wissenschaftlichen Format nach rechts (>0) oder links (<0)
///                 verschoben wird.
/// \param forceSci Extra fuer den Casio fx-85v: er hat zwei Modi, wie kleine
///                 Zahlen dargestellt werden: Ab einer bestimmtem "Kleinheit"
///                 (<0.01 oder <0.0000001) wird im Standardformat zwangsweise
///                 in die SCI-Darstellung gewechselt. Diese Grenze kann man in
///                 forceSci uebergeben.
/// \return         Ergebnis-String
//  fse, 22.04.02


string Number::makeStr(int format, int digits, int shiftNum, double forceSci) const
{
	string str;

	if ((format == 3 || format == 4) && !mIsFrc) format = 0;
	switch (format)
	{
		case 0:
		case 1:
		case 2: str = Util::doubleToStr(getVal(), format, digits, shiftNum, forceSci); break;
		case 3: str = mFrc.getFraction(); break;
		case 4: str = mFrc.getMixedFraction(); break;
		case 5: str = Util::doubleToSxgStr(getVal()); break;
		default: str = Util::doubleToStr(mVal, 0, digits, shiftNum, forceSci);
	}
	return str;
}

//******************************************************************************

/// Ueberladener Zuweisungsoperator.

Number& Number::operator= (const Number& n)
{
	if (&n == this/* || n == *this*/) return *this;
	if (n.mIsFrc)
	{
		mFrc = n.mFrc;
		mIsFrc = true;
	}
	else
	{
		mVal = n.mVal;
		mIsFrc = false;
	}
	return *this;
}

//******************************************************************************

/// Multiplikation und gleichzeitig Zuweisung.

Number& Number::operator*=(const Number& n)
{
	if (calcBaseOp(operator*, n)) return *this;
	mVal *= n.getVal();
	return *this;
}

//************************************** ***************************************

/// Allgemeine Funktion zur Berechnung einer Operation (Grundrechenart) mit dem
/// Objekt *this und einer weiteren Zahl n.
/// Es wird versucht, die Operation *opP mit Bruechen durchzufuehren. Falls das
/// nicht geht, werden beide Operanden in double gewandelt.
/// \param opP Zeiger auf den auszufuehrenden Operator
/// \param n   zweiter Operand neben *this
//  Integer-Arithmetik verursacht in MSVC Absturz bei /0, Float-Arithmetik nicht;
//  daher wird auch in double gerechnet, wenn Operation "/" und der Divisor 0.
//  Als Alternative koennte man Ausnahmebehandlung einfuehren.
//  fse, 11.05.02

bool Number::calcBaseOp(FrcOpFuncp opP, Number n)
{
	// Groessenbegrenzung, weil sonst evtl. Integer-Ueberlauf:
	if (  getVal() > 10000.0 ||   isFrcOverflow()
	 || n.getVal() > 10000.0 || n.isFrcOverflow())
	{
		(void)frcToDbl();
		(void)n.frcToDbl();
	}

	if (mIsFrc || n.isFrc())
	{ // einer der Operanden ist Bruch; falls moeglich, anderen in Bruch wandeln:
		if (mIsFrc) (void)n.dblToFrc(); else (void)dblToFrc();
	}
	if (mIsFrc && n.isFrc()
	   && !(opP == (FrcOpFuncp)operator/ && n.getVal() == 0.0))
	{
		mFrc = (*opP)(mFrc, n.mFrc);
		return true;
	}
	(void)frcToDbl();
	(void)n.frcToDbl();
	return false;
}

//******************************************************************************

/// Division und gleichzeitig Zuweisung.

Number& Number::operator/=(const Number& n)
{
	if (calcBaseOp(operator/, n)) return *this;
	mVal /= n.getVal();
	return *this;
}

//******************************************************************************

/// Addition und gleichzeitig Zuweisung.

Number& Number::operator+=(const Number& n)
{
	if (calcBaseOp(operator+, n)) return *this;
	mVal += n.getVal();
	return *this;
}

//******************************************************************************


/// Subtraktion und gleichzeitig Zuweisung.

Number& Number::operator-=(const Number& n)
{
	if (calcBaseOp(operator-, n)) return *this;
	mVal -= n.getVal();
	return *this;
}

//******************************************************************************
//******************************************************************************

/// Vergleichsoperator.

bool operator==(const Number& n1, const Number& n2)
{
	if (n1.mIsFrc && n2.mIsFrc)
		return n1.mFrc == n2.mFrc;
	else
	{
		double v1 = n1.getVal();
		double v2 = n2.getVal();
		return fabs(fabs(v1) - fabs(v2)) < 1e-10; // !!! numeric_limits<double>::epsilon();
	}
}

//******************************************************************************

/// Pruefung auf Ungleichheit.

bool operator!=(const Number& n1, const Number& n2)
{
	if (n1.mIsFrc && n2.mIsFrc)
		return n1.mFrc != n2.mFrc;
	else
	{
		double v1 = n1.getVal();
		double v2 = n2.getVal();
		#ifdef _MSC_VER
			string strng;              // fuer MSVC sind die Zahlen -1.#IND und z.B.
			ostringstream str;         // -2 gleich!? Dafuer -1.#IND ungleich -1#IND!?
			str << v1 << v2; //<<ends; // => wenn '#', per Def. ungleich, da
			strng = str.str();         // in DisplaySci::makeValue() so benoetigt.
			if (strng.find('#') != string::npos) return true; 
		#endif
		return v1 != v2;
	}
}

//******************************************************************************

/// Operator "kleiner als".

bool operator< (const Number& n1, const Number& n2)
{
	if (n1.mIsFrc && n2.mIsFrc)
		return n1.mFrc < n2.mFrc;
	else
	{
		double v1 = n1.getVal();
		double v2 = n2.getVal();
		return v1 < v2;
	}
}

//******************************************************************************

/// Operator "kleiner als oder gleich".

bool operator<=(const Number& n1, const Number& n2)
{
	if (n1.mIsFrc && n2.mIsFrc)
		return n1.mFrc <= n2.mFrc;
	else
	{
		double v1 = n1.getVal();
		double v2 = n2.getVal();
		return v1 <= v2;
	}
}

//******************************************************************************

/// Operator "groesser als".

bool operator> (const Number& n1, const Number& n2)
{
	if (n1.mIsFrc && n2.mIsFrc)
		return n1.mFrc > n2.mFrc;
	else
	{
		double v1 = n1.getVal();
		double v2 = n2.getVal();
		return v1 > v2;
	}
}

//******************************************************************************

/// Operator "groesser als oder gleich".

bool operator>=(const Number& n1, const Number& n2)
{
	if (n1.mIsFrc && n2.mIsFrc)
		return n1.mFrc >= n2.mFrc;
	else
	{
		double v1 = n1.getVal();
		double v2 = n2.getVal();
		return v1 >= v2;
	}
}

//******************************************************************************

/// Ausgabeoperator.

ostream& operator<<(ostream& os, const Number& n)
{
	if (n.mIsFrc)
		os << n.mFrc;

	else
		os << n.mVal;
	return os;
}

//******************************************************************************
//************************************** ***************************************
//************************* ************************* **************************
//****************** ******************* ******************* *******************
//************** *************** *************** *************** ***************
//************ ************ ************ ************ ************ *************
