//******************************************************************************
// Copyright (c) 2002-2004 by Friedemann Seebass, Germany.
// Dieses Programm ist freie Software. Siehe GNU GPL Notiz in Datei main.cpp.
// This program is free software. See the GNU GPL notice in file main.cpp.
// Projekt: Almetare - Alle meine Taschenrechner
//******************************************************************************
/// \file symbols.cpp
///       Symbole, die alle Taschenrechner haben (Modul base)
//******************************************************************************
// Wann       Wer      Was
// ---------- -------- ---------------------------------------------------------
// 11.11.2003 fse      SymOtherFunction::process(): lastInP auf != NULL pruefen
// 10.08.2003 fse      NumDmy::process(): Aenderung in Ausnahmebehandlung
// 02.08.2003 fse      SymOperator::processCommon(): Kontrolle auf leeren
//                     Zahlen-Stack.
//                     SymOtherFunction::process() u. SymTrigFunction::process():
//                     Kontrolle, ob zuvor Operator eingegeben wurde.
// 01.05.2003 fse      processCommon(): Kontrolle, ob INV-Taste gedrueckt.
// 23.03.2002 fse      erzeugt
//******************************************************************************

#include <iostream>
#include "calculator.h"
#include "stack.cpp" // GNU kennt Schluesselwort "export" nicht

//******************************************************************************
// Klasse Symbol und Unterklassen:
//******************************************************************************

/// Ausgabeoperator fuer ein Symbol.
/// \param os Output-Stream
/// \param s  auszugebendes Symbol
/// \return   Output-Stream-Referenz
//  fse, 26.09.01

ostream& operator<<(ostream& os, const Symbol& s)
{
	os << s.mSym;
	return os;
}

//******************************************************************************

/// Ausgabeoperator fuer einen Symbolpointer.
/// \param os Output-Stream
/// \param s  auszugebender Symbolpointer
/// \return   Output-Stream-Referenz
//  fse, 26.09.01

ostream& operator<<(ostream& os, const Symbol* s)
{
	os << (*s).mSym;
	return os;
}

//******************************************************************************
// Klasse Symbol:
//******************************************************************************

/// Liefert das letzte eingegebene Symbol zurueck. Dieses ist in Klasse
/// ComputerBase gespeichert.
/// \return Zeiger auf das zuletzt eingegebene Symbol
//  Die Funktion kann nicht inline definiert werden, weil in der Header-Datei
//  symbols.h sonst auch calculator.h inkludiert werden msste. calculator.h
//  inkludiert aber symbols.h...
//  fse, 08.04.02

Symbol* Symbol::getLastInP()
{
	return mCompP->getLastInP();
}

//******************************************************************************

/// Liefert das letzte eingegebene Operator-Symbol zurueck. Dieses ist in 
/// Klasse ComputerBase gespeichert.
/// \return Zeiger auf das zuletzt eingegebene Symbol
//  Die Funktion kann nicht inline definiert werden, weil in der Header-Datei
//  symbols.h sonst auch calculator.h inkludiert werden msste. calculator.h
//  inkludiert aber symbols.h...
//  fse, 21.06.14

Symbol* Symbol::getLastOpP()
{
	return mCompP->getLastOpP();
}

//******************************************************************************

/// Liefert das vorletzte eingegebene Operator-Symbol zurueck. Dieses ist in 
/// Klasse ComputerBase gespeichert.
/// \return Zeiger auf das zuletzt eingegebene Symbol
//  Die Funktion kann nicht inline definiert werden, weil in der Header-Datei
//  symbols.h sonst auch calculator.h inkludiert werden msste. calculator.h
//  inkludiert aber symbols.h...
//  fse, 20.06.14
//  fse, 22.06.14: Funktion ist doch ueberfluessig und wird aktuell nicht 
//                 benutzt.

Symbol* Symbol::getLastButOneOpP()
{
	return mCompP->getLastButOneOpP();
}

//******************************************************************************
// Klasse SymNumber und Kinder:
//******************************************************************************

/// Holt die vorherige Eingabe wieder vom Zahlenstack, wenn es kein Operator war.
//  fse, 07.10.01

void SymNumber::processCommon() const
{
	mCompP->popNumDepOnLastSym();
}

//******************************************************************************

/// Eine eingegebene Zahl wird prozessiert, indem sie einfach auf den Stack
/// gelegt wird.
/// War die letzte Eingabe auch schon eine Zahl, wird sie durch die neue
/// ersetzt.
/// Ausserdem muss untersucht werden, ob es sich um einen Bruch handelt, indem
/// nach dem Zeichen "/" gesucht wird.
//  fse, 07.10.01
//  fse, 10.08.03: Das ganze If-Else-Konstrukt in den Try-Block eingeschlossen,
//                 weil auch isFraction()-Funktion eine Ausnahme
//                 werfen kann (z.B. bei Eingabe von "3" "abc" "sin").

void NumDmy::process()
{
	Number val;
	double dVal, zero = 1.0;
	long   numr, dnom;

	try
	{
		if (isFraction(mSym, dVal, numr, dnom))
			val = Fraction(numr, dnom);
		else
			val = dVal;
	}
	catch (ZeroDivide)
	{
		zero = log10(zero); // Float-Fehler div by 0 provozieren, aber Compiler-
		val = 1.0/zero;     // Warnung "division: function singularity" vermeiden
	}
	processCommon();
	mCompP->pushNum(val);
}

//************************************** ***************************************

/// Untersucht, ob der eingegebene String als Bruch interpretiert werden kann.
/// - es darf kein "e" vorkommen
/// - es darf kein "." vorkommen
/// - Zahl muss kleiner als 100000 sein, da sonst ein Overflow passieren koennte.
/// - wenn "/" gefunden, muessen Zaehler und Nenner kleiner 100000 sein.
/// \param  sym  die eingegebene Zahl als String
/// \retval val  der Wert der Zahl
/// \retval numr falls Bruch: Zaehler, sonst 0
/// \retval dnom falls Bruch: Nenner, sonst 1
/// \return      true, wenn Zahl als Bruch dargestellt werden kann, sonst false
//  fse, 25.04.02
//  fse, 30.09.02: Umgestellt auf Konstruktor Number(string). Mal sehen, ob's
//                 funktioniert. !!!

bool NumDmy::isFraction(const string& sym, double& val, long& numr, long& dnom)
{
	Number num(sym);
	Fraction frc = num.frc();

	numr = frc.getNumr();
	dnom = frc.getDnom();
	val = num.getVal();
	return num.isFrc();

	/*
	const double maxNum = 100000.0; // numr oder dnom mehr als 5 Stellen => kein Bruch
	double dWhole, dNumr, dDnom;
	string::size_type slashPos    = sym.find("/");
	string::size_type slashPos2nd = sym.find("/", slashPos+1);
	string::size_type pointPos    = sym.find(".");
	string::size_type ePos        = sym.find("e");

	if (slashPos != string::npos)
	{ // "/" enthalten => Bruch, falls Zaehler und Nenner klein genug
		if (slashPos2nd == string::npos)
		{ // normaler Bruch ohne ganzzahligen Anteil:
			dNumr = atof(sym.substr(0, slashPos).c_str());
			dDnom = atof(sym.substr(slashPos+1 ).c_str());
		}
		else
		{ // gemischter Bruch:
			dWhole = atof(sym.substr(0, slashPos).c_str());
			dNumr = atof(sym.substr(slashPos+1, slashPos2nd-slashPos).c_str());
			dDnom = atof(sym.substr(slashPos2nd+1 ).c_str());
			dNumr += dWhole*dDnom;
		}
		if (fabs(dNumr) >= maxNum || fabs(dDnom) >= maxNum || pointPos != string::npos)
		{ // doch kein Bruch:
			val = dNumr/dDnom;
			return false;
		}
		else
		{ // tatsaechlich Bruch:
			numr = static_cast<long>(dNumr);
			dnom = static_cast<long>(dDnom);
			return true;
		}
	}
	else
	{ // zwar kein "/", aber vielleicht doch als Bruch darstellbar:
		val = atof(sym.c_str());
		if (fabs(val) >= maxNum || pointPos != string::npos || ePos != string::npos)
		{ // zu gross oder "." enthalten => kein Bruch:
			return false;
		}
		else
		{ // klein genug => Bruch mit Nenner = 1:
			numr = static_cast<long>(val);
			dnom = 1;
			return true;

		}
	}
	*/
}

//******************************************************************************
// Klasse SymOperator:
//******************************************************************************

/// Prozessiert einen eingegebenen Operator.
/// Wenn der Symbolstack leer ist, wird der Operator einfach auf den Stack
/// gelegt. Wenn nicht, muss erst ueberprueft werden, ob der vorherige Operator
/// hoehere oder gleiche Prioritaet hat; falls ja, wird diese Operation zuerst
/// ausgefuehrt; daraufhin ist ggf. ein neuer Operator oben, mit dem der
/// Vergleich rekursiv erneut durchgefuehrt wird.
//  fse, 30.09.01
//  fse, 01.05.03: Pruefung, ob letzte Taste INV war; wenn ja, steht der letzte
//                 Operator an vorletzter Stelle (gilt fuer x^y, x^(1/y)).
//                 Ausserdem ist aufgefallen:
//                 Aufruf von checkConstMode() in der Rekursion, schaltet
//                 einen gerade eingestellten Konstantenmodus gleich wieder ab.
//                 Dies koennte durch Einfuehren einer statischen Variablen
//                 beobachtet und verhindert werden. Allerdings kann man die
//                 Casio-Rechner auch nicht in den Konstantenmodus schalten,
//                 wenn man z.B. "2+3**" eingibt; nur "3**" funktioniert.
//  fse, 02.08.03: Pruefung, ob der Zahlen-Stack leer ist. Wenn ja, wurde
//                 direkt nach dem Einschalten oder nach "AC" ein Operator
//                 eingegeben und in der Anzeige steht "0", der Stack ist
//                 aber noch leer. Daher wird die 0 auf den Stack gelegt.

void SymOperator::processCommon()
{
	SymOperator* lastSymP = mCompP->topSym();
	Symbol* lastInP       = getLastInP();
	Symbol* lastButOneInP = mCompP->getLastButOneInP(); // fse, 01.05.03

	if (lastInP && lastInP->getSym() == "inv") // falls letzte Taste INV, dann
		lastInP = lastButOneInP;// steht der letzte Operator an vorletzter Stelle
	if (dynamic_cast<SymOperator*>(lastInP) || mCompP->getConstOpP())
		(void)mCompP->popSym(); // falls letzte Eingabe auch schon Operator => entfernen
	(void)mCompP->checkConstMode(this);
	if (mCompP->isSymEmpty())
	{
		mCompP->pushSym(this);
		if (mCompP->isNumEmpty()) // fse, 02.08.03
			mCompP->pushNum(0L);   //
	}
	else
	{
		lastInP = mCompP->topSym();
		if (mPrio <= lastSymP->getPrio())
		{
			(void)mCompP->compute();
			(void)processCommon();
		}
		else
		{
			mCompP->pushSym(this);
		}
	}
}

//******************************************************************************

/// Prozessiert "+" und "-".
/// Hier ist ein Sonderfall implementiert: Wenn unmittelbar vor Eingabe des
/// Operators eine Prozentrechnung ausgefuehrt wurde, wird sofort ein Aufschlag
/// oder Abschlag berechnet.
/// Beispiel: "100*3%+" ergibt sofort 103, ohne dass "=" gedrueckt werden
/// muesste. Nach dem "%" steht im Zahlen-Stack nur die 3. In AcPrc::process()
/// wurde in mPrcWhole die 100 gespeichert. Diese beiden Operanden werden in der
/// richtigen Reihenfolge auf den Stack gebracht, "+" oder "-" auf den Symbol-
/// Stack und OthEqu::process() ausgefuehrt.
//  fse, 05.10.01
//  fse, 22.06.14: Der Sonderfall darf nur ausgefuehrt werden, wenn der letzte
//                 Operator "*" oder "/" war ("%" ist kein SymOperator, sondern 
//                 eine SymAction), weil sonst nach einer "normalen" Betaetigung
//                 der Prozenttaste nicht mehr weitergerechnet werden kann. S.
//                 E-Mail von Kelly vom 19.6.2014.
//  fse, 24.06.14: Den Sonderfall herausgenommen und in neue Funktion
//                 processAddSubCasioSci() verlagert; der TI-30 ruft jetzt
//                 diese Funktion hier auf, die Casios die andere, indem
//                 sie eigene Symbole in ihrer jeweiligen Symboltabelle haben.

void SymOperator::processAddSub()
{
//	Symbol* lastInP = getLastInP();
//	Symbol* lastOpP = getLastOpP();
//	OthEqu oe;
//	Number val;
	
//	if (dynamic_cast<AcPrc*>(lastInP) && (dynamic_cast<OpMul*>(lastOpP) || dynamic_cast<OpDiv*>(lastOpP)))
//	{
//		val = mCompP->popNum();
//		mCompP->pushNum(mCompP->getPrcWhole());
//		mCompP->pushNum(val);
//		mCompP->pushSym(this);
//		(void)oe.process();
//	}
//	else
//	{
		(void)processCommon();
//	}
}

//******************************************************************************

/// Prozessiert "+" und "-".
/// Hier ist ein Sonderfall implementiert: Wenn unmittelbar vor Eingabe des
/// Operators eine Prozentrechnung ausgefuehrt wurde, wird sofort ein Aufschlag
/// oder Abschlag berechnet.
/// Beispiel: "100*3%+" ergibt sofort 103, ohne dass "=" gedrueckt werden
/// muesste. Nach dem "%" steht im Zahlen-Stack nur die 3. In AcPrc::process()
/// wurde in mPrcWhole die 100 gespeichert. Diese beiden Operanden werden in der
/// richtigen Reihenfolge auf den Stack gebracht, "+" oder "-" auf den Symbol-
/// Stack und OthEqu::process() ausgefuehrt.
//  fse, 05.10.01
//  fse, 22.06.14: Der Sonderfall darf nur ausgefuehrt werden, wenn der letzte
//                 Operator "*" oder "/" war ("%" ist kein SymOperator, sondern 
//                 eine SymAction), weil sonst nach einer "normalen" Betaetigung
//                 der Prozenttaste nicht mehr weitergerechnet werden kann. S.
//                 E-Mail von Kelly vom 19.6.2014.

void SymOperator::processAddSubCasioSci()
{
	Symbol* lastInP = getLastInP();
	Symbol* lastOpP = getLastOpP();
	OthEqu oe;
	Number val;
	static bool isAdditionDone = false;
	
	if (dynamic_cast<AcPrc*>(lastInP) && (dynamic_cast<OpMul*>(lastOpP) || dynamic_cast<OpDiv*>(lastOpP)))
	{
		val = mCompP->popNum();
		mCompP->pushNum(mCompP->getPrcWhole());
		mCompP->pushNum(val);
		mCompP->pushSym(this);
		(void)oe.process();
		isAdditionDone = true; // merken, dass gerade Proezentrechnung gemacht wurde
	}
	else
	{
		if (isAdditionDone)           // bei letzter Operation wurde Prozentrechnung gemacht => verhindern, dass bei 
			mCompP->setLastInP(NULL); // der zweiten Eingabe desselben Operators die Konstantenrechnung aktiviert wird
		(void)processCommon();
		isAdditionDone = false;
	}
}

//******************************************************************************
// Klasse SymOther und Kinder:
//******************************************************************************

/// Berechnet die letzte Operation auf dem Stack. Ruft sich rekursiv selbst auf,
/// bis der Stapel leer ist.
//  fse, 07.10.01

void OthEqu::process()
{
	if (mCompP->isSymEmpty()) return;
	(void)mCompP->compute();
	process();
}

//******************************************************************************

/// Prozessiert Symbol "(".
/// Neue Stacks werden auf die Stacklisten gelegt, auf den neuen Zahlen-Stack
/// wird vorsorglich eine 0.0 gelegt.
/// Falls die letzte Eingabe auch schon "(" war, muss diese 0.0 zuvor wieder
/// entfernt werden.
//  fse, 07.10.01
//  fse, 16.11.02: Entfernen der Null nicht noetig und schaedlich (fuer die
//                 Optik) beim TI-30.

void OthBra::process()
{
	Symbol* lastInP = getLastInP();
	
//	if (dynamic_cast<OthBra*>(lastInP)) (void)mCompP->popNum(); // falls letzte Eingabe "(" => 0.0 entfernen
	if (dynamic_cast<NumDmy*>(lastInP)) (void)mCompP->popNum(); // falls letzte Eingabe Zahl => entfernen
	mCompP->incBraLev();
	mCompP->handleDisplayAfterBra();
}

//******************************************************************************

/// Prozessiert Symbol ")".
/// Die obersten Stacks auf den Stack-Listen werden abgearbeitet, die obersten
/// Stacks von den Listen entfernt und das Ergebnis auf den jetzt oberen
/// Zahlen-Stack gelegt.
//  fse, 07.10.01

void OthKet::process()
{
	Number val = 0.0;
	OthEqu othEqu;

	othEqu.process();
	if (mCompP->getBraLev() <= 0) return;
	val = mCompP->topNum();
	mCompP->decBraLev();
	mCompP->popStacks();
	mCompP->pushNum(val);
}

//******************************************************************************

/// Prozessiert alle Nicht-Winkelfunktionen.
/// Eine Funktion wird prozessiert, indem das oberste Element vom Zahlen-Stack
/// geholt wird und mit dieser als Argument die ueberladene Funktion calcVal()
/// ausgefuehrt wird. Das Ergebnis wird wieder auf dem Zahlen-Stack abgelegt.
//  fse, 02.08.03: Wenn unmittelbar vor der Funktion ein Operator eingegeben
//                 wurde, darf die Zahl vom Stack nicht gepopped, sondern nur
//                 getopped werden. Denn: Sowohl Casio als auch TI liefern bei
//                 unvollstaendigen Rechnungen wie
//                 5 * ln = 8.047... = 5 * ln(5)
//                 Ohne diese Neuerung wuerde
//                 5 * ln = 2.59...  = ln(5)*ln(5) berechnet werden, da die 5
//                 vom Stack genommen, ln(5) = 1.609... berechnet und dies auf
//                 den Stack gelegt werden wuerde. Bei der Prozessierung der
//                 Multiplikation fehlte dann ein Operand und die 1.609...
//                 wuerde gemaess handleEmptyNumStack() dupliziert werden.
//  fse, 04.08.03: Korrektur, falls STO oder INV das letzte Symbol.
//  fse, 11.11.03: lastInP auf != NULL pruefen

void SymOtherFunction::process()
{
// alt:
//	Number arg = mCompP->popNum();
//	mCompP->pushNum(calcVal(arg));
// neu:
	Number arg;
	Symbol* lastInP = getLastInP();

	if (lastInP && lastInP->getSym() == "inv") // Abfrage mit dynamic_cast<AcInv> geht
		lastInP = mCompP->getLastButOneInP();   // hier nicht, da AcInv erst in calc_sci.h
	if (dynamic_cast<SymOperator*>(lastInP) ||
	    dynamic_cast<AcSto*      >(lastInP))
		arg = mCompP->topNum();
	else
		arg = mCompP->popNum();
	mCompP->pushNum(calcVal(arg));
}

//******************************************************************************
// Klasse SymAction und Kinder:
//******************************************************************************

/// Zeigt alle Symbole auf der Konsole.
//  fse, 17.03.02

void AcHelp::process()
{
	mCompP->printSymHelp();
}

//******************************************************************************

/// Ersetzt die letzte Eingabe durch 0 (Taste "C" - clear).
//  fse, 30.06.02

void AcC::process()
{
	NumDmy num;

	num.process();
}

//******************************************************************************

/// Initialisiert den Rechner, ohne die Speicher zu loeschen (Taste "AC" - all
/// clear).
//  fse, 30.06.02

void AcAc::process()

{
	AcInitWoMem acInitWoMem;

	acInitWoMem.process();
}

//******************************************************************************

/// Prozessiert das Gleichheitszeichen.
/// Dazu muss die letzte Operation ausgefuehrt werden. Das geschieht hier durch
/// Prozessierung von OthEqu, das auch bei OthKet benutzt wird.
//  fse, 01.10.01

void AcEqu::process()
{
	Number val, topButOneNum;
	OthEqu othEqu;
	OthKet othKet;
	
	while (mCompP->getBraLev() > 0)
		othKet.process(); // erst alle Klammern zu machen
	if (mCompP->getConstOpP())
		mCompP->pushNum(mCompP->getConstVal());
	othEqu.process();
	if (!mCompP->isNumEmpty())
		val = mCompP->popNum();
	topButOneNum = mCompP->getTopButOneNum(); // merken fuer y-Register (Casio)
	mCompP->initStacks();
	mCompP->setTopButOneNum(topButOneNum);
	mCompP->pushNum(val);
	if (mCompP->getConstOpP())
		mCompP->pushSym(mCompP->getConstOpP());
}

//******************************************************************************

/// Prozessiert das Prozentzeichen. Prozentrechnung wird von jedem Rechner
/// speziell behandelt, daher ist die wesentliche Funktion virtuell in den
/// Klassen Computer... untergebracht.
//  fse, 14.04.02

void AcPrc::process()
{
	mCompP->handlePercent();
}

//******************************************************************************

/// Initialisiert die Stacks und Members des Computers.
//  fse, 12.10.01

void AcInit::process()
{
	mCompP->initStacks();
	mCompP->initMembers();
}

//******************************************************************************

/// Initialisiert die Stacks und Members des Computers, loescht aber nicht den
/// Memory-Speicher.
//  fse, 24.12.01

void AcInitWoMem::process()
{
	mCompP->initStacks();
	mCompP->initMembers(false);
	mCompP->setConstOpP(NULL);
	mCompP->setConstVal(0);
	mCompP->setIsToInit(true); // fse, 22.04.18: Flag wird von ComputerFx3600p::execPrg() ausgewertet
	mDsplP->setValue("0");     // fse, 20.04.18: sonst bekommt man einen Error waehrend Programmausfuehrung nicht mehr weg
}

//******************************************************************************

/// Vorzeichenwechsel.
//  fse, 12.10.01

void AcSign::process()
{
	mCompP->pushNum(0.0-mCompP->popNum());
}

//******************************************************************************

/// Speichern im Memory-Register.
//  fse, 24.03.02

void AcSto::process()
{
	mCompP->setMemory(mCompP->topNum());
}

//******************************************************************************

/// Ersetzen von oberstem Wert im Zahlen-Stack durch Memory-Register, falls
/// zuletzt Zahl eingegeben; sonst push() auf Zahlen-Stack.
//  fse, 24.03.02

void AcRcl::process()
{
	mCompP->popNumDepOnLastSym();
	mCompP->pushNum(mCompP->getMemory());
}

//******************************************************************************

/// Addieren von aktueller Zahl zum Memory-Register.
//  fse, 24.03.02

void AcSum::process()
{
	mCompP->setMemory(mCompP->getMemory() + mCompP->topNum());
}

//******************************************************************************

/// Subtrahieren von aktueller Zahl vom Memory-Register.
//  fse, 24.03.02

void AcSub::process()
{
	mCompP->setMemory(mCompP->getMemory() - mCompP->topNum());
}

//******************************************************************************

/// Austausch von Memory-Register und oberstem Wert im Zahlen-Stack.
//  fse, 24.03.02

void AcExc::process()
{
	Number memory = mCompP->getMemory();
	
	mCompP->setMemory(mCompP->topNum());
	mCompP->popNumDepOnLastSym();
	mCompP->pushNum(memory);

}

//******************************************************************************
//************************************** ***************************************
//************************* ************************* **************************
//****************** ******************* ******************* *******************
//************** *************** *************** *************** ***************
//************ ************ ************ ************ ************ *************
