//******************************************************************************
// Copyright (c) 2002-2004 by Friedemann Seebass, Germany.
// Dieses Programm ist freie Software. Siehe GNU GPL Notiz in Datei main.cpp.
// This program is free software. See the GNU GPL notice in file main.cpp.
// Projekt: Almetare - Alle meine Taschenrechner
//******************************************************************************
/// \file symbols.h
///       Symbole, die alle Taschenrechner haben (Modul base)
//******************************************************************************
// Wann       Wer      Was
// ---------- -------- ---------------------------------------------------------
// 23.03.2002 fse      erzeugt
//******************************************************************************

#ifndef SYMBOLS_H
#define SYMBOLS_H

#ifndef __cplusplus
#error symbols.h is only for C++!
#endif

//******************************************************************************

#include <string>

#include <math.h>
#include "number.h"

using namespace std;

//******************************************************************************

class ComputerBase; // Forward-Deklaration
class DisplayBase;  // Forward-Deklaration

//******************************************************************************
// Klasse Symbol und Ableitungen:
//******************************************************************************

/// Basisklasse fuer saemtliche Symbole.
/// Jedes Symbol weiss selbst, was es zu tun hat und veraendert die Stacks
/// entsprechend. Daher wird hier ein Zeiger auf mComputer in Klasse Calculator
/// bereitgestellt, ueber den die Stacks manipuliert werden koennen.
//  fse, 06.10.01

class Symbol
{
protected:
	string mSym;                 ///< das Symbol als String ("+", "sto", ...)
	string mHlp;                 ///< kurzer Hilfetext
	static ComputerBase* mCompP; ///< Zeiger auf Computer mit der Symboltabelle
	static DisplayBase*  mDsplP; ///< Zeiger auf Display (f. wissenschftl. Rechner)
	static Symbol* getLastInP();
	static Symbol* getLastOpP();
	static Symbol* getLastButOneOpP();
public:
	explicit Symbol(string sym = "", string hlp = "no help available") : mSym(sym), mHlp(hlp) {}
	string getSym() const { return mSym; }
	string getHlp() const { return mHlp; }
	virtual ~Symbol() {}
	virtual void process() = 0;
	static void setCompP(ComputerBase* cP) { mCompP = cP; }
	static void setDsplP(DisplayBase*  cP) { mDsplP = cP; }
	friend ostream& operator<<(ostream& os, const Symbol& s);
	friend ostream& operator<<(ostream& os, const Symbol* s);
};

//************************************** ***************************************

/// Basisklasse fuer Zahlen.
/// Neben allen "normalen" Zahlen gibt es auch noch Konstanten wie pi und e.

class SymNumber : public Symbol
{
protected:
	explicit SymNumber(string sym = "", string hlp = "any number")
		: Symbol(sym, hlp) {}
public:
	void processCommon() const;
};

//************************* ************************* **************************

/// Klasse der "normalen" Zahlen. Z.B. 1.23, 4, -5 etc.
/// Die Objekte dieser Klasse sind Dummies, da nicht Objekte dieser Klasse,
/// sondern direkt die Double-Werte (nicht ganz richtig: Objekte der Klasse
/// \c Number) auf den Zahlen-Stack gelegt werden.

class NumDmy : public SymNumber
{
public:
	explicit NumDmy(string num = "0", string hlp = "any number") : SymNumber(num) { hlp = hlp; }
	void setSym(const string& num) { mSym = num; }
	void process();
	static bool isFraction(const string& sym, double& val, long& numr, long& dnom);
};

//************************************** ***************************************

/// Basisklasse fuer die Operatoren.
//  fse, 06.10.01

class SymOperator : public Symbol
{
protected:
	int mPrio;
public:
	explicit SymOperator(string op = "", int prio = 2, string hlp = "no help available")
		: Symbol(op, hlp), mPrio(prio) {}
	int getPrio() const { return mPrio; }
	void processCommon();
	void processAddSub();
	void processAddSubCasioSci();
	SymOperator* checkConstMode(const SymOperator* lastSymP) const;
	virtual Number calcVal(const Number& a, const Number& b) = 0;
};

//************************* ************************* **************************

/// Operator fuer Addition.

class OpAdd : public SymOperator
{
public:
	OpAdd() : SymOperator("+", 2, "adding operator") {};
	Number calcVal(const Number& a, const Number& b) { return a + b; }
	void process() { processAddSub(); }
};

//****************** ******************* ******************* *******************

/// Operator fuer Addition.

class OpAddCasioSci : public OpAdd
{
public:
	OpAddCasioSci() : OpAdd() {};
	void process() { processAddSubCasioSci(); }
};

//************************* ************************* **************************

/// Operator fuer Subtraktion.

class OpSub : public SymOperator
{
public:
	OpSub() : SymOperator("-", 2, "subtracting operator") {}
	Number calcVal(const Number& a, const Number& b) { return a - b; }
	void process() { processAddSub(); }
};

//****************** ******************* ******************* *******************

/// Operator fuer Subtraktion.

class OpSubCasioSci : public OpSub
{
public:
	OpSubCasioSci() : OpSub() {};
	void process() { processAddSubCasioSci(); }
};

//************************* ************************* **************************

/// Operator fuer Multiplikation.

class OpMul : public SymOperator
{
public:
	OpMul() : SymOperator("*", 4, "multiplying operator") {}

	Number calcVal(const Number& a, const Number& b) { return a * b; }
	void process() { processCommon(); }
};

//************************* ************************* **************************

/// Operator fuer Division.

class OpDiv : public SymOperator
{
public:
	OpDiv() : SymOperator("/", 4, "dividing operator") {}
	Number calcVal(const Number& a, const Number& b) { return a / b; }
	void process() { processCommon(); }
};

//************************************** ***************************************

/// Basisklasse fuer weitere Symbole ("(", etc.)
//  fse, 06.10.01

class SymOther : public Symbol
{
protected:
	static int mBraLev;
public:
	explicit SymOther(string sym = "", string hlp = "no help available")
		: Symbol(sym, hlp) {}
	static int getBraLev() { return mBraLev; }
};

//************************* ************************* **************************

/// Symbol "(" ("Bra" von Bracket).

class OthBra : public SymOther
{
public:
	OthBra() : SymOther("(", "parentheses open") {}
	void process();
};

//************************* ************************* **************************

/// Symbol ")" ("ket" von Bracket).

class OthKet : public SymOther
{
public:
	OthKet() : SymOther(")", "parentheses close") {}
	void process();
};

//************************* ************************* **************************

/// Symbol "=".

class OthEqu : public SymOther
{
public:
	OthEqu() : SymOther("#") {}
	void process();
};

//************************************** ***************************************

/// Basisklasse aller Funktionen.

class SymFunction : public Symbol
{
public:
	explicit SymFunction(string sym = "", string hlp = "no help available")
		: Symbol(sym, hlp) {}
};

//************************* ************************* **************************

/// Basisklasse der Nicht-Winkelfunktionen.

class SymOtherFunction : public SymFunction
{
public:
	explicit SymOtherFunction(string sym = "", string hlp = "no help available")
		: SymFunction(sym, hlp) {}
	void process();
	virtual Number calcVal(const Number& a) = 0;
};

//****************** ******************* ******************* *******************

/// Wurzelfunktion.

class FnSqrt : public SymOtherFunction
{
public:
	FnSqrt() : SymOtherFunction("sqrt", "square root") {}
	Number calcVal(const Number& a) { return sqrt(a.getVal()); }
};

//************************************** ***************************************

/// Basisklasse fuer alle Aktionen ("=", "%", ...).
//  fse, 06.10.01

class SymAction : public Symbol
{
public:
	explicit SymAction(string sym = "", string hlp = "no help available")
		: Symbol(sym, hlp) {}
};

//************************* ************************* **************************

/// Symbol fuer Ausgabe der Hilfe.

class AcHelp : public SymAction
{
public:
	AcHelp() : SymAction("help", "print this help page") {}
	void process();
};

//************************* ************************* **************************

/// Loeschen der letzten Eingabe
//  fse, 30.06.02

class AcC : public SymAction
{
public:
	AcC() : SymAction("c", "clear") {}
	void process();
};

//************************* ************************* **************************

/// Gesamtloeschung.
//  fse, 30.06.02

class AcAc : public SymAction
{
public:
	AcAc() : SymAction("ac", "clear all") {}
	void process();
};

//************************* ************************* **************************

/// Berechnung durchfuehren.

class AcEqu : public SymAction
{
public:
	AcEqu() : SymAction("=", "execute calculation") {}
	void process();
};

//************************* ************************* **************************

/// Prozent.

class AcPrc : public SymAction
{
public:
	AcPrc() : SymAction("%", "percent calculation") {}
	void process();
};

//************************* ************************* **************************

/// Anzeige speichern (store).

class AcSto : public SymAction
{
public:
	AcSto() : SymAction("sto", "store display in memory") {}
	void process();
};

//************************* ************************* **************************

/// Speicher wieder in die Anzeige holen (recall).

class AcRcl : public SymAction
{
public:
	AcRcl() : SymAction("rcl", "recall memory to display") {}
	void process();
};

//************************* ************************* **************************

/// Im Speicher addieren.

class AcSum : public SymAction
{
public:
	AcSum() : SymAction("sum", "sum display to memory") {}
	void process();
};

//************************* ************************* **************************

/// Im Speicher subtrahieren.

class AcSub : public SymAction
{
public:
	AcSub() : SymAction("sub", "subtract display from memory") {}
	void process();
};

//************************* ************************* **************************

/// Anzeige und Speicher austauschen (exchange).

class AcExc : public SymAction
{
public:
	AcExc() : SymAction("exc", "exchange memory and display") {}
	void process();
};

//************************* ************************* **************************

/// Rechner initialisieren.

class AcInit : public SymAction
{
public:
	AcInit() : SymAction("init", "initialize calculator") {}
	void process();
};

//************************* ************************* **************************

/// Rechner initialisieren, ohne Speicher zu loeschen (init without memory).

class AcInitWoMem : public SymAction
{
public:
	AcInitWoMem() : SymAction("initwomem", "initialize without clearing memory") {}
	void  process();
};

//************************* ************************* **************************

/// Vorzeichenwechsel.

class AcSign : public SymAction
{
public:
	AcSign() : SymAction("sign", "change sign of display") {}
	void process();
};

//******************************************************************************
//************************************** ***************************************
//************************* ************************* **************************

#endif // !SYMBOLS_H

