//******************************************************************************
// Copyright (c) 2002-2005 by Friedemann Seebass, Germany.
// Dieses Programm ist freie Software. Siehe GNU GPL Notiz in Datei main.cpp.
// This program is free software. See the GNU GPL notice in file main.cpp.
// Projekt: Almetare - Alle meine Taschenrechner
//******************************************************************************
/// \file confg.cpp
///       Klasse zum Einlesen/Schreiben von Windows-Ini-Dateien (Modul lib).
///       Die Datei kann leider nicht config.cpp heissen, weil es sonst eine
///       Kollision mit der Datei config.h des Tools autoconf gibt.
//******************************************************************************
// Wann       Wer      Was
// ---------- -------- ---------------------------------------------------------
// 28.05.2005 fse      init(): auch '\r' als Endezeichen
// 27.10.2003 fse      readParameter(): neuer Parameter isStripSpc
// 06.09.2003 fse      nextParameter(): Kommentarzeilen ueberlesen
// 22.08.2003 fse      Anpassungen fuer MM1
// 11.05.2003 fse      Datei von config.cpp in confg.cpp umbenannt
// 26.04.2003 fse      zu viele Leerzeichen am Ende eines Parameters werden
//                     entfernt
// 18.11.2001 fse      Schreiben von Parametern
// 18.02.2001 fse      erzeugt
//******************************************************************************

#include <iostream>
#include <sstream>
#include "confg.h"

//******************************************************************************
// Konfigurationsdateiverwaltung
//******************************************************************************

/// Initialisierungsfunktion wird von allen Konstruktoren aufgerufen.
/// \param fileName Name der Initialisierungsdatei
/// \return OK, falls Oeffnen der Datei geklappt hat, sonst FILE_NOT_FOUND

//  Um zu testen, ob die Datei vorhanden ist, muss zunaechst ein ifstream
//  geoeffnet werden. (Das Oeffnen des fstreams mFile liefert keinen Fehler,
//  falls die Datei noch nicht angelegt ist; stattdessen wird die Datei
//  angelegt.)
//  fse, 17.11.01
//  fse, 28.01.05: auch '\r' als Endesymbol

Config::Err Config::init(const string& fileName)
{
	ifstream ifs(fileName.c_str());
	
	mCsAndEol = "#\r\n\0";  // <- Kommentarsymbol hier zentral aenderbar   <-
	mAssigns = " =\t";      // <- Zuweisungszeichen hier zentral aenderbar <-
	mCs = mCsAndEol[0];           
	mIsFileRead = false;
	mFileName = fileName;
	mErr = OK;
	if (!ifs)
	{
		mErr = FILE_NOT_FOUND;
		return mErr;
	}
	ifs.close();
	mFile.clear();
	mFile.open(fileName.c_str(), ios::in|ios::out);
	if (!mFile) mErr = FILE_NOT_FOUND;
	return mErr;
}

//******************************************************************************

/// Sucht von Beginn der Konfigurationsdatei an nach Sektion sectionName. Der
/// Filepointer wird dann an diese Stelle gesetzt.
/// \param sectionName Name der zu suchenden Sektion
/// \return OK, falls Sektion gefunden, sonst SECTION_NOT_FOUND
// fse, 19.02.01

Config::Err Config::seekSection(const string& sectionName)
{
	string line, section, dummy;

	mErr = OK;
	mSection = sectionName;
	mFile.clear(); // ggf. eof zuruecksetzen
	(void)mFile.seekg(0);
	while (!mFile.eof())
	{
		(void)getline(mFile, line);
		splitLine(line, dummy, section, dummy, dummy, dummy, dummy);
		if (section != sectionName) continue;
		return mErr;
	}
	mErr = SECTION_NOT_FOUND;
	return mErr;
}

//************************************** ***************************************

/// Trennt eine Konfigurationszeile in ihre  Bestandteile auf.
/// \param  line        zu trennende Zeile
/// \retval space       alle Zeichen vor dem Parameter
/// \retval sectionName Sektionsname (falls die Zeile einen solchen enthaelt),
///                     ohne die eckigen Klammern
/// \retval parName     Parametername, von seinem ersten bis zum letzten
///                     Buchstaben
/// \retval assignOp    Zuweisungsoperator, alle Zeichen zwischen dem letzten
///                     Buchstaben des Parameters und dem ersten Buchstaben
///                     des Wertes
/// \retval value       Wert des Parameters, Zeichen ab dem ersten
///                     Nicht-Space-Zeichen nach dem Zuweisungsoperator bis zum
///                     Zeilenende bzw. Kommentarzeichen
/// \retval comment     Kommentar, alle Zeichen ab dem Kommentarsymbol bis zum
///                     Ende der Zeile
///
/// Hier ein Beispiel (die "^" zeigen jeweils auf das erste und letzte Zeichen
/// des Bestandteils):
/// \code
///       Parameter   =    Wert       # Kommentar
/// ^    ^^       ^^      ^^         ^^  
/// space  parName assignOp   value      comment
///
/// oder
///
///       [sektion]  Kommentar
/// ^    ^ ^     ^ ^
/// space  section comment
/// \endcode
///

// fse, 15.11.01

void Config::splitLine(string line,
                       string& space,
                       string& sectionName,
                       string& parName,
                       string& assignOp,
                       string& value,
                       string& comment)
{
	string::size_type start, end;
	char c;

	space = sectionName = parName = assignOp = value = comment = "";
	start = line.find_first_not_of(" \t"); // Leerzeichen ueberlesen
	if (start == string::npos)
	{ // enthaelt nichts ausser " " oder "\t":
		space = line;
		return;
	}
	c = line[start];
	if (c == mCs)
	{ // enthaelt nur Kommentar und evtl. space
		space = line.substr(0, start);
		comment = line.substr(start);
		return;
	}
	space = line.substr(0, start);
	if (c == '[')
	{ // Sektionsbeginn gefunden
		end = line.find(']', start);
		if (end == string::npos) return;
		sectionName = line.substr(start+1, (end-start)-1);
		comment = line.substr(end+1);
		return;
	}
	end = line.find_first_of(mAssigns, start);
	if (end == string::npos) end = line.size();
	parName = line.substr(start, end-start);
	start = end;
	line += mCs; // damit line.find_first_of(mCsAndEol, start) garantiert was findet
	end = line.find_first_not_of(mAssigns, start);
	assignOp = line.substr(start, end-start);
	start = end;
	end = line.find_first_of(mCsAndEol, start);
	value = line.substr(start, end-start);
	start = end;
	end = line.size() - 1;  // mCs wieder weglassen
	comment = line.substr(start, end-start);
	return;
}

//******************************************************************************

/// Liest den Parameter parName ein und gibt den Wert in value zurueck.
/// Wenn section = "", wird ab der aktuellen Zeile gesucht, sonst wird zunaechst
/// die Section ab Dateianfang gesucht.
/// \param  par        der zu lesende Parameter
/// \param  section    Name der Sektion, in dem der Parameter gesucht wird
/// \retval val        der Wert des Parameters (grundsaetzlich ein String)
/// \param  isStripSpc wenn true, werden endstaendige Leerzeichen des Parameters
///                    entfernt
/// \return OK, falls Lesen des Parameters geklappt hat, sonst PAR_NOT_FOUND
//  fse, 29.07.99
//  fse, 27.10.03: neuer Parameter isStripSpc

Config::Err Config::readParameter(const string& par, string& val, const string& section, bool isStripSpc)
{
	string actPar;

	mErr = OK;
	mParName = par;
	if (section != "")
	{
		if (seekSection(section) != OK) return mErr;
	}
	while (nextParameter(actPar, val) == OK)
	{
		if (par == actPar)
		{
			if (isStripSpc)
				val = val.substr(0, val.find_last_not_of(" ")+1);
			return mErr;
		}
	}
	mErr = PAR_NOT_FOUND;
	return mErr;
}

//******************************************************************************

/// Holt naechsten Parameter ab der aktuellen Zeile.
/// \retval parName Parametername der naechsten Zeile
/// \retval value   zugehoeriger Parameterwert
/// \return OK, falls Lesen des Parameters geklappt hat, sonst NO_FURTHER_PAR
//  fse, 19.02.01
//  fse, 06.09.03: Kommentarzeilen ueberlesen

Config::Err Config::nextParameter(string& parName, string& value)
{
	string line, dummy;

	mErr = OK;
	if (mFile.eof())
	{
		mErr = NO_FURTHER_PAR;
		return mErr;
	}
	do
	{
		(void)getline(mFile, line);
	} while (line[0] == mCs); // alle Kommentarzeilen ueberlesen
	splitLine(line, dummy, dummy, parName, dummy, value, dummy);
	if (parName == "")
	{
		mErr = NO_FURTHER_PAR;
		return mErr;
	}
	return mErr;
}

//******************************************************************************

/// Weist dem Parameter setPar in der Sektion setSection den Wert setVal zu.
/// Sonst wird an der Datei nichts veraendert, Kommentare etc. bleiben
/// erhalten. Ausnahme: Wenn der Platz zwischen Parametername und Kommentar
/// nicht reicht, wird der Kommentar nach rechts verschoben.
///
/// Beim ersten Aufruf wird die ganze Datei in mLines[] eingelesen, weitere
/// Aufrufe arbeiten dann auf diesem Puffer. Er wird im Destruktor oder mit
/// flush() dann wieder geschrieben.
///
/// Wenn setSection = "", wird ab der aktuellen Zeilenposition direkt nach dem
/// Parameter gesucht.
///
/// \param setPar     Name des zu schreibenden Parameters
/// \param setVal     Wert des zu schreibenden Parameters
/// \param setSection Sektion, in der der zu schreibende Parameter steht
/// \return OK, falls Schreiben des Parameters geklappt hat, sonst
///         SECTION_NOT_FOUND
//  fse, 18.11.01

Config::Err Config::writeParameter(const string& setPar, 
                                   const string& setVal,
                                   const string& setSection)
{
	string line;
	static unsigned int i = 0;
	string space, section, par, assignOp, val, comment;

	mErr = OK;
	mParName = setPar;
	if (setSection == "")
	{ // wir sind angeblich schon in der richtigen Section, Param. setzen:
		if (!mIsFileRead)
		{
			mSection = "";
			mErr = SECTION_NOT_FOUND;
		}
		else
		{
			mErr = setParameter(setPar, setVal, ++i);
		}
		return mErr;
	}
	else
	{ // erst Datei lesen:
		mSection = setSection;
		if (!mIsFileRead) readFile();
		i = 0;
		while (i < mLines.size())
		{ // erst Section suchen:
			splitLine(mLines[i], space, section, par, assignOp, val, comment);
			if (section == setSection)
			{ // evtl. Parameter setzen:
				mErr = setParameter(setPar, setVal, ++i);
				return mErr;
			}
			++i;
		}
		mErr = SECTION_NOT_FOUND;
		return mErr;
	}
}

//************************************** ***************************************

/// Liest die ganze Konfigurationsdatei in das String-Array \c mLines ein.
//  fse, 14.11.01

void Config::readFile()
{
	string line;

	mFile.clear(); // ggf. eof zuruecksetzen
	(void)mFile.seekg(0);
	while (!mFile.eof())
	{
		(void)getline(mFile, line);
		mLines.push_back(line);
	}
	mIsFileRead = true;
}

//************************************** ***************************************

/// Beschreibt den Parameter setPar mit Wert setVal.
/// Es wird davon ausgegangen, dass i bereits eine Zeile in der passenden Sektion
/// ist und dann zeilenweise nach dem Parameter gesucht.
/// \param  setPar zu schreibender Parameter
/// \param  setVal zu schreibender Wert
/// \retval i      Nummer der Zeile, in der der Parameter steht
/// \return OK, falls Schreiben des Parameters geklappt hat, sonst
///         PAR_NOT_FOUND
// fse, 18.11.01
// fse, 26.04.03: Wenn mehr als 20 Leerzeichen am Ende eines Parameters, werden
//                alle Leerzeichen entfernt; Grund: beim Casio fx-3600P wird
//                man sonst den von den Programmen eingenommenen Platz nicht
//                mehr los.

Config::Err Config::setParameter(const string& setPar, const string& setVal, unsigned int& i)
{
	string space, section, par, assignOp, val, comment, newVal;
	string::size_type cPos;

	splitLine(mLines[i], space, section, par, assignOp, val, comment);
	while (i < mLines.size() && section == "")
	{
		if (par == setPar)
		{
			ostringstream str;
			(void)str.width((int)val.size());
			(void)str.setf(ios::left, ios::adjustfield);
			str << setVal.c_str();// << ends; // width funktioniert bei GNU nicht mit strings!?
			newVal = str.str();
			// wenn mehr als 20 Leerzeichen am Ende, Leerzeichen entfernen:
			cPos = newVal.find_last_not_of(' ');
			if (cPos != string::npos)
				if (newVal.size() - cPos >= 20) newVal = newVal.substr(0, cPos+1);
			mLines[i] = space + par + assignOp + newVal + comment;
			return OK;
		}
		++i;
		splitLine(mLines[i], space, section, par, assignOp, val, comment);
	}
	return PAR_NOT_FOUND;
}

//******************************************************************************

/// Schreibt die Konfigurationsdatei neu auf Platte.
/// \return OK, falls Schreiben des Parameters geklappt hat, sonst
///         FILE_NOT_FOUND
//  fse, 18.11.01

Config::Err Config::flush()
{
	unsigned int i;

	mErr = OK;
	if (!mIsFileRead) return mErr;
	mFile.close();
	mFile.clear();
	mFile.open(mFileName.c_str(), ios::in|ios::out);
	if (!mFile)
	{
		mErr = FILE_NOT_FOUND;
		return mErr;
	}
	for (i = 0; i < mLines.size(); i++)
	{
		mFile << mLines[i] << "\n";
		//cout << mLines[i] << "\n";
	}
	mFile.close();
	mFile.clear();
	mFile.open(mFileName.c_str(), ios::in|ios::out);
	return OK;
}

//******************************************************************************

/// Wandelt Error-Code in String um. 
/// \return String mit dem Text des Error-Codes
//  fse, 17.11.01

string Config::errToStr() const
{
	switch(mErr)
	{
		case OK               : return "OK";
		case FILE_NOT_FOUND   : return      "File '" + mFileName + "' not found!";
		case SECTION_NOT_FOUND: return   "Section '" + mSection  + "' not found!";
		case PAR_NOT_FOUND    : return "Parameter '" + mParName  + "' in section '" + mSection + "' not found!";
		case NO_FURTHER_PAR   : return "No further parameter in section '" + mSection +"'!";
		default: return "Unknown error in class Config!";
	}
}

//******************************************************************************
//******************************************************************************
/*
int test()
{
	string fileName = "sources\\test.ini";
	Config::Err err;
	Config cfg("sources\\test.ini");
	string section = "section a";
	string parName = "par2";

	string value = "aasdhgg";

	if (cfg.getErr() == Config::OK)
		cout << "File '" << fileName << "' opened.\n";
	else
		cout << cfg.errToStr() << "\n---\n";

	err = cfg.seekSection("saction");
	if (err != Config::OK) cout << cfg.errToStr() << "\n";
	while (cfg.nextParameter(parName, value) == Config::OK)
		cout << parName << ", " << value << "<\n";
	cout << cfg.errToStr() << "\n---\n";

	err = cfg.seekSection("saction");
	err = cfg.readParameter("toast", value);
	cout << cfg.errToStr() << "\n---\n";

	cfg.writeParameter("par2", "neu", "section a");
	cfg.writeParameter("par8", "neu2");
	cout << cfg.errToStr() << "\n---\n";
	cfg.flush();

	cfg.readParameter("par2", value, "section a");
	cout << value << ", " << cfg.errToStr() << "\n---\n";
	cfg.readParameter("par6", value, "saction");
	cout << value << ", " << cfg.errToStr() << "\n---\n";
	cfg.readParameter("par1", value, "section a");
	cout << value << ", " << cfg.errToStr() << "\n---\n";
	cfg.readParameter("par2", value);
	cout << value << ", " << cfg.errToStr() << "\n---\n";

	return 0;
}
*/
//******************************************************************************
//************************************** ***************************************
//************************* ************************* **************************
//****************** ******************* ******************* *******************
//************** *************** *************** *************** ***************
//************ ************ ************ ************ ************ *************
