//******************************************************************************
// Copyright (c) 2002-2004 by Friedemann Seebass, Germany.
// Dieses Programm ist freie Software. Siehe GNU GPL Notiz in Datei main.cpp.
// This program is free software. See the GNU GPL notice in file main.cpp.
// Projekt: Almetare - Alle meine Taschenrechner
//******************************************************************************
/// \file confg.h
///       Klasse zum Einlesen/Schreiben von Windows-Ini-Dateien (Modul lib).
///       Die Datei kann leider nicht config.h heissen, weil es sonst eine
///       Kollision mit der Datei config.h des Tools autoconf gibt.
//******************************************************************************
// Wann       Wer      Was
// ---------- -------- ---------------------------------------------------------
// 26.08.2003 fse      anderer Default-Konstruktor
// 22.08.2003 fse      Anpassungen fuer MM1
// 11.05.2003 fse      Datei von config.h in confg.h umbenannt
// 01.03.2003 fse      neue Funktion remove()
// 18.11.2001 fse      Schreiben von Parametern
// 18.02.2001 fse      erzeugt
//******************************************************************************

#ifndef CONFG_H
#define CONFG_H

#ifndef __cplusplus
#error confg.h is only for C++!
#endif

//******************************************************************************

#include <string>
#include <fstream>
#include <vector>

using namespace std;

//******************************************************************************
// Kofigurations-Klasse
//******************************************************************************

/// Klasse zum Einlesen/Schreiben von Windows-Ini-Dateien.
///
/// Die Klasse Config dient zum Lesen und zerstoerungsfreien Schreiben von
/// Initialisierungsdateien, wie sie frueher unter Windows ueblich waren.
/// Zerstoerungsfrei bedeutet dabei, dass die Formatierung der Datei auch beim
/// Schreiben von Parameteren (z.B. Kommentare am Zeilenende) erhalten bleibt.
///
/// Eine Initialisierungsdatei gliedert sich in Sektionen, in denen die
/// Parameter aufgelistet sind; Kommentare sind am Zeilenende erlaubt. Beispiel:
///
/// \code
/// # configuration file for Almetare's Casio fx-85v
/// 
/// [PROGRAM]
///    name         = Casio fx-85v
///    version      = V.1.0 beta    
///    nvRam        = ||||0|||||||  # non volatile RAM - you should not edit this entry
///    displayMode  = moreRealistic # allowed: 'debug', 'simple', 'realistic' or 'moreRealistic'
///    xPos         = 50            # x position of the dialog in pixels
///    yPos         = 20            # y position of the dialog in pixels
///    width        = 230           # width  of the dialog in pixels (0: default)
///    height       = 445           # height of the dialog in pixels (0: default)
///    showHint     = off           # allowed: 'on' or 'off'
///    showToolTips = on            # allowed: 'on' or 'off'
///    logMode      = off           # allowed: 'on' or 'off'
///    browserPath  = default       # allowed: 'default' or path to browser
/// 
/// [LOGGING]
///    onlFile      = log.txt       # name of file for normal logging
///    dbgFile      = log.txt       # name of file for debug logging
///    errFile      = log.txt       # name of file for error logging
///    maxLogLen    = 50            # max. length of files in kByte (if longer: -> *.bak)
///    onlChannels  = gnl cns       # allowed: 'ctl' 'mem' 'gnl' 'hgh' 'med' 'low' 'all' 'cns'
///    dbgChannels  =               # allowed: 'ctl' 'mem' 'gnl' 'hgh' 'med' 'low' 'all' 'cns'
///    errChannels  = all cns       # allowed: 'ctl' 'mem' 'gnl' 'hgh' 'med' 'low' 'all' 'cns'
///    fileToLog    =               # name of the exclusively logged file
/// \endcode
///
/// Der Sektionsname muss in eckigen Klammern oberhalb der zugehoerigen
/// Parameter stehen. Leerzeichen und Tabs am Zeilenanfang werden ignoriert.
///
/// Trennzeichen zwischen Parameternamen und Parametern koennen in der
/// Klassenvariablen \c mAssigns definiert werden (z.Zt. Tab, Space und "=").
/// Das Kommentarsymbol wird in \c mCsAndEol definiert (z.Zt. "#").
/// Die Werte der Parameter sind grundsaetzlich Strings und muessen ggf. vom
/// Code, der die Klasse verwendet, in andere Typen umgewandelt werden.
//  fse, 18.04.02
//  fse, 01.03.03: neue Funktion remove();
//  fse, 26.08.03: Default-Konstruktor
//                   Config(const string& fileName = "default.ini") { (void)init(fileName); }
//                 durch
//                   Config() {}
//                 ersetzt, damit man ein Config-Objekt erzeugen kann, ohne
//                 sofort den Dateinamen zu wissen.

class Config
{
public:
	enum Err { OK = 0, FILE_NOT_FOUND, SECTION_NOT_FOUND, PAR_NOT_FOUND, NO_FURTHER_PAR };
	Config() {};
	Config(const string& fileName, Err& err) { err = init(fileName); }
	~Config() { (void)flush(); mFile.close(); }
	Err init(const string& fileName);
	Err getErr() const { return mErr; }
	string errToStr() const;
	Err seekSection(const string& section);
	Err readParameter(const string& par, string& val, const string& section = "", bool isStripSpc = true);
	Err nextParameter(string& par, string& val);
	Err writeParameter(const string& par, const string& val, const string& section = "");
	Err flush();
	void remove() { mFile.close(); ::remove(mFileName.c_str()); }
private:
	char mCs;              ///< Kommentarsymbol
	string mAssigns;       ///< Liste von Zuweisungszeichen
	string mCsAndEol;      ///< String mit Kommentar und EOL-Symbol
	string mFileName;      ///< Name der Konfigurationsdatei
	string mSection;       ///< Name der Sektion
	string mParName;       ///< Name des Parameters
	fstream mFile;         ///< zu oeffnende Konfigurationsdatei
	bool mIsFileRead;      ///< Flag, ob Datei schon gelesen
	Err mErr;              ///< Fehlercode
	vector<string> mLines; ///< gepufferte Datei

	void splitLine(string line,
	               string& space,
	               string& sectionName,
	               string& parName,
	               string& assignOp,
	               string& value,
	               string& comment);
	void readFile();
	Err setParameter(const string& setPar, const string& setVal, unsigned int& i);
};

//******************************************************************************

#endif // !CONFG_H
