//******************************************************************************
// Copyright (c) 2002-2004 by Friedemann Seebass, Germany.
// Dieses Programm ist freie Software. Siehe GNU GPL Notiz in Datei main.cpp.
// This program is free software. See the GNU GPL notice in file main.cpp.
// Projekt: Almetare - Alle meine Taschenrechner
//******************************************************************************
/// \file fraction.cpp
///       Klassen zum Rechnen mit Bruechen - Implementierung (Modul lib)
//******************************************************************************
// Wann       Wer      Was
// ---------- -------- ---------------------------------------------------------
// 17.11.2003 fse      strstream durch sstream ersetzt
// 20.04.2002 fse      erzeugt
//******************************************************************************

#include <sstream>
#include "fraction.h"

//******************************************************************************
// Klasse Fraction:
//******************************************************************************

bool Fraction::mIsShortened = true; // Brueche normalerweise kuerzen

//******************************************************************************

/// Konstruktor konstruiert Brueche aus Zaehler und Nenner. Kann ausserdem
/// zur Konvertierung von long-Parametern in Brueche verwendet werden.
/// \param n Zaehler ("Nominator")
/// \param d Nenner  ("Denominator")
/// \exception class ZeroDevide, wenn der Nenner 0 ist
//  fse, 09.06.02

Fraction::Fraction(long n, long d) throw (ZeroDivide): mNumr(n), mDnom(d)
{
	if (mDnom == 0) throw ZeroDivide();
}

//******************************************************************************

/// Kuerzt den Bruch durch Division von Zaehler und Nenner durch den ggT.
/// Ein negatives Vorzeichen wird grundsaetzlich dem Zaehler zugeordnet.
//  fse, 20.04.02

void Fraction::shorten()
{
	long gcd_ = gcd();

	mNumr /= gcd_;
	mDnom /= gcd_;
	if (mDnom < 0)
	{
		mDnom = -mDnom;
		mNumr = -mNumr;
	}

}

//******************************************************************************

/// Berechnet den groessten gemeinsamen Teiler (ggT) des Bruchs.
///
/// Der ggT (gcd - greatest common devisor) wird nach dem "Euklidischen
/// Algorithmus" berechnet.
/// \return ggT
//  Aus: R. u. P. Gaitzsch, "Computer-Loesungen fuer Schule, Studium und Alltag",
//  mvg Verlag, 1985.
//  fse, 15.04.02

long Fraction::gcd() const
{
	long a = mNumr;
	long b = mDnom;
	long r;

	for (;;)
	{
		r = a % b; // Rest
		//cout << "a=" << a << ", b=" << b << ", r=" << r << "\n";
		if (r == 0) break;
		a = b;
		b = r;
	}
	return b;
}

//******************************************************************************

/// Berechnet das kleinste gemeinsame Vielfache (kgV) des Bruchs.
///
/// Das kgV (lcm - least common multiple) wird nach der Formel
/// \code
///              a*b
/// kgV(a,b) = -------
///            ggT(a,b)
/// \endcode
/// berechnet.
/// \return kgV
//  Aus: R. u. P. Gaitzsch, "Computer-Loesungen fuer Schule, Studium und Alltag",
//  mvg Verlag, 1985.
//  fse, 15.04.02

long Fraction::lcm() const
{
	return mNumr*mDnom/gcd();
}

//******************************************************************************

/// Wandelt Bruch in einen gemischten Bruch um.
///
/// Ein gemischter Bruch besteht aus einem ganzen Teil und einem Bruch, z.B.:
///
/// 4/3 = 1_1/3.
///
/// Falls der ganzzahlige Teil vorhanden ist (nicht Null), erhaelt er das
/// Vorzeichen.
/// \param whole ganzzahliger Anteil des Bruchs
/// \param f     der Restbruch nach Abspalten des ganzzahligen Anteils
//  fse, 21.04.02

void Fraction::getMixedFraction(long& whole, Fraction& f) const
{
	whole = mNumr/mDnom;
	f.mNumr = mNumr % mDnom;
	f.mDnom = mDnom;
	f.shorten();
	if (whole != 0)
	{
		f.mNumr = abs(f.mNumr);
		f.mDnom = abs(f.mDnom);
	}
}

//******************************************************************************

/// Gibt den Bruch als gemischten Bruch in String-Form zurueck.
/// \return Der gemischte Bruch wird als Aufzaehlung, bestehend aus ganzzahligem
/// Anteil, Zaehler des Restbruchs und Nenner des Restbruchs mit "/" als
/// Trennzeichen zurueckgegeben: z.B. 4/3 = 1/1/3
//  fse, 21.04.02

string Fraction::getMixedFraction() const
{
	long whole;
	Fraction f;
	ostringstream str;

	getMixedFraction(whole, f);
	if (whole == 0)
		str <<                 f.mNumr << "/" << f.mDnom; // << ends;
	else
		str << whole << "/" << f.mNumr << "/" << f.mDnom; // << ends;
	return str.str();
}

//******************************************************************************

/// Gibt Bruch als String zurueck.
/// \return Zaehler und Nenner, durch "/" getrennt.
//  fse, 11.07.02

string Fraction::getFraction() const
{
	ostringstream str;

	str << mNumr << "/" << mDnom; // << ends;
	return str.str();
}

//******************************************************************************

/// Multiplikation von Bruechen.
/// \code
///  a     c      a*c
/// --- * ---  =  ---
///  b     d      b*d
/// \endcode
/// \param f Multiplikator (*this ist der Multiplikand)
/// \return  Referenz auf sich selbst - den Produkt-Bruch
//  fse, 20.04.02

Fraction& Fraction::operator*=(const Fraction& f)
{
	mNumr *= f.mNumr;
	mDnom *= f.mDnom;
	if (mIsShortened) shorten();
	return *this;
}

//******************************************************************************

/// Division von Bruechen.
/// \code
///  a     c       a     d      a*d
/// --- : ---  =  --- * ---  =  ---
///  b     d       b     c      b*c
/// \endcode
/// \param f Divisor (*this ist der Dividend)
/// \return  Referenz auf sich selbst - den Quotienten-Bruch
//  fse, 20.04.02

Fraction& Fraction::operator/=(const Fraction& f)  throw (ZeroDivide)
{
	mNumr *= f.mDnom;
	mDnom *= f.mNumr;
	if (mDnom == 0) throw ZeroDivide();
	if (mIsShortened) shorten();
	return *this;
}

//******************************************************************************

/// Addition von Bruechen.
/// \code
///  a     c       a*d   c*b     a*d + c*b
/// --- + ---  =  ---- + ---  =  ---------
///  b     d       b*d   d*b        b*d
/// \endcode
/// \param f zweiter der Summanden (*this ist der erste Summand)
/// \return  Referenz auf sich selbst - den Summen-Bruch
//  fse, 20.04.02

Fraction& Fraction::operator+=(const Fraction& f)
{
	mNumr = mNumr*f.mDnom + f.mNumr*mDnom;
	mDnom *=f.mDnom; // Hauptnenner
	if (mIsShortened) shorten();
	return *this;
}

//******************************************************************************

/// Subtraktion von Bruechen.
/// \code
///  a     c       a*d   c*b     a*d - c*b
/// --- - ---  =  ---- - ---  =  ---------
///  b     d       b*d   d*b        b*d
/// \endcode
/// \param f Subtrahend (*this ist der Minuend)
/// \return  Referenz auf sich selbst - den Differenz-Bruch
//  fse, 20.04.02

Fraction& Fraction::operator-=(const Fraction& f)
{
	mNumr = mNumr*f.mDnom - f.mNumr*mDnom;
	mDnom *=f.mDnom; // Hauptnenner
	if (mIsShortened) shorten();
	return *this;
}

//******************************************************************************
//******************************************************************************

/// Ausgabeoperator fuer einen Bruch.
/// \param os Output-Stream
/// \param f  auszugebender Bruch
/// \return   Output-Stream-Referenz
//  fse, 20.04.02

ostream& operator<<(ostream& os, const Fraction& f)
{
	os << f.getMixedFraction();
//	os << f.mNumr << "/" << f.mDnom;
	return os;
}

//******************************************************************************
