//******************************************************************************
// Copyright (c) 2002-2004 by Friedemann Seebass, Germany.
// Dieses Programm ist freie Software. Siehe GNU GPL Notiz in Datei main.cpp.
// This program is free software. See the GNU GPL notice in file main.cpp.
// Projekt: Almetare - Alle meine Taschenrechner
//******************************************************************************
/// \file fraction.h
///       Klassen zum Rechnen mit Bruechen - Deklarationen (Modul lib)
//******************************************************************************
// Wann       Wer      Was
// ---------- -------- ---------------------------------------------------------
// 08.03.2003 fse      neue Operatoren <= und >=
// 20.04.2002 fse      erzeugt
//******************************************************************************

#ifndef FRACTION_H
#define FRACTION_H

#ifndef __cplusplus
#error fraction.h is only for C++!
#endif

//******************************************************************************

#include <iostream>
#include <string>

using namespace std;

//******************************************************************************
// Klasse ZeroDevide:
//******************************************************************************

/// Klasse zur Ausnahmebehandlung "Division durch 0". 
///
/// Sowohl GNUs gcc als auch MSVC erzeugen standardmaessig keine Exception bei
/// Fliesskommafehlern (Ueberlauf, Division durch Null etc.; stattdessen wird,
/// wenn man eine bei einem Fehler erzeugte Fliesskommavariable in einen
/// String-Stream schreibt, ein Text wie "-1.#IND" oder "nan" erzeugt:
///
/// \code
/// double nix = 0.0;
/// double val = 3.0/nix;
/// strstream str;
/// str << val << ends;
/// string valStr = str.str();
/// => valStr = "nan"
/// \endcode
///
/// Daher wird in Almetare zur Fehlererkennung geprueft, ob bei einer Operation
/// ein solcher String erzeugt wurde.
/// Bei Integer-Arithmetik ist das allerdings anders, wie sich leider erst
/// spaeter herausgestellt hat: Hier wird eine Exception ausgeloest und das
/// Programm stuerzt ab. Daher ist hierfuer eine C++-Ausnahmebehandlung noetig.
/// Die Klasse ZeroDivide ist der Ausnahmedatentyp und hat sonst keine
/// Funktion.
//  fse, 09.06.02

class ZeroDivide
{
};

//******************************************************************************
// Klasse Fraction:
//******************************************************************************

/// Klasse zum Rechnen mit Bruechen.
///
/// Die Klasse repraesentiert einen Bruch, der aus einem Zaehler und einem
/// Nenner besteht. Zaehler und Nenner sind long-Variablen, dadurch kann bei
/// Berechnungen kein Genauigkeitsverlust entstehen; allerdings wird ein
/// Ueberlauf nicht automatisch erkannt, so dass ein Anwender selbst
/// sicherstellen muss, dass Zaehler und/oder Nenner nicht zu gross werden.
///
/// Die Operatoren der vier Grundrechenarten und die ueblichen
/// Vergleichsoperatoren stehen als ueberladene Operatorfunktionen zur
/// Verfuegung. Weitere Funktionen kuerzen den Bruch, ermitteln kgV und ggT,
/// erzeugen einen gemischten Bruch u.a.
///
/// Ein Element dieser Klasse wird als Member in der Klasse Number verwendet.

//fse, 18.04.02

class Fraction
{
	long        mNumr;        ///< Zaehler (numerator)
	long        mDnom;        ///< Nenner  (denominator)
	static bool mIsShortened; ///< gibt an, ob Bruch gekuerzt werden soll
public:
	Fraction() : mNumr(0), mDnom(1) {};
	Fraction(long n, long d = 1) throw (ZeroDivide);
	Fraction(const Fraction& f) : mNumr(f.mNumr), mDnom(f.mDnom) {}
	long getNumr() const { return mNumr; }
	long getDnom() const { return mDnom; }
	long gcd() const;
	long lcm() const;
	void shorten();
	Fraction rcp() const throw (ZeroDivide) { Fraction f(mDnom, mNumr); return f; }
	double getVal() const { return static_cast<double>(mNumr)/static_cast<double>(mDnom); }
	void getMixedFraction(long& whole, Fraction& f) const;
	string getMixedFraction() const;
	string getFraction() const;
	bool   isOverflow() const { return abs(mNumr) > 10000 || abs(mDnom) > 10000; }
	static void setShortened(bool is) { mIsShortened = is; }
	// Operatoren:
	Fraction& operator*=(const Fraction& f);
	Fraction& operator/=(const Fraction& f) throw (ZeroDivide);
	Fraction& operator+=(const Fraction& f);
	Fraction& operator-=(const Fraction& f);
	friend Fraction operator* (const Fraction& f1, const Fraction& f2)                    { Fraction ret(f1); ret *= f2; return ret; }
	friend Fraction operator/ (const Fraction& f1, const Fraction& f2) throw (ZeroDivide) { Fraction ret(f1); ret /= f2; return ret; }
	friend Fraction operator+ (const Fraction& f1, const Fraction& f2)                    { Fraction ret(f1); ret += f2; return ret; }
	friend Fraction operator- (const Fraction& f1, const Fraction& f2)                    { Fraction ret(f1); ret -= f2; return ret; }
	friend bool     operator==(const Fraction& f1, const Fraction& f2) { return f1.mNumr*f2.mDnom == f1.mDnom*f2.mNumr; }
	friend bool     operator!=(const Fraction& f1, const Fraction& f2) { return f1.mNumr*f2.mDnom != f1.mDnom*f2.mNumr; }
	friend bool     operator< (const Fraction& f1, const Fraction& f2) { return f1.mNumr*f2.mDnom <  f1.mDnom*f2.mNumr; }
	friend bool     operator<=(const Fraction& f1, const Fraction& f2) { return f1.mNumr*f2.mDnom <= f1.mDnom*f2.mNumr; }
	friend bool     operator> (const Fraction& f1, const Fraction& f2) { return f1.mNumr*f2.mDnom >  f1.mDnom*f2.mNumr; }
	friend bool     operator>=(const Fraction& f1, const Fraction& f2) { return f1.mNumr*f2.mDnom >= f1.mDnom*f2.mNumr; }
	friend ostream& operator<<(ostream& os, const Fraction& f);
};

//******************************************************************************

#endif // !FRACTION_H
