//******************************************************************************
// Copyright (c) 2002-2005 by Friedemann Seebass, Germany.
// Dieses Programm ist freie Software. Siehe GNU GPL Notiz in Datei main.cpp.
// This program is free software. See the GNU GPL notice in file main.cpp.
// Projekt: Almetare - Alle meine Taschenrechner
//******************************************************************************
/// \file logging.cpp
///       Klasse zum Loggen von Daten in drei verschiedene Dateien (Modul lib)
//******************************************************************************
// Wann       Wer      Was
// ---------- -------- ---------------------------------------------------------
// 17.01.2005 fse      im Konstruktor mIsSuppressed auf true statt false
// 17.11.2003 fse      strstream durch sstream ersetzt
// 20.03.2003 fse      dbgLog->gDbgLog, errLog->gErrLog, onlLog->gOnlLog
// 29.11.2001 fse      channelsToStr(), strToChannels() zugefuegt
// 10.11.2001 fse      writeHeader() zugefuegt
// 19.12.2000 fse      erzeugt
//******************************************************************************

#include <stdio.h>   // remove(), rename()
#include <time.h>    // fuer Uhrzeit
#include <errno.h>
#include <iostream>
#include <sstream>
#include "logging.h"

//******************************************************************************

Log gDbgLog; // |
Log gErrLog; // | globale Instanzen
Log gOnlLog; // |

//******************************************************************************
// Logging-Klasse
//******************************************************************************

/// Vorlaeufige Konstruktion eines Log-Objektes. Die eigentliche Initialisierung
/// kann nicht im Konstruktor gemacht werden, da beim Aufruf des Konstruktors
/// (Log-Objekte sind global) die CFG-Datei noch nicht gelesen ist. Dies muss
/// daher durch Aufruf der init()-Funktion nachgeholt werden.
/// \param fname     Name der zu schreibenden Log-Datei
/// \param channels  eingeschaltete Kanaele (Bitmuster)
/// \param maxLogLen maximal erlaubte Laenge der Log-Datei in Bytes
/// \param fileToLog Quell-Code-Datei, die ausschliesslich geloggt werden soll;
///                  wenn "", dann werden alle Quell-Code-Dateien geloggt
/// \param progName  Name der Applikation (nur fuer Titelzeile der Log-Datei)
/// \param version   Version der Applikation (nur fuer Titelzeile der Log-Datei)
//  fse, 19.12.00
//  fse, 17.01.05: mIsSuppressed auf true statt false

Log::Log(const string& fname,
         int   channels,
         int   maxLogLen,
         const string& fileToLog,
         const string& progName,
         const string& version)
{
	mFname = fname;
	mFstreamP = NULL;
	mFileToLog = fileToLog;
	mConfChannels = channels; // vorlaeufig, bis init() aufgerufen
	mMaxLogLen = maxLogLen;   // maximale Laenge der Logdatei in Bytes
	mIsSuppressed = true;
	mProgName = progName;
	mVersion = version;
	if (mFname != "") (void)init(fname, channels, maxLogLen, fileToLog, progName, version);
}

//************************************** ***************************************

/// Initialisiert die Dateinamen und Channels.
/// Kann nicht im Konstruktor gemacht werden, da beim Aufruf des Konstruktors
/// (Log-Objekte sind global) die CFG-Datei noch nicht gelesen ist.
/// Es ist moeglich, alle Log-Objekte mit derselben Datei zu verknuepfen.
/// Dazu wird geprueft, ob eine Datei mit gleichem Namen schon vorher geoeffnet
/// wurde. Daher ist die Reihenfolge beim Initialisieren wichtig! Zuerst muss
/// die Online-Log-Datei geoeffnet werden, damit mit den Membervariablen
/// verglichen werden kann.
/// \param fname     Name der zu schreibenden Log-Datei
/// \param channels  eingeschaltete Kanaele (Bitmuster)
/// \param maxLogLen maximal erlaubte Laenge der Log-Datei in Bytes
/// \param fileToLog Quell-Code-Datei, die ausschliesslich geloggt werden soll;
///                  wenn "", dann werden alle Quell-Code-Dateien geloggt
/// \param progName  Name der Applikation (nur fuer Titelzeile der Log-Datei)
/// \param version   Version der Applikation (nur fuer Titelzeile der Log-Datei)
/// \return 0, wenn OK; -1, wenn sich Datei nicht anlegen laesst; -2, wenn sich
///         der Filestream nicht oeffnen laesst.
//  fse, 30.09.99

int Log::init(const string& fname,
              int   channels, 
              int   maxLogLen,
              const string& fileToLog,
              const string& progName,
              const string& version)
{
	bool isOnlEqual = fname == gOnlLog.mFname;
	bool isErrEqual = fname == gErrLog.mFname;

	if (mFstreamP != NULL) stopLogging();
	mMaxLogLen = maxLogLen * 1024;
	mConfChannels = channels;
	mIsSuppressed = false;
	mProgName = progName;
	mVersion = version;
	mFname = fname;
	mFileToLog = fileToLog;
	if (fname == "")
	{
		mIsSuppressed = true;
		return 0;
	}
  	if (isOnlEqual) // Dateiname wie Online-Log-Datei => Filestreams gleichsetzen
  	{
  		mFstreamP = gOnlLog.mFstreamP;
  	}
  	else
  	{
  		if (isErrEqual) // Name wie Error-Log-Datei => Filepointer gleichsetzen
  		{
  			mFstreamP = gErrLog.mFstreamP;
  		}
  		else // noch nicht offen => Datei oeffnen
  		{
			mFstreamP = new ofstream;
			if (mFstreamP == NULL) return -1;
  			mFstreamP->open(fname.c_str(), ios::app | ios::out);
  			if (*mFstreamP)
  			{
  				writeHeader();
  			}
  			else
  			{
  				cerr << "Opening of stream of'" << fname << "' failed!\n";
  				mIsSuppressed = true;
  				return -2;
  			}
  		}
  	}
	return 0;
}

//************************* ************************* **************************

/// Schliesst Datei und loescht den allokierten ofstream.
/// Problematisch ist die Sache, wenn gemeinsame Log-Dateien fuer Online-,
/// Debug- und/oder Error-Logging benutzt werden. Dann sind naemlich die Pointer
/// mFstreamP in allen drei Objekten gleich, das zuerst aufgerufene stopLogging()
/// wuerde aber schon fuer alle Objekte den ofstream entfernen, so dass beim
/// zweiten Objekt mFstreamP->close() einen Speicherfehler verursachen wuerde.
/// Daher muss verglichen werden, ob in einem anderen Log-Objekt derselbe Pointer
/// benutzt wird und dieser dann auch gleich genullt werden.
//  fse, 24.11.01

void Log::stopLogging()
{
	ofstream* fstreamP = mFstreamP; // speichern, damit man vergleichen kann
	
	if (mFstreamP) mFstreamP->close();
	delete mFstreamP;
	if (gOnlLog.mFstreamP == fstreamP) gOnlLog.mFstreamP = NULL;
	if (gErrLog.mFstreamP == fstreamP) gErrLog.mFstreamP = NULL;
	if (gDbgLog.mFstreamP == fstreamP) gDbgLog.mFstreamP = NULL;
	mFstreamP = NULL;
	mFname = "";
	mIsSuppressed = true;
}

//************************* ************************* **************************

/// Schreibt Datei-Header.
/// Format z.B.:
/// \code
/// ----------------------------------------------------------------------------
/// Casio fx-85v (V.1.0 beta) logfile 'log.txt' opened at Sat Sep 28 2002
/// ----------------------------------------------------------------------------
/// \endcode
//  fse, 10.11.01

void Log::writeHeader()
{
	string date = getDate();

	*this << "------------------------------------------------------------------------------"   << NL
	      << mProgName << " (" << mVersion << ") logfile '" << mFname << "' opened at " << date << NL
	      << "------------------------------------------------------------------------------"   << NL;
}

//****************** ******************* ******************* *******************

/// Gibt String mit dem aktuellen Datum der Systemuhr im Format
/// "Wed Dec 21 2000" zurueck.
/// \return String mit dem Datum
//  fse, 20.12.00

string Log::getDate()
{
	time_t ltime;
	string ret;

	ltime = time(&ltime);
	ret = ctime(&ltime); // = "Wed Jan 02 02:03:55 1980\n\0"
	ret = ret.substr(0, 11) + ret.substr(20, 4);
	return ret;
}

//******************************************************************************

Log::~Log()
{
	stopLogging();
}

//******************************************************************************

/// Prueft, ob die geforderten Kanaele ueberhaupt "eingeschaltet" sind und
/// die Groesse des Logfiles. Wenn zu gross, wird BAK-Datei angelegt.
/// Ausserdem wird der Header der Zeile, bestehend aus Uhrzeit, Dateiname und
/// Zeilennummer generiert und als erstes Element der Zeile ausgegeben.
/// \param channels Kanaele (Bitmuster), in die geschrieben werden soll
/// \param srcFileP Zeiger auf Name der Quell-Code-Datei (aus Makro _FILE_)
/// \param lineNr   Zeilennummer des Logging-Kommandos (aus Makro _LINE_)
/// \return false, wenn einer der gewuenschten Kanaele eingeschaltet ist und die
///         Ausgabe nicht unterdrueckt wird (kommt z.B. vor, wenn stopLogging()
///         aufgerufen wurde), sonst true
//  fse, 21.12.00

bool Log::configured(int channels, const char* srcFileP, int lineNr)
{
	if (!(mConfChannels & channels) || mIsSuppressed) return false;

	const unsigned int maxLen = 16; // maximale Laenge des Dateinamens

	string srcFile(srcFileP);
	string::size_type slashPos = srcFile.find_last_of("/\\");
	ostringstream hdr;     // ^ Slash (Linux) oder Backslash (Windows) suchen ^

	if (slashPos != string::npos) srcFile = srcFile.substr(++slashPos);
	if (mFileToLog != "" && srcFile != mFileToLog) return false;
	if (mFstreamP && mFstreamP->tellp() >= mMaxLogLen) (void)backupLogFile();
	hdr << getTime() << ": ";
	if (srcFile.size() > maxLen) srcFile[maxLen-1] = '?';
	// Gnu ignoriert width() mit strings, daher .c_str():
	(void)hdr.width(maxLen); hdr << srcFile.substr(0, maxLen).c_str() << ", ";
	(void)hdr.width(     4); hdr << lineNr << ", ";
	(void)hdr.width(     3); hdr << channelsToStr(channels).c_str() << ": "; // << ends;
	*this << hdr.str();
	return true;
}

//************************************** ***************************************

/// Benennt die Logdatei *.txt in *.bak um und loescht zuvor eine aeltere
/// *.bak-Datei, falls vorhanden.
/// \return  0, wenn OK;
///         -1, wenn der Filestream gar nicht offen ist;
///         -2, wenn sich ein aelteres Log-File nicht entfernen laesst;
///         -3, wenn sich das alte File nicht in *.bak umbenennen laesst;
//  fse, 21.12.00

int Log::backupLogFile()
{
	string bkupFname;
	string::size_type pointPos;

	if (!mFstreamP) return -1;
	mFstreamP->close();
	bkupFname = mFname;
	// Datei-Extension suchen:
	pointPos = bkupFname.find_last_of(".");
	if (pointPos != string::npos) bkupFname = bkupFname.substr(0, pointPos);
	bkupFname += ".bak";
	if (remove(bkupFname.c_str()) != 0)
	{
		if (errno != ENOENT)
		{
			cerr << "File '" << bkupFname << "' couldn't be removed!\n";
			return -2;
		}
	}

	if (rename(mFname.c_str(), bkupFname.c_str()) != 0)
	{
		if (errno != ENOENT)
		{
			cerr << "File '" << mFname    << "' couldn't be renamed to '"
			                 << bkupFname << "'\n";
			return -3;
		}
	}
	mFstreamP->open(mFname.c_str(), ios::app | ios::out);
	if (mFstreamP->fail())
		cerr << "Opening of stream of'" << mFname << "' failed!\n";
	else
		writeHeader();
	return 0;
}

//************************************** ***************************************

/// Gibt die Systemuhrzeit als String im Format "hh:mm:ss" zurueck.
/// \return String mit der Uhrzeit
//  fse , 20.12.00

string Log::getTime()
{
	time_t ltime;
	string ret;

	ltime = time(&ltime);
	ret = ctime(&ltime); // = "Wed Jan 02 02:03:55 1980\n\0"
	return ret.substr(11, 8);
}

//******************************************************************************

// Fuehrt Funktionsaufrufe aus, falls ein Pointer auf eine void-Funktion
// empfangen wurde.
// fse, 21.12.00

//Log& Log::operator<<(const funcp fp)
//{

//	(*fp)();
//	return *this;
//}

//******************************************************************************

/// Wandelt channels ch in einen String um.
/// \param  ch Kanaele (Bitmuster), die in einen String gewandelt werden sollen
/// \return String mit allen Konfigurierten Kanaelen
//  fse, 29.11.01

string Log::channelsToStr(int ch)
{
	string str = "";
	
	if (ch & ctl) str += "ctl ";
	if (ch & mem) str += "mem ";
	if (ch & gnl) str += "gnl ";
	if (ch & hgh) str += "hgh ";
	if (ch & med) str += "med ";
	if (ch & low) str += "low ";
	if (ch&ctl && ch&mem && ch&gnl && ch&hgh && ch&med && ch&low) str = "all ";
	if (ch & cns) str += "cns ";
	str = str.substr(0, str.size()-1); // letztes Leerzeichen wieder weg
	return str;
}

//******************************************************************************

/// Wandelt String str in channels um.
/// \param  str String mit den Kanalbezeichnungen
/// \return Integer mit den Kanaelen (Bitmuster)
//  fse, 28.11.01

int Log::strToChannels(const string& str)
{
	 int channels = 0;
	
	if (str.find("ctl") != string::npos) channels |= ctl;
	if (str.find("mem") != string::npos) channels |= mem;
	if (str.find("gnl") != string::npos) channels |= gnl;
	if (str.find("hgh") != string::npos) channels |= hgh;
	if (str.find("med") != string::npos) channels |= med;
	if (str.find("low") != string::npos) channels |= low;
	if (str.find("cns") != string::npos) channels |= cns;
	if (str.find("all") != string::npos) channels |= all;
	return channels;
}

//******************************************************************************
// Ausgabeoperatoren: (ersetzt durch Template in logging.h)
//******************************************************************************

//Log& Log::operator<<(const char*        out) { if (mFstreamP) *mFstreamP << out; if (mConfChannels & cns) cout << out; return *this; }
//Log& Log::operator<<(const string&      out) { if (mFstreamP) *mFstreamP << out; if (mConfChannels & cns) cout << out; return *this; }
//Log& Log::operator<<(string&            out) { if (mFstreamP) *mFstreamP << out; if (mConfChannels & cns) cout << out; return *this; }
//Log& Log::operator<<(const int          out) { if (mFstreamP) *mFstreamP << out; if (mConfChannels & cns) cout << out; return *this; }
//Log& Log::operator<<(const float        out) { if (mFstreamP) *mFstreamP << out; if (mConfChannels & cns) cout << out; return *this; }
//Log& Log::operator<<(const double       out) { if (mFstreamP) *mFstreamP << out; if (mConfChannels & cns) cout << out; return *this; }
//Log& Log::operator<<(const void*        out) { if (mFstreamP) *mFstreamP << out; if (mConfChannels & cns) cout << out; return *this; }
//Log& Log::operator<<(const unsigned int out) { if (mFstreamP) *mFstreamP << out; if (mConfChannels & cns) cout << out; return *this; }
//Log& Log::operator<<(Symbol&            out) { if (mFstreamP) *mFstreamP << out; if (mConfChannels & cns) cout << out; return *this; }
//Log& Log::operator<<(Computer&          out) { if (mFstreamP) *mFstreamP << out; if (mConfChannels & cns) cout << out; return *this; }
//Log& Log::operator<<(CalcDisplay&       out) { if (mFstreamP) *mFstreamP << out; if (mConfChannels & cns) cout << out; return *this; }
//Log& Log::operator<<(funcp); // Implementierung mit Kommentar oben
// geht leider nicht mit VC++, nur mit GNU (s. Kommentar bei Node::new()):
//#ifdef _MSC_VER
//Log& Log::operator<<(_Smanip<int>       out) { if (mFstreamP) *mFstreamP << out; if (mConfChannels & cns) cout << out; return *this; }
//#else
//Log& Log::operator<<(smanip<int>        out) { if (mFstreamP) *mFstreamP << out; if (mConfChannels & cns) cout << out; return *this; }
//#endif

//******************************************************************************
//******************************************************************************
//************************************** ***************************************
//************************* ************************* **************************
//****************** ******************* ******************* *******************
//************** *************** *************** *************** ***************
//************ ************ ************ ************ ************ *************
