//******************************************************************************
// Copyright (c) 2002-2004 by Friedemann Seebass, Germany.
// Dieses Programm ist freie Software. Siehe GNU GPL Notiz in Datei main.cpp.
// This program is free software. See the GNU GPL notice in file main.cpp.
// Projekt: Almetare - Alle meine Taschenrechner
//******************************************************************************
/// \file logging.h
///       Klasse zum Loggen von Daten in drei verschiedene Dateien (Modul lib)
//******************************************************************************
// Wann       Wer      Was
// ---------- -------- ---------------------------------------------------------
// 20.03.2003 fse      dbgLog->gDbgLog, errLog->gErrLog, onlLog->gOnlLog
// 28.03.2002 fse      Alle speziellen Ausgabeoperatoren entfernt und durch
//                     Template ersetzt. 
// 29.03.2002 fse      Ausgabeoperatoren als Template
// 29.11.2001 fse      channelsToStr(), strToChannels() zugefuegt
// 10.11.2001 fse      writeHeader() zugefuegt
// 19.12.2000 fse      erzeugt
//******************************************************************************

#ifndef LOGGING_H
#define LOGGING_H

#ifndef __cplusplus
#error logging.h is only for C++!
#endif

//******************************************************************************

#include <iostream>
#include <string>
#include <fstream>

using namespace std;

//******************************************************************************

#define DBG_LOG(channel) if (gDbgLog.configured(channel, __FILE__, __LINE__)) gDbgLog
#define ERR_LOG(channel) if (gErrLog.configured(channel, __FILE__, __LINE__)) gErrLog
#define ONL_LOG(channel) if (gOnlLog.configured(channel, __FILE__, __LINE__)) gOnlLog
#define NL "\n"

typedef void (*funcp)(void);

//******************************************************************************
// Log-Klasse
//******************************************************************************

enum
{
	ctl = 1,  ///< "Control" - Kontrollfluss
	mem = 2,  ///< "Memory"  - Speicherverwaltung
	gnl = 4,  ///< "General" - allgemeine Ausgaben
	hgh = 8,  ///< "High"    - wichtige Ausgaben
	med = 16, ///< "Medium"  - normale Ausgaben
	low = 32, ///< "Low"     - unwichtige Ausgaben
	all = ctl + mem + gnl + hgh + med + low,
	cns = 64  ///< "Console" - Ausgaben zusaetzlich auf Konsole
};

//******************************************************************************

/// Klasse zum Loggen von Laufzeitausgaben in eine Datei.
/// Ein Objekt der Klasse Log erlaubt das Loggen von beliebigen Datentypen in
/// eine Datei und/oder die Konsole unter Verwendung von deren Ausgabeoperatoren
/// und der von cout bekannten "<<"-Syntax. Uhrzeit und Zeilennummern werden der
/// Ausgabe automatisch vorangestellt. Z.B.:
/// \code
/// ------------------------------------------------------------------------------
/// Casio fx-85v (V.1.0 beta) logfile 'log.txt' opened at Sat Sep 28 2002
/// ------------------------------------------------------------------------------
/// 21:12:01: calc_dlg_base.c?,  503, gnl: button '5' mapped to symbol '5'
/// 21:12:02: calc_dlg_base.c?,  503, gnl: button '2' mapped to symbol '2'
/// 21:12:02: calc_dlg_base.c?,  503, gnl: button 'Mul' mapped to symbol '*'
/// ...
/// \endcode
/// Die Groesse der Datei wird ueberwacht; bei Erreichen der Maximalgroesse wird
/// die Datei geschlossen, in "*.bak" umbenannt und eine neue Datei desselben
/// Namens geoeffnet.
///
/// Die Logausgaben werden im Quelltext hart in einen "Kanal" geleitet ("hgh",
/// "gnl", "low" etc.), die Ausgaben werden von der Klasse nur dann gelogged,
/// wenn dieser Kanal auch eingeschaltet ist; Ein- und Ausschalten von Kanaelen
/// ist bei der Initialisierung oder dynamisch zur Laufzeit moeglich.
/// Drei Log-Objekte werden als globale Variablen in logging.cpp definiert und 
/// durch die Makros ONL_LOG, DBG_LOG und ERR_LOG gekapselt. Diese drei
/// Log-Objekte stehen also in der ganzen Applikation ohne jede weitere
/// Initialisierung zur Verfuegung.
///
/// Weitere Log-Objekte koennen nicht ohne weiteres hinzugefuegt werden, weil
/// sie wahlweise auch alle mit derselben Datei verknuepft werden koennen. Daher
/// muessen sie in der richtigen Reihenfolge initialisiert und die Dateien
/// geschlossen werden, was in den Funktionen init() und stopLogging() durch
/// Abfrage genau dieser drei Objekte geschieht.
///
/// Beispiele:
///
/// Stack<double> st;
///
/// ERR_LOG(hgh) << "Stack empty!" << NL;
///
/// ONL_LOG(med) << st << NL;
///
/// DBG_LOG(gnl) << 3.14 << NL;
//  fse, 19.12.02

class Log
{
	string    mFname;        ///< Name d. Datei, in die geloggt wird ("" = Konsole)
	ofstream* mFstreamP;     ///< Filestream der Datei
	int       mConfChannels; ///< Kanaele, die konfiguriert sind (z. B. per Ini-Datei)
	string    mFileToLog;    ///< Name der Quelltextdatei, die als einzige geloggt wird
	int       mMaxLogLen;    ///< maximale Laenge der Log-Dateien in Byte
	bool      mIsSuppressed; ///< Ausgabe unterdrueckt
	string    mProgName;     ///< Name des Programms (fuer Titelzeile der Log-Datei)
	string    mVersion;      ///< Version des Programms (fuer Titelzeile der Log-Datei)
	void writeHeader();
	int backupLogFile();
public:
	Log(const string& fname    = "",
	    int   channels = hgh | cns | med,
	    int   maxLen   = 50000,
	    const string& toLog    = "",
	    const string& progName = "Unknown",
	    const string& version = "Vx.x");
	~Log();
	int init(const string& fname,
	         int channels,
	         int maxLogLen,
	         const string& fileToLog,
	         const string& progName,
	         const string& version);
	ofstream* getFstreamP() { return mFstreamP; }
	int       getConfChannels() const { return mConfChannels; }
	void   stopLogging();
	void   setConfChannels(int chans) { mConfChannels = chans; }
	static string getDate();
	static string getTime();
	static string channelsToStr(int channels);
	static int strToChannels(const string& str);
	bool configured(int channels, const char* file, int line);
	ios::fmtflags setf(ios::fmtflags format) { return mFstreamP->setf(format); }
	// Ausgabeoperatoren:
//	Log& operator<<(const char*        out);
//	Log& operator<<(const string&      out);
//	Log& operator<<(string&            out);
//	Log& operator<<(const int          out);
//	Log& operator<<(const float        out);
//	Log& operator<<(const double       out);
//	Log& operator<<(const void*        out);
//	Log& operator<<(const unsigned int out);
//	Log& operator<<(Symbol&            out);
//	Log& operator<<(Computer&          out);
//	Log& operator<<(CalcDisplay&       out);
//	Log& operator<<(const funcp);
	// geht leider nicht mit VC++, nur mit GNU (s. Kommentar bei Node::new()):
//#ifdef _MSC_VER 
//	Log& operator<<(_Smanip<int>       out);
//#else
//	Log& operator<<(smanip<int>        out);
//#endif
};

//******************************************************************************

/// Globales Funktions-Template. Diese Funktionen Geben alle Datentypen per
/// Ausgabeoperator auf ein Log-Objekt aus. Ersetzt die oben auskommentierten
/// Member-Funktionen.
//  fse, 28.03.02

template<class T> Log& operator<<(Log& log, const T& out)
{
	if (log.getFstreamP()          ) *log.getFstreamP() << out;
	if (log.getConfChannels() & cns) cout               << out;
	return log;
}

//******************************************************************************

extern Log gDbgLog;
extern Log gErrLog;
extern Log gOnlLog;
void logFlush(void);
void errFlush(void);
void onlFlush(void);

#endif // !LOGGING_H
