//******************************************************************************
// Copyright (c) 2002-2005 by Friedemann Seebass, Germany.
// Dieses Programm ist freie Software. Siehe GNU GPL Notiz in Datei main.cpp.
// This program is free software. See the GNU GPL notice in file main.cpp.
// Projekt: Almetare - Alle meine Taschenrechner
//******************************************************************************
/// \file util.cpp
///       Implementierung allgemeiner Funktionen (Modul lib)
//******************************************************************************
// Wann       Wer      Was
// ---------- -------- ---------------------------------------------------------
// 20.01.2005 fse      neue Funktion eraseCrs() in Funktion stripLines()
// 22.12.2004 fse      Funktion getNextSym() aus Klasse Fx3600pPrgDlg hierher
//                     verlegt, da jetzt auch in CasioSciDatDlg benoetigt
// 17.11.2003 fse      strstream durch sstream ersetzt
// 15.03.2003 fse      neue Funktion replace
// 22.04.2002 fse      erzeugt
//******************************************************************************

#include <iostream>
#include <sstream>
#include <iomanip>   // setprecision()
#include <math.h>    // pow()
#include "util.h"

//******************************************************************************
// Klasse Util: Allgemeine Funktionen:
//******************************************************************************

/// Wandelt String in Kleinbuchstaben um.
/// \param  str String, der umgewandelt werden soll
/// \return umgewandelter String
//  fse, 22.04.02

string Util::toLower(const string& str)
{
	string ret = "";
	
	for (unsigned int i = 0; i < str.length(); i++)
		ret += tolower(str[i]);
	return ret;
}

//******************************************************************************

/// Wandelt val in einen String im Sexagesimalformat um:
/// z.B. 6030'20" <=> 60.505556
/// \param  val Wert, der umgewandelt werden soll
/// \return Ergebnis-String
//  fse, 13.07.02
//  fse, 29.05.05: Rundungsproblem: Eingabe von 220 wird flschlich zu 2160
//                 umgewandelt, da rem zu 1.9999999999999929 statt 2 wird. Daher
//                 wird ab sofort rem durch Umwandeln in einen String auf die
//                 standardmaessigen 6 Nachkommastellen gerundet.

string Util::doubleToSxgStr(double val)
{
	string sign = val >= 0 ? "" : "-";
	string remStr;
	int deg, min;
	double sec, rem; // remainder
	ostringstream strm, remStrm;
	string str;

	val = fabs(val);
	deg = (int)floor(val);
	rem = (val-deg)*60.0;
	remStrm.setf(ios::fixed);   // runden auf 6 Nachkommastellen durch
	remStrm << rem;             // Umwandeln in String
	remStr = remStrm.str();     //
	rem = atof(remStr.c_str()); // zurueck in double
	min = (int)floor(rem);
	sec = (rem-min)*60.0;
	sec *= 100.0;               // runden auf 2 Nachkommastellen
	sec = floor(sec);           //
	sec /= 100.0;               //
	strm << sign << deg << "" << min << "'";
	strm.setf(ios::fixed);
	strm << sec;
	str = strm.str();
	return str;
}

//******************************************************************************

/// Wandelt double val mit Formatierung in einen String.
/// \param val      der umzuwandelnde Wert
/// \param format   Format, in das umgewandelt werden soll:
///                 - format = 0: Standardformat
///                 - format = 1: wissenschaftliches Format
///                 - format = 2: fixes Format
/// \param digits   Anzahl der Nachkommastellen (fix) oder der Nachkommastellen der
///                 Mantisse (sci)
/// \param shiftNum Anzahl der Stellen, um die das Komma der Mantisse einer Zahl
///                 im wissenschaftlichen Format nach rechts (>0) oder links (<0)
///                 verschoben wird.
/// \param forceSci Extra fuer den Casio fx-85v: er hat zwei Modi, wie kleine
///                 Zahlen dargestellt werden: Ab einer bestimmtem "Kleinheit"
///                 (<0.01 oder <0.0000001) wird im Standardformat zwangsweise
///                 in die SCI-Darstellung gewechselt. Diese Grenze kann man in
///                 forceSci uebergeben.
/// \return         Ergebnis-String
//  fse, 14.04.02

string Util::doubleToStr(double val, int format, int digits, int shiftNum, double forceSci)
{
	string result;

	if (shiftNum == 0)
	{
		ostringstream str;
		if (format == 0 && val != 0.0 && forceSci > 0.0)
		{ // wenn Zahl kleiner forceSci, SCI erzwingen, sonst FIX erzwingen:
			if (fabs(val) < forceSci)
				format = 1;
			else
				format = 2;
		}
		if (format == 2 && fabs(val) > 1.0e21)
		{ // Fixed-Format nur bis zu bestimmter Groesse (sonst String zu lang):
			format = 0;
			digits = 15;
		}
		if (digits >= 0)
			str.precision(digits);
		else
			str.precision(15);
		switch(format)
		{
			case 0: break;
			case 1: str.setf(ios::scientific); break;
			case 2: str.setf(ios::fixed     ); break;
			default: ;
		}
		str << val;// << ends;
		result =  str.str();
	}
	else
	{
		result = shiftPoint(val, digits, shiftNum);
	}
	if (digits < 0)
		result = stripTrailingOs(result);
	return result;
}

//************************************** ***************************************

/// Erzeugt einen String in Exponentialschreibweise mit der Zahl, die in num
/// uebergeben wird.
/// \param num      die Zahl, deren Komma verschoben werden soll
/// \param digits   Anzahl der Nachkommastellen (fix) oder der Nachkommastellen der
///                 Mantisse (sci)
/// \param shiftNum Anzahl der Stellen, um die das Komma der Mantisse einer Zahl
///                 im wissenschaftlichen Format nach rechts (>0) oder links(<0)
///                 verschoben wird.
/// \return         Ergebnis-String
//  fse, 28.10.01

string Util::shiftPoint(double num, int digits, int shiftNum)
{
	ostringstream numStrStr, strstr, strstr2;
	string numStr, mnt, exp;
	string::size_type ePos;
	double mantisse;
	int exponent;

//	numStrStr << ios::scientific << setprecision(15) << num << ends; // geht nicht mit GNU!?
	numStrStr.setf(ios::scientific); numStrStr << setprecision(15) << num; // << ends;
	numStr = numStrStr.str();
	ePos = numStr.find('e');
	mnt = numStr.substr(0, ePos);
	exp = numStr.substr(ePos+1);
	mantisse = atof(mnt.c_str());
	exponent = atoi(exp.c_str());
	mantisse *= pow(10.0, shiftNum);
	exponent -= shiftNum;
	strstr2.setf(ios::fixed);
	if (digits >= 0)
		strstr2 << setprecision(digits);
	else
		strstr2 << setprecision(15);
	strstr2 << mantisse; // << ends;
	mnt = strstr2.str();
	if (digits >= 0) mnt = formatFixNumStr(mnt, digits);
	strstr << mnt << "e";
	if (exponent >= 0) strstr << "+" << setw(3); else strstr << setw(4);
	strstr.setf(ios::internal); strstr << setfill('0') << exponent; // << ends;
	return strstr.str();
}

//************************************** ***************************************

/// Erwartet in numStr den String einer Zahl in SCI- oder FIX-Notation und
/// entfernt endstaendige Nachkomma-Nullen der Mantisse.
/// \param  numStr String mit der Zahl, deren Nullen entfernt werden sollen
/// \return        Ergebnis-String
//  fse, 21.07.02

string Util::stripTrailingOs(const string& numStr)
{
	string mnt, exp;
	bool isSci   = getMantExp(numStr, mnt, exp);
	bool isPoint = mnt.find('.') != string::npos;
	int last     = (int)mnt.size() - 1;

	if (isPoint)
	{
		while (mnt[last] == '0')
		{
			(void)mnt.replace(last,1,"");
			last = (int)mnt.size() - 1;
		}
	}
	if (isSci)
		return mnt + "e" + exp;
	else
		return mnt;
}

//************************* ************************* **************************

/// Extrahiert aus dem String numStr die Mantisse mnt und den Exponenten exp
/// als Strings.
/// Wenn ein Exponent vorhanden ist, wird true zurueckgegeben.
/// Beispiele: 
/// \code
/// - "2.345e+002"   => mnt = "2.345"    , exp = "+002"
/// - "3.4567"       => mnt = "3.4567"   , exp = ""
/// - "123.45000"    => mnt = "123.45000", exp = ""
/// - "-23.400e-032" => mnt = "-23.400"  , exp = "-032"
/// \endcode
/// \param  numStr der String, aus dem Mantisse und Exponent extrahiert werden
///                sollen
/// \retval mnt    String mit der Mantisse
/// \retval exp    String mit dem Exponenten
/// \return        true, wenn numStr tatsaechlich in Sci-Notation ist,
///                sonst false
// fse, 21.07.02

bool Util::getMantExp(const string& numStr, string& mnt, string& exp)
{
	string::size_type ePos  = numStr.find('e');
	bool              isSci = ePos != string::npos;

	if (isSci)
	{
		mnt = numStr.substr(0, ePos);
		exp = numStr.substr(ePos+1 );
	}
	else
	{
		mnt = numStr;
		exp = "";
	}
	return isSci;
}

//******************************************************************************



/// Erwartet in numStr den String einer Zahl in Fix- oder Sci-Notation (keinen
/// Bruch) und wandelt ihn in eine Zahl in "Engineering"-Notation, d.h., der
/// Exponent ist ein Vielfaches von 3.
/// Je nach \c dir wird das Komma der Mantisse dabei nach rechts oder links
/// verschoben.
/// Gesteuert wird das Ganze durch den statischen Zaehler \c eng, der zunaechst 0
/// sein muss. Jeder Druck auf ENG inkrementiert ihn, jeder Druck auf ENG<
/// dekrementiert ihn, die 0 selbst wird immer uebersprungen.
/// Problematisch ist, dass je nach dem, wie die Zahl im "Urzustand"
/// aussieht, verschieden verfahren werden muss (in Klammern: eng, shiftNum):
/// \code 
/// Urzustand  1. ENG        2. ENG           1. ENG<           2. ENG<
/// 3.1        3.1+00 (1,0)    3100-03 (2,3)   3.1+00 ( 1, 0)   0.0031+03 (-2,-3)
/// 3.1                                        3.1+00 (-1, 0)   0.0031+03 (-2,-3)
/// 3.1-01     310-03 (1,2)  310000-06 (2,5)   310-03 ( 1, 2)     0.31+00 ( 0,-1)
/// 3.1-01                                    0.31+00 (-1,-1)  0.00031+03 (-2,-4)
/// \endcode
/// \param  numStr Eingabe-String
/// \param  dir    Richtung, in die das Komma verschoben werden soll
///                (Util::LEFT oder Util::RIGHT)
/// \param  digits Anzahl der Nachkommastellen (fix) oder der Nachkommastellen der
///                Mantisse (sci)
/// \return        Ergebnis-String
//  fse, 14.07.02

string Util::shiftPointEng(const string& numStr, Direction dir, int digits)
{
	string::size_type ePos      = numStr.find('e');
	bool              isSci     = ePos != string::npos;
	static string     memNumStr = "";
	static int        eng       = 0;
	int               delta     = (dir == Util::LEFT) ? -1 : 1;
	int               shiftNum;
	double            m;
	int               e;
	ostringstream     str;
	static string     lastResult = "";
	string            result;

	// evtl. Initialisierung:
	if (digits == 0)
	{ 
		memNumStr = numStr;
		if (isSci)
		{
			eng = 0;
		}
		else
		{
			m = atof(numStr.c_str());
			m = log10(m);
			m /= 3.0;
			m = m >= 0.0 ? floor(m) : ceil(m);
			eng = -int(m) - 2;
			if (dir == LEFT) ++eng; // das stimmt nur manchmal mit dem Original-Casio ueberein
		}
//		eng = 0; // frueher fuer alle Faelle Null setzen
		return "";
	}
	getMantExp(memNumStr, m, e);
	eng += delta;
	// evtl. Zaehlerkorrektur bei Ueberschreiten der 0:
	if (eng == 0)
	{
		if (e == 0)
			eng += 2*delta; // damit beim Richtungswechsel nicht zweimal e+000
		else
			eng += delta;   // hier kommt e+000 nicht vor
	}
	// Stellenverschiebung ermitteln und Mantisse und Exponent anpassen:
	shiftNum = getEngShiftNum(e, eng);
	m *= pow(10.0, shiftNum);
	e -= shiftNum;
	// Pruefung, ob Ergebnis noch in die Anzeige passt:
	result = formatFixNum(m, digits);
	if (result[0] == '-') ++digits;                 // Vorzeichen und Komma zaeh-
	if (result.find('.') != string::npos) ++digits; // len bei Laenge nicht mit
	if (result.size() > (unsigned int)digits || result == "0." && m != 0.0)
	{ // zu lang:
		eng -= delta;        // Zaehler soll bleiben wo er war
		result = lastResult; // letztes Ergebnis wird zurueckgeliefert
	}
	else
	{ // passt:
		str << result << "e";
		if (e >= 0) str << "+" << setw(3); else str << setw(4);
		str.setf(ios::internal); str << setfill('0') << e; // << ends;
		result = str.str();
	}
	if (m == 0.0) eng = 0; // erlaubt 0.e000, aber verhindert 0.e003 etc.
	lastResult = result;

	return result;
}

//************************************** ***************************************

/// Extrahiert aus dem String numStr die Mantisse m und den Exponenten e
/// als double bzw. int.
/// \param  numStr der String, aus dem Mantisse und Exponent extrahiert werden
///                sollen
/// \retval m      Mantisse als double
/// \retval e      Exponent als double

//  fse, 15.07.02

void Util::getMantExp(const string& numStr, double& m, int& e)
{
	string::size_type ePos      = numStr.find('e');
	bool              isSci     = ePos != string::npos;

	if (isSci)
	{
		m = atof(numStr.substr(0, ePos).c_str());
		e = atoi(numStr.substr(ePos+1 ).c_str());
	}
	else
	{
		m = atof(numStr.c_str());
		e = 0;
	}
}

//************************************** ***************************************

/// Funktion ermittelt die Anzahl der Stellen, um die das Komma einer Zahl
/// in Folge eines Drucks auf die "ENG"- oder "ENG<"-Taste verschoben werden
/// muss.
/// \param e   Exponent einer Zahl
/// \param eng Anzahl der Drcke auf "ENG"- oder "ENG<"-Taste
/// \return    Anzahl der Stellen, um die das Komma verschoben werden muss
//  fse, 15.07.02

int Util::getEngShiftNum(int e, int eng)
{
	int baseShiftNum = 0;
	int shiftNum;
	int signEng = eng > 0 ? 1 : -1;

	if (e != 0)
	{ // wenn Grundzustand schon mit e, kann Shiften sofort beginnen:
		while (e % 3 != 0)
		{ // wenn Grundzustand z.B. e-2, "Basis-Shift" auf -3 um -1,
		  // baseShiftNum ist anders je nach Richtung!:
			e -= signEng;
			baseShiftNum += signEng;
		}
		if (baseShiftNum == 0)
		{
			shiftNum = eng * 3;
		}
		else
		{ 
			if (eng > 0)
				shiftNum = baseShiftNum + (eng-1) * 3;
			else
				shiftNum = baseShiftNum + (eng+1) * 3;
		}
	}
	else

	{ // wenn Grundzustand ohne e, muss erst mal e+000 erscheinen, also
	  // wird bei eng = 1 oder eng = -1 (erster Tastendruck) shiftNum = 0:
		if (eng > 0)
			shiftNum = (eng-1) * 3;
		else
			shiftNum = (eng+1) * 3;
	}
	return shiftNum;

}

//******************************************************************************

/// Formatiert String numStr so, dass er genau len Zeichen lang ist,
/// rechtsbuendig ist und der Exponent - falls vorhanden - mit Vorzeichen eine
/// Laenge von expLen Zeichen hat.
/// Leerstellen werden mit dem Fuellzeichen fillChar gefuellt.
/// \param numStr   Eingabe-String
/// \param len      diese Laenge soll der Ergebnis-String haben
/// \param expLen   diese Laenge soll der Ergebnis-Exponent haben
/// \param fillChar Zeichen mit dem der Ergebnis-String aufgefuellt wird, wenn
///                 er kuerzer als len ist
/// \return         Ergebnis-String
//  fse, 11.07.02

string Util::formatSeven(string& numStr, int len, int expLen, char fillChar)
{
	int               maxExp = (int)pow(10.0, expLen-1) - 1;
	int               mntLen = len - expLen;
	string            mnt(mntLen, fillChar);
	string            exp(expLen, fillChar);
	string::size_type ePos = numStr.find('e');
	string::size_type pPos = numStr.find('.');
	bool              isSci     = ePos != string::npos;

	bool              isNoPoint = pPos == string::npos;
	string            mant, expo;
	int               mantLen, expoLen, iExp;
	ostringstream     str;
	string            result;

	if (isSci)
	{
		mant = numStr.substr(0, ePos);
		if (isNoPoint) mant += "."; // Dezimalpunkt anhngen
		mantLen = mant.size();
		if (mantLen > len - expLen) mant = mant.substr(0, len-expLen);
		expo = numStr.substr(ePos+1 ); expoLen = expo.size();
		iExp = atoi(expo.c_str());
		if (abs(iExp) > maxExp) return "E";
		if (expo[0]=='-') exp[0] = '-'; // Vorzeichen
		expo = expo.substr(expoLen-expLen+1, expLen-1);
		exp.replace(1, expLen-1, expo);
	}
	else
	{
		mant = numStr;
		if (isNoPoint) mant += "."; // Dezimalpunkt anhngen
	}
	mantLen = mant.size();
	mnt.replace(mntLen - mantLen, mantLen, mant);
	result = mnt + exp;

	return result;
}

//******************************************************************************

/// Funktion bringt die Mantisse der Zahl in numStr auf hoechstens digits
/// Stellen, falls sie mehr Stellen hat.
/// Die Zahl in numStr liegt eigentlich bereits in der gewuenschten Form (SCI,
/// FIX, shiftNum, Anzahl Stellen, Anzahl Stellen der Mantisse) vor und wird in
/// der Regel unveraendert wieder zurueckgegeben. Es kann aber
/// sein, dass die Zahl (bzw. die Mantisse) zu viele Stellen hat, um noch in der
/// Anzeige dargestellt werden zu koennen. Dann muss die Zahl (bzw. nur die 
/// Mantisse) auf die passende Stellenanzahl neu gerundet werden. Dabei kann es
/// auch passieren, dass von FIX- in SCI-Notation gewechselt werden muss. In
/// diesem Fall veraendert sich aber die Bedeutung von digits! Denn die Mantisse
/// darf dann natuerlich nur die Zahl von Stellen abzueglich der Stellen fuer den
/// Exponenten haben. Ein Vorzeichen zaehlt bei der Stellenanzahl nicht mit.
/// Ein besonderer Fall besteht auch, wenn die Zahl mit einer Anzahl von Nullen
/// beginnt (z.B. 0.00012345678); entweder wird dann in die SCI-Notation
/// gewechselt (wenn die Zahl kleiner als forceSci ist), oder die Zahl der Stellen
/// muss um die Zahl der fuehrenden Nullen reduziert werden, um den ganzen String
/// noch unterzubringen. Wenn die resultierende Stellenzahl kleiner wird als eins,
/// wird ebenfalls in SCI-Notation gewechselt.
/// \param numStr   Eingabe-String
/// \param digits   Anzahl der Nachkommastellen (fix) oder der Nachkommastellen der
///                 Mantisse (sci)
/// \param forceSci wenn die Zahl kleiner als forceSci ist, wird sie garantiert
///                 in SCI-Notation zurueckgegeben
/// \return         Ergebnis-String
//  fse, 11.07.02

string Util::roundMantisse(string& numStr, int digits, double forceSci)
{
	string::size_type ePos = numStr.find('e');
	string::size_type pPos = numStr.find('.');
	string::size_type aPos = numStr.find("'");
	bool              isSci     = ePos != string::npos;
	bool              isSxg     = aPos != string::npos;
	bool              isNoPoint = pPos == string::npos;
	string            mant;
	int               mantLen;
	unsigned int      i = 0;
	int               wholeLen = pPos;    // Anzahl Ziffern vor dem Komma
	string            result = numStr;
	char              c;
	int               leadOs = 0;
	int               roundedDigits = abs(digits);
//	bool              isNewSci = false;

	if (isSci)
		mant = numStr.substr(0, ePos);
	else
		mant = numStr;
	if (isNoPoint) mant += ".";    // Dezimalpunkt anhngen
	mantLen = mant.size() - 1;     // Dezimalpunkt zaehlt nicht mit
	if (mant[0] == '-')
	{ // Vorzeichen   zaehlt nicht mit
		--mantLen;
		--wholeLen;
	}

	// fuehrende Nullen der Mantissse zaehlen:
	if (digits > 0) // TI-30 uebergibt -digits
	{
		c = mant[i];
		while ((c == '0' || c == '-' || c == '.') && i < mant.size())

		{
			if (c == '0') ++leadOs;
			c = mant[++i];
		}
		if (mantLen > digits)
			roundedDigits -= leadOs; // fuehrende Nullen zaehlen als Stellen mit!
	}

	if (mantLen <= roundedDigits)
		return result; // String passt, alles paletti

	// wenn Mantisse groesser als erlaubte Stellen, runden:
	if (wholeLen > abs(digits))
	{ // wenn mehr Vorkommastellen als digits, SCI erzwingen:
		result = roundNumStr(numStr, abs(digits)-1, true);
	}
	else
	{
		if (isSxg)
			result = roundSxgNumStr(numStr, roundedDigits);
		else
		{
			if (isSci)
			{ // nur Mantisse runden, Exponent wieder anhaengen:
				result = roundNumStr(mant, roundedDigits, false);
				if (result.find('e') != string::npos)
					result = roundNumStr(numStr, roundedDigits, true); // Mantisse braucht alleine schon SCI-Notation
				else
					result += numStr.substr(ePos);
			}
			else
				result = roundNumStr(numStr, abs(digits), false, forceSci);
		}
	}
	result = stripTrailingOs(result);
	return result;
}

//************************************** ***************************************

/// Rundet die Zahl in in numStr auf digits Stellen. Die Zahl kann dabei sowohl
/// in FIX- als auch in SCI-Notation gegeben sein. Wenn isSci true ist, wird
/// die Zahl auf jeden Fall in SCI-Notation zurueckgeliefert; wenn isSci false
/// ist, wird zusaetzlich forceSci ausgewertet:
/// Die Zahl wird in FIX-Notation zurueckgegeben, wenn sie groesser oder gleich
/// forceSci ist, sonst als SCI.
/// kleiner als forceSci ist.
/// \param numStr   Eingabe-String
/// \param digits   Anzahl der Nachkommastellen (fix) oder der Nachkommastellen der
///                 Mantisse (sci)
/// \param isSci    wenn true, wird die Zahl garantiert in SCI-Notation
///                 zurueckgegeben
/// \param forceSci wenn die Zahl kleiner als forceSci ist, wird sie garantiert
///                 in SCI-Notation zurueckgegeben
/// \return         Ergebnis-String
//  fse, 13.07.02

string Util::roundNumStr(string& numStr, int digits, bool isSci, double forceSci)
{
	string::size_type ePos = numStr.find('e');
	string::size_type pPos = numStr.find('.');
	double            num = atof(numStr.c_str());
	int               precision = digits;
	ostringstream     str;
	string            result;

	isSci = isSci || ePos != string::npos;
	if (isSci)
	{
		str.setf(ios::scientific);
	}
	else
	{
		if (forceSci >= 0.0)
		{
			if (fabs(num) < forceSci)
			{
				isSci = true;
				str.setf(ios::scientific);
			}
			else
			{
				str.setf(ios::fixed);

			}
		}
	}
	precision = digits;
	if (!isSci)
	{
		// noch uebrige Nachkommastellen ermitteln:
		if (num >= 0)
			precision = digits-(int)pPos;
		else 
			precision = digits+1-(int)pPos;
	}
	str.precision(precision);
	str << num; // << ends;

	result = str.str();
	return result;
}

//************************************** ***************************************

/// Rundet die Sexagesimal-Zahl in numStr so, dass sie die Laenge maxLen nicht
/// ueberschreitet (ohne Vorzeichen).
/// Format der Sexagesimalzahl: 1234'56.78 (Grad, Minuten, Sekunden).
/// Die Stellenanzahl der Minuten kann 2 nicht ueberschreiten, die Anzahl der
/// Vorkommastellen der Sekunden ebenfalls nicht.
/// \param numStr Eingabe-String
/// \param maxLen maximale Laenge des Ergebnis-Strings
/// \return       Ergebnis-String
//  fse, 13.07.02

string Util::roundSxgNumStr(string& numStr, int maxLen)
{
	string::size_type dPos = numStr.find('');
	string::size_type mPos = numStr.find("'");
	int               sign = (numStr[0] == '-') ? 1 : 0;
	string            deg = numStr.substr(sign, dPos-sign);
	string            min = numStr.substr(dPos+1, mPos-dPos-1);
	string            sec = numStr.substr(mPos+1);
	double            d = atof(deg.c_str());
	double            m = atof(min.c_str());
	double            s = atof(sec.c_str());
	int               remainingDigits;
	ostringstream     str;
	string            result;

	result = deg;
	if (result.size() > (unsigned int)maxLen - 3)
	{
		double num = d + m/60.0 + s/60.0/60.0;
		str.precision(maxLen);
		str << num; // << ends;
		result = str.str();
		goto ret;
	}
	result = result + "" + min;
	remainingDigits = maxLen -1 - result.size(); // -1 wg. Trennzeichens "'"
	if (remainingDigits < 1) goto ret;

	str.precision(remainingDigits);
	str << s; // << ends;
	result = result + "'" + str.str();
ret:
	if (sign) result = "-" + result;
	return result;
}

//******************************************************************************

/// Funktion verchiebt alle Zeichen des Strings str um ein Zeichen nach links
/// oder rechts, abhaengig von dir. Dabei muss ein neues Zeichen recruit von der
/// entsprechenden Seite nachruecken.
/// Wenn das Zeichen "" ist, wird das auf der anderen Seite herausfallende
/// Zeichen als Nachruecker verwendet.
/// \retval str     zu bearbeitender String, gleichzeitig Ergebnis-String
/// \param  recruit nachrueckendes Zeichen
/// \param  dir     Richtung der Verschiebung (Util::LEFT, Util::RIGHT)
//  fse, 16.03.02

void Util::shiftStr(string& str, string recruit, Direction dir)
{
	string::size_type len = str.size();


	
	if (len == 0) return;
	if (recruit == "")

		if (dir == LEFT) recruit = str[0]; else recruit = str[len-1];
	if (dir == LEFT)
		str = str.substr(1) + recruit;
	else
		str = recruit + str.substr(0, len-1);
}

//******************************************************************************

/// Liefert String mit len Zeichen Laenge zurueck, der num an der Position dir
/// enthaelt. Leerzeichen werden mit fillChar aufgefuellt.
/// \param num      zu trimmende Zahl (int)
/// \param len      Laenge des Ergebnisstrings
/// \param dir      Position der Zahl im String: Util::LEFT, CENTER, RIGHT
/// \param fillChar Zeichen, mit dem aufgefuellt wird
/// \return         Ergebnis-String
// fse, 13.07.02

string Util::trimStr(int num, int len, Direction dir, char fillChar)
{
	ostringstream str;

	str << num; // << ends;
	return trimStr(str.str(), len, dir, fillChar);
}

//******************************************************************************

/// Liefert String mit len Zeichen Laenge zurueck, der str an der Position dir
/// enthaelt. Leerzeichen werden mit fillChar aufgefuellt.
/// \param str      zu trimmender String
/// \param len      Laenge des Ergebnisstrings
/// \param dir      Position des Strings im Ergebnis-String: Util::LEFT,
///                 Util::CENTER, Util::RIGHT
/// \param fillChar Zeichen, mit dem aufgefuellt wird
/// \return         Ergebnis-String
// fse, 13.07.02

string Util::trimStr(string str, int len, Direction dir, char fillChar)
{
	string result(len, fillChar);
	int strLen = str.size();

	if (strLen > len) str.erase(len, strLen - len);
	strLen = str.size();
	switch (dir)
	{
		case LEFT: result.replace(0, strLen, str); break;
		case RIGHT: result.replace(len - strLen, strLen, str); break;
		case CENTER: result.replace(len/2-strLen/2, strLen, str); break;
	}
	return result;
}

//******************************************************************************
/*
// Formatiert den String numStr, der eine Zahl in wissenschaftlicher
// Schreibweise enthalten muss, so, dass die Mantisse eine Genauigkeit von
// digits Stellen hat und der Dezimalpunkt um shiftNum Stellen verschoben ist.
// Der Exponent hat 2 Stellen (+Vorzeichen). Falls er groesser ist, wird "E"
// zurueckgegeben.
// Wenn isOstripped true, werden endstaendige Nullen der Mantisse entfernt.
// fse, 10.07.02

string Util::formatSciNumStr(string& numStr, int digits, int shiftNum, bool isOstripped)
{
	string::size_type ePos = numStr.find('e');
	string mnt = numStr.substr(0, ePos);
	string exp = numStr.substr(ePos+1);
	string result;
	int exponent;

	double mant;
	strstream str, str2;

	if (exp.size() > 3 && exp[1] != '0') return "E";
	exponent = atoi(exp.c_str());
	mant     = atof(mnt.c_str());
	// Dezimalpunkt verschieben:
	if (shiftNum != 0)
	{
		exponent -= shiftNum;
		mant *= pow(10.0, shiftNum);
		str2.setf(ios::fixed);
		str2 << setprecision(15) << mant << ends;
		mnt = str2.str();
	}
	mnt = Util::formatFixNumStr(mnt, digits, 100, isOstripped); // 100: digits Stellen insgesamt
	str << mnt;
	if (exponent >= 0) str << "#" << setw(2); else str << setw(3);
	str.setf(ios::internal); str << setfill('0') << exponent << ends;
	result = str.str();
	return result;
}
*/
//******************************************************************************

/// Formatiert String numStr, der Gleitkommazahl im fixed Format enthaelt, auf
/// digitsAfterPoint Nachkommastellen. Wenn die Gesamtzahl der Stellen dann aber
/// digits ueberschreitet, wird auf digits Stellen insgesamt formatiert (Vor- +
/// Nachkommastellen ohne Komma).
/// Wenn isOstripped true, werden abschliessende Nullen nach dem Komma entfernt.
/// \param numStr           Eingabe-String
/// \param digits           maximale Anzahl der Stellen insgesamt
/// \param digitsAfterPoint Anzahl der Stellen nach dem Komma
/// \param isOstripped      wenn true, werden endstaendige Nullen nach dem Komma
///                         entfernt
/// \return                 Ergebnis-String
//  fse, 31.10.01

string Util::formatFixNumStr(string& numStr, int digits, int digitsAfterPoint, bool isOstripped)
{
	ostringstream     str;
	double            num = atof(numStr.c_str());
	int               precision, last;
	string::size_type pPos = numStr.find('.');
	int               digitsBeforePoint = pPos;
	
	if (num < 0) --digitsBeforePoint; // Vorzeichen zaehlt nicht als Stelle
	precision = digitsAfterPoint;
	if (precision+digitsBeforePoint > digits) precision = digits - digitsBeforePoint;
	// noch uebrige Nachkommastellen ermitteln:
//	if (num >= 0)
//		precision = digits-(int)pPos;
//	else 
//		precision = digits+1-(int)pPos;
	// formatieren:
	str.setf(ios::fixed); str << setprecision(precision) << num; // << ends;
	numStr = str.str();
	if (numStr.find('.') == string::npos) numStr += '.';
	if (isOstripped)
	{ // nachstaendige Nullen entfernen:
		last = (int)numStr.size() - 1;
		while (numStr[(unsigned int)last] == '0')
		{
			(void)numStr.replace(last,1,"");
			last = (int)numStr.size() - 1;
		}
	}
	return numStr;
}

//******************************************************************************

/// Formatiert double num im fixed Format auf insgesamt digits echte Stellen
/// (Vor- + Nachkommastellen ohne Komma) ohne abschliessende Nullen nach dem
/// Komma.
/// \param num    Eingabewert
/// \param digits maximale Anzahl der Stellen insgesamt
/// \return       Ergebnis-String
// fse, 31.10.01

string Util::formatFixNum(double num, int digits)
{
	ostringstream str;
	string numStr;

	// double in String wandeln:
	str.setf(ios::fixed);

	str.precision(digits);
	str << num; // << ends;
	numStr = str.str();
	return Util::formatFixNumStr(numStr, digits);
}

//******************************************************************************

/// Formatiert String numStr, der Gleitkommazahl im fixed Format enthaelt, auf
/// insgesamt digits echte Stellen (Vor- + Nachkommastellen ohne Komma) ohne
/// abschliessende Nullen nach dem Komma.
/// \param numStr           Eingabe-String
/// \param digits           maximale Anzahl der Stellen insgesamt
/// \return                 Ergebnis-String
//  fse, 31.10.01


string Util::formatFixNumStr(string numStr, int digits)
{
	ostringstream str;
	double num = atof(numStr.c_str());
	int precision, last;
	string::size_type pPos = numStr.find('.');

	// noch uebrige Nachkommastellen ermitteln:
	if (num >= 0)
		precision = digits-(int)pPos;
	else 
		precision = digits+1-(int)pPos;
/*
	// fuehrende Nullen der Mantissse zaehlen:
	c = mant[i];
	while ((c == '0' || c == '-' || c == '.') && i < mant.size())
	{
		if (c == '0') ++leadingOs;
		c = mant[++i];
	}

	digits -= leadingOs; // fuehrende Nullen zaehlen als Stellen mit!
*/
	// formatieren:
	if (precision < 0) precision = 0;
	str.setf(ios::fixed); str << setprecision(precision) << num; // << ends;
	numStr = str.str();
	if (numStr.find('.') == string::npos) numStr += '.';
	// nachstaendige Nullen entfernen:
	last = (int)numStr.size() - 1;
	while (numStr[(unsigned int)last] == '0')
	{
		(void)numStr.replace(last,1,"");
		last = (int)numStr.size() - 1;
	}
	return numStr;
}


//******************************************************************************

/// Haengt an numStr gegebenenfalls Dezimalpunkt an und bringt ihn rechtsbuendig
/// auf maxLen Zeichen Laenge.
/// Beispiele mit maxLen = 10:
/// \code
/// "123"          3 => "      123."
/// "0.12345678"  10 => "0.12345678"
/// "-0.12345678" 11 => "-.12345678"
/// "-12345678"    9 => "-12345678."
/// "-12.345678"  10 => "-12.345678"
/// "-0.123"       6 => "    -0.123"
/// "-1.2345-12"  10 => "-1.2345-12"
/// \endcode
/// \param numStr Eingabe-String
/// \param maxLen Laenge des Ergebnis-Strings
/// \return       Ergebnis-String
//  fse, 01.11.01

string Util::trimNumStr(string numStr, int maxLen)
{
	string result(maxLen, '#'); // maxLen Zeichen langer Leerstring
	bool isNotExp = numStr.find('e') == string::npos;
	bool hasNoPnt = numStr.find('.') == string::npos;
	int len;

	if (hasNoPnt && isNotExp) numStr += "."; // Dezimalpunkt anhngen
	len = numStr.size();
	if (len <= maxLen)
	{
		result.replace(maxLen-len, len, numStr);
		return result;
	}
	if (numStr[0] == '-' && numStr[1] == '0') (void)numStr.replace(1, 1, "");
	return numStr;
}

//******************************************************************************

/// Berechnet rekursiv die Fakultaet n! der Zahl n.
/// \param n Eingabezahl
/// \return  Fakultaet der Eingabezahl
//  fse, 15.07.02

double Util::fact(double n)
{
	double res, dummy = -1.0;
	
	if (n < 0.0 || int(n) != n) return log(dummy); // Fehler erzwingen
	if (n == 0.0 || n == 1.0) return 1.0;
	if (n == 2.0) return 2.0;
	res = n * fact(n-1);
	return res;
}

//******************************************************************************

/// Berechnet x^(1/y).
/// Als Besonderheit sind auch die Wurzeln negativer Radikanden erlaubt, sofern
/// y ganzzahlig und ungerade.
/// \param x Radikand
/// \param y Wurzelexponent
/// \return  Wurzel

//  fse, 15.07.02

double Util::root(double x, double y)

{
	if (x < 0.0 && double(int(y)) == y && int(y) % 2 == 1)
		return -pow(fabs(x), 1.0/y);
	else
		return pow(x, 1.0/y);
}

//******************************************************************************

/// Funktion ersetzt im String srcStr jedes Vorkommen des Strings oldStr durch
/// den String newStr und gibt das Ergebnis als Return-Wert zurueck.
/// \param srcStr String in dem gesucht wird
/// \param oldStr Teilstring, nach dem in srcStr gesucht wird
/// \param newStr String, durch den jedes Vorkommen von oldStr ersetzt wird
//  fse, 15.03.03


string Util::replace(const string& srcStr, const string& oldStr, const string& newStr)
{
	string res = srcStr;
	string::size_type len = oldStr.size();
	string::size_type pos = res.find(oldStr);

	while (pos != string::npos)
	{
		res.replace(pos, len, newStr);
		pos = res.find(oldStr, pos);
	}
	return res;
}

//******************************************************************************

/// Entfernt alle Kommentare (Kommentarsymbol ';') und fuehrende
/// Leerzeichen/Tabulatoren einer Zeile aus dem String str und ersetzt alle
/// Zeilenumbrueche durch ein Leerzeichen.
/// Vom String wird angenommen, dass er aus mehreren Zeilen besteht, die durch
/// '\n' voneinander getrennt sind.
/// Die Funktion wird benoetigt, um die Programme des Casio fx-3600P von allen
/// Kommentaren zu befreien und so aufzubereiten, dass das Programm ausgefuehrt
/// werden kann.
/// \param str zu behandelnder String
/// \return der gestrippte String
// fse, 26.04.03

string Util::stripLines(string str)
{
	string line, strippedStr;  
	string::size_type nlPos;

	eraseCrs(str);
	str += "\n";
	nlPos = str.find("\n");
	while (nlPos != string::npos)
	{
		line = str.substr(0, nlPos);
		stripLine(line);
		if (line != "") strippedStr += " " + line;
		str = str.substr(nlPos+1);
		nlPos = str.find("\n");
	}
	if (strippedStr.size() >= 1)
		strippedStr = strippedStr.substr(1); // erstes Leerzeichen wieder entfernen  */
	return strippedStr;
}

//************************************** ***************************************

/// Entfernt aus dem uebergebenen String alle Carriage Returns ("\r").
/// Noetig unter Windows, da dort am Zeilenende "\r\n" steht.
/// \param str zu strippender String
//  fse, 20.01.05

void Util::eraseCrs(string& str)
{
	string::size_type crPos;

	crPos = str.find('\r');
	while (crPos != string::npos)
	{
		str.erase(crPos, 1);
		crPos = str.find('\r', crPos);
	}
	return;
}

//************************************** ***************************************

/// Entfernt aus dem uebergebenen String fuehrende Leer- und Tabulatorzeichen,
/// abschliessende Leerzeichen und Kommentare, die durch ein Kommentarzeichen
/// (';') abgetrennt sind, und ersetzt mehrfache Leerzeichen durch nur eines.
/// \param line zu strippender String
//  fse, 26.04.03

void Util::stripLine(string& line)
{
	string::size_type start, end;
	string::size_type len;

	start = line.find_first_not_of(" \t"); // Leerzeichen ueberlesen
	if (start == string::npos) return;
	end  = line.find_first_of(';', start); // lesen bis zu erstem ';'
	line = line.substr(start, end-start);  // und Kommentar abschneiden
	end  = line.find_last_not_of(' ');     // rueckwaerts bis zu erstem Nicht-Space
	if (end == string::npos) return;
	line = line.substr(0, end+1);          // und nach erstem Nicht-Space abschneiden
	// alle mehrfachen Leerzeichen durch nur eines ersetzen:
	do
	{
		len = line.size();
		line = replace(line, "  ", " ");
	} while (len != line.size());
	if (line == " ")
		line = "";
	return;
}

//************************* ************************* **************************

/// Holt aus dem String expr das erste Symbol.
/// \param expr String mit einer ganzen Tastenfolge (z.B. "3.4 * 3 sin ="
/// \return     String mit dem ersten Symbol

string Util::getNextSym(string& expr)
{
	string::size_type pos = expr.find(" ");
	string sym(expr, 0, pos);
	string newExpr(expr, pos+1, string::npos);

	expr = newExpr;
	if (pos == string::npos) expr = "";
	return sym;
}

//******************************************************************************
/*
// Formatiert den String numStr, der eine Zahl in normaler oder
// Exponentialschreibweise enthaelt, so, dass die Mantisse nicht laenger als
// mntDigits Zeichen ist (ohne Vorzeichen). Ist die Mantisse laenger, wird die
// Zahl in Exponentialdarstellung umgewandelt.
// Folgende Faelle sind moeglich:
// - "1.23456789012" => runden auf mntLen Stellen
// - "1234567890.12" => Zahl in Exponentialdarstellung wandeln
// Ist die Zahl in Exponentialdarstellung, wird der Exponent mit fuehrenden
// Nullen auf expLen Zeichen gebracht (ohne Vorzeichen); wenn nicht, werden
// expLen+1 Leerzeichen angehaengt.
// fse, 11.07.02

string Util::formatNumStr(string& numStr, int mntDigits, int expDigits)
{
	string::size_type ePos = numStr.find('e');
	bool   isSci = ePos != string::npos;
	string mnt = numStr;
	string exp = "";
	string result;
	double num = atof(numStr.c_str());
	strstream str;

	if (isSci)
	{

		mnt = numStr.substr(0, ePos);
		exp = numStr.substr(ePos+1);
	}
	if (mnt.size() <= mntDigits) return numStr;

	str.precision(mntDigits);
	str << num << ends;
	return str.str();
}
*/
//******************************************************************************
/*
string Util::formatFixNumStr(string& numStr, int digits)
{
	strstream str;
	double num = atof(numStr.c_str());
	int precision, last;
	string::size_type pPos = numStr.find('.');


	str << setprecision(digits) << num << ends;
	numStr = str.str();
	if (numStr.find('.') == string::npos) numStr += '.';
	if (isOstripped)
	{ // nachstaendige Nullen entfernen:
		last = (int)numStr.size() - 1;
		while (numStr[(unsigned int)last] == '0')
		{
			(void)numStr.replace(last,1,"");
			last = (int)numStr.size() - 1;
		}
	}
	return numStr;
}
*/
//******************************************************************************
//************************************** ***************************************
//************************* ************************* **************************
//****************** ******************* ******************* *******************
//************** *************** *************** *************** ***************
//************ ************ ************ ************ ************ *************
