//******************************************************************************
// Copyright (c) 2002-2004 by Friedemann Seebass, Germany.
// Dieses Programm ist freie Software. Siehe GNU GPL Notiz in Datei main.cpp.
// This program is free software. See the GNU GPL notice in file main.cpp.
// Projekt: Almetare - Alle meine Taschenrechner
//******************************************************************************
/// \file calc_casio_sci.cpp
///       Gemeinsame Klassen fuer wissenschaftliche Casio-Rechner
///       (Modul special).
///       In dieser Datei befinden sich die Implementierungen der Klassen, die
///       alle wissenschaftlichen Casio-Taschenrechner benoetigen.
//******************************************************************************
// Wann       Wer      Was
// ---------- -------- ---------------------------------------------------------
// 11.06.2004 fse      ComputerCasioSci::ComputerCasioSci(): Member-Initialis.
// 17.11.2003 fse      strstream durch sstream ersetzt
// 02.08.2003 fse      ComputerCasioSci::handleEmptyNumStack(): Behandlung
//                     weiterer Operatoren
// 01.05.2003 fse      checkConstMode(): Pruefung auch des vorletzten Operators,
//                     falls der letzte INV war.
//                     mapInvCasioSci(): sum nach ComputerFx85v::map()
//                     verschoben
// 27.04.2003 fse      Fehler beim Rechnen mit Sexagesimalzahlen behoben.
// 17.03.2003 fse      Trennzeichen '#' bei Parametern durch '@' ersetzt. Siehe
//                     Kommentar in calculator.cpp.
// 16.03.2003 fse      mapInvCasioSci(): rcp, sqr und sqrt nach
//                     ComputerFx85v::map() verschoben
// 09.04.2002 fse      erzeugt
//******************************************************************************

#include <sstream>
#include <iomanip>   // setw()
#include "calc_casio_sci.h"
#include "util.h"    // formatSciNumStr()

//******************************************************************************
// Klasse DisplayCasioSci:
//******************************************************************************

/// Konstruktor der Casio-Display-Klasse.
/// \param cbP Zeiger zum Computer, entweder fx-85v oder fx-3600p
   
DisplayCasioSci::DisplayCasioSci(ComputerBase* cbP)
	: DisplaySci(cbP), mSd(".."), mLr(".."), mMode(".")
{
	mCompP = static_cast<ComputerCasioSci*>(DisplayBase::mCompP);
	setModesCasioSci();
	makeSeven();
}

//************************************** ***************************************

/// Schaut im Computer, welche Modes gesetzt sind, und baut den Mode-String des
/// Displays entsprechend zusammen. Entspricht Casio fx-3600P.
//  fse, 12.04.02

void DisplayCasioSci::setModesCasioSci()
{
	CalcMode cm    = mCompP->getCalcMode();
	bool     isHyp = mCompP->getIsHyp   ();

	setModesSci();
	setHyp(isHyp);
	setCalcMode(cm);
	mModes =  mInv + " " + mMem  + " " + mK   + " " + mHyp + " "
	        + mLr  + " " + mSd   + " "
		     + mDeg + " " + mRad  + " " + mGra + " ";
	mModesShort = mInv + mMem + mHyp[0] + mK + mSd[0] + mLr[0] + mDeg[0] + mRad[0] + mGra[0];
}

//************************************** ***************************************

/// Setzt den Modus, in dem sich der Rechner jetzt befinden soll. Es gibt den
/// normalen Modus, Statistikmodus und Regressionsmodus.
/// \param cm der zu setzende Modus: CMP, SD oder LR

void DisplayCasioSci::setCalcMode(CalcMode cm)
{
	switch(cm)
	{
		case CMP: setCmp(); break;
		case SD : setSd (); break;
		case LR : setLr(); break;
		default: setCmp();
	}
}

//******************************************************************************
// Klasse ComputerCasioSci:
//******************************************************************************

/// Konstruktor fuegt die neuen Symbole zur Symboltabelle hinzu.
//  fse, 11.06.04: Initialisierung der Members - frueher vergessen :-(
//                 das fuehrte erst unter Knoppix 3.4 zu fehlerhaftem Setzen
//                 der Werte z.B. in CalculatorFx3600p::restoreNvRam(), da dort
//                 oefters kein else-Zweig.

ComputerCasioSci::ComputerCasioSci()
: mDisplayCasioP(NULL),
  mIsMode(false), mIsKinInput(false), mIsKoutInput(false), mIsXkInput(false),
  mCalcMode(CMP)
{
	for (int i = 0; i < 6; ++i)
		mKmem[i] = 0;
	addCasioSymsToSymTab();
}

//************************************** ***************************************

/// Fuegt die Symbole der wissenschaftlichen Casio-Rechner der Symboltabelle
/// hinzu.

void ComputerCasioSci::addCasioSymsToSymTab()
{	
	string sym;
	Symbol* symP;

	symP = new AcMode ; sym = symP->getSym(); mSymTab[sym] = symP;	
	symP = new AcHyp  ; sym = symP->getSym(); mSymTab[sym] = symP;	
	symP = new AcKin  ; sym = symP->getSym(); mSymTab[sym] = symP;	
	symP = new AcKout ; sym = symP->getSym(); mSymTab[sym] = symP;	
	symP = new AcKac  ; sym = symP->getSym(); mSymTab[sym] = symP;	
	symP = new AcXk   ; sym = symP->getSym(); mSymTab[sym] = symP;	
	symP = new AcXy   ; sym = symP->getSym(); mSymTab[sym] = symP;	
	symP = new AcCmp  ; sym = symP->getSym(); mSymTab[sym] = symP;	
	symP = new AcSd   ; sym = symP->getSym(); mSymTab[sym] = symP;	
	symP = new AcLr   ; sym = symP->getSym(); mSymTab[sym] = symP;	
	symP = new AcXd   ; sym = symP->getSym(); mSymTab[sym] = symP;	
	symP = new AcData ; sym = symP->getSym(); mSymTab[sym] = symP;	
	symP = new AcDel  ; sym = symP->getSym(); mSymTab[sym] = symP;	
	symP = new AcXave ; sym = symP->getSym(); mSymTab[sym] = symP;	
	symP = new AcYave ; sym = symP->getSym(); mSymTab[sym] = symP;	
	symP = new AcXsgm ; sym = symP->getSym(); mSymTab[sym] = symP;	
	symP = new AcYsgm ; sym = symP->getSym(); mSymTab[sym] = symP;	
	symP = new AcXsgm1; sym = symP->getSym(); mSymTab[sym] = symP;	
	symP = new AcYsgm1; sym = symP->getSym(); mSymTab[sym] = symP;	
	symP = new AcA    ; sym = symP->getSym(); mSymTab[sym] = symP;	
	symP = new AcB    ; sym = symP->getSym(); mSymTab[sym] = symP;	
	symP = new AcR    ; sym = symP->getSym(); mSymTab[sym] = symP;	
	symP = new AcX    ; sym = symP->getSym(); mSymTab[sym] = symP;	
	symP = new AcY    ; sym = symP->getSym(); mSymTab[sym] = symP;	
	symP = new AcRan  ; sym = symP->getSym(); mSymTab[sym] = symP;	
	symP = new AcIsxg ; sym = symP->getSym(); mSymTab[sym] = symP;	
	symP = new AcSxg  ; sym = symP->getSym(); mSymTab[sym] = symP;	
	symP = new FnFact ; sym = symP->getSym(); mSymTab[sym] = symP;	
	symP = new FnRnd  ; sym = symP->getSym(); mSymTab[sym] = symP;	
	symP = new OpRp   ; sym = symP->getSym(); mSymTab[sym] = symP;	
	symP = new OpPr   ; sym = symP->getSym(); mSymTab[sym] = symP;	
}

//******************************************************************************

/// Mapped einige Symbole auf andere Symbole um.
/// Das ist z.B. noetig, wenn die Invers-Taste gedrueckt wurde (sin wird zu
/// asin).
/// \retval sym zu mappendes Symbol, enthaelt bei der Rueckkehr das neue Symbol.
/// \return true, wenn Mapping stattgefunden hat, sonst false
//  fse, 16.03.03: Mapping "sum" auf "data" nach

bool ComputerCasioSci::mapCasioSci(string& sym)
{
	bool isMapped = mapSci(sym);

	if (mIsInv ) isMapped = mapInvCasioSci(sym);
	if (mIsMode) isMapped = mapModeCasioSci(sym);
	if (mIsKinInput || mIsKoutInput || mIsXkInput)
		isMapped = mapKinKoutInputCasioSci(sym);
//	if (mCalcMode != CMP)
//	{ // SD- oder LR-Modus
//		if (sym == "sum") { sym = "data"; isMapped = true; }
//	}
	if (mCalcMode == LR)
	{
		if (sym == "("  ) { sym = "xd"  ; isMapped = true; }
		if (sym == ")"  ) { sym = "y"   ; isMapped = true; }
	}
	return isMapped;
}

//************************************** ***************************************

/// Fuehrt spezielles Mapping von Ziffern oder "+", "-", "*", "/" aus, falls
/// zuvor "Kin" oder "Kout" eingegeben wurde.
/// "Kin", "Kout" oder "X<->K" wurden ohne Parameter eingegeben. Die naechste
/// Ziffer muss nun als Parameter interpretiert werden.
/// Es gibt noch eine Sonderfunktion der Kin-Taste: Mit der Tastenkombination
/// "Kin + 2" wird z.B. der aktuelle Anzeigewert zum Speicher 2 addiert.
/// Der Zusammenbau des endgueltigen Symbols ist also dreistufig.
/// Um das Konzept zu erhalten, dass ein Symbol immer nur einen numerischen
/// Parameter erhalten kann, wird folgende Vereinbarung getroffen:
/// Die Symbole "kin@1", "kin@2", ..., "kin@6" bedeuten: Speichern in entspr.
/// Speicher.                                                          \n
/// "kin@431": addiere in Speicher 1                                   \n
/// "kin@432": addiere in Speicher 2                                   \n
/// ...                                                                \n
/// "kin@436": addiere in Speicher 6       (43 ist ASCII-Wert von '+') \n
///                                                                    \n
/// "kin@451": subtrahiere in Speicher 1                               \n
/// "kin@452": subtrahiere in Speicher 2                               \n
/// ...                                                                \n
/// "kin@456": subtrahiere in Speicher 6   (45 ist ASCII-Wert von '-') \n
///                                                                    \n
/// "kin@421": multipliziere in Speicher 1                             \n
/// "kin@422": multipliziere in Speicher 2                             \n
/// ...                                                                \n
/// "kin@426": multipliziere in Speicher 6 (42 ist ASCII-Wert von '*') \n
///                                                                    \n
/// "kin@471": dividiere in Speicher 1                                 \n
/// "kin@472": dividiere in Speicher 2                                 \n
/// ...                                                                \n
/// "kin@476": dividiere in Speicher 6     (47 ist ASCII-Wert von '/') \n
///                                                                    \n
/// Der Parameter wird in der process()-Funktion des Symbols Kin entsprechend
/// ausgewertet.
/// \retval sym Ziffer oder "+", "-", "*", "/", bei Rueckkehr neues Symbol
/// \return true, wenn Mapping stattgefunden hat, sonst false
//  fse, 06.07.02
//  fse, 17.03.03: Trennzeichen '#' durch '@' ersetzt

bool ComputerCasioSci::mapKinKoutInputCasioSci(string& sym)
{
	char          c = sym[0];
	bool          ret = false;
	static string catSym = "";
	ostringstream str;

	if (sym == "processed") return true;
	if (sym.size() != 1) goto lbl;
	if (mIsKinInput)
	{
		if (c == '+' || c == '-' || c == '*' || c == '/')
		{
			str << "kin@" << int(c); // << ends;
			catSym = str.str();
			sym = "processed";
			return true;
		}
		else
		{
			if (c >= '0' && c <= '9')
			{
				if (catSym == "")
					sym = "kin@";
				else
					sym = catSym;
				sym += c;
			}
		}
	}
	else
	{
		if (c >= '0' && c <= '9')
		{
			sym = "";
			if (mIsKoutInput) sym = "kout@";
			if (mIsXkInput) sym = "xk@";
			sym += c;
			ret = true;
		}
	}
lbl:
	mIsKinInput = mIsKoutInput = mIsXkInput = false;
	catSym = "";
	return ret;
}

//************************************** ***************************************

/// Ordnet einem Symbol dasjenige Symbol zu, das ihm im Invers-Modus entspricht.
/// \retval sym zu mappendes Symbol, bei Rueckkehr neues Symbol
/// \return true, wenn Mapping stattgefunden hat, sonst false
//  fse, 26.05.02
//  fse, 16.03.03: Mapping von rcp, sqr und sqrt herausgenommen, da dies nicht
//                 fuer fx-3600P gilt.
//  fse, 01.05.03: Mapping von sum auf del ebenfalls herausgenmommen.

bool ComputerCasioSci::mapInvCasioSci(string& sym)
{
	if (mCalcMode == SD || mCalcMode == LR)
	{
		if (sym == "1"  ) { sym = "xave" ; return true; }
		if (sym == "2"  ) { sym = "xsgm" ; return true; }
		if (sym == "3"  ) { sym = "xsgm1"; return true; }
//		if (sym == "sum") { sym = "del"  ; return true; } // fse, 16.03.03: verschoben nach ComputerFx85v::map()
	}
	if (mCalcMode == LR)
	{
		mIsInv = false;
		if (sym == "4"  ) { sym = "yave" ; return true; }
		if (sym == "5"  ) { sym = "ysgm" ; return true; }
		if (sym == "6"  ) { sym = "ysgm1"; return true; }
		if (sym == "7"  ) { sym = "a"    ; return true; }
		if (sym == "8"  ) { sym = "b"    ; return true; }
		if (sym == "9"  ) { sym = "r"    ; return true; }
		if (sym == ")"  ) { sym = "x"    ; return true; }
//		if (sym == "sum") { sym = "del"  ; return true; } // fse, 16.03.03: verschoben nach ComputerFx85v::map()
	}
	if (sym == "ac"  ) { sym = "kac" ; return true; }
	if (sym == "="   ) { sym = "%"   ; return true; }
	if (sym == "sum" ) { sym = "sub" ; return true; }
	if (sym == "0"   ) { sym = "rnd" ; return true; }
	if (sym == "."   ) { sym = "ran" ; return true; }
	if (sym == "isxg") { sym = "sxg" ; return true; }
//	if (sym == "rcp" ) { sym = "fact"; return true; } // fse, 16.03.03: verschoben 
//	if (sym == "sqr" ) { sym = "xk"  ; return true; } // nach ComputerFx85v::map()
//	if (sym == "sqrt") { sym = "xy"  ; return true; } //
	if (sym == "+"   ) { sym = "rp"  ; return true; }
	if (sym == "-"   ) { sym = "pr"  ; return true; }
	return false;
}

//************************************** ***************************************

/// Ordnet einem Symbol dasjenige Symbol zu, das ihm im Mode-Modus entspricht.
/// \retval sym zu mappendes Symbol, bei Rueckkehr neues Symbol
/// \return true, wenn Mapping stattgefunden hat, sonst false
// fse, 06.07.02

bool ComputerCasioSci::mapModeCasioSci(string& sym)
{
	if (sym == "2") { sym = "lr" ; return true; }
	if (sym == "3") { sym = "sd" ; return true; }
	if (sym == "4") { sym = "deg"; return true; }
	if (sym == "5") { sym = "rad"; return true; }
	if (sym == "6") { sym = "gra"; return true; }
	if (sym == "7") { sym = "fix"; return true; }
	if (sym == "8") { sym = "sci"; return true; }
	if (sym == "9") { sym = "nrm"; return true; }
	return false;
}

//******************************************************************************

/// Erfindet einen zweiten Operanden, falls ein Operator mit nur einem berechnet
/// werden soll.
/// Casio fx-85v und fx-3600P verhalten sich so:
/// "3+=" => 3, "3-=" => 3, "3*=" => 9, "3/=" => 1.
/// \param soP Zeiger auf Operator   
/// \retval a   erster Operand
/// \retval b   zweiter Operand
//  fse, 13.04.02
//  fse, 02.08.03: a=b auch bei den Operatoren x^y und x^1/y

void ComputerCasioSci::handleEmptyNumStack(Number& a, SymOperator* soP, Number& b)
{
	if (dynamic_cast<OpSub*>(soP))
		b = 0L - b;
	if (dynamic_cast<OpMul*>(soP) || dynamic_cast<OpDiv* >(soP) ||
		 dynamic_cast<OpPow*>(soP) || dynamic_cast<OpRoot*>(soP))
		a = b;
}

//******************************************************************************

/// Siehe Kommentar ComputerBase::handleDisplayAfterBra().
/// Casio schreibt "0" ins Display nach "(".
//  fse, 13.04.02

void ComputerCasioSci::handleDisplayAfterBra()
{
	pushNewStacks();
	pushNum(0.0); // Casio schreibt nach Klammer "0" ins Display
}

//************************************** ***************************************

/// Untersucht, ob in den Konstanten-Modus gewechselt werden muss oder dieser
/// abgeschaltet werden muss.
///
/// Bedingungen fuers Einschalten:
/// - Es muss zwei Mal unmittelbar hintereinander derselbe Operator eingegeben
///   worden sein.
/// - Es duerfen keine Klammern eingegeben worden sein.
///
/// Bedingung fuers Ausschalten:
/// - Irgendein Operator wurde eingegeben.
///
/// \param opP Zeiger auf den Operator
/// \return falls Konstantenmodus, Zeiger auf den Konstantenoperator, sonst NULL
//  fse, 03.10.01
//  fse, 01.05.03: Auch den vorletzten Operator pruefen, wenn der letzte INV
//                 war; noetig, da auch x^y und x^(1/y) Konstantenoperatoren
//                 sein koennen.

SymOperator* ComputerCasioSci::checkConstMode(SymOperator* opP)
{
	Symbol* lastInP       = getLastInP();
	Symbol* lastButOneInP = getLastButOneInP();

	// falls letzte Taste INV, dann vorletzte als Operator nehmen:
	if (dynamic_cast<AcInv*>(lastInP))
		lastInP = lastButOneInP;
	if (mConstOpP)
	{ // wenn der Konstanten-Mode eingeschaltet, fuehrt neuer Operator zum Abschalten
		mConstOpP = NULL;
	}
	else
	{ // vielleicht einschalten?:
		if (opP == lastInP)
		{ // letzte beiden Operatoren gleich => einschalten:
			mConstVal = topNum();
			mConstOpP = opP; // Konstanten-Operator speichern
		}
	}
	return mConstOpP;
}

//******************************************************************************

/// Berechnet aus den K-Speichern den y-Achsenabschnitt A bei
/// Regressionsrechnung.
/// \return y-Achsenabschnitt A
//  fse, 27.04.02

Number ComputerCasioSci::calcA() const
{
	Number sumX  = mKmem[1]; // K2
	Number n     = mKmem[2]; // K3
	Number sumY  = mKmem[4]; // K5
	Number B     = calcB();

	return (sumY - B*sumX)/n;
}

//******************************************************************************

/// Berechnet Regressionskoeffizient B aus den K-Speichern.
/// \return Regressionskoeffizient B
//  fse, 27.04.02

Number ComputerCasioSci::calcB() const
{
	Number sumX2 = mKmem[0]; // K1
	Number sumX  = mKmem[1]; // K2
	Number n     = mKmem[2]; // K3
//	Number sumY2 = mKmem[3]; // K4
	Number sumY  = mKmem[4]; // K5
	Number sumXY = mKmem[5]; // K6

	return (n*sumXY-sumX*sumY)/(n*sumX2-sumX*sumX);
}

//******************************************************************************

/// Berechnet Korrelationskoeffizient r aus den K-Speichern.
/// \return Korrelationskoeffizient r
//  fse, 27.04.02

Number ComputerCasioSci::calcR() const
{
	Number sumX2 = mKmem[0]; // K1
	Number sumX  = mKmem[1]; // K2
	Number n     = mKmem[2]; // K3
	Number sumY2 = mKmem[3]; // K4
	Number sumY  = mKmem[4]; // K5
	Number sumXY = mKmem[5]; // K6
	Number root  = (n*sumX2-sumX*sumX)*(n*sumY2-sumY*sumY);

	return (n*sumXY-sumX*sumY)/sqrt(root.getVal());
}

//******************************************************************************
//******************************************************************************

/// Prozessiert das Prozentzeichen.
///
/// Prozentrechnung des Casio fx-3600P:
///	 \code
///  1: a* b%: a*b/100
///  2: a**b%: a*b/100
///  3: a/ b%: a*100/b
///  4: a//b%: b*100/a
///  5: a+ b%: (a+b)*100/b
///  6: a++b%: (a+b)*100/a
///  7: a- b%: (a-b)*100/b
///  8: a--b%: (b-a)*100/a
///
///  9: a%   : a
/// 10: a*%  : a*a/100          : wie oben mit b = a
/// 11: a**% : a*a/100          : wie oben mit b = a
/// 12: a/%  : a*100/a = 100    : wie oben mit b = a
/// 13: a//% : a*100/a = 100    : wie oben mit b = a
/// 14: a+%  : a                :
/// 15: a++% : (a+a)*100/a = 200: wie oben mit b = a
/// 16: a-%  : a
/// 17: a--% : (a-a)*100/a = 0  : wie oben mit b = a
/// \endcode
// fse, 05.10.01

void ComputerCasioSci::handlePercent()
{
	SymOperator* constOpP = getConstOpP();
	SymOperator* soP = popSym();
	Number a, b;
	Number val;
	string op;

	if (soP == NULL || isNumEmpty()) return;
	op = soP->getSym();
	b = popNum();
	if (isNumEmpty() || constOpP != NULL)
	{
		if (constOpP == NULL)
		{
			if (op == "+" || op == "-")
			{ // Faelle 14 und 16:
				pushNum(b);
				return;
			}
			else
			{
				a = b;
			}
		}
		else
		{
			a = getConstVal();
		}
	}
	else
	{
		a = popNum();
	}
	setPrcWhole(a);
	if (constOpP)
	{
		if      (op == "*") val = a / 100.0 * b;     // Faelle 2 u. 11
		else if (op == "/") val = b * 100.0 / a;     // Faelle 4 u. 13
		else if (op == "+") val = (a+b) * 100.0 / a; // Faelle 6 u. 15
		else if (op == "-") val = (b-a) * 100.0 / a; // Faelle 8 u. 17
		pushSym(soP);
	}
	else
	{
		if      (op == "*") val = a / 100.0 * b;     // Faelle 1 u. 10
		else if (op == "/") val = a * 100.0 / b;     // Faelle 3 u. 12
		else if (op == "+") val = (a+b) * 100.0 / b; // Fall   5
		else if (op == "-") val = (a-b) * 100.0 / b; // Fall   7
	}
	pushNum(val);
}

//******************************************************************************

/// Siehe gleichartige Funktion in ComputerBase.
//  fse, 27.04.02

void ComputerCasioSci::popNumDepOnLastSym()
{
	bool isPopped = popNumDepOnLastSymBase();

	if (!isPopped) (void)popNumDepOnLastSymCasio();
}

//************************************** ***************************************

/// Siehe gleichartige Funktion in ComputerBase.
//  fse, 27.04.02

bool ComputerCasioSci::popNumDepOnLastSymCasio()
{
	Symbol* lastInP = getLastInP();
	
	if (isSymEmpty()
	   || dynamic_cast<AcKout*>(lastInP)
	   || dynamic_cast<AcRan* >(lastInP))
	{
		(void)popNum();
		return true;
	}
	return false;
}

//******************************************************************************

/// Fuehrt Aktionen nach Prozessierung des Symbols in aus.
/// Schaltet die Modes zurueck.
/// \param in Symbol der aktuellen Taste
//  fse, 05.07.02

void ComputerCasioSci::aftProcess(string& in)
{
	if (in != "inv" && in != "hyp") mIsInv  = false;
	if (in != "hyp" && in != "inv") mIsHyp  = false;
	if (in != "mode") mIsMode = false;
}

//******************************************************************************

/// Erzeugt fuers Debugging String mit den wichtigen Member-Variablen.
/// \return String mit den wichtigen Member-Variablen
//  fse, 13.07.02

string ComputerCasioSci::makeCompCasioSciInfoStr()
{
	ostringstream str;
	string tmp;

	str << makeCompSciInfoStr() << "\n";
	str << "      mIsMode: " << Util::trimStr(mIsMode  )          << "  mIsKinInput: " << mIsKinInput                       << "\n";
	str << "    mCalcMode: " << Util::trimStr(mCalcMode)          << " mIsKoutInput: " << mIsKoutInput                      << "\n";
	str << "     mKmem[0]: " << Util::trimStr(mKmem[0].makeStr()) << "     mKmem[3]: " << Util::trimStr(mKmem[3].makeStr()) << "\n";
	str << "     mKmem[1]: " << Util::trimStr(mKmem[1].makeStr()) << "     mKmem[4]: " << Util::trimStr(mKmem[4].makeStr()) << "\n";
	str << "     mKmem[2]: " << Util::trimStr(mKmem[2].makeStr()) << "     mKmem[5]: " << Util::trimStr(mKmem[5].makeStr()) << "\n";
	str << "       mXdata: " << Util::trimStr(mXdata  .makeStr());
//	str << ends;
	return str.str();
}

//******************************************************************************
// weitere Symbole von Casio-Rechnern:
//******************************************************************************

/// Setzt Mode-Modus.
//  fse, 01.07.02

void AcMode::process()
{
	ComputerCasioSci* compP = static_cast<ComputerCasioSci*>(mCompP);
	bool isMode = compP->getIsMode();
	
	compP->setIsMode(!isMode);
}

//******************************************************************************

/// Schreiben der Speicher K1 bis K6.
/// Siehe ComputerCasioSci::mapKinKoutInputCasioSci().
//  fse, 27.04.02

void AcKin::process()
{
	ComputerCasioSci* compP = static_cast<ComputerCasioSci*>(mCompP);
	int par = compP->getParam();
	int idx = par, op;
	Number num;

	if (idx == NO_PARAM)
	{
		compP->setIsKinInput(true);
		compP->setLastInPisNotToSet(true);
	}
	else
	{
		if (idx <= 6)
		{
			compP->setKmem(idx-1, mCompP->topNum());
			return;
		}
		op = int(par/10);  // das sind die ersten zwei Ziffern des Parameters
		idx = par - op*10; // das ist die letzte Ziffer des Parameters
		num = compP->getKmem(idx-1);
		switch (op)
		{
			case '+': compP->setKmem(idx-1, num + mCompP->topNum()); break;
			case '-': compP->setKmem(idx-1, num - mCompP->topNum()); break;
			case '*': compP->setKmem(idx-1, num * mCompP->topNum()); break;
			case '/': compP->setKmem(idx-1, num / mCompP->topNum()); break;
			default: ERR_LOG(hgh) << "Wrong parameter in switch!" << NL;
		}
	}
}

//******************************************************************************

/// Lesen der Speicher K1 bis K6.
//  fse, 27.04.02

void AcKout::process()
{
	ComputerCasioSci* compP = static_cast<ComputerCasioSci*>(mCompP);
	int idx = compP->getParam();
	Number num;

	if (idx == NO_PARAM)
	{
		compP->setIsKoutInput(true);
		compP->setLastInPisNotToSet(true);
	}
	else
	{
		num = compP->getKmem(idx-1);
		compP->popNumDepOnLastSym();
		compP->pushNum(num);
	}
}

//******************************************************************************

/// Loeschen aller Speicher K1 bis K6.
//  fse, 27.04.02

void AcKac::process()
{
	ComputerCasioSci* ccP = static_cast<ComputerCasioSci*>(mCompP);
	AcAc ac;

	ac.process();
	for (int i = 0; i < 6; ++i)
	{
		ccP->setKmem(i, 0);
	}
}

//******************************************************************************

/// Vertauschen der Speicher K1 bis K6 mit der Anzeige.
//  fse, 15.07.02

void AcXk::process()
{
	ComputerCasioSci* compP = static_cast<ComputerCasioSci*>(mCompP);
	int idx = compP->getParam();
	Number k, s;

	if (idx == NO_PARAM)
	{
		compP->setIsXkInput(true);
		compP->setLastInPisNotToSet(true);
	}
	else
	{
		k = compP->getKmem(idx-1);
		s = compP->popNum();
		compP->setKmem(idx-1, s);
		mCompP->pushNum(k);
	}

}

//******************************************************************************

/// Vertauschen des obersten Nummern-Stack-Elements mit der Anzeige.
//  fse, 15.07.02

void AcXy::process()
{
	ComputerCasioSci* compP = static_cast<ComputerCasioSci*>(mCompP);
	Number tmp1, tmp2;

	tmp1 = compP->popNum();
	if (compP->isNumEmpty())
	{
		tmp2 = compP->getTopButOneNum();
		compP->setTopButOneNum(tmp1);
	}
	else
	{
		tmp2 = compP->popNum();
		mCompP->pushNum(tmp1);
	}
	mCompP->pushNum(tmp2);
}

//******************************************************************************

/// Enter fuer Statistik- und Regressionsmodus.
//  fse, 27.04.02

void AcData::process()
{
	ComputerCasioSci* ccP = static_cast<ComputerCasioSci*>(mCompP);

	switch(ccP->getCalcMode())
	{
		case SD: processSd(); break;
		case LR: processLr(); break;
		default: cerr << "switch par neither SD nor LR!\n";
	}
}

//************************************** ***************************************

/// Prozessiert "Data" im Statistikmodus.
/// Besonderheit: Bei Eingabe von z.B. "2.3 * 10 data" wird 2.3 zehn Mal
/// eingegeben. Bei anderen Operatoren Fehler.
//  fse, 04.05.02

void AcData::processSd()
{
	ComputerCasioSci* ccP = static_cast<ComputerCasioSci*>(mCompP);
	Symbol* lastSymP = mCompP->popSym();
	OpMul*  mulP = dynamic_cast<OpMul*>(lastSymP);
	Number x, k, n = 1;

	if (lastSymP != NULL && mulP == NULL)
	{
		mDsplP->setError();
		return;
	}
	if (mulP) n = ccP->popNum();
	x = ccP->topNum();
	k = ccP->getKmem(0); ccP->setKmem(0, k + x*x * n);	
	k = ccP->getKmem(1); ccP->setKmem(1, k + x   * n);
	k = ccP->getKmem(2); ccP->setKmem(2, k +       n);
}

//************************************** ***************************************

/// Prozessiert "Data" im Regressionsmodus.
//  fse, 04.05.02

void AcData::processLr()
{
	ComputerCasioSci* ccP = static_cast<ComputerCasioSci*>(mCompP);
	Number x = ccP->getXdata();
	Number y = ccP->topNum();
//	Number x = ccP->popNum();
	Number k;

//	ccP->popSym ( ); // um den Dummy-Operator zu entfernen
//	ccP->pushNum(y); // um die letzte Eingabe wieder zu restaurieren
	k = ccP->getKmem(0); ccP->setKmem(0, k+x*x);	
	k = ccP->getKmem(1); ccP->setKmem(1, k+x  );
	k = ccP->getKmem(2); ccP->setKmem(2, k+1  );
	k = ccP->getKmem(3); ccP->setKmem(3, k+y*y);	
	k = ccP->getKmem(4); ccP->setKmem(4, k+y  );
	k = ccP->getKmem(5); ccP->setKmem(5, k+x*y);
//	ccP->setXdata(0);
}

//******************************************************************************

/// Delete fuer Statistik- und Regressionsmodus.
//  fse, 13.07.02

void AcDel::process()
{
	ComputerCasioSci* ccP = static_cast<ComputerCasioSci*>(mCompP);

	switch(ccP->getCalcMode())
	{
		case SD: processSd(); break;
		case LR: processLr(); break;
		default: cerr << "switch par neither SD nor LR!\n";
	}
}

//************************************** ***************************************

/// Prozessiert "Delete" im Statistikmodus.
/// S. auch AcData::processSd().
//  fse, 13.07.02

void AcDel::processSd()
{
	ComputerCasioSci* ccP = static_cast<ComputerCasioSci*>(mCompP);
	Symbol* lastSymP = mCompP->popSym();
	OpMul*  mulP = dynamic_cast<OpMul*>(lastSymP);
	Number x, k, n = 1;

	if (lastSymP != NULL && mulP == NULL)
	{
		mDsplP->setError();
		return;
	}
	if (mulP) n = ccP->popNum();
	x = ccP->topNum();
	k = ccP->getKmem(0); ccP->setKmem(0, k - x*x * n);	
	k = ccP->getKmem(1); ccP->setKmem(1, k - x   * n);
	k = ccP->getKmem(2); ccP->setKmem(2, k -       n);
}

//************************************** ***************************************

/// Prozessiert "Delete" im Regressionsmodus.
//  fse, 04.05.02

void AcDel::processLr()
{
	ComputerCasioSci* ccP = static_cast<ComputerCasioSci*>(mCompP);
	Number x = ccP->getXdata();
	Number y = ccP->topNum();
//	Number x = ccP->popNum();
	Number k;

//	ccP->popSym ( ); // um den Dummy-Operator zu entfernen
//	ccP->pushNum(y); // um die letzte Eingabe wieder zu restaurieren
	k = ccP->getKmem(0); ccP->setKmem(0, k-x*x);	
	k = ccP->getKmem(1); ccP->setKmem(1, k-x  );
	k = ccP->getKmem(2); ccP->setKmem(2, k-1  );
	k = ccP->getKmem(3); ccP->setKmem(3, k-y*y);	
	k = ccP->getKmem(4); ccP->setKmem(4, k-y  );
	k = ccP->getKmem(5); ccP->setKmem(5, k-x*y);
	ccP->setXdata(0);
}

//******************************************************************************

/// Durchschnitt der x-Werte.
//  fse, 27.04.02

void AcXave::process()
{
	ComputerCasioSci* ccP = static_cast<ComputerCasioSci*>(mCompP);
	Number sum = ccP->getKmem(1); // K2
	Number n   = ccP->getKmem(2); // K3
	Number ave = sum/n;
	
	ccP->popNumDepOnLastSym();
	ccP->pushNum(ave);
}

//******************************************************************************

/// Durchschnitt der y-Werte.
//  fse, 27.04.02

void AcYave::process()
{
	ComputerCasioSci* ccP = static_cast<ComputerCasioSci*>(mCompP);
	Number sum = ccP->getKmem(4); // K5
	Number n   = ccP->getKmem(2); // K3
	Number ave = sum/n;
	
	ccP->popNumDepOnLastSym();
	ccP->pushNum(ave);
}

//******************************************************************************

/// Grundgesamtheits-Standardabweichung x.
//  fse, 27.04.02

void AcXsgm::process()
{
	ComputerCasioSci* ccP = static_cast<ComputerCasioSci*>(mCompP);
	Number sum2 = ccP->getKmem(0); // K1
	Number sum  = ccP->getKmem(1); // K2
	Number n    = ccP->getKmem(2); // K3
	Number ss = sum*sum;
	Number ssn = ss/n;
	Number s2ssn = sum2 - ssn;
	Number r = s2ssn/n;
	Number sgm  = sqrt(fabs(r.getVal()));
	
	ccP->popNumDepOnLastSym();
	ccP->pushNum(sgm);
}

//******************************************************************************

/// Grundgesamtheits-Standardabweichung y.
//  fse, 27.04.02

void AcYsgm::process()
{
	ComputerCasioSci* ccP = static_cast<ComputerCasioSci*>(mCompP);
	Number sum2 = ccP->getKmem(3); // K4
	Number sum  = ccP->getKmem(4); // K5
	Number n    = ccP->getKmem(2); // K3
	Number root = (sum2 - sum*sum/n) / n;
	Number sgm  = sqrt(fabs(root.getVal()));
	
	ccP->popNumDepOnLastSym();
	ccP->pushNum(sgm);
}

//******************************************************************************

/// Stichproben-Standardabweichung x.
//  fse, 27.04.02

void AcXsgm1::process()
{
	ComputerCasioSci* ccP = static_cast<ComputerCasioSci*>(mCompP);
	Number sum2 = ccP->getKmem(0); // K1
	Number sum  = ccP->getKmem(1); // K2
	Number n    = ccP->getKmem(2); // K3
	Number root = (sum2 - sum*sum/n) / (n-1);
	Number sgm  = sqrt(fabs(root.getVal()));
	
	ccP->popNumDepOnLastSym();
	ccP->pushNum(sgm);
}

//******************************************************************************

/// Stichproben-Standardabweichung y.
//  fse, 27.04.02

void AcYsgm1::process()
{
	ComputerCasioSci* ccP = static_cast<ComputerCasioSci*>(mCompP);
	Number sum2 = ccP->getKmem(3); // K4
	Number sum  = ccP->getKmem(4); // K5
	Number n    = ccP->getKmem(2); // K3
	Number root = (sum2 - sum*sum/n) / (n-1);
	Number sgm  = sqrt(fabs(root.getVal()));
	
	ccP->popNumDepOnLastSym();
	ccP->pushNum(sgm);

}

//******************************************************************************

/// y-Achsenabschnitt A bei Regression.
//  fse, 27.04.02

void AcA::process()
{
	ComputerCasioSci* ccP = static_cast<ComputerCasioSci*>(mCompP);
	
	ccP->popNumDepOnLastSym();
	ccP->pushNum(ccP->calcA());
}

//******************************************************************************

/// Regressionskoeffizient B.
//  fse, 27.04.02

void AcB::process()
{
	ComputerCasioSci* ccP = static_cast<ComputerCasioSci*>(mCompP);
	
	ccP->popNumDepOnLastSym();
	ccP->pushNum(ccP->calcB());
}

//******************************************************************************

/// Korrelationskoeffizient r.
//  fse, 27.04.02

void AcR::process()
{
	ComputerCasioSci* ccP = static_cast<ComputerCasioSci*>(mCompP);
	
	ccP->popNumDepOnLastSym();
	ccP->pushNum(ccP->calcR());
}

//******************************************************************************


/// Berechnungswert x bei Regression.
//  fse, 27.04.02

void AcX::process()
{
	ComputerCasioSci* ccP = static_cast<ComputerCasioSci*>(mCompP);
	Number y = ccP->topNum();
	Number a = ccP->calcA();
	Number b = ccP->calcB();
	Number x = (y-a)/b;

	ccP->popNumDepOnLastSym();
	ccP->pushNum(x);
}

//******************************************************************************

/// Berechnungswert y bei Regression.
//  fse, 27.04.02

void AcY::process()
{
	ComputerCasioSci* ccP = static_cast<ComputerCasioSci*>(mCompP);


	Number x = ccP->topNum();
	Number a = ccP->calcA();
	Number b = ccP->calcB();
	Number y = a + b*x;

	ccP->popNumDepOnLastSym();
	ccP->pushNum(y);
}

//******************************************************************************

/// Dreistellige Zufallszahl zwischen 0 und 1.
//  fse, 12.10.01

void AcRan::process()
{
	double ran = (double(rand())/RAND_MAX) * 999.0; // Zufallszahl zwischen 0 und 999
	
	ran = int(ran + .5); // Nachkommastellen mit runden abschneiden
	ran /= 1000.0;       // Zufallszahl zwischen 0.000 und 0.999
	mCompP->popNumDepOnLastSym();
	mCompP->pushNum(ran);
}

//******************************************************************************

/// Konvertierung Sexagesimal-/Dezimalsystem.
/// Z.B. 1425'35" <=> 14.426389
//  fse, 12.07.02
//  fse, 27.04.03: zusaetzliches popNum() bei case 1 und case 2

void AcIsxg::process()
{
	cout << "mCompP: " << mCompP << endl;
	ComputerCasioSci* compP = static_cast<ComputerCasioSci*>(mCompP);
	Symbol* lastButOneSymP = compP->getLastButOneInP();
	Symbol* lastSymP       = compP->getLastInP();
	static int num = 1;
	static Number val;

	if (!dynamic_cast<AcIsxg*>(lastButOneSymP) && !dynamic_cast<AcIsxg*>(lastSymP))
		 num = 1;
	switch (num)
	{
		case 1: val = mCompP->topNum(); val.frcToDbl(); break;
		case 2: val += mCompP->popNum()/60.0     ; mCompP->popNum(); mCompP->pushNum(val); break;
		case 3: val += mCompP->popNum()/60.0/60.0; mCompP->popNum(); mCompP->pushNum(val); break;
	}
	if (++num >= 4) num = 1;
}

//******************************************************************************

/// Wandelt rechtwinklige Koordinaten (a,b) in Polarkoordinaten (r,phi) um.
/// \code
/// Formeln:  r = Sqrt(a*a + b*b)
///         phi = atan b/a
/// \endcode
//  fse, 16.07.02

Number OpRp::calcVal(const Number& a, const Number& b)
{
//	Number phi;
	FnAtan atan;

	mCompP->pushNum(b/a);
	atan.process();
	return sqrt((a*a + b*b).getVal());
}

//******************************************************************************

/// Wandelt Polarkoordinaten (r,phi) in rechtwinklige Koordinaten (a,b) um.
/// \code
/// Formeln:  a = r * cos(phi)
///           b = r * sin(phi)
/// \endcode
// fse, 16.07.02

Number OpPr::calcVal(const Number& r, const Number& phi)
{
	Number a, b;
	FnSin  sin;
	FnCos  cos;

	mCompP->pushNum(phi);
	cos.process();
	a = r * mCompP->popNum();
	mCompP->pushNum(phi);
	sin.process();
	b = r * mCompP->popNum();
	mCompP->pushNum(b);
	return a;
}

//******************************************************************************

/// Rundet den Wert in der Anzeige so, wie er aktuell in der Aneige erscheint.
//  fse, 16.07.02

Number FnRnd::calcVal(const Number& a)
{
	DisplayCasioSci* dsplP = static_cast<DisplayCasioSci*>(mDsplP);
	string value = dsplP->getValue();
	Number n = a;

	if (value.find('/') == string::npos)
	{ // kein Bruch => Anzeige nehmen:
		n = Number(atof(value.c_str()));
		//mCompP->popNum();
	}
	return Number(n);
}

//******************************************************************************

//******************************************************************************
//************************************** ***************************************
//************************* ************************* **************************
//****************** ******************* ******************* *******************
//************** *************** *************** *************** ***************
//************ ************ ************ ************ ************ *************

