//******************************************************************************
// Copyright (c) 2002-2005 by Friedemann Seebass, Germany.
// Dieses Programm ist freie Software. Siehe GNU GPL Notiz in Datei main.cpp.
// This program is free software. See the GNU GPL notice in file main.cpp.
// Projekt: Almetare - Alle meine Taschenrechner
//******************************************************************************
/// \file casio_sci_dat_dlg.cpp
///       Implementierung der Dateneingabe-Dialogbox (Modul special)
//******************************************************************************
// Wann       Wer      Was
// ---------- -------- ---------------------------------------------------------
// 23.01.2005 fse      Verbesserungen
// 11.09.2004 fse      erzeugt
//******************************************************************************

#include <sstream>
#include <qpushbutton.h>
#include <qcheckbox.h>
#include <qmessagebox.h>
#include <qlabel.h>
#include <qfiledialog.h>
#include <qregexp.h>
#include "logging.h"
#include "util.h"
#include "casio_sci_dat_dlg.h"

//******************************************************************************

/// Konstruktor.
/// \param calcP          Zeiger auf den Computer
/// \param applIconPixmap Pixmap mit dem Application-Icon links oben im Dialog
/// \param parentP        Zeiger auf Vaterdialog
/// \param name           Name des Dialogs
/// \param modal          Flag, ob Dialog modal oder nicht
/// \param fl             diverse Qt-Dialog-Flags
//  fse, 05.09.04

CasioSciDatDlg::CasioSciDatDlg(CalculatorBase*   calcP,
                               QPixmap&          applIconPixmap,
                               QWidget*          parentP,
                               const char*       name,
                               bool              modal,
                               WFlags            fl)
    : CasioSciDatBaseDlg(parentP, name, modal, fl),
      mApplIconPixmap(applIconPixmap)
{
//	WFlags symDlgFl = WStyle_Customize | WStyle_NormalBorder | WStyle_Title | WStyle_SysMenu;
	mCompP = (ComputerCasioSci*)calcP->getComputerP();
	CalcMode cm = mCompP->getCalcMode();

	mAccPath = calcP->getAccPath();
	mDat = cm == SD ? mCompP->getSdDat() : mCompP->getLrDat();
	QString dat2(mDat.c_str());
	mMleP->setText(dat2);
	if (!applIconPixmap.isNull())
		setIcon(applIconPixmap); // kleines Icon links oben im Fenster setzen
}

//******************************************************************************

/// Destruktor.

CasioSciDatDlg::~CasioSciDatDlg()
{
    // no need to delete child widgets, Qt does it all for us
}

//******************************************************************************

/// Oeffnet Fileselectbox und gibt den Inhalt des gewaehlten Files im
/// Rueckgabe-String zurueck.
/// \param  oldDat die alten Daten, werden im Fehlerfall wieder zurueckgegeben
/// \return der Inhalt des Files

string CasioSciDatDlg::loadDat(const string& oldDat)
{
	static string dataPath = mAccPath + "/data";
	static QDir   dir(dataPath.c_str());
	CalcMode      cm    = mCompP->getCalcMode();
	QString       ext   = cm == SD ? "(*.sd)" : "(*.lr)";
	QString       type1 = "Datafiles " + ext;;
	QString       type2 = "All files (*)";
	string        dat   = oldDat;
	QStringList   list  = QStringList() << type1 << type2;
	QFileDialog   fileDlg(dir.path(), "", this, "", true);
	QString       fileName;
	int rc;
	ostringstream error;
	char c;

	// Dateidialog oeffnen:
	fileDlg.setCaption("Load data file");
	fileDlg.setFilters(list);
	rc = fileDlg.exec();
	if (rc != 1) return oldDat;
	// Dateinamen holen und Pfad fuers naechste Mal merken:
	fileName = fileDlg.selectedFile();
	dir = *fileDlg.dir();
	// Datei lesen, falls vorhanden:
	if (!fileName.isEmpty())
	{
		ifstream datFile(fileName.latin1());
		if (!datFile)
		{
			error << "Error opening File '" << fileName.latin1() << "'!"; // << ends;
			dat = error.str();
			QMessageBox::critical(this, "Error reading file", dat.c_str(), "&OK");
			return oldDat;
		}
		// Datei lesen:
		dat = "";
		while(datFile.get(c))
			dat += c;
		datFile.close();
	}
	return dat;
}

//******************************************************************************

/// Oeffnet Fileselectbox und schreibt den String dat in die ausgewaehlte
/// Datei.
/// \param  dat Inhalt der zu schreibenden Datei

void CasioSciDatDlg::saveDat(const string& dat)
{
	static string dataPath = mAccPath + "/data";
	static QDir   dir(dataPath.c_str());
	CalcMode      cm    = mCompP->getCalcMode();
	QString       ext   = cm == SD ? ".sd" : ".lr";
	QString       type1 = "Datafiles (*" + ext + ")";
	QString       type2 = "All files (*)";
	QStringList   list = QStringList() << type1 << type2;
	QFileDialog   fileDlg(dir.path(), "", this, "", true);
	QString fileName, dirStr;
	int rc, pointPos;
	string err;

	// Dateidialog oeffnen:
	fileDlg.setCaption("Save data file");
	fileDlg.setFilters(list);
	rc = fileDlg.exec();
	if (rc != 1) return;
	// Dateinamen holen und Pfad fuers naechste Mal merken:
	fileName = fileDlg.selectedFile();
	dir = *fileDlg.dir();
	// wenn keine Extension, Extension anhaengen:
	pointPos = fileName.findRev('.');
	if (pointPos == -1) fileName += ext;
	// Datei speichern:
	if (!fileName.isEmpty())
	{
		rc = 1;
		// pruefen, ob Datei schon vorhanden:
		ifstream datFile(fileName.latin1());
		if (datFile)
		{ // wenn ja, Warnung ausgeben:
			ostringstream error;
			error << "File '" << fileName.latin1() << "' already exists!"; // << ends;
			err = error.str();
			rc = QMessageBox::warning(this, "File exists", err.c_str(), "&Cancel", "&Save");
		}
		datFile.close();
		if (rc == 1)
		{
			ofstream datFile(fileName.latin1());
			if (!datFile)
			{
				ostringstream error;
				error << "Error writing File '" << fileName.latin1() << "'!\n";
				error << "Write protection active?"; // << ends;
				err = error.str();
				QMessageBox::critical(this, "Error writing file", err.c_str(), "&OK");
				return;
			}
			datFile << dat;
			datFile.close();
		}
	}
	return;
}

//******************************************************************************

/// Slot fuer den OK-Button.

void CasioSciDatDlg::accept()
{
	string datOrg = mMleP->text().latin1();
	bool isOk;
	CalcMode cm = mCompP->getCalcMode();

	mDatStripped = Util::stripLines(datOrg);
	isOk = isAllSymbolsOk(mDatStripped);
	if (!isOk)
		return;
	if (cm == SD)
		mCompP->setSdDat(datOrg);
	else
		mCompP->setLrDat(datOrg);
	processData(mDatStripped, cm);
	done(1);
}

//************************************** ***************************************

/// Prueft den uebergebenen Daten-String auf fehlerhafte Symbole.
/// \param dat zu pruefender Programm-String
//  fse, 31.03.03

bool CasioSciDatDlg::isAllSymbolsOk(string dat)
{
	string        sym;
	ostringstream str;
	string        msg;
	bool          isNumber;

	while (dat != "")
	{
		sym = Util::getNextSym(dat);
		isNumber = sym.size() >= 2 && sym[0] == '-' && (isdigit(sym[1]) || sym[1] == '.');
		isNumber = isNumber || isdigit(sym[0]) || sym[0] == '.';
		if (isNumber) continue;
		//if (isParamSym(sym)) continue;
		//if (mCompP->isSym(sym)) continue;
		str << "'" << sym << "' is no number!"; // << ends;
		msg = str.str();
		QMessageBox::critical(this, "Wrong symbol", msg.c_str(), "&OK");
		return false;
	}
	return true;
}

//************************************** ***************************************

/// Traegt alle Daten im String dat in die K-Speicher ein.
/// \param dat zu prozessierender Daten-String
/// \param cm  Calculator-Modus ("LR" oder "SD")
//  fse, 12.09.04

void CasioSciDatDlg::processData(string dat, CalcMode cm)
{
	string sym;
	Number x, y, k;

	for (int i = 0; i < 6; ++i)
	{
		mCompP->setKmem(i, 0);
	}
	while (dat != "")
	{
		sym = Util::getNextSym(dat);
		x = sym;
		k = mCompP->getKmem(0); mCompP->setKmem(0, k + x*x);	
		k = mCompP->getKmem(1); mCompP->setKmem(1, k + x  );
		k = mCompP->getKmem(2); mCompP->setKmem(2, k + 1  );
		if (cm == LR && dat != "")
		{
			sym = Util::getNextSym(dat);
			y = sym;
			k = mCompP->getKmem(3); mCompP->setKmem(3, k+y*y);	
			k = mCompP->getKmem(4); mCompP->setKmem(4, k+y  );
			k = mCompP->getKmem(5); mCompP->setKmem(5, k+x*y);
		}
	}
}

//******************************************************************************
//************************************** ***************************************
//************************* ************************* **************************
//****************** ******************* ******************* *******************
//************** *************** *************** *************** ***************
//************ ************ ************ ************ ************ *************
