//******************************************************************************
// Copyright (c) 2002-2019 by Friedemann Seebass, Germany.
// Dieses Programm ist freie Software. Siehe GNU GPL Notiz in Datei main.cpp.
// This program is free software. See the GNU GPL notice in file main.cpp.
// Projekt: Almetare - Alle meine Taschenrechner
//******************************************************************************
/// \file calc_fx3600p.cpp
///       Klassen fuer Casio-fx-3600P-Taschenrechner (Modul special).
///       In dieser Datei befinden sich die Implementierungen der Klassen, die
///       speziell fuer den Casio fx-3600P benoetigt werden.
//******************************************************************************
// Wann       Wer      Was
// ---------- -------- ---------------------------------------------------------
// 29.05.2005 fse      V.1.11
//                     CalculatorFx3600p::CalculatorFx3600p(): mIniFileName auch
//                     unter Windows in mHomeDir
// 11.06.2004 fse      ComputerFx3600p::ComputerFx3600p(): Member-Initialisier.
//                     V.1.10
// 17.11.2003 fse      strstream durch sstream ersetzt
// 11.08.2003 fse      ComputerFx3600p::qsimp() und ComputerFx3600p::integrate()
//                     verbessert
// 08.08.2003 fse      V.0.99
// 04.08.2003 fse      V.0.98
// 02.08.2003 fse      V.0.97
// 01.03.2003 fse      erzeugt
//******************************************************************************

#include <sstream>
#include <iomanip>   // setw()
#ifndef _WIN32
#include <unistd.h>  // STDIN_FILENO
#endif
#include "util.h"
#include "calc_fx3600p.h"
#include "qapplication.h" // qApp fuer unterbrechbare Programmausfuehrung

//******************************************************************************

typedef double (*funcP)(double);

//CalculatorFx3600p* AcPrgCtrl::mCalcP;

//******************************************************************************
// Klasse DisplayFx3600p:
//******************************************************************************

/// Konstruktor des Displays.
/// Setzt den Zeiger auf den Computer und initialisiert die Anzeige der Modes.
/// \param cbP Zeiger auf den Computer, zu dem das Display gehoert

DisplayFx3600p::DisplayFx3600p(ComputerBase* cbP)
	: DisplayCasioSci(cbP), mLrn("..."), mIdx("..."), mEnt("..."), mP1(".."), mP2("..")
{
	mCompP = static_cast<ComputerFx3600p*>(DisplayBase::mCompP);
	setModes();
}

//******************************************************************************

/// Schaut im Computer, welche Modes gesetzt sind, und baut den Mode-String des
/// Displays entsprechend zusammen.
//  fse, 05.03.03

void DisplayFx3600p::setModes()
{
	setModesCasioSci();
	setModesCasioFx3600p();
	ostringstream str;

	mModes =  mOn  + " " + mInv + " " + mMem + " " + mK   + " "
	        + mLrn + " " + mIdx + " " + mLr  + " " + mSd  + " "
	        + mDeg + " " + mRad + " " + mGra + " "
	        + mEnt + " " + mP1  + " " + mP2  + " ";
	str << "m" // Flag fuer setLcd()
	    << mOn [0] << mInv[0] << mMem[0] << mK [0]
	    << mLrn[2] << mIdx[1] << mLr [0] << mSd[0]
	    << mDeg[0] << mRad[0] << mGra[0]
	    << mEnt[0] << mP1 [1] << mP2 [1]; // << ends; // mOIMKNdLSDRGE12
	mModesShort = str.str();
}                            

//************************************** ***************************************

/// Schaut im Computer, welche Modes gesetzt sind, und baut den Mode-String des
/// Displays entsprechend zusammen.
//  fse, 05.03.03

void DisplayFx3600p::setModesCasioFx3600p()
{
	bool isIdxMode = mCompP->getIsIdxMode();
	bool isLrnMode = mCompP->getIsLrnMode();
	bool isEntMode = mCompP->getIsEntMode();
	bool isHltMode = mCompP->getIsHltMode();
	int  actPrg    = mCompP->getActivePrg();

	setModesSci();
	setOn(true);
	setIdx(isIdxMode);
	setLrn(isLrnMode);
	setEnt(isEntMode);
	setP1(false);
	setP2(false);
	switch(actPrg)
	{
		case 1: if (isLrnMode || isEntMode || isHltMode) setP1(true ); setP2(false); break;
		case 2: if (isLrnMode || isEntMode || isHltMode) setP2(true ); setP1(false); break;
		case 3: setP1(true ); setP2(true ); break;
	}
}

//******************************************************************************

/// Erzeugt Anzeige-String.
/// Erzeugt aus dem Wert in mVal bzw. dem uebergebenen String displayStr einen
/// String, der so aussieht wie die Anzeige des Casio Fx-3600P und schreibt ihn
/// in mSeven.
/// Normalerweise wird versucht, die Zahl in Fixed-Notation darzustellen und zwar
/// mit mDigits Nachkommastellen, wenn der Display-Mode FIX eingestellt ist.
/// Grundstzlich koennen hoechstens 10 Stellen dargestellt werden, wenn die
/// Zahl positiv ist, wenn sie negativ ist, nur 9.
/// Folgende Faelle koennen auftreten:
/// \code
///                 wahre Zahl         auf 10 Stellen    Anzeige
///                                    bzw. 4/5 Stellen
/// - Modus NRM:                       
///              - "1.234"         => "1.234000000"    => "1.234"
///              - "1.234567891"   => "1.234567891"    => "1.2345679"
///              - "-123.4567896"  => "-123.4567896"   => "-123.456789"
///              - "1234"          => "1234.000000"    => "1234"
///              - "123456789.061" => "123456789.1"    => "123456789.1"
///              - "-123456789.061"=> "-123456789.1"   => "-123456789."
///              - "0.00000000012" => "1.200000000e-10"=> "1.2e-10"
///              - "1.23e9"        => "1.2300000e9"    => "1.23e9"
/// - Modus FIX (z.B. 4 Nachkommastellen):
///              - "1.234"         => "1.2340"         => "1.2340"
///              - "1.23456789"    => "1.2346"         => "1.2346"
///              - "123456.789012" => "123456.79"      => "123456.79"
///              - "0.00000000012" => "0.0000"         => "0.0000"
///              - "1.23e9"        => "1.23e9"         => "1.23e9"
/// - Modus SCI (z.B. 5 Stellen)
///              - "1.234"         => "1.2340e0"       => "1.2340e0"
///              - "1.23456789"    => "1.2346e0"       => "1.2346e0"
///              - "123456.789012" => "1.2346e5"       => "1.2346e5"
///              - "0.00000000012" => "1.2000e-10"     => "1.2000e-10"
///              - "1.23e9"        => "1.2300e9"       => "1.2300e9"
/// - Modus FRC:
///              - 12              => "12/1"           => "12"
///              - 12_3/4          => "51/4"           => "12/3/4"
///              - 12.3            => "12.3"           => "12.3"
/// \endcode
/// Aus den Beispielen lassen sich folgende Regeln ablesen:
/// - In den Modi FIX und SCI ist die Anzeige identisch mit der gerundeten
///   Zahl
/// - Im Modus NRM ist die Anzeige ebenfalls identisch mit der gerundeten
///   Zahl, es werden allerdings endstaendige Nullen der Mantisse entfernt
/// - Es muss grundsaetzlich gerundet werden: Entweder auf mDigits (FIX/SCI)
///   oder 10 bzw. 9 Stellen, wenn Zahl positiv bzw. negativ (NRM)
/// Die Anzeige besteht grundsaetzlich aus einem String von 14 Zeichen Laenge:
/// z.B. "-1.23456789-90".
/// \param displayStr darzustellender String
//  fse, 09.03.03

void DisplayFx3600p::makeSeven(string displayStr)
{
	int rnd = mValue[0] == '-' ? 9 : 10; // Anzahl der Stellen, auf die gerundet wird
	
	if (isError())
	{
		mSeven = ".#########E.###"; // das Komma braucht setLcd(); sonst ein Zeichen zu lang
		return;
	}
	if (displayStr == "")
	{
		mSeven = Util::roundMantisse(mValue, rnd, mForceSci);
	}
	else
	{
		mSeven = displayStr;
		mSeven = Util::roundMantisse(mSeven, rnd);
		if (mSeven[0] == '.') mSeven = "0" + mSeven;
		if (mSeven.substr(0, 2) == "-.") mSeven.replace(0, 1, "-0");
	}
	mSevenRaw = mSeven;
	mSeven = Util::formatSeven(mSeven, 14, 3, '#');
	if (mSeven == "E") mSeven = ".#########E.###";
}

//******************************************************************************
// Klasse ComputerFx3600p:
//******************************************************************************

/// Konstruktor fuegt der Symboltabelle spezielle fx-3600P-Symbole hinzu.
//  fse, 11.06.04: Initialisierung der Members - frueher vergessen :-(
//                 das fuehrte erst unter Knoppix 3.4 zu fehlerhaftem Setzen
//                 der Werte z.B. in CalculatorFx3600p::restoreNvRam(), da dort
//                 oefters kein else-Zweig.

ComputerFx3600p::ComputerFx3600p(CalculatorFx3600p* cP)
: mIsForceSciTypeB(false), mIsIdxMode(false), mIsLrnMode(false),
  mIsEntMode(false), mIsHltMode(false), mActivePrg(0), mRunningPrg(""),
  mIsPrgBreak(false), mIdxStep(0), mDspMode(NRM), mDigits(0)
{
	mPrgMem[0] = mPrgMem[1] = mPrgMemOrg[0] = mPrgMem[1] = "";
	mCalcFx3600pP = cP;
	addCasioFx3600SymsToSymTab();
}

//************************************** ***************************************

/// Fuegt der Symboltabelle spezielle fx-3600P-Symbole hinzu.
//  fse, 24.06.14: Operatoren "+" und "-" des Basisrechners sind neu und werden
//                 hier wieder durch die Casio-Versionen ersetzt.

void ComputerFx3600p::addCasioFx3600SymsToSymTab()
{	
	string sym;         
	Symbol* symP;
	OpAdd oa;
	OpSub os;

	// Symbole "+" und "-" des Basisrechners entfernen und durch die fuer Casio
	// ersetzen:
	mSymTab.erase(oa.getSym());
	mSymTab.erase(os.getSym());
	symP = new OpAddCasioSci; sym = symP->getSym(); mSymTab[sym] = symP;
	symP = new OpSubCasioSci; sym = symP->getSym(); mSymTab[sym] = symP;

	// weitere zusaetzliche Symbole:
	symP = new AcIdx        ; sym = symP->getSym(); mSymTab[sym] = symP;
	symP = new AcLrn        ; sym = symP->getSym(); mSymTab[sym] = symP;
	symP = new AcRnm        ; sym = symP->getSym(); mSymTab[sym] = symP;
	symP = new AcPcl        ; sym = symP->getSym(); mSymTab[sym] = symP;
	symP = new AcStrps      ; sym = symP->getSym(); mSymTab[sym] = symP;
	symP = new PrgCtrlP1    ; sym = symP->getSym(); mSymTab[sym] = symP;
	symP = new PrgCtrlP2    ; sym = symP->getSym(); mSymTab[sym] = symP;
	symP = new PrgCtrlRun   ; sym = symP->getSym(); mSymTab[sym] = symP;
	symP = new PrgCtrlEnt   ; sym = symP->getSym(); mSymTab[sym] = symP;
	symP = new PrgCtrlHlt   ; sym = symP->getSym(); mSymTab[sym] = symP;
	symP = new PrgCtrlXgt0  ; sym = symP->getSym(); mSymTab[sym] = symP;
	symP = new PrgCtrlXleM  ; sym = symP->getSym(); mSymTab[sym] = symP;
	symP = new PrgCtrlRtn   ; sym = symP->getSym(); mSymTab[sym] = symP;
}

//******************************************************************************

/// Fuehrt Pre-Processing des Symbols durch.
/// Falls der Programmiermodus aktiviert ist, werden alle Eingaben in mPrgMem[]
/// gespeichert.
/// \retval in zu mappendes Symbol
//  fse, 08.03.03

bool ComputerFx3600p::preProcess(string& in)
{
	bool isMapped, isProcessed = false;
//	char c = in[0];

	if (in == "sd" || in == "lr") 
	{
		mActivePrg = 0;
		mIsLrnMode = mIsIdxMode = mIsEntMode = false;
	}
	if (in == "ac")
	{
		mIsEntMode = false;
		if (!mIsLrnMode) mActivePrg = 0;
		mIsPrgBreak = true;
	}
	isMapped    = map(in);
	isProcessed = false;
	if (in == "processed")
		isProcessed = true;
	if (mIsLrnMode && mActivePrg != 3)
		addPrgStep(in);
	return isProcessed;
}

//************************************** ***************************************

/// Mapped das uebergebene Symbol gegebenenfalls auf ein anderes Symbol um.
/// Das ist z.B. noetig, wenn der Computer gerade im Invers-Modus ist.

bool ComputerFx3600p::map(string& in)
{
//	DisplayFx3600p* dsplP = static_cast<DisplayFx3600p*>(mDsplP);
//	DspMode dm = dsplP->getDspMode();
	bool isMapped;

	if (mIsLrnMode)
	{
		if (mIsInv)
		{
			if (in == "mode") { in = "pcl" ; mIsInv = false; return true; }
			if (in == "7"   ) { in = "xgt0"; mIsInv = false; return true; }
			if (in == "8"   ) { in = "xlem"; mIsInv = false; return true; }
			if (in == "9"   ) { in = "rtn" ; mIsInv = false; return true; }
			if (in == "run" ) { in = "hlt" ; mIsInv = false; return true; }
		}
		if (in == "run" ) { in = "ent"    ; return true; }
	}
	if (mIsMode)
	{
		if (in == "1"   ) { in = "idx"    ; return true; }
		if (in == "0"   ) { in = "lrn"    ; return true; }
		if (in == "."   )
		{
			if (mIsLrnMode || mIsIdxMode) { in = "rnm"; return true; }
			if (mCalcMode != CMP        ) { in = "cmp"; return true; }
		}
	}
	if (mCalcMode != CMP)
	{ // SD- oder LR-Modus
		if (in == "run" && mIsInv) { in = "del" ; mIsInv = false; return true; }
		if (in == "run" ) { in = "data"   ; return true; }
	}
	if (mIsInv)
	{
		if (in == "p1"  ) { in = "p2"     ; mIsInv = false; return true; }
		if (in == "abc" ) { in = "rcp"    ; mIsInv = false; return true; }
		if (in == "eng" ) { in = "engback"; mIsInv = false; return true; }
		if (in == "sign") { in = "sqr"    ; mIsInv = false; return true; }
		if (in == "kin" ) { in = "xy"     ; mIsInv = false; return true; }
		if (in == "kout") { in = "xk"     ; mIsInv = false; return true; }
		if (in == "rcl" ) { in = "sto"    ; mIsInv = false; return true; }
		if (in == "*"   ) { in = "pow"    ; mIsInv = false; return true; }
		if (in == "/"   ) { in = "root"   ; mIsInv = false; return true; }
		if (mCalcMode != LR)
		{
			if (in =="(") { in = "sqrt"   ; mIsInv = false; return true; }
			if (in ==")") { in = "fact"   ; mIsInv = false; return true; }
		}
		if (mIsIdxMode && in == "run") { in = "strps"; mIsInv = false; return true; }
	}
	isMapped = mapCasioSci(in);
	return isMapped;
}

//************************************** ***************************************

/// Speichert im Programmiermodus den naechsten Programmschritt.
/// Dabei ist zu beruecksichtigen, dass nicht alle Eingaben gespeichert werden
/// duerfen. Z.B. wird "inv" nicht gespeichert, weil das Symbol ohnehin durch
/// das entsprechende "ge-inv-te" ersetzt wird und dieses gespeichert wird. Alle
/// Symbole, die nicht gespeichert werden duerfen, sind im String notToStore
/// aufgezaehlt.
/// Ein Sonderfall ist auch bei der Eingabe von Sexagesimalzahlen zu beachten:
/// Wenn die Zahl eine Eingabe ist (also nach der "ENT"-Taste eingegeben wird),
/// duerfen die bis zu drei "isxg"-Symbole nicht gespeichert werden, vielmehr
/// muss die Sexagesimalzahl als eine einzige Eingabe betrachtet werden. Dabei
/// ist ausreichend, nur die letzte Eingabezahl aufzuzeichnen, da diese bei der
/// Ausfuehrung des Programms ohnehin durch den aktuellen Eingabewert ersetzt
/// wird.
/// \param in Symbol der aktuellen Taste
//  fse, 03.05.03

void ComputerFx3600p::addPrgStep(const string& in)
{
	string&     mem    = mPrgMem   [mActivePrg-1]; // zur Abkuerzung
	string&     memOrg = mPrgMemOrg[mActivePrg-1]; //
	string      sep    = mem == "" ? "" : " ";
	string      sepOrg = mem == "" ? "" : "\n";
	char        c = in[0];
	string      notToStore = " mode inv hyp fix sci ee ac c kin kout xk processed p1 p2 rnm idx ";
	bool        isToStore  = notToStore.find(" "+in+" ") == string::npos;
	string      num;
	int         entPos;

	if (in == "isxg" && mIsEntMode)
	{ // wenn gerade Enter, dann nicht "isxg" speichern, sondern den Platzhalter:
		entPos  = mem.rfind("ent");           // nach dem letzten "ent" suchen
		mem     = mem   .substr(0, entPos+3); // Programm hinter "ent" abschneiden
		memOrg  = memOrg.substr(0, entPos+3); // 
		num     = topNum().makeStr();         // als Platzhalter halt letzten Wert nehmen
		mem    += sep    + num;               // und diesen speichern
		memOrg += sepOrg + num;               // auch im Speicher mit den Kommentaren
		isToStore = false;                    // Speichern von "isxg" verhindern
	}
	if (isToStore)
	{
		mem    += sep    + in; // dieser String wird als Programm ausgefuehrt
		memOrg += sepOrg + in; // auch in den Speicher mit Kommentaren schreiben
		if (!(isdigit(c) || c == '.' || in == "pi")) mIsEntMode = false;
	}
}

//******************************************************************************

/// Fuehrt das Programm actPrg aus.
/// Ein Programm besteht aus einem String, in dem alle aufgezeichneten Symbole
/// durch ein Leerzeichen getrennt aufgeschrieben sind. Die Symbole werden bei
/// der Programmausfuehrung aus dem String geholt und in einer Schleife der
/// normalen process()-Funktion des Calculators uebergeben, so wie es immer
/// geschieht.
/// Bloed: Die process()-Funktion gehoert dem Calculator, dieser benutzt
///        eigentlich den Computer; hier muss jetzt der Computer den Calculator
///        benutzen, was eigentlich so nicht vorgesehen war. Diese gegenseitige
///        Abhaengigkeit ist unschoen, aber scheint mir immer noch besser als
///        im Computer die process()-Funktion zu duplizieren.
///
/// Ein Problem sind die Symbole "ent" und "hlt". Bei diesen muss naemlich der
/// Programmfluss unterbrochen werden, damit der Benutzer eine Eingabe machen
/// kann. Um auch hier die ganz normale Algorithmik beim Eingeben von Zahlen
/// benutzen zu koennen, wird das Programm tatsaechlich voellig beendet und
/// die Kontrolle kehrt zur Hauptereignisschleife zurueck. Wenn der User seine
/// Eingabe gemacht hat, drueckt er wieder die "RUN"-Taste und das Programm
/// wird fortgesetzt, indem es erneut gestartet wird.
///
/// Das funktioniert nun so:
/// - Der Benutzer startet ein Programm durch "p1" oder "p2".
///   => Die Funktion AcPrgCtrl::processP1P2() holt zuerst den passenden
///      Programm-String aus dem Speicher, kopiert ihn in die Membervariable
///      mRunningPrg, setzt das Flag mActivePrg von 0 auf die Nummer des
///      aktiven Programms (1 bzw. 2) und ruft execPrg() auf.
/// - execPrg() holt Symbol fuer Symbol aus dem String mRunningPrg; der
///   String wird dabei abgebaut, so dass immer das naechste Symbol vorne
///   steht.
/// - Wenn das Symbol "ent" oder "hlt" im Programm auftaucht, wird das
///   Flag mIsEntMode bzw. mIsHltMode gesetzt und execPrg() verlassen.
///   => Die Kontrolle geht damit wieder ganz normal an die Hauptschleife
///      ueber.
/// - Der Anwender kann den Rechner jetzt ganz normal bedienen, insbesondere
///   kann er im Display eine Eingabe machen. Sobald er die "RUN"-Taste
///   drueckt, ruft PrgCtrlRun::process() wieder execPrg() auf.
///   => Das Programm wird fortgesetzt, und zwar genau an der Stelle, an der
///      es unterbrochen wurde, da der String mRunningPrg nicht veraendert
///      wurde.
/// - Ein gerade laufendes Programm muss durch Druecken von "AC" abgebrochen
///   werden koennen. Dies geschieht, indem in chkOnPrgBreak() die Funktion
///   qApp->processEvents() aufgerufen wird, die kurz zur Qt-Hauptschleife
///   zurueckkehrt bis alle Events abgearbeitet sind. Wenn "AC" gedrueckt
///   wurde, wird nun wie immer auch die preProcess()-Funktion aufgerufen.
///   Wenn diese das Symbol "ac" findet und gerade ein Programm laeuft (was
///   an mActivePrg festgestellt wird), wird dort das Flag mIsPrgBreak gesetzt.
///   Anschliessend kehrt die Kontrolle hierher zurueck, und das Flag wird
///   abgefragt; falls es gesetzt ist, wird execPrg() verlassen.
///
/// Version 1.15: 
/// Auf Anwenderwunsch soll es moeglich sein, dass man aus einem Programm
/// das jeweils andere als Unterprogramm aufruft. Das ist
/// kompliziert, weil execPrg() nicht nur von AcPrgCtrl::processP1P2()
/// aufgerufen wird, sondern nach einem ENT oder HLT auch von 
/// PrgCtrlRun::process(); der Aufruf durch intgrInputAb() ist dagegen
/// unkritisch.
/// Ziel ist es daher, dass alle noetigen Informationen, um den Vorgang zu
/// steuern, lokal hier in execPrg() ermittelt und gespeichert werden. Das
/// geschieht nun in einigen statischen Variablen, aus denen ersichtlich ist,
/// ob gerade ein Unterprogramm laeuft oder/und ein solches gerade durch ENT
/// oder HLT unterbrochen wurde und bein naechsten Aufruf fortgesetzt werden
/// muss:
/// - isToContinue wird auf true gesetzt, bevor ein Programm wegen ENT oder
///   HLT unterbrochen wird.
/// - isSubPrgRunning gibt an, ob gerade ein Unterprogramm laeuft.
/// - mainRunningPrg ist eine Kopie des im Rahmen der Abarbeitung sich
///   verkuerzenden Hauptprogramms.
/// - mainActivePrg ist die Nummer (1 oder 2) des Hauptprogramms.
/// 
/// \return das oberste Element des Number-Stacks.
//  fse, 08.03.03
//  fse, 22.04.18: Ein Programm kann jetzt das jeweils andere als
//                 Unterprogramm aufrufen.

Number ComputerFx3600p::execPrg()
{
	static int    mainActivePrg     = 0;
	static string mainRunningPrg    = "";
	static bool   isToContinue      = false; // true, wenn das Programm durch ENT oder HLT unterbrochen wurde
	static bool   isSubPrgRunning   = false; // true, wenn ein Unterprogramm laeuft
	bool          isWrongPrg        = !(mActivePrg == 1 || mActivePrg == 2);
	bool          isSubPrgToExecute;
	bool          isRecursionError;
	string in;
	char c;

	if (mIsToInit) // Flag wird von Ac gesetzt
	{
		mainActivePrg   = 0;
		mainRunningPrg  = "";
		isToContinue    = false;
		isSubPrgRunning = false;
		mIsToInit       = false;
	}
	isSubPrgToExecute = !isSubPrgRunning && mainActivePrg != 0 && mainActivePrg != mActivePrg;
//	isRecursionError  = isSubPrgToExecute && mainActivePrg != 0 && mainActivePrg == mActivePrg || isSubPrgRunning && !isToContinue;
	isRecursionError  = isSubPrgRunning && !isToContinue;
	if (isWrongPrg || isRecursionError)
	{
		mDsplP->setError();
		goto ret;
	}
	mIsPrgBreak = mIsEntMode = mIsHltMode = false;

	if (isToContinue)
	{
		isToContinue = false; // dann normale Wiederaufnahme der While-Schleife unten
	}
	else
	{
		if (isSubPrgToExecute)
			isSubPrgRunning = true;
		else
			mainActivePrg = mActivePrg; // das ist der ganz normale Start eines Programms
	}

	continueMainPrg:
	// Programm aus String auslesen:
	in = mCalcFx3600pP->getNextSym(mRunningPrg);
	if (storeMainPrgAndCheckOnRecursionError(isSubPrgRunning, in, mainRunningPrg))
		goto ret;
	while (in != "")
	{
		if (in == "ent" || in == "hlt")
		{
			isToContinue= true;
			if (in == "hlt")
			{
				mIsHltMode = true;
			}
			else
			{
				mIsEntMode = true;
				// Eingabezahl der Programmierung verwerfen, wenn vorhanden:
				in = mCalcFx3600pP->getNextSym(mRunningPrg);
				if (storeMainPrgAndCheckOnRecursionError(isSubPrgRunning, in, mainRunningPrg))
					goto ret;
				c = in[0];
				if (!(isdigit(c) || c == '.' || c == '-' || in == "pi"))
					mRunningPrg = in + " " + mRunningPrg;
			}
			if (getIsIdxMode())
			{ // wir sind bei der Integration:
				if (mIdxStep < 2)
				{ // wir sind gerade bei der Eingabe von a oder b:
					goto ret;
				}
				else
				{ // wir berechnen gerade Funktionswerte, etwaige
				  // ENTs oder HLTs fuehren jetzt wie beim Original zum Error:
					mIsEntMode = mIsHltMode = false;
					mDsplP->setError();
					goto ret;
				}
			}
			else
				goto ret;
		}
		mCalcFx3600pP->process(in);
		if (isToContinue)
			goto ret; // Unterprogramm ist im ENT- oder HLT-Modus
		in = mCalcFx3600pP->getNextSym(mRunningPrg);
		if (storeMainPrgAndCheckOnRecursionError(isSubPrgRunning, in, mainRunningPrg))
			goto ret;
		if (checkOnPrgBreak()) break;
		if (mIsPrgBreak || mDsplP->isError()) break;
	}
	if (mIdxStep == 0) mActivePrg = 0;
	mIsEntMode = mIsLrnMode = false;
	if (isSubPrgRunning)
	{
		mRunningPrg = mainRunningPrg;
		isSubPrgRunning = false;
		goto continueMainPrg;
	}
	else
		mainActivePrg = 0;
	ret:
	return topNum();
}

//************************************** ***************************************

/// Falls gerade kein Unterprogramm laeuft, wird prophylaktisch das sich
/// abbauende Hauptprogramm immer in den zweiten String mainRunningPrg kopiert.
/// Falls dann irgendwann ein Unterprogramm aufgerufen wird, kann man nach
/// dessen Abarbeitung das Hauptprogramm an der aktuellen Stelle wieder
/// fortsetzen.
/// Ausserdem wird geprueft, ob in einem laufenden Programm noch einmal
/// dasselbe Programm aufgerufen wird. Falls das der Fall ist, muss der Rechner
/// in den Error-State versetzt werden.
/// Die beiden Dinge haben eigentlich nichts miteinander zu tun, muessen aber
/// beide nach jedem getNextSym() gemacht werden; daher in einem Unterprogramm
/// zusammengefasst.
/// \param  isSubPrgRunning true, wenn gerade ein Unterprogramm laeuft
/// \param  string          das letzte gelesene Symbol aus dem laufenden
///                         Programm
/// \retval mainRunningPrg  die Kopie des abgearbeiteten Hauptprogramms
/// \return true, wenn dasselbe Programm noch einmal aufgerufen werden soll,
///         sonst false
//  fse, 20.04.18

bool ComputerFx3600p::storeMainPrgAndCheckOnRecursionError(bool isSubPrgRunning, const string& in, string& mainRunningPrg)
{
	bool isError = false;

	if (!isSubPrgRunning)
		mainRunningPrg = mRunningPrg;
	if (in == "p1" && mActivePrg == 1 || in == "p2" && mActivePrg == 2)
	{
		mDsplP->setError();
		isError = true;
	}
	return isError;
}

//************************************** ***************************************

/// Funktion dient zum Abbrechen eines laufenden Programmes. Sie wird bei jedem
/// Schleifendurchlauf von execPrg() aufgerufen. Einzelheiten siehe dort.
/// Als Konsolenanwendung wird die Standardeingabe abgefragt (nur Linux).
/// \return true, wenn im Konsolenmodus eine Eingabe gemacht wurde, sonst false

bool ComputerFx3600p::checkOnPrgBreak()
{
	static int cnt = 0;

	if (mCalcFx3600pP->getIsDlgApp())
	{
		if (++cnt > 100)
		{
			qApp->processEvents(); // Abbrechen mit "AC" erlauben
			cnt = 0;
		}
	}
	else
	{
		#ifndef _WIN32
		// select-Parameter vorbereiten:
		fd_set readSet;
		timeval timeout;
		timeout.tv_sec = 0;
		timeout.tv_usec = 0;
		FD_ZERO(&readSet);
		FD_SET(STDIN_FILENO, &readSet);
		// Programm verlassen, wenn irgendeine Eingabe gemacht wird:
		int rc = select(STDIN_FILENO+1, &readSet, NULL, NULL, &timeout);
		if (rc > 0) return true;
		#endif
	}
	return false;
}

//******************************************************************************

/// Eingabe der Integrationsgrenzen.
/// Es wird temporaer ein Programm angelegt und ausgefuehrt, in dem man zwei
/// Zahlen angeben kann, die in K1 und K2 abgelegt werden. Alle anderen Speicher
/// werden geloescht. Siehe auch integrate().
//  fse, 28.03.03

void ComputerFx3600p::intgrInputAb()
{
	mIdxStep = 1;
	mRunningPrg = "0 = 0 = ent 0 kin@1 ent 0 kin@2"; // zweimal "0=", um im Falle von P2 das INV abzuschalten
	mKmem[0] = mKmem[1] = mKmem[2] = mKmem[3] = mKmem[4] = mKmem[5] = 0;
	execPrg();
}

//******************************************************************************

/// Funktion zur Berechnung von Bestimmten Integralen.
/// Zur Berechnung werden die normalen Programmspeicher benutzt. Daher muss hier
/// ein etwas kompliziertes Hin- und Herspringen erfolgen, das vom Flag mIdxStep
/// gesteuert wird. Dieser Zaehler wird immer inkrementiert, wenn ein Schritt
/// beendet wurde.
///
/// - Zunaechst wird der Rechner vom Anwender in den Intagrationsmodus
///   geschaltet ("Mode" "1"), das heisst, das Flag mIsIdxMode ist also gesetzt.
/// - Anschliessend wird eine der beiden Funktionen P1 oder P2 ausgewaehlt, die
///   zugehoerige process()-Funktion AcPrgCtrl::processP1P2() wird daher
///   aufgerufen. Diese ruft wiederum zur Eingabe der Integrationsgrenzen
///   intgrInputAb() auf.
/// - intgrInputAb() erzeugt ein neues Programm und schreibt es in mRunningPrg.
///   Dies besteht aus der Eingabe von zwei Zahlen, die in K1 und K2 gespeichert
///   werden. Bei jeder Eingabe wird dieses Programm wie ueblich unterbrochen
///   und fortgesetzt, sobald die "RUN"-Taste gedrueckt wird. Die zugehoerige
///   process()-Funktion PrgCtrlRun::process() fragt mIdxStep ab, das von
///   intgrInputAb() auf 1 gesetzt wurde, und inkrementiert es. Auch bei der
///   Eingabe der oberen Integrationsgrenze wird mIdxStep inkrementiert, so dass
///   es dann auf 3 steht. Nach dem zweimaligen Inkrementieren ist die Eingabe
///   beendet und PrgCtrlRun::process() ruft intgegrate() auf.
/// - Zuvor kann durch die Eingabe von 'Zahl' "INV" "RUN" noch eine
///   Streifenanzahl festegelegt werden. In diesem Fall wird das Symbol "run"
///   von preProcess() auf das Symbol "strps" gemapped und AcStrps::process()
///   schreibt die Streifenanzahl in K3.
/// - integrate() hat nun alle Informationen, um die Integration
///   durchzufuehren: Die Grenzen des Integrals stehen in K1 und K2, falls mit
///   'Zahl' "INV" "RUN" eine Streifenzahl vorgegeben wurde, steht diese in K3
///   (sonst dort 0). Die Integration wird jetzt mit der Simpson-Naeherung
///   durchgefuehrt. Fuer jeden Funktionswert wird die Funktion f() aufgerufen.
///   Sie schreibt das aktuelle Programm jedesmal in mRunningPrg und ruft
///   execPrg() auf. Dies darf in diesem Fall die Nummer des aktuellen
///   Programms in mActivePrg nicht wie sonst ueblich nach Abarbeitung auf
///   0 setzen (mIdxStep != 0 wird abgefragt).
/// - Schliesslich ist das Integral berechnet. Jetzt hat der Casio fx-3600P
///   noch die Eigenheit, das Ergebnis zunaechst im wissenschaftlichen
///   Format mit sechs Stellen zurueckzugeben, wobei aber der Modus "SCI" nicht
///   aktiv ist; wenn man "=" drueckt, wird die Darstellung umgewandelt.
///   Dieser Effekt wird hier erzielt, indem integrate() vor Ausgabe des
///   Ergebnisses in den SCI-Modus schaltet. Der muss nun aber sofort wieder
///   abgeschaltet (bzw. auf den alten Modus zurueckgesetzt) werden.
///   Das macht die neue Funktion aftProcess2(), die von
///   CalculatorBase::process() nach der Ausgabe ins Display aufgerufen
///   wird. Da diese Funktion aber natuerlich nach jedem prozessierten Symbol
///   aufgerufen wird, aber nur nach Ende der Integration aktiv werden
///   soll, wird wieder die Steuervariable mIdxStep abgefragt.
//  fse, 28.03.03
//  fse, 11.08.03: Rundung abhaengig von der Genauigkeit

void ComputerFx3600p::integrate()
{
	Number a, b;
	int n, digits = 9;
	double res, dif;
	ostringstream str;
	string resStr;
	DisplayFx3600p* dsplP = static_cast<DisplayFx3600p*>(mDsplP);

	// aktuelle Display-Einstellungen sichern:
	mDspMode = dsplP->getDspMode();
	mDigits  = dsplP->getDigits();

	// Integrationsparameter aus Speichern holen:
	a = mKmem[0];
	b = mKmem[1];
	n = (int)mKmem[2].getVal(); // dyadischer Log. der Streifenanzahl

	// integrieren und Streifenanzahl in K3 schreiben:
	res = qsimp(a.getVal(), b.getVal(), n, dif);
	mKmem[2] = pow(2, (double)(n-2));
	//cout << "integrate(): qsimp(): " << res << endl;

	// falls Fehler, abbrechen:
	if (mDsplP->isError())
	{
		//cout << "integrate(): qsimp(): " << res << endl;
		mIdxStep = 0;
		mActivePrg = 0;
		return;
	}

	// Ergebnis nach K6 schreiben und im SCI-Modus ausgeben:
	if (dif > 0.0 && dif < 1.0)
		digits = abs(int(log10(dif)));
	if (digits > 9)
		digits = 9;
	str.precision(digits);
	str << res; // << ends;
	resStr = str.str();
	resStr = Util::roundNumStr(resStr, digits, true, 0.01);
	mKmem[5] = atof(resStr.c_str());
	dsplP->setDspMode(SCI);
	dsplP->setDigits(digits);
	mCalcFx3600pP->process(resStr);
	mIdxStep = 4; // Flag fuer aftProcess2()
	mActivePrg = 0;
}

//************************************** ***************************************

/// Berechnet das Integral der Funktion funcP in den Grenzen a bis b nach der
/// Simpson-Naeherung.
/// Aus: Press u.a.: Numerical Recipes in C, S. 139
/// \param  a   untere Grenze des Integrals
/// \param  b   obere Grenze des Integrals
/// \retval j   Verfeinerungsstufe, die noetig war
/// \retval dif Differenz zwischen Endergebnis und der vorherigen Verfeinerung
/// \return Integral
//  fse, 11.08.03: nur wenn's wirklich sehr ungenau, Fehler ausgeben.

double ComputerFx3600p::qsimp(double a, double b, int& j, double& dif)
{
	int          jSet = j;
	const double eps  = 1.0e-5;
	const int    jmax = jSet == 0 ? 13 : jSet+2;
	double       s = 0.0, st, ost, os;

	ost = os = -1.0e30;
	for (j = 1; j <= jmax; j++)
	{
		st = trapzd(a, b, j);
		s = (4.0*st - ost)/3.0;
		dif = fabs(s - os);
		DBG_LOG(hgh) << "qsimp(): j: " << j << ", s: " << s << ", dif: " << dif << NL;
		if (dif < eps * fabs(os))
		{
			if (jSet == 0)
				return s;
			else
				if (j == jmax) return s;
		}
		if (s == 0.0 && os == 0.0 && j > 6) return s;
		os = s;
		ost = st;
	}
	if (dif > 100*eps * fabs(os)) // fse, 11.08.03
		mDsplP->setError();
	j = jmax;
	return s;
}

//************************* ************************* **************************

/// Berechnet die n-te Verfeinerungsstufe der Erweiterten Trapezregel als
/// Naeherungsformel der Integration.
/// Aus: Press u.a.: Numerical Recipes in C, S. 137
/// \param a untere Grenze des Integrals
/// \param b obere Grenze des Integrals
/// \param n Verfeinerungsstufe
/// \return Integral

double ComputerFx3600p::trapzd(double a, double b, int n)
{
	double x, tnm, sum, del, fa, fb;
	static double s;

	int it, j;

	if (n == 1)
	{
		fa = f(a); mKmem[3] = fa;
		fb = f(b); mKmem[4] = fb;
		return (s = 0.5 * (b-a) * (f(a)+f(b)));
	}
	else
	{
		for (it=1, j=1; j<n-1; j++) it <<=1;
		tnm = it;
		del = (b-a)/tnm;
		x = a + 0.5 * del;
		for (sum=0.0, j=1; j<=it; j++, x+=del) sum += f(x);
		s = 0.5 * (s+(b-a)*sum/tnm);
		//cout << n << ": delta: " << del << ", s: " << s << endl;
		return s;
	}
}

//****************** ******************* ******************* *******************

/// Ermittelt den Funktionswert y der Funktion im aktuellen Programmspeicher an
/// der Stelle x.
/// \param x x-Wert, an dem die Funktion berechnet wird
/// \return Funktionswert an der Stelle x

double ComputerFx3600p::f(double x)
{
	ostringstream str;
	Number num;
	int actPrg = mActivePrg;

	str << x; // << ends;
	mCalcFx3600pP->process("ac");
	mCalcFx3600pP->process(str.str());
	mActivePrg = actPrg;
	mRunningPrg = mPrgMem[mActivePrg-1];
	num = execPrg();
	return num.getVal();
}

//******************************************************************************

/// Fuehrt Aktionen nach Prozessierung des Symbols aus. Im Unterschied zu
/// aftProcess wird diese Funktion von Calculator::process() aufgerufen,
/// NACHDEM das Display aufgefrischt wurde.
/// Noetig bei der Integration, um die Anzeige wieder von der SCI-Notation,
/// in der die Ergebnisse zunaechst ausgegeben werden, wieder zurueckzustellen.
/// \param in Symbol der aktuellen Taste; wird hier zunaechst nicht benoetigt,
///           aber vielleicht spaeter mal
//  fse, 20.04.03

void ComputerFx3600p::aftProcess2(string& in)
{
	DisplayFx3600p* dsplP = static_cast<DisplayFx3600p*>(mDsplP);

	in = in; // sonst Compiler-Warnung
	if (mIdxStep == 4)
	{
		dsplP->setDspMode(mDspMode);
		dsplP->setDigits(mDigits);
		mIdxStep = 0;
	}
}

//******************************************************************************
// Klasse CalculatorFx3600p:
//******************************************************************************

/// Konstruktor des Calculators.
/// Achtung: Aufraeumen wird im wesentlichen von ~CalculatorBase() uebernommen!
//  fse, 11.05.02
//  fse, 29.05.05: mIniFileName auch unter Windows in mHomeDir

CalculatorFx3600p::CalculatorFx3600p(int argc, char** argv) : CalculatorBase(argc, argv), mSep('|')
{
	#ifdef _WIN32
//	mIniFileName = mProgPath + "fx3600p.ini";
	mIniFileName = mHomeDir + "\\fx3600p.ini";
	#else
	mIniFileName = mHomeDir + "/.fx3600p";
	#endif
	mProgName    = "Casio fx-3600P";
	mProgVers    = "V.1.16";
	mAccPath     = mProgPath + "fx3600p_acc";
	mCompP       = new ComputerFx3600p(this);
	mDisplayP    = new DisplayFx3600p(mCompP);
	mCfgsP       = new Cfgs(mAccPath, mIniFileName, mProgName, mProgVers);
	mComputerP   = mCompP;
	mComputerP->setDsplP(mDisplayP);

//	mCompP->setCalcP(this);
	Symbol::setCompP(mComputerP);
	Symbol::setDsplP(mDisplayP );
//	AcPrgCtrl::setCalcP(this);
	mDisplayP->setCompP(mComputerP);
	restoreNvRam();
};

//************************************** ***************************************

/// Der NV-RAM-String aus der Konfigurationsdatei wird geparsed und die Daten in
/// die entsprechenden Variablen geschrieben.
/// ToDo: evtl. den Force-Parameter, der vom fx-85v stammt, entfernen.
//  fse, 09.03.03

void CalculatorFx3600p::restoreNvRam()
{
	ComputerFx3600p* compP = static_cast<ComputerFx3600p*>(mComputerP);
	DisplayFx3600p*  dsplP = static_cast<DisplayFx3600p* >(mDisplayP );
	string::size_type i = 0; // Laufindex im nvRam-String
	string::size_type sPos;  //Position des naechsten Separators
	string nvRam = mCfgsP->mNvRam;
	string am, cm, dm, dt, fs, m, k1, k2, k3, k4 , k5, k6, lm, im, p1, p2;

	                                                                        // cout << "nvRam: " << nvRam << endl;
	sPos = nvRam.find(mSep, i); am = nvRam.substr(i, sPos-i); i = sPos + 1; // cout << "am: " << am << "<" << endl;
	sPos = nvRam.find(mSep, i); cm = nvRam.substr(i, sPos-i); i = sPos + 1; // cout << "cm: " << cm << "<" << endl;
	sPos = nvRam.find(mSep, i); dm = nvRam.substr(i, sPos-i); i = sPos + 1; // cout << "dm: " << dm << "<" << endl;
	sPos = nvRam.find(mSep, i); dt = nvRam.substr(i, sPos-i); i = sPos + 1; // cout << "dt: " << dt << "<" << endl;
	sPos = nvRam.find(mSep, i); fs = nvRam.substr(i, sPos-i); i = sPos + 1; // cout << "fs: " << fs << "<" << endl;
	sPos = nvRam.find(mSep, i); m  = nvRam.substr(i, sPos-i); i = sPos + 1; // cout << " m: " <<  m << "<" << endl;
	sPos = nvRam.find(mSep, i); k1 = nvRam.substr(i, sPos-i); i = sPos + 1; // cout << "k1: " << k1 << "<" << endl;
	sPos = nvRam.find(mSep, i); k2 = nvRam.substr(i, sPos-i); i = sPos + 1; // cout << "k2: " << k2 << "<" << endl;
	sPos = nvRam.find(mSep, i); k3 = nvRam.substr(i, sPos-i); i = sPos + 1; // cout << "k3: " << k3 << "<" << endl;
	sPos = nvRam.find(mSep, i); k4 = nvRam.substr(i, sPos-i); i = sPos + 1; // cout << "k4: " << k4 << "<" << endl;
	sPos = nvRam.find(mSep, i); k5 = nvRam.substr(i, sPos-i); i = sPos + 1; // cout << "k5: " << k5 << "<" << endl;
	sPos = nvRam.find(mSep, i); k6 = nvRam.substr(i, sPos-i); i = sPos + 1; // cout << "k6: " << k6 << "<" << endl;
	sPos = nvRam.find(mSep, i); lm = nvRam.substr(i, sPos-i); i = sPos + 1; // cout << "lm: " << lm << "<" << endl;
	sPos = nvRam.find(mSep, i); im = nvRam.substr(i, sPos-i); i = sPos + 1; // cout << "im: " << im << "<" << endl;
	sPos = nvRam.find(mSep, i); p1 = nvRam.substr(i, sPos-i); i = sPos + 1; // cout << "p1: " << p1 << "<" << endl;
	sPos = nvRam.find(mSep, i); p2 = nvRam.substr(i, sPos-i);               // cout << "p2: " << p2 << "<" << endl;
	if (am != "") compP->setAngMode(AngMode(atoi(am.c_str())));
	if (cm != "") compP->setCalcMode(CalcMode(atoi(cm.c_str())));
	if (dm != "") dsplP->setDspMode(DspMode(atoi(dm.c_str())));
	if (dt != "") dsplP->setDigits (        atoi(dt.c_str()) );
	if (fs != "") dsplP->setForceSci(0.01); else dsplP->setForceSci(0.0000001);
	if (m  != "") compP->setMemory (Number (m ));
	if (k1 != "") compP->setKmem(0, Number (k1));
	if (k2 != "") compP->setKmem(1, Number (k2));
	if (k3 != "") compP->setKmem(2, Number (k3));
	if (k4 != "") compP->setKmem(3, Number (k4));
	if (k5 != "") compP->setKmem(4, Number (k5));
	if (k6 != "") compP->setKmem(5, Number (k6));
	if (lm != "") { compP->setIsLrnMode(true); compP->setActivePrg(3); } 
	if (im != "") compP->setIsIdxMode(true);
	if (p1 != "") restorePrg(0, p1);
	if (p2 != "") restorePrg(1, p2);
}

//************************* ************************* **************************

/// Programme aus der Konfigurationsdatei werden in die einzeilige Darstellung
/// umgewandelt, die vom Rechner ausgefuehrt werden koennen. Alle CRs werden in
/// Leerzeichen gewandelt, alle Kommentare entfernt.
/// \param i   Nummer des Programmspeichers
/// \param prg der Programmstring
//  fse, 17.05.03

void CalculatorFx3600p::restorePrg(int i, string prg)
{
	ComputerFx3600p* compP = static_cast<ComputerFx3600p*>(mComputerP);
	string p = Util::replace(prg, "\\n", "\n");  // wieder echte CRs erzeugen

	compP->setPrgMemOrg(i, p);                   // Origale speichern
	p = Util::stripLines(p);                     // alle Kommentare/CRs entfernen
	compP->setPrgMem(i, p);                      // Kurzform speichern
}           

//******************************************************************************

/// Achtung: Aufraeumen wird im wesentlichen von ~CalculatorBase() uebernommen!
/// Hier werden nur die persistenten Daten in die Konfigurationsdatei gesichert.
//  fse, 11.05.02

CalculatorFx3600p::~CalculatorFx3600p()
{
	mCfgsP->mNvRam = nvRamToStr();
	mCfgsP->updtCfgFile();
};

//************************************** ***************************************

/// Schreibt die permanenten Daten (K-Register, M-Speicher) in String.
/// Holt hierzu alle Daten, die nach dem Einschalten wieder vorhanden sein sollen,
/// aus den entsprechenden Variablen und legt sie im zurueckzugebenden String ab.
/// Meistens wird 0 als Default-Wert betrachtet, der dann nicht (bzw. als 
/// Leerstring) gespeichert wird. Dies geschieht nur, damit die
/// Konfigurationsdatei schoener aussieht.
/// \return String, der die Werte der permanenten Daten enthaelt
//  fse, 09.03.03

string CalculatorFx3600p::nvRamToStr()
{
	ComputerFx3600p* compP     = static_cast<ComputerFx3600p*>(mComputerP);
	DisplayFx3600p*  dsplP     = static_cast<DisplayFx3600p* >(mDisplayP );
	AngMode          am        = compP->getAngMode();
	CalcMode         cm        = compP->getCalcMode();
	DspMode          dm        = dsplP->getDspMode();
	int              dt        = dsplP->getDigits();
	bool             isLowFs   = dsplP->getForceSci() < 1e-4; // Grenzen: 1e-2 und 1e-7
	bool             isLrnMode = compP->getIsLrnMode();
	bool             isIdxMode = compP->getIsIdxMode();
	Number           m         = compP->getMemory();
	string           p1        = compP->getPrgMemOrg(0);
	string           p2        = compP->getPrgMemOrg(1);
	ostringstream    str;
	string           res;
	int              i;

	str.precision(15);
	if (am != DEG) str << am; str << mSep;
	if (cm != CMP) str << cm; str << mSep;
	if (dm != NRM) str << dm; str << mSep;
	if (dt != -1 ) str << dt; str << mSep;
	if (!isLowFs ) str << isLowFs; str << mSep;
	if (m  != 0  ) str << m ; str << mSep;
	for (i = 0; i < 6; ++i)
	{
		m = compP->getKmem(i);
		if (m != 0) str << m;
		str << mSep;
	}
	if (isLrnMode) str << isLrnMode; str << mSep;
	if (isIdxMode) str << isIdxMode; str << mSep;
	if (p1 != "") { res = Util::replace(p1, "\n", "\\n"); str << res << mSep; }
	if (p2 != "") { res = Util::replace(p2, "\n", "\\n"); str << res;         }
//	str << ends;
	res = str.str();
	return res;
}

//******************************************************************************
// weitere Symbole des Casio fx-3600P:
//******************************************************************************

/// Schaltet den Rechner in den Integrations-Modus.
//  fse, 04.03.03

void AcIdx::process()
{
	ComputerFx3600p* compP = static_cast<ComputerFx3600p*>(mCompP);

	compP->setCalcMode(CMP);
	compP->setActivePrg(0);
	compP->setIsLrnMode(false);
	compP->setIsIdxMode(true);
}

//******************************************************************************

/// Schaltet den Rechner in den "Learn"-Modus, um ihn zu programmieren.
//  fse, 04.03.03

void AcLrn::process()
{
	ComputerFx3600p* compP = static_cast<ComputerFx3600p*>(mCompP);

	compP->setCalcMode(CMP);
	compP->setActivePrg(3);
	compP->setIsLrnMode(true);
	compP->setIsIdxMode(false);
}

//******************************************************************************

/// Schaltet den Rechner in den normalen Run-Modus.
/// Wenn der Rechner als Konsolenanwendung gestartet wurde, muss nur der 
/// Modus von "lrn" auf "rnm" gestellt werden.
/// In der Dialogversion wird diese Funktion durch die Tastenfolge "mode" "."
/// aufgerufen und das "mode" wurde bereits in den Programmspeicher geschrieben.
/// Dieses muss daher wieder entfernt werden.
//  fse, 08.03.03

void AcRnm::process()
{
	ComputerFx3600p* compP = static_cast<ComputerFx3600p*>(mCompP);

	// Lern-/Integral-Modus abschalten:
	compP->setActivePrg(0);
	compP->setIsLrnMode(false);
	compP->setIsIdxMode(false);
}

//******************************************************************************

/// Loescht einen oder beide der Programmspeicher.
//  fse, 04.03.03

void AcPcl::process()
{
	ComputerFx3600p* compP = static_cast<ComputerFx3600p*>(mCompP);
	int actPrg = compP->getActivePrg();

	if (actPrg == 1) compP->setPrgMem   (0, "");
	if (actPrg == 1) compP->setPrgMemOrg(0, "");
	if (actPrg == 2) compP->setPrgMem   (1, "");
	if (actPrg == 2) compP->setPrgMemOrg(1, "");
	if (actPrg == 3)
	{
		compP->setPrgMem   (0, "");
		compP->setPrgMem   (1, "");
		compP->setPrgMemOrg(0, "");
		compP->setPrgMemOrg(1, "");
	}
}

//******************************************************************************

/// Ermittelt die beim Integrieren eingegebene Streifenanzahl.
//  fse, 23.03.03

void AcStrps::process()
{
	ComputerFx3600p* compP = static_cast<ComputerFx3600p*>(mCompP);
	Number val = mDsplP->getVal();
	Number stripes;

	if (!(val.isFrc() && val >= 1 && val <= 9))
	{
		mDsplP->setError();
		return;
	}
	stripes = pow(2, val.getVal());
	compP->setKmem(2, val); // dyadischen Logarithmus der Streifenanzahl merken
	mCompP->popNumDepOnLastSym();
	mCompP->pushNum(stripes);
}


//******************************************************************************

/// Prozessiert die Tasten P1 und P2.
//  fse, 08.03.03

void AcPrgCtrl::processP1P2(int actPrg)
{
	ComputerFx3600p* compP = static_cast<ComputerFx3600p*>(mCompP);

	compP->setActivePrg(actPrg);
	if (compP->getIsLrnMode())
	{
		compP->setPrgMem   (actPrg-1, "");
		compP->setPrgMemOrg(actPrg-1, "");
	}
	else
	{
		if (compP->getIsIdxMode())
		{
			compP->intgrInputAb();
		}
		else
		{
			string runningPrg = compP->getPrgMem(actPrg-1);
			compP->setRunningPrg(runningPrg);
			compP->execPrg();
		}
	}
}

//******************************************************************************

/// Prozessiert die Taste P1.
//  fse, 04.03.03

void PrgCtrlP1::process()
{
	processP1P2(1);
}

//******************************************************************************

/// Prozessiert die Taste P2.
//  fse, 04.03.03

void PrgCtrlP2::process()
{
	processP1P2(2);
}

//******************************************************************************

/// Prozessiert das "RUN"-Symbol.
/// Das Symbol wird gesendet, wenn gerade ein Programm ablaeuft und dieses
/// unterbrochen wurde, weil der Benutzer eine Eingabe ("ENT") gemacht hat.
/// Anschliessend wird das Programm fortgesetzt. Das geht deswegen, weil der
/// Programmstring abgebaut wurde und nun genau am Anfang des aktuellen Strings
/// (= aktuelle Programmposition) weitergefuehrt wird.
/// Im Integrationsmodus wird das Symbol ebenfalls benutzt: Auch hier wird
/// das aktuelle Programm ausgefuehrt, das so gepatched wurde, dass die
/// Eingaben entgegengenommen werden. Wenn alle Eingaben gemacht sind (dann ist
/// mIdxStep >= 2), wird aber nicht mehr die Programmausfuehrung fortgesetzt,
/// sondern die Integration gestartet. Siehe auch Kommentar bei integrate().
//  fse, 04.03.03

void PrgCtrlRun::process()
{
	ComputerFx3600p* compP = static_cast<ComputerFx3600p*>(mCompP);
	int idxStep = compP->getIdxStep();
	ostringstream str;

	if (!compP->getIsLrnMode())
	{
		compP->setIsHltMode(false); // "P1" oder "P2" im Display abschalten
		str << mDsplP->getValue() << " "; // << ends; // aktuellen Anzeigewert holen
		//cout << str.str() << "<<" << endl;
		compP->execPrg();           // Programm fortsetzen
	}
	if (idxStep >= 1)
	{
		compP->setIdxStep(idxStep+1);
		if (idxStep >= 2) compP->integrate();
	}
}

//******************************************************************************

/// Prozessiert das "ent"-Symbol.
/// Mit dem ent-Symbol kann man bei der Programmierung erreichen, dass ein
/// neuer Wert von der Anzeige gelesen und in das Programm eingefuegt wird.
/// Bei der Programmierung wird das "ent" einfach an den Programm-String
/// angehaengt und ebenso die Zahl, die bei der Programmierung als Platzhalter
/// dient. Diese wird von der execPrg()-Funktion des Calculators aus dem String
/// entfernt und dann hier durch die aktuelle Eingabe ersetzt. Anschliessend
/// wird das Programm fortgesetzt.
//  fse, 04.03.03

void PrgCtrlEnt::process()
{
	ComputerFx3600p* compP = static_cast<ComputerFx3600p*>(mCompP);

	if (compP->getIsLrnMode())
	{
		compP->setIsEntMode(true);                // "ENT" im Display einschalten
	}
	else
	{
		ostringstream str;
		compP->setIsEntMode(false);                // "ENT" im Display abschalten
		str << mDsplP->getValue() << " ";//<<ends; // aktuellen Anzeigewert holen
		compP->precedeRunningPrg(str.str());       // akt. Wert vor lfd. Programm setzen
		compP->execPrg();                          // Programm fortsetzen
	}
}

//******************************************************************************

//  fse, 04.03.03

void PrgCtrlHlt::process()
{
	// nothing to do
}


//******************************************************************************

/// Prozessiert das xgt0-Symbol.
/// Wenn gerade ein Programm ausgefuehrt wird, muss das Programm am Beginn
/// fortgesetzt werden, falls die aktuelle Anzeige groesser ist als 0.0.
/// Dies wird dadurch erreicht, dass einfach der String des gerade ausgefuehrten
/// Programms (der ja bei der Ausfuehrung Symbol fuer Symbol abgebaut wird)
/// wieder in den Startzustand versetzt wird.
//  fse, 04.03.03

void PrgCtrlXgt0::process()
{
	ComputerFx3600p* compP = static_cast<ComputerFx3600p*>(mCompP);
	Number  val = mCompP->topNum();

	if (!compP->getIsLrnMode())
	{
		if (val > 0.0)
			compP->resetRunningPrg();
	}
}

//******************************************************************************

/// Prozessiert das xlem-Symbol.
/// Siehe Funktion PrgCtrlXgt0::process().
//  fse, 04.03.03

void PrgCtrlXleM::process()
{
	ComputerFx3600p* compP = static_cast<ComputerFx3600p*>(mCompP);
	Number  val = mCompP->topNum();
	Number  mem = mCompP->getMemory();

	if (!compP->getIsLrnMode())
	{
		//cout << "val=" << val << ", mem = " << mem << endl;
		if (val <= mem)
			compP->resetRunningPrg();
	}
}

//******************************************************************************

/// Prozessiert das rtn-Symbol.
/// Siehe Funktion PrgCtrlXgt0::process().
//  fse, 04.03.03

void PrgCtrlRtn::process()
{
	ComputerFx3600p* compP = static_cast<ComputerFx3600p*>(mCompP);

	if (!compP->getIsLrnMode())
	{
		compP->resetRunningPrg();
	}
}

//******************************************************************************
//******************************************************************************
//************************************** ***************************************
//************************* ************************* **************************
//****************** ******************* ******************* *******************
//************** *************** *************** *************** ***************
//************ ************ ************ ************ ************ *************
