//******************************************************************************
// Copyright (c) 2002-2017 by Friedemann Seebass, Germany.
// Dieses Programm ist freie Software. Siehe GNU GPL Notiz in Datei main.cpp.
// This program is free software. See the GNU GPL notice in file main.cpp.
// Projekt: Almetare - Alle meine Taschenrechner
//******************************************************************************
/// \file fx3600p_main_dlg.cpp
///       Implementierung der Casio-fx-3600P-Dialogbox (Modul special).
//******************************************************************************
// Wann       Wer      Was
// ---------- -------- ---------------------------------------------------------
// 28.01.2005 fse      Fx3600pMainDlg(): QSound-Objekte nur anlegen, wenn
//                     MORE_REALISTIC gesetzt
// 28.01.2005 fse      butPrg(): Signatur Aufruf von Programm-Editor geaendert
// 05.09.2004 fse      butPrg(): Aufruf von Dateneditor
// 17.11.2003 fse      strstream durch sstream ersetzt
// 15.11.2003 fse      Fx3600pMainDlg::Fx3600pMainDlg(): wflags an boot() ueberg.
// 11.08.2003 fse      Fx3600pMainDlg::butExp(): Aenderung wg. isFrac
// 10.08.2003 fse      Fx3600pMainDlg::butAbc(): lastInP wird nicht mehr gesetzt
// 04.08.2003 fse      Fx3600pMainDlg::butSum(): INV immer abschalten
// 01.03.2003 fse      erzeugt
//******************************************************************************

#include <sstream>
#ifdef _WIN32
	#include <time.h>        // clock_t
#else
	#include <unistd.h>      // usleep()
#endif
#include <qpushbutton.h>
#include <qapplication.h>   // QApplication::palette()
#include <qevent.h>         // QMouseEvent
#include <qlabel.h>
#include <qtoolbutton.h>    // sonst geht sendEvent(mCalcTypeP, &me) nicht!
#include <qobjectlist.h>    // QObjectList
#include <qwmatrix.h>
#include <qpixmap.h>
#include "fx3600p_main_dlg.h"
#include "cfg_dlg.h"
#include "fx3600p_prg_dlg.h"
#include "casio_sci_dat_dlg.h"

//******************************************************************************
//******************************************************************************

#ifdef _WIN32

/// Funktion pausiert unter Windows die angegebene Zahl von Millisekunden.
/// Bloed: Es wird mit voller Prozessorlast gewartet. Leider gibt es unter
/// Windows die richtige usleep()-Funktion nicht.
/// Ausserdem sind hier die 150 ms, die unter Linux passen, etwas zu viel;
/// daher werden die uebergebenen 150 ms etwas gekuerzt.
//  fse, 02.04.03

void usleep(clock_t wait)
{
   clock_t goal;
   goal = wait-50 + clock();
   while( goal > clock() )
      ;
}

#endif

//******************************************************************************

/// Konstruktion des Dialogs.
/// Initialisierung der Members.
/// \param parentP Zeiger auf Vaterdialog
/// \param name    Name des Dialogs
/// \param modal   Flag, ob Dialog modal oder nicht
/// \param fl      diverse Qt-Dialog-Flags
//  fse, 10.03.03
//  fse, 15.11.03: Aufruf von boot() mit getWFlags()
//  fse, 28.01.05: QSound-Objekte nur anlegen, wenn MORE_REALISTIC gesetzt

Fx3600pMainDlg::Fx3600pMainDlg(QWidget* parentP, const char* name, bool modal, WFlags fl)
	: Fx3600pMainBaseDlg(parentP, name, modal, fl), mCalc(qApp->argc(), qApp->argv())
{
	string  onSoundStr = mCalc.getAccPath() + "/wav/fx3600p_on.wav";
	string offSoundStr = mCalc.getAccPath() + "/wav/fx3600p_off.wav";

	mCompP = static_cast<ComputerFx3600p*>(mCalc.getComputerP());
	mDsplP = static_cast<DisplayFx3600p *>(mCalc.getDisplayP());

	QRect qr = frameGeometry();
	int yDspl = mLbWgModeP->y();      // letztes Label des Info-Feldes  zur Bestimmung 
	int hDspl = mLbWgModeP->height(); // der Groesse der dunklen "Blende" hinter Display

	// spezielle numerische Werte der Rechnergeometrie:
	mDefaultWidth  = 229; // Default-Breite des Rechners (falls keine Pixmap, 50% der Pixmap-Breite)
	mDefaultHeight = 426; // Default-Hoehe  des Rechners (falls keine Pixmap, 50% der Pixmap-Hoehe)
	mOrgLcdRect.setCoords(62, 54, 400, 130); // Geometrie des LCD (Koord. und Groesse in calc.png)

	// Kopieren von Widgetpointern des Dialogs in Members der Basisklasse:
	mLcdStrBaseP  = mLcdStrP;
	mModesBaseP   = mModesP;
	mDisplayBaseP = mDisplayP;
	mSeebassBaseP = mSeebassP;

	// globale Farbdefinitionen:
	mColLgtbrown.setRgb(146, 127, 131); // Farbe um das Display herum
	mColDrkbrown.setRgb( 84, 83 ,  72); // dunkleres Braun f. Mode-Labels
	mColLgtblue. setRgb(220, 220, 255); // helleres Blau fuer LR- und SD-Label
	mColDrkblue. setRgb( 70, 124, 188); // dunkleres Blau fuer Typbezeichnung
	mColOrngInv. setRgb(142, 103,  24); // Orange fuer INV-Button
	mColOrngLbl. setRgb(149,  48,   9); // Orange fuer Button-Label
	mColBeige.   setRgb(237, 236, 216); // Beige fuer die kleinen Buttons
	mColOrngAc.  setRgb(250, 183, 150); // Farbe "C" und "AC"

	// "Booten" des Rechners:
	mLbBackgroundP->setGeometry(0, 0, qr.width(), yDspl + hDspl*2 - 1);
	mLbBackgroundP->lower();
	boot(&mCalc, this, getWFlags()); // getWFlags() ist protected, daher Aufruf nicht in boot()
	mOnSoundP  = NULL;
	mOffSoundP = NULL;
	if (mCalcP->getCfgsP()->mDisplayMode == MORE_REALISTIC)
	{
		mOnSoundP  = new QSound( onSoundStr.c_str());
		mOffSoundP = new QSound(offSoundStr.c_str());
	}
	setLabelColors();
	mIsHardOff = false;
	mMaxDigitNum = 10;

	// Timer:
	connect(&mTimOff  , SIGNAL(timeout()), this, SLOT(timeoutOff  ()));
	connect(&mTimBlink, SIGNAL(timeout()), this, SLOT(timeoutBlink()));
	if (mCompP->getActivePrg() == 3) mTimBlink.start(400);
}

//************************************** ***************************************

/// Geht alle Widgets durch und setzt die Farben der Labels.
/// Die Farben sind im Label-Namen kodiert:
/// "mLbB_PiP"    => Vodergrundfarbe schwarz, keine spezielle Hintergrundfarbe
/// "mLbR_FctP"   => Vodergrundfarbe rot, keine spezielle Hintergrundfarbe
/// "mLbRbXaveP"  => Vodergrundfarbe rot, Hintergrundfarbe blau
/// "mLbBbXaveP"  => Vodergrundfarbe schwarz, Hintergrundfarbe blau
/// Die Kodierung stammt vom fx-85v, hier ist es etwas anders:
/// Rot ist orange, grau ist braun
//  fse, 18.07.02

void Fx3600pMainDlg::setLabelColors()
{
	const QObjectList*     qolP = children();
	QListIterator<QObject> it(*qolP);
	QObject*               qoP;
	QLabel*                qlP;
	QPalette p = QApplication::palette();
	QPalette pRed, pBlackBlue, pRedBlue, pWhiteGray, pWightLgtbrown;
	string name; // in Qt-Designer, hier fuer mSeebassP

	pRed = pWightLgtbrown = pBlackBlue = pRedBlue = pWhiteGray = p;

	pRed      .setColor (QColorGroup::Foreground, mColOrngLbl);
	pBlackBlue.setColor (QColorGroup::Background, mColLgtblue);

	pWightLgtbrown.setColor (QColorGroup::Foreground, Qt::white);
	pWightLgtbrown.setColor (QColorGroup::Background, mColLgtbrown);

	pRedBlue  .setColor (QColorGroup::Foreground, mColOrngLbl);
	pRedBlue  .setColor (QColorGroup::Background, mColLgtblue);

	pWhiteGray.setColor (QColorGroup::Foreground, Qt::white);
	pWhiteGray.setColor (QColorGroup::Background, mColDrkbrown);

	// alle Widgets durchgehen:
	qoP = it.current();
	while (qoP != NULL)
	{
		qlP = dynamic_cast<QLabel*>(qoP);
		if (qlP)
		{
			name = qlP->name();
			if (name[3] != 'W') qlP->setAlignment(Qt::AlignCenter);
			if (name[3] == 'R' && name[4] == 'b') { qlP->setPalette(pRedBlue  ); goto lbl; }
			if (name[3] == 'B' && name[4] == 'b') { qlP->setPalette(pBlackBlue); goto lbl; }
			if (name[3] == 'W' && name[4] == 'g') { qlP->setPalette(pWhiteGray); goto lbl; }
			if (name[3] == 'R'                  ) { qlP->setPalette(pRed      ); goto lbl; }
		}
		lbl:
		++it;
		qoP = it.current();
	}

	// Korrektur einiger Labels, die obige Schleife auch aendert:
//	mSeebassP->setAlignment(Qt::AlignCenter);
//	mLbWgModeP->setAlignment(Qt::AlignCenter);
	mSeebassP->setPalette(pWightLgtbrown);
	mLbBbSdP->setAlignment(Qt::AlignLeft);
	mLbBbLrP->setAlignment(Qt::AlignLeft);
}

//******************************************************************************

/// Speichert vor dem Verlassen die Groessendaten des Dialogs in der
/// Konfigurationsdatei.
//  fse, 17.07.02

Fx3600pMainDlg::~Fx3600pMainDlg()
{
	Cfgs* cfgsP = mCalcP->getCfgsP();

	QRect rect = geometry();
	cfgsP->mXpos   = rect.x();
	cfgsP->mYpos   = rect.y();
	cfgsP->mWidth  = rect.width();
	cfgsP->mHeight = rect.height();
	cfgsP->updtCfgFile();
}

//******************************************************************************

/// "Abschalten" des Rechners nach Ablauf des 6-Minuten Timers.
//  fse, 15.06.02

void Fx3600pMainDlg::timeoutOff()
{
	switchOff();
}

//******************************************************************************

/// 0,4-Sekunden Timer fuers Blinken von P1P2.
//  fse, 16.03.03

void Fx3600pMainDlg::timeoutBlink()
{
	static bool isOff = true;
	string modesShort = mDsplP->getModesShort();

	if (mIsHardOff || mIsOff) return;
	if (isOff)
	{
		modesShort[13] = ' '; // "mOIMKNdLSDRGE12"
		modesShort[14] = ' ';
	}
	clearLcd();
	setLcd(mDsplP->getSeven());
	setLcd(modesShort);
	update(mLcdRect);
	isOff = !isOff;
}

//******************************************************************************
// virtuelle Funktionen:
//******************************************************************************

/// Wird von extractSym() aufgerufen, erlaubt Vorverarbeitung.
/// Die Funktion wird benoetigt, um das Flackern der Anzeige bei jedem
/// Tastendruck zu erzeugen. Ausserdem wird das Display immer verdunkelt, wenn
/// ein Programm ausgefuehrt wird.
/// \param sym Symbol auf der Taste (ungempapped)
//  fse, 28.03.03

void Fx3600pMainDlg::preExtractSym(string sym)
{
	DisplayMode dm = mCalcP->getCfgsP()->mDisplayMode;

	if (dm == MORE_REALISTIC || sym == "p1" || mCompP->getIdxStep() >= 1)
	{
		clearLcdButOn();
	}
}

//************************************** ***************************************

/// Loescht die Anzeige bis auf das "ON"-Zeichen.
//  fse, 16.03.03

void Fx3600pMainDlg::clearLcdButOn()
{
	clearLcd();
	setLcd("mO             ");
	update(mLcdRect);
	qApp->processEvents();
	usleep(150);
}

//******************************************************************************

/// Veraendert die Tool-Buttons.
//  fse, 30.03.03

void Fx3600pMainDlg::adjustRealisticSkin()
{
	mCalcTypeP->setAutoRaise(true);
	mPrgP     ->setAutoRaise(true);
	mOffP     ->show();
}

//******************************************************************************

/// Setzt die Farben der Buttons.
/// Die kleinen werden beige, die grossen schwarz.
//  fse, 06.06.02

void Fx3600pMainDlg::adjustSimpleSkin()
{
	const QObjectList*     qolP = children();
	QListIterator<QObject> it(*qolP);
	QObject*               qoP;
	QButton*               qbP;
	QPalette               p = QApplication::palette();
	QPalette               pWhiteLgtbrown, pWhiteBlue, pWhiteOrngInv, pBlackBeige;
	QPalette               pWhiteBlack   , pOrngBrown, pWhiteDrkbrown;
	int                    h, w;
	string name;

	pWhiteLgtbrown = pWhiteBlue  = pWhiteOrngInv = pBlackBeige
		           = pWhiteBlack = pOrngBrown    = pWhiteDrkbrown = p;

	pWhiteLgtbrown.setColor (QColorGroup::Button    , mColLgtbrown);
	pWhiteLgtbrown.setColor (QColorGroup::ButtonText, Qt::white );
	pWhiteLgtbrown.setColor (QColorGroup::Foreground, Qt::white   );

	pWhiteBlue.setColor (QColorGroup::Button    , mColDrkblue);
	pWhiteBlue.setColor (QColorGroup::ButtonText, Qt::white );
	pWhiteBlue.setColor (QColorGroup::Foreground, Qt::white   );

	pWhiteOrngInv.setColor (QColorGroup::Button    , mColOrngInv);
	pWhiteOrngInv.setColor (QColorGroup::ButtonText, Qt::white );

	pBlackBeige.setColor (QColorGroup::Button    , mColBeige);
	pBlackBeige.setColor (QColorGroup::ButtonText, Qt::black   );

	pWhiteBlack.setColor (QColorGroup::Button    , Qt::black   );
	pWhiteBlack.setColor (QColorGroup::ButtonText, Qt::white   );

	pOrngBrown.setColor (QColorGroup::Button    , mColDrkbrown);
	pOrngBrown  .setColor (QColorGroup::ButtonText, mColOrngAc   );

	pWhiteDrkbrown.setColor (QColorGroup::Button    , mColDrkbrown);
	pWhiteDrkbrown.setColor (QColorGroup::ButtonText, Qt::white );

	// alle Widgets durchgehen:
	qoP = it.current();
	while (qoP != NULL)
	{
		qbP = dynamic_cast<QButton*>(qoP);
		if (qbP)
		{
			w = qbP->width();
			h = qbP->height();

			if (double(w)/double(h) < 1.4)   // Seitenverhaeltnis eines Buttons
				qbP->setPalette(pBlackBeige); // die kleinen sind schmaler, das sind die grauen
			else
				qbP->setPalette(pWhiteBlack); // das ist ein grosser, der wird schwarz
		}
		++it;
		qoP = it.current();
	}

	mCalcTypeP->setPalette(pWhiteLgtbrown);
//	mSeebassP->setPalette(pWhiteBrown);
	mPrgP    ->setPalette(pWhiteBlue);
	mInvP    ->setPalette(pWhiteOrngInv);

	mCP ->setPalette(pOrngBrown);
	mAcP->setPalette(pOrngBrown);

	mMulP->setPalette(pWhiteDrkbrown);
	mDivP->setPalette(pWhiteDrkbrown);
	mAddP->setPalette(pWhiteDrkbrown);
	mSubP->setPalette(pWhiteDrkbrown);
	mExpP->setPalette(pWhiteDrkbrown);
	mEquP->setPalette(pWhiteDrkbrown);

	mRunP->setPalette(pBlackBeige);

	mCalcTypeP->setAutoRaise(false);
	mPrgP     ->setAutoRaise(false);
	mOffP     ->hide();
}

//******************************************************************************

/// Entfernt die schwarze "Blende" hinter dem Display.
//  fse, 06.06.02

void Fx3600pMainDlg::adjustDebugSkin()
{
	mLbBackgroundP->hide();
	mOffP         ->show();
}

//******************************************************************************

/// Erneuert die LCD-Anzeige.
/// Da diese Prozedur nacb jedem Tastendruck aufgerufen wird, koennen hier auch
/// die Timer gestartet werden.
/// Das Argument isDoPerhapsNot wird nur vom TI-30 benoetigt im Rahmen seiner
/// "Thinking"-Sequenz.
//  fse, 15.06.02

void Fx3600pMainDlg::setDisplayFields(bool isDoPerhapsNot)
{
	DisplayMode  dm    = mCalcP->getCfgsP()->mDisplayMode;

	isDoPerhapsNot = isDoPerhapsNot; // Compiler-Warnung eliminieren
	if (dm == MORE_REALISTIC)
	{
		mTimOff.start(360000); // aus nach 6 Minuten
		//mTimOff.start(5000); // zum Testen
	}
	else
	{ // Timer stoppen, falls Konfigurationsdialog Display-Mode geaendert hat
		mTimOff.stop();
	}
	if (dm != SIMPLE && mIsPixLoaded)
	{
		string modesShort = mDsplP->getModesShort();
		clearLcd();
		setLcd(mDsplP->getSeven());
		setLcd(modesShort);
		//setLcd("mOIMKNdLSDRGE12"); // nur zum Testen
	}
}

//************************* ************************* **************************

/// Setzt den String str in die passenden Buchstaben des LCDs um.
/// \param str auszugebender String
//  fse, 08.06.02

void Fx3600pMainDlg::setLcd(string str)
{
	QPixmap*  pmP  = const_cast<QPixmap*>(backgroundPixmap());
	int       xSrc, ySrc, wSrc, hSrc, yRow1, wMntDgt, wExpDgt, dummy;
	int       xDst, yDst, xOffset = (int)(8 * mXfact);
	double    xx   = mLcdRect.x() + xOffset;
	char      c;
	bool      isPointSucc;
	bool      isMode = str[0] == 'm'; // gibt an, ob String mit den Modes uebergeben wurde
	const unsigned int mantDigits = 10;

	if (isMode) str.replace(0, 1, "");
	rowColToPix(1, 0, dummy, yRow1, wMntDgt, dummy);
	rowColToPix(3, 0, dummy, dummy, wExpDgt, dummy);
	str += " "; // fuer Vergleich mit Nachfolger '.'
	yDst = mLcdRect.y() + int(yRow1 * mYfact); // Zahlen-Zeile
	for (unsigned int i = 0; i< str.size()-1; i++)
	{
		c = str[i];
		if (c == '.' || c == ' ') continue;
		if (str[i+1] == '.') isPointSucc = true; else isPointSucc = false;
		getSourceCoords(c, isPointSucc, i>mantDigits, isMode, xSrc, ySrc, wSrc, hSrc);
		xSrc = int(xSrc * mXfact);
		ySrc = int(ySrc * mYfact);
		wSrc = int(wSrc * mXfact);
		hSrc = int(hSrc * mYfact);
		xDst = int(xx);
		if (isMode)
		{ // Modes-Zeile:
			switch (c)
			{
				case 'E': xDst = int(mLcdRect.x() + xOffset + 258*mXfact); // "ENT"
				          yDst = int(mLcdRect.y()           + 48 *mYfact); break;
				case '1': xDst = int(mLcdRect.x() + xOffset + 286*mXfact); // "P1"
				          yDst = int(mLcdRect.y()           + 48 *mYfact); break;
				case '2': xDst = int(mLcdRect.x() + xOffset + 302*mXfact); // "P2"
				          yDst = int(mLcdRect.y()           + 48 *mYfact); break;
				default:  xDst = int(mLcdRect.x() + xOffset+ xSrc);
				          yDst = int(mLcdRect.y());
			}
		}
		if (i <= mantDigits)
			xx += wMntDgt * mXfact;
		else
			xx += wExpDgt * mXfact;
		if (c != '#')
			bitBlt(pmP, xDst, yDst, &mScaledDigitsPixmap, xSrc, ySrc, wSrc, hSrc);
	}
	update(mLcdRect);
}

//****************** ******************* ******************* *******************

/// Liefert zu einem Buchstaben die zugehoerigen Pixelkoordinaten in der Pixmap.
///        
/// Die Pixmap hat folgenden Aufbau:
/// \code
/// Zeile |   0   1   2   3   4   5   6   7   8   9  10  11  12  13     Spalte
/// ------|-------------------------------------------------------------------
///    0  |   On INV  M   K  LRN Idx  LR  SD DEG RAD GRA ENT  P1  P2
///    1  |   0   1   2   3   4   5   6   7   8   9   -   E   _   [
///    2  |   0.  1.  2.  3.  4.  5.  6.  7.  8.  9.  -.  E.  _.  
///    3  |   0   1   2   3   4   5   6   7   8   9   -
/// \endcode
/// \param c           gesuchter Character
/// \param isPointSucc gibt an, ob dem Zeichen ein Punkt folgt
/// \param isExp       gibt an, ob das Zeichen im Exponentialbereich der Anzeige
/// \param isMode      gibt an, ob es sich um ein Mode-Zeichen handelt
/// \retval x          x-Koordinate des Zeichens in der Pixmap
/// \retval y          y-Koordinate des Zeichens in der Pixmap
/// \retval w          Breite des Zeichens in der Pixmap
/// \retval h          Hoehe des Zeichens in der Pixmap
// fse, 27.06.02

void Fx3600pMainDlg::getSourceCoords(char c, bool isPointSucc, bool isExp, bool isMode, int& x, int& y, int& w, int& h)
{
//	int a = 0;
	string cols = "OIMKNdLSDRGE12";       // alle moeglichen Symbole fuer die 0. Zeile
	string::size_type col = cols.find(c); // Spaltennummer in der 0. Zeile

	// Zeile 0 (Modes):
	if (isMode)
	{
		if (col != string::npos) rowColToPix(0, col, x, y, w, h);
		return;
	}

	// Zeilen 1 bis 3 (Wert):
	if (c == '' || c == '\'') { rowColToPix(2, 13, x, y, w, h); return; }
	if (c == '[') { rowColToPix(1, 13, x, y, w, h); return; }
	if ((c == '_' || c == '/') && isPointSucc) { rowColToPix(2, 12, x, y, w, h); return; }
	if (c == '_' || c == '/') { rowColToPix(1, 12, x, y, w, h); return; }
	if (c == '#' || c == ' ')
	{
		if      (isExp      ) rowColToPix(3, 11, x, y, w, h);
		else                  rowColToPix(2, 13, x, y, w, h);
		return;
	}
	if (c >= '0' && c <= '9')
	{
		if      (isExp      ) rowColToPix(3, c-'0', x, y, w, h);

		else if (isPointSucc) rowColToPix(2, c-'0', x, y, w, h);
		else                  rowColToPix(1, c-'0', x, y, w, h);
		return;
	}
	if (c == '-')
	{
		if      (isExp      ) rowColToPix(3, 10, x, y, w, h);
		else if (isPointSucc) rowColToPix(2, 10, x, y, w, h);
		else                  rowColToPix(1, 10, x, y, w, h);
		return;
	}
	if (c == 'E')
	{
		if (isPointSucc)      rowColToPix(2, 11, x, y, w, h);
		else                  rowColToPix(1, 11, x, y, w, h);
		return;
	}
	return; 
}

//************** *************** *************** *************** ***************

/// Liefert zu einer Zeilen-/Spalten-Position die zugehoerigen Pixelkoordinaten
/// in der nicht skalierten Digits-Pixmap.
/// \param  row        Zeile in der Tabelle
/// \param  col        Spalte in der Tabelle
/// \retval x          x-Koordinate des Zeichens in der Pixmap
/// \retval y          y-Koordinate des Zeichens in der Pixmap
/// \retval w          Breite des Zeichens in der Pixmap
/// \retval h          Hoehe des Zeichens in der Pixmap
//  fse, 29.06.02

void Fx3600pMainDlg::rowColToPix(int row, int col, int& x, int& y, int& w, int& h)
{
	const int yy[ 4] = {  0, 22, 68, 114 }; // y-Koord.       0., 1., 2., 3. Zeile
	const int ww[ 4] = {  0, 26, 26,  20 }; // Zeichenbreiten     1., 2., 3. Zeile
	const int hh[ 4] = { 22, 46, 46,  28 }; // Zeilenhoehen   0., 1., 2., 3. Zeile
	const int x0[15] = {  0, 34, 70, 92, 112, 148, 176, 200, 228, 262, 294, 328, 361, 381, 402 }; // 0. Zeile

	x = col * ww[row];
	y = yy[row];
	w = ww[row];
	h = hh[row];
	if (row == 0)
	{
		x = x0[col];
		w = x0[col+1] - x0[col];
	}
}

//******************************************************************************
// Button-Slots:
//******************************************************************************

/// Slot fr Schiebeschalter (ein/aus).
//  fse, 28.01.05: Sounds nur spielen, wenn geladen

void Fx3600pMainDlg::butPon()
{
	// Initialisierung geht davon aus, dass der Rechner eingeschaltet ist, d.h.
	// der groessere Bereich ist links, der kleinere rechts:
	DisplayMode dm = mCalcP->getCfgsP()->mDisplayMode;
	QPixmap     pmb(23, 29);          // groesserer  Teil des Schalterbereichs
	QPixmap     pms(21, 29);          // kleinerer Teil des Schalterbereichs
	QRect       leftIfOn (47, 291, 23, 29); // Koordinaten linker  Bereich, wenn an
	QRect       rigtIfOn (70, 291, 21, 29); // Koordinaten rechter Bereich, wenn an
	QRect       leftIfOff(47, 291, 21, 29); // Koordinaten linker  Bereich, wenn aus
	QRect       rigtIfOff(70, 291, 23, 29); // Koordinaten rechter Bereich, wenn aus
	string      ac = "ac", dummy;
	AcNrm       acNrm;

	if (!mIsPixLoaded || dm == SIMPLE) return;

	if (mIsHardOff)
	{ // Einschalten:
		if (mOnSoundP) mOnSoundP->play();
		// linken und rechten Teil des Schalterbereichs holen:
		bitBlt(&pms, QPoint(0, 0), &mBackgroundPixmap, leftIfOff); // schmaler Teil links
		bitBlt(&pmb, QPoint(0, 0), &mBackgroundPixmap, rigtIfOff); // breiter  Teil rechts
		// beide Regionen austauschen:
		bitBlt(&mBackgroundPixmap, leftIfOff.topLeft(), &pmb, QRect(0, 0, rigtIfOff.width(), rigtIfOff.height()));
		bitBlt(&mBackgroundPixmap, rigtIfOff.topLeft(), &pms, QRect(0, 0, leftIfOff.width(), leftIfOff.height()));
	}
   else
   { // Ausschalten:
		if (mOffSoundP) mOffSoundP->play();

		// linken und rechten Teil des Schalterbereichs holen:
		bitBlt(&pmb, QPoint(0, 0), &mBackgroundPixmap, leftIfOn); // breiter  Teil links
		bitBlt(&pms, QPoint(0, 0), &mBackgroundPixmap, rigtIfOn); // schmaler Teil rechts
		// beide Regionen austauschen:
		bitBlt(&mBackgroundPixmap, leftIfOn.topLeft(), &pms, QRect(0, 0, rigtIfOn.width(), rigtIfOn.height()));
		bitBlt(&mBackgroundPixmap, rigtIfOn.topLeft(), &pmb, QRect(0, 0, leftIfOn.width(), leftIfOn.height()));
	}

	acNrm.process(); // evtl. SCI oder FIX abschalten
	resizeAllWidgets(this);
	mIsHardOff = !mIsHardOff;
	mIsOff = mIsHardOff;
	if (mIsHardOff)
		clearLcd();
	else
		butActionBase(ac, dummy);
}

//******************************************************************************

/// Fuehrt den Programmierdialog aus.
//  fse, 05.09.04: Aufruf auch vom Datendialog
//  fse, 28.01.05: prgDlg() Adresse vom mCalc statt mCompP uebergeben

void Fx3600pMainDlg::butPrg()
{
	CalcMode cm = mCompP->getCalcMode();
	int      ret;

	ONL_LOG(gnl) << "button: prg" << NL;
	if (cm == CMP)
	{ // Programmdialog aufrufen:
		//WFlags fl = WStyle_Customize | WStyle_NormalBorder | WStyle_Title | WStyle_SysMenu;
		string pOrg1 = mCompP->getPrgMemOrg(0);
		string pOrg2 = mCompP->getPrgMemOrg(1);
		Fx3600pPrgDlg prgDlg(&mCalc, mApplIconPixmap, pOrg1, pOrg2, this, 0, true);//, fl);
		ret = prgDlg.exec();
		if (ret == 1)
		{
			mCompP->setPrgMemOrg(0, prgDlg.getP1Org());
			mCompP->setPrgMemOrg(1, prgDlg.getP2Org());
			mCompP->setPrgMem   (0, prgDlg.getP1   ());
			mCompP->setPrgMem   (1, prgDlg.getP2   ());
		}
	}
	else
	{
		CasioSciDatDlg datDlg(&mCalc, mApplIconPixmap, this, 0, true);
		ret = datDlg.exec();
	}
}

//******************************************************************************

/// Slot-Funktionen fuer die meisten Buttons.
//  fse, 23.02.02

void Fx3600pMainDlg::butAction()
{
	QPushButton* senderP = (QPushButton*)(sender());
	string sym = senderP->name();
	string dummy;

	if (isError() || mIsOff || isPrgRunning()) return;
	sym = extractSym(senderP);
	butActionBase(sym, dummy);
	if (sym != "sd" && sym != "lr"
	 && sym != "deg" && sym != "rad" && sym != "gra")
		setLastType(B_ACTION);
	if (mCompP->getActivePrg() != 3)
		mTimBlink.stop();
}

//******************************************************************************

/// Button AC loescht alles bis auf die Speicher.
//  fse, 12.07.02

void Fx3600pMainDlg::butAc()
{
	QPushButton* senderP = (QPushButton*)(sender());
	string sym, dummy;

	if (mIsHardOff) return;
	mIsOff = false;
	sym = extractSym(senderP);
	mDsplP->setFrcMode(STD);
	butActionBase(sym, dummy);
	setLastType(B_ACTION);
}

//******************************************************************************

/// Slot-Funktionen fuer die Mode-Buttons.
//  fse, 23.02.02

void Fx3600pMainDlg::butAnyMode()
{
	QPushButton* senderP = (QPushButton*)(sender());
	string sym;

	if (isError() || mIsOff || isPrgRunning()) return;
	sym = extractSym(senderP);
	butAnyModeBase(sym);
}

//******************************************************************************

/// Slot-Funktion fuer die Zifferntasten.

void Fx3600pMainDlg::butDigit()
{
	QPushButton* senderP = (QPushButton*)(sender());
	string sym;

	if (isError() || mIsOff || isPrgRunning()) return;
	sym = extractSym(senderP);
	resetAllModes();
	butDigitBase(sym);
	if (mCompP->getActivePrg() == 3) mTimBlink.start(400);
}

//************************************** ***************************************

/// Setzt alle Shift-Ebenen zurueck.
//  fse, 12.07.02

void Fx3600pMainDlg::resetAllModes()
{
	mCompP->setIsInv (false);
	mCompP->setIsHyp (false);
	mCompP->setIsMode(false);
}

//******************************************************************************

/// +/--Taste.
//  fse, 26.07.02

void Fx3600pMainDlg::butSign()
{
	QPushButton* senderP = (QPushButton*)(sender());
	string       sym, dummy;

	if (isError() || mIsOff || isPrgRunning()) return;
	sym = extractSym(senderP);
	if (sym == "sqr")
		butActionBase(sym, dummy);
	else	
		butSignBase();
}

//******************************************************************************

/// M+-Taste.
/// Noetig, weil der Casio vor dem Summieren erst das Ergebnis einer offenen
/// Rechnung ermittelt.
//  fse, 28.07.02
//  fse, 04.08.03: INV immer abschalten nach extractSym()

void Fx3600pMainDlg::butSum()
{
	QPushButton* senderP = (QPushButton*)(sender());
	string       sym, equ = "=", dummy;

	if (isError() || mIsOff || isPrgRunning()) return;
	sym = extractSym(senderP);
	// INV ist auf jeden Fall schon verarbeitet; wenn man es stehen liesse,
	// ergaebe z.B. "5 * INV M-" 0.25 statt 25, da das INV auf das kuenstliche
	// "=" wirkt und daher die Prozentrechnung durchfuehrte;
	mCompP->setIsInv(false);
	if (sym == "sum" || sym == "sub")
	{
		butActionBase(equ, dummy); // bei "data" und "del": nicht erst ausrechnen
	}
	butActionBase(sym, dummy);
}

//******************************************************************************

/// Taste zum Bruchrechnen.
//  fse, 10.08.03: "abc" ist kein Symbol der Symboltabelle, deshalb darf es
//                 nicht mit mCompP->setLastInP(mCompP->getSymP(sym)) als
//                 Symbol eingetragen werden.

void Fx3600pMainDlg::butAbc()
{
	QPushButton*      senderP     = (QPushButton*)(sender());
	const int         maxLen      = 8;
	int               len         = mDisplayStr.size();
	string::size_type sPos        = mDisplayStr.find('/');
	bool              isSci       = mDisplayStr.find('e') != string::npos;
	bool              isPoint     = mDisplayStr.find('.') != string::npos;
	bool              isSlashLast = mDisplayStr[len-1] == '/';
	bool              isSlash     = false;
	bool              isSlash2    = false;
	bool              isInv       = mCompP->getIsInv();
	string            sym;
	ostringstream     str;

	if (isError() || mIsOff || isPrgRunning()) return;
	if (isInv)
	{
		butAction();
		return;
	}
	sym = extractSym(senderP);
	resetAllModes();
	isSlash  =                                     sPos != string::npos;
	isSlash2 = isSlash && mDisplayStr.find('/', sPos+1) != string::npos;
	if (mLastType == B_DIGIT && !isInv)
	{
		if (isSci || isPoint || isSlash2 || isSlashLast)
		{
			updateDisplay(mDisplayStr); // fse, 10.08.03
			return;
		}
		len = mCalcP->getNumStrLen(mDisplayStr);
		if (len < maxLen) mDisplayStr += "/";
		mDsplP->setFrcMode(MIX);
		updateDisplay(mDisplayStr);
	}
	else
	{
		if (isSlash)
		{
			if (isInv)
			{
				if (isSlash2) { mDsplP->setFrcMode(FRC); mCalcP->process(mDisplayStr); }
			}
			else
			{
				mDsplP->setFrcMode(STD);

			}
		}
		else
		{
			if (!isInv) mDsplP->setFrcMode(MIX);
		}
		//mCompP->setLastInP(mCompP->getSymP(sym)); // fse, 10.08.03: "abc" ist kein Symbol
		updateDisplay();
		mDisplayStr = mDsplP->getValue();
		if (mDsplP->getFrcMode() == FRC)
			mDsplP->setFrcMode(MIX); // FRC-Mode immer nur fuer einmal Anzeigen
	}
}

//******************************************************************************

/// Button EXP.
//  fse, 10.07.02
//  fse, 12.08.03: Wenn isFrac darf Funktion nur dann verlassen werden, wenn
//                 zuvor eine Ziffer eingegeben wurde. Sonst kann man z.B. die
//                 Rechnung "3_4*pi" nicht durchfuehren.

void Fx3600pMainDlg::butExp()
{
	QPushButton* senderP = (QPushButton*)(sender());
	bool         isFrac = mDisplayStr.find('/') != string::npos;
	string       sym, dummy;

	if (isFrac && mLastType == B_DIGIT) return;
	if (isError() || mIsOff || isPrgRunning()) return;
	sym = extractSym(senderP);
	if (mLastType == B_DIGIT)
	{
		if (mDisplayStr.find('e') != string::npos)
		{
			updateDisplay(mDisplayStr);
			return;
		}
		if (mDisplayStr.find('.') == string::npos) mDisplayStr += ".";
		mDisplayStr += "e+000";
		resetAllModes();
		updateDisplay(mDisplayStr);
	}
	else
	{
		butActionBase(sym, dummy);
	}
}

//******************************************************************************

/// Schaltet die Anzeige in den Engineering-Modus bzw. verschiebt das Komma um
/// jeweils drei Stellen nach rechts.
//  fse, 15.07.02

void Fx3600pMainDlg::butEng()
{
	DisplayMode  dm      = mCalcP->getCfgsP()->mDisplayMode;
	QPushButton* senderP = (QPushButton*)(sender());
	string       sym, dummy;
	bool         isInv = mCompP->getIsInv();

	if (isError() || mIsOff || isPrgRunning()) return;
	sym = extractSym(senderP);
	mCompP->aftProcess(sym); // evtl. INV-Modus wieder ausschalten
	if (dm == SIMPLE) return;
	if (isInv)
		butEngBoth(Util::LEFT);
	else
		butEngBoth(Util::RIGHT);
}

//************************************** ***************************************

/// Schaltet die Anzeige in den Engineering-Modus bzw. verschiebt das Komma um
/// jeweils drei Stellen nach rechts oder links.
//  fse, 15.07.02

void Fx3600pMainDlg::butEngBoth(Util::Direction dir)
{
	string sevenRaw = mDsplP->getSevenRaw();
	string seven;
	NumDmy num;
	double val;
	long   numr, dnom;

	if (mLastType != B_ENG) 
	{
		if (num.isFraction(sevenRaw, val, numr, dnom))
		{ // Bruchdarstellung aktiv => in double wandeln:
			Number number(double(numr)/double(dnom));
			sevenRaw = number.makeStr(0);
		}
		seven = Util::shiftPointEng(sevenRaw, dir, 0);
	}
	seven = Util::shiftPointEng("", dir, 10);
	/*if (seven != sevenRaw)*/ updateDisplayAfterDigit("e"+seven); // e ist Flag
	setLastType(B_ENG);
}

//******************************************************************************

/// Slot-Funktion ist nur noetig, um in der Anzeige die Klammer erscheinen zu
/// lassen.
//  fse, 16.07.02

void Fx3600pMainDlg::butBra()
{
	QPushButton*  senderP = (QPushButton*)(sender()); // nur noetig fuer extractSym()
	DisplayMode   dm = mCalcP->getCfgsP()->mDisplayMode;
	CalcMode      cm = mCompP->getCalcMode();
	bool          isInv = mCompP->getIsInv();
	string        dummy, sym = "(";
	int           braLev;
	ostringstream str;
	string        seven;

	if (isError() || mIsOff || isPrgRunning()) return;
	if (isInv)
	{ // es muss nur sqrt berechnet werden:
		butAction();
		return;
	}
	ONL_LOG(gnl) << "button: bra" << NL;
	dummy = extractSym(senderP); // nur damit die Anzeige flackert
	butActionBase(sym, dummy);
	setLastType(B_ACTION);
	if (cm != LR && cm != SD && dm != SIMPLE)
	{
		braLev = mCompP->getBraLev();
		str.width(2); str.fill('0');
		str.setf(ios::internal); //str << setfill('0') << braLev; // << ends;
		str << braLev; // << ends;
		seven = str.str();
		seven = "[" + seven + "######0.";
		updateDisplayAfterDigit(seven);
	}
}

//******************************************************************************

/// Schaltet den Rechner im More-Realistic-Modus "aus", wenn der 6-Minuten-Timer
/// abgelaufen ist.
//  fse, 17.07.02

void Fx3600pMainDlg::switchOff()
{

	DisplayMode  dm = mCalcP->getCfgsP()->mDisplayMode;
	AcNrm       acNrm;

	if (mCompP->getActivePrg()) return; // nicht abschalten, wenn gerade Programm laeuft
	mIsOff = true;
	if (dm != SIMPLE && mIsPixLoaded)
	{
		acNrm.process(); // evtl. SCI oder FIX abschalten
		mTimOff.stop();
		clearLcd();
		update(mLcdRect);
	}
}

//******************************************************************************

/// Funktion erzeugt die Tabelle mit allen Shortcuts.
//  fse, 17.08.02

void Fx3600pMainDlg::makeShortCutTab()
{
	mShortCutTab['0' ] = m0P       ;
	mShortCutTab['1' ] = m1P       ;
	mShortCutTab['2' ] = m2P       ;
	mShortCutTab['3' ] = m3P       ;
	mShortCutTab['4' ] = m4P       ;
	mShortCutTab['5' ] = m5P       ;
	mShortCutTab['6' ] = m6P       ;
	mShortCutTab['7' ] = m7P       ;
	mShortCutTab['8' ] = m8P       ;
	mShortCutTab['9' ] = m9P       ;
	mShortCutTab['.' ] = mPointP   ;
	mShortCutTab[',' ] = mPointP   ;
	mShortCutTab['+' ] = mAddP     ;
	mShortCutTab['-' ] = mSubP     ;
	mShortCutTab['*' ] = mMulP     ;
	mShortCutTab['/' ] = mDivP     ;
	mShortCutTab['=' ] = mEquP     ;
	mShortCutTab['\r'] = mEquP     ;
	mShortCutTab['i' ] = mInvP     ;
	mShortCutTab['k' ] = mKinP     ;
	mShortCutTab['K' ] = mKoutP    ;
	mShortCutTab['n' ] = mEngP     ;
	mShortCutTab['m' ] = mModeP    ;
	mShortCutTab['l' ] = mLogP     ;
	mShortCutTab['L' ] = mLnP      ;
	mShortCutTab['A' ] = mAbcP     ;
	mShortCutTab['g' ] = mIsxgP    ;
	mShortCutTab['h' ] = mHypP     ;
	mShortCutTab['S' ] = mSinP     ;
	mShortCutTab['C' ] = mCosP     ;
	mShortCutTab['T' ] = mTanP     ;
	mShortCutTab['_' ] = mSignP    ; 
	mShortCutTab['(' ] = mBraP     ;
	mShortCutTab[')' ] = mKetP     ;
	mShortCutTab['l' ] = mLogP     ;
	mShortCutTab['r' ] = mRclP     ;
	mShortCutTab['c' ] = mCP       ;
	mShortCutTab['a' ] = mAcP      ;
	mShortCutTab['e' ] = mExpP     ;
	mShortCutTab['p' ] = mSumP     ;
	mShortCutTab['t' ] = mCalcTypeP;
	mShortCutTab['P' ] = mP1P      ;
	mShortCutTab['R' ] = mRunP     ;
	mShortCutTab['O' ] = mOffP     ;
	mShortCutTab['f' ] = mPrgP     ;
}                       

//******************************************************************************
// Tastaturbedienung:
//******************************************************************************

/// Mapped alle Tastatureingaben auf Tastendruecke in der Dialogbox.
/// \param evP Zeiger auf das Qt-Tasten-Event
//  fse, 25.12.01

void Fx3600pMainDlg::keyPressEvent(QKeyEvent* evP)
{
	QMouseEvent me(QEvent::MouseButtonPress, QPoint(5,5), LeftButton, LeftButton);

	execKeyEvent(evP, me);
}

//******************************************************************************

/// Mapped alle Tastatureingaben auf Tastendruecke in der Dialogbox.
/// \param evP Zeiger auf das Qt-Tasten-Event
//  fse, 25.12.01

void Fx3600pMainDlg::keyReleaseEvent(QKeyEvent* evP)
{
	QMouseEvent me(QEvent::MouseButtonRelease, QPoint(5,5), LeftButton, NoButton);

	execKeyEvent(evP, me);
	return;
}

//******************************************************************************
//******************************************************************************
//************************************** ***************************************
//************************* ************************* **************************
//****************** ******************* ******************* *******************
//************** *************** *************** *************** ***************
//************ ************ ************ ************ ************ *************
