//******************************************************************************
// Copyright (c) 2002-2004 by Friedemann Seebass, Germany.
// Dieses Programm ist freie Software. Siehe GNU GPL Notiz in Datei main.cpp.
// This program is free software. See the GNU GPL notice in file main.cpp.
// Projekt: Almetare - Alle meine Taschenrechner
//******************************************************************************
/// \file calc_ti_sci.cpp
///       Gemeinsame Klassen fuer wissenschaftliche TI-Rechner (Modul special).
///       In dieser Datei befinden sich die Implementierungen der Klassen, die
///       alle wissenschaftlichen TI-Taschenrechner benoetigen.
//******************************************************************************
// Wann       Wer      Was
// ---------- -------- ---------------------------------------------------------
// 11.06.2004 fse      ComputerTiSci::ComputerTiSci(): Initialisierung von
//                     mDisplayTiP
// 24.03.2002 fse      erzeugt
//******************************************************************************

#include <iomanip>   // setw()
#include "util.h"
#include "calc_ti_sci.h"

//******************************************************************************
// Klasse ComputerTiSci:
//******************************************************************************

/// Konstruktor erweitert lediglich die Symboltabelle um die TI-Symbole.
//  fse, 11.06.04: Initialisierung von mDisplayTiP

ComputerTiSci::ComputerTiSci() : mDisplayTiP(NULL)
{
	addTiSciSymsToSymTab();
}

//************************************** ***************************************

/// Erweitert die Symboltabelle um das Exponentialmodussymbol.

void ComputerTiSci::addTiSciSymsToSymTab()
{
	string sym;
	Symbol* symP;

	symP = new AcEe; sym = symP->getSym(); mSymTab[sym] = symP;
}

//******************************************************************************

/// Ordnet einem Symbol dasjenige Symbol zu, das ihm im Invers-Modus entspricht.
/// \retval neues Symbol
/// \return true, wenn das Symbol geaendert wurde, sonst false
//  fse, 26.05.02

bool ComputerTiSci::mapInv(string& sym)
{
	return mapInvSci(sym);
}

//******************************************************************************

/// Erfindet einen zweiten Operanden, falls ein Operator mit nur einem berechnet
/// werden soll.
/// TI-30 und TI-45 verhalten sich so:
/// "3+=" => 3, "3-=" => 3, "3*=" => 3, "3/=" => 3
/// \retval a   erster Operand
/// \retval soP Zeiger auf das Operatorsymbol
/// \retval b   zweiter Operand
//  fse, 13.04.02

void ComputerTiSci::handleEmptyNumStack(Number& a, SymOperator* soP, Number& b)
{
	if (dynamic_cast<OpMul*>(soP)) a = 1;
	if (dynamic_cast<OpDiv*>(soP)) a = b * b;
	if (dynamic_cast<OpSub*>(soP)) b = -b;
}

//******************************************************************************

/// Siehe Kommentar ComputerBase::handleDisplayAfterBra().
/// TI behaelt Wert im Display bei nach "(".
//  fse, 13.04.02

void ComputerTiSci::handleDisplayAfterBra()
{
	Number lastNum = topNum();
	pushNewStacks();
	pushNum(lastNum);
}

//******************************************************************************

/// Prozessiert das Prozentzeichen.
/// Prozentrechnung des TI-30:
/// \code
///  1: a*b%: a und * bleiben auf den Stacks, b/100 auf Stack
///  2: a/b%: a und / bleiben auf den Stacks, b/100 auf Stack
///  3: a+b%: a und + bleiben auf den Stacks, a*b/100 auf Stack
///  4: a-b%: a und - bleiben auf den Stacks, a*b/100 auf Stack
///  5: a* %: a und * bleiben auf den Stacks, a/100 auf Stack
///  6: a/ %: a und / bleiben auf den Stacks, a/100 auf Stack
///  7: a+ %: a und + bleiben auf den Stacks, a*a/100 auf Stack
///  8: a- %: a und - bleiben auf den Stacks, a*a/100 auf Stack
///  9: a  %: a/100 auf Stack
/// \endcode
// fse, 11.10.01

void ComputerTiSci::handlePercent()
{
	SymOperator* soP = topSym();
	Number a, b, val;
	string op;

	b = popNum();
	if (soP == NULL)
	{ // Fall 9:
		val = b/100.0;
	}
	else
	{ // Faelle 1-8:
		op = soP->getSym();
		if (isNumEmpty())
		{ // Faelle 5-8:
			pushNum(b);
			if (op == "+" || op == "-")
				val = b*b/100.0; // Faelle 7 und 8
			else
				val = b/100.0;   // Faelle 5 und 6
		}
		else
		{ // Faelle 1-4:
			a = topNum();
			if (op == "+" || op == "-")
				val = a*b/100.0; // Faelle 3 und 4
			else

				val = b/100.0;   // Faelle 1 und 2
		}
	}
	pushNum(val);
	return;

}

//******************************************************************************
// Klasse DisplayTiSci:
//******************************************************************************

/// Konstruktor der TI-Display-Klasse.
/// \param cbP Zeiger zum Computer, entweder TI-30 oder TI-45

DisplayTiSci::DisplayTiSci(ComputerBase* cbP) : DisplaySci(cbP)
{
	mCompP = static_cast<ComputerTiSci*>(DisplayBase::mCompP);
	makeSeven();
}

//******************************************************************************
/*
/// Ermittelt die Laenge eines Zahlenstrings in Fixed-Notation, wie sie fuer den
/// TI-30 wesentlich ist.
/// Der TI-30 kann 8 Zeichen darstellen, es gibt aber noch eine neunte Stelle,
/// die ein Vorzeichen, die Winkelmodus-Anzeigen und einen Dezimalpunkt aufnehmen
/// kann.
/// Der Dezimalpunkt beansprucht keine eigene Stelle.
/// Ausserdem wird eine Zahl, die mit "0." beginnt i.d.R. mit der fuehrenden Null
/// dargestellt; es sei denn, es folgen noch 8 Ziffern, dann entfaellt die Null.
/// Daher duerfen Dezimalpunkt, Vorzeichen und fuehrende Null nicht mitgezaehlt
/// werden.
/// Erst wenn dann die Laenge immer noch groesser als 8 Zeichen ist, kann die Zahl
/// nicht mehr in Fixed-Notation dargestellt werden.
/// Beispiele:
/// \code
///      C++      Laenge      TI-30      wesentliche Laenge
///          12      2             12.          2
///         -12      3            -12.          2
///        12.3      4            12.3          3
///       -12.3      5           -12.3          3
///   1.2345678      9       1.2345678          8
///  -1.2345678     10      -1.2345678          8
///   0.1234567      9       0.1234567          8
///  0.12345678     10       .12345678          8
/// -0.12345678     11      -.12345678          8
/// \endcode
/// \param  numStr String, dessen Laenge zu ermitteln ist
/// \return Laenge des Strings

int DisplayTiSci::getNumStrLen(string numStr) const
{
	unsigned int len = numStr.size();

	if (numStr.find('.') != string::npos) --len;
	if (numStr[0] == '-')
	{
		--len;
		if (numStr[1] == '0') --len;
	}
	if (numStr[0] == '0') --len;
	return (int)len;
}
*/
//******************************************************************************

/// Erzeugung des Ausgabe-Strings.
/// Erzeugt aus dem Wert in mVal einen String, der so aussieht wie die
/// Anzeige des TI-30 und schreibt ihn in mSeven.
/// \param displayStr wenn displayStr == "", wird der aktuelle String in mValue
///                   verwendet, sonst der uebergebene String
//  fse, 05.05.02

void DisplayTiSci::makeSeven(string displayStr)
{
	string::size_type ePos;
	bool isSci = false;
	int digits, maxLen = 10;

	if (isError())
	{
		mSeven = "    Error.";
		return;
	}
	if (displayStr == "")
	{
		ePos = mValue.find('e');
		isSci = ePos != string::npos;
		if (isSci) digits = 5; else digits = 8;
		if (mValue[0] == '0' || mValue.substr(0, 2) == "-0") ++digits;
		if (isSci)
		{
			mSeven = Util::roundMantisse(mValue, -digits);
		}
		else
		{
			mSeven = Util::roundMantisse(mValue, digits);
			// vielleicht hat sich jetzt SCI ergeben? => neu runden:
			ePos = mSeven.find('e');
			isSci = ePos != string::npos;
			if (isSci)
			{
				digits = 5;
				if (mSeven[0] == '0' || mSeven.substr(0, 2) == "-0") ++digits;
				mSeven = Util::roundMantisse(mValue, -digits);
			}
		}
	}
	else
	{
		mSeven = displayStr;
		ePos = mSeven.find('e');
		isSci = ePos != string::npos;
		if (isSci) digits = 5; else digits = 8;
		//if (mValue[0] == '0' || mValue.substr(0, 2) == "-0") ++digits;
		if (mSeven[0] == '0' || mSeven.substr(0, 2) == "-0") ++digits;
		mSeven = Util::roundMantisse(mSeven, digits);
	}
	mSevenRaw = mSeven;

//	if (mSeven.substr(0, 2) == "-0") ++maxLen;
	if (mSeven[0] == '.') mSeven = "0" + mSeven;
	if (mSeven.substr(0, 2) == "-."  && mSeven.size() < 10)
		mSeven.insert(1, "0");
	if (mSeven.substr(0, 3) == "-0." && mSeven.size() > 10)
	{
		mSeven.erase(1, 1);
//		--maxLen;
	}
	if (isSci)
		mSeven = Util::formatSeven(mSeven, maxLen, 3, '#');
	else
		mSeven = Util::formatSeven(mSeven, maxLen, 0, '#');
	//if (mSeven.substr(0, 2) == "-0") mSeven.replace(0, 2, "-");
	if (mSeven[0] == '0') mSeven.replace(0, 1, "#");
	if (mSeven == "E") mSeven = "    Error.";
 	mSeven = prefixAngMode(mSeven);
}

//************************************** ***************************************

/// Erzeugt vor dem Ausgabe-String ein Zeichen, das den Winkelmodus
/// symbolisiert.
/// \param  numStr urspruenglicher String
/// \return String mit vorangestelltem Winkelmoduszeichen

string DisplayTiSci::prefixAngMode(string numStr) const
{
	if (numStr[0] == '-')
	{
		switch(mCompP->getAngMode())
		{
			case DEG:                ; break;
			case RAD: numStr[0] = '~'; break;
			case GRA: numStr[0] = '='; break;
			default: ;
		}
	}
	else
	{
		switch(mCompP->getAngMode())
		{
			case DEG:                 ; break;
			case RAD: numStr[0] = '\''; break;
			case GRA: numStr[0] = '\"'; break;
			default: ;
		}
	}
	return numStr;
}

//******************************************************************************

/// Schaut im Computer, welche Modes gesetzt sind, und baut den Mode-String des
/// Displays entsprechend zusammen.
//  fse, 12.04.02

void DisplayTiSci::setModes()
{
	setModesSci();
	mModes =  mInv + " " + mMem + " " + mK + " " + mDeg + " " + mRad + " " + mGra + " " + mSci;
}

//******************************************************************************
// zusaetzliche Symbole in TI-Taschenrechnern:
//************************* ************************* **************************

//******************************************************************************

