//******************************************************************************
// Copyright (c) 2002-2004 by Friedemann Seebass, Germany.
// Dieses Programm ist freie Software. Siehe GNU GPL Notiz in Datei main.cpp.
// This program is free software. See the GNU GPL notice in file main.cpp.
// Projekt: Almetare - Alle meine Taschenrechner
//******************************************************************************
/// \file ti30_main_dlg.cpp
///       Implementierung der TI-30-Dialogbox (Modul special).
//******************************************************************************
// Wann       Wer      Was
// ---------- -------- ---------------------------------------------------------
// 15.11.2003 fse      Ti30mainDlg::Ti30mainDlg(): wflags an boot() ueberg.
// 05.08.2003 fse      butEe(): Anpassungen
// 16.03.2003 fse      Konstruktor mCalc() mit Argument qApp->argc()
// 03.03.2003 fse      updtCfgFile() jetzt im Destruktor des Calculators statt
//                     in switchOff()
// 01.03.2003 fse      Konstruktor Ti30mainDlg mit Argument **argv
// 08.10.2001 fse      erzeugt
//******************************************************************************

#include <qpushbutton.h>
#include <qapplication.h>  // QApplication::palette()
#include <qevent.h>        // QMouseEvent
#include <qlabel.h>
#include <qtoolbutton.h>   // sonst geht sendEvent(mCalcTypeP, &me) nicht!
#include "util.h"
#include "ti30_main_dlg.h"
#include "cfg_dlg.h"

//******************************************************************************
//******************************************************************************

/// Konstruktion des Dialogs.
/// Initialisierung der Members.
/// \param parentP Zeiger auf Vaterdialog
/// \param name    Name des Dialogs
/// \param modal   Flag, ob Dialog modal oder nicht
/// \param fl      diverse Qt-Dialog-Flags
//  fse, 26.10.01
//  fse, 01.03.03: Initialisierung von mCalc
//  fse, 15.11.03: Aufruf von boot() mit getWFlags()

Ti30mainDlg::Ti30mainDlg(QWidget* parentP, const char* name, bool modal, WFlags fl)
	: Ti30mainBaseDlg(parentP, name, modal, fl), mCalc(qApp->argc(), qApp->argv())
{
	mCompP = static_cast<ComputerTi30*>(mCalc.getComputerP());
	mDsplP = static_cast<DisplayTi30 *>(mCalc.getDisplayP());

	// spezielle numerische Werte der Rechnergeometrie:
	mDefaultWidth  = 242; // Default-Breite des Rechners (falls keine Pixmap)
	mDefaultHeight = 446; // Default-Hoehe  des Rechners (falls keine Pixmap)
	mDigitWidth    =  32; // natuerliche Breite einer Ziffer im Display
	mDigitHeight   =  24; // natuerliche Hoehe  einer Ziffer im Display
	mOrgLcdRect.setCoords(105, 45, mDigitWidth*9, 24); // Geometrie des LED

	// Kopieren von Widgetpointern des Dialogs in Members der Basisklasse:
	mLcdStrBaseP  = mLcdStrP;
	mModesBaseP   = mModesP;
	mDisplayBaseP = mDisplayP;
	mSeebassBaseP = mSeebassP;

	// "Booten" des Rechners:
	boot(&mCalc, this, getWFlags()); // getWFlags() ist protected, daher Aufruf nicht in boot()

	// Timer:
	connect(&mTim8min    , SIGNAL(timeout()), this, SLOT(timeout8min    ()));
	connect(&mTim45s     , SIGNAL(timeout()), this, SLOT(timeout45s     ()));
	connect(&mTim150ms   , SIGNAL(timeout()), this, SLOT(timeout150ms   ()));
	connect(&mTimThinking, SIGNAL(timeout()), this, SLOT(timeoutThinking()));
}

//******************************************************************************

/// 8-Minuten Timer zum "Abschalten" des Rechners.
//  fse, 15.06.02

void Ti30mainDlg::timeout8min()
{
	switchOff();
}

//******************************************************************************

/// 45-Sekunden Timer fuer Einschalten des "Stromspar"-Modus'.
//  fse, 15.06.02

void Ti30mainDlg::timeout45s()
{
	mTim45s.stop();
	mTim150ms.start(150);
}

//******************************************************************************

/// 150-Millisekunden-Timer fuer den laufenden Punkt im "Stromspar"-Modus.
//  fse, 15.06.02

void Ti30mainDlg::timeout150ms()
{
	static int len = mIdleStr.size();

	setLed(mIdleStr.substr(0, 10));
	mIdleStr = mIdleStr[len-1] + mIdleStr.substr(0, len-1);
}

//******************************************************************************
// virtuelle Funktionen:
//******************************************************************************

/// Setzt fuer einige Buttons rote und blaue Farbe.
//  fse, 06.06.02

void Ti30mainDlg::adjustSimpleSkin()
{
	QPalette p = QApplication::palette();
	QPalette pBlue = p;
	QPalette pRed = p;

	pBlue.setColor(QColorGroup::ButtonText, Qt::darkBlue);
	pRed.setColor (QColorGroup::ButtonText, Qt::darkRed );
	mAddP  ->setPalette(pRed );
	mEquP  ->setPalette(pRed );

	mSubP  ->setPalette(pRed );
	mMulP  ->setPalette(pRed );
	mDivP  ->setPalette(pRed );
	mPowP  ->setPalette(pBlue);
	mConstP->setPalette(pRed );
	mInvP  ->setPalette(pBlue);
	mEeP   ->setPalette(pBlue);
	mSinP  ->setPalette(pBlue);
	mLnP   ->setPalette(pBlue);
	mTanP  ->setPalette(pBlue);
	mCosP  ->setPalette(pBlue);
	mLogP  ->setPalette(pBlue);
}

//******************************************************************************

/// Erneuert die LED-Anzeige.
/// Da diese Prozedur nach jedem Tastendruck aufgerufen wird, koennen hier auch
/// die Timer gestartet werden.
/// Besonderheit:
/// Bei der realistischen Darstellung soll der TI-30 die "Thinking-Sequenz"
/// zeigen. Um zu verhindern, dass das korrekte Ergebnis schon kurz aufblitzt
/// [ausgeloest durch den Aufruf in butActionBase()] bevor die Sequenz
/// durchgelaufen ist, muss es in diesem Fall durch Setzen des Flags
/// isDoPerhapsNot kurz deaktiviert werden. setDisplayFields() wird dann
/// abschliessend vom Timer aufgerufen.
/// \param isDoPerhapsNot wenn true und MORE_REALISTIC, wird die Funktion sofort
///                       verlassen
//  fse, 15.06.02

void Ti30mainDlg::setDisplayFields(bool isDoPerhapsNot)
{
	DisplayBase* dsplP = mCalcP->getDisplayP();
	DisplayMode  dm    = mCalcP->getCfgsP()->mDisplayMode;

	if (dm == MORE_REALISTIC && isDoPerhapsNot)
		return;
	if (dm == MORE_REALISTIC)
	{
		mTim150ms.stop();       // "Stromsparanzeige" aus
		mTim45s.start(45000);   // "Stromsparen" nach 45 Sekunden
		mTim8min.start(480000); // aus nach 8 Minuten

		//mTim45s.start(5000);  // zum Testen
		//mTim8min.start(15000);
		mIdleStr = ".              ";
	}
	else
	{ // Timer stoppen, falls Konfigurationsdialog Display-Mode geaendert hat
		mTim150ms.stop();
		mTim45s.stop();
		mTim8min.stop();
	}
	if (dm != SIMPLE && mIsPixLoaded)
		setLed(dsplP->getSeven());
}

//************************* ************************* **************************

/// Setzt den String str in die passenden Buchstaben der Leuchtdiodenanzeige um.
/// \param str String, der dargestellt werden soll.
// fse, 08.06.02

void Ti30mainDlg::setLed(string str)
{
	QPixmap*  pmP  = const_cast<QPixmap*>(backgroundPixmap());
	const int wSrc = int(mDigitWidth  * mXfact);
	const int hSrc = int(mDigitHeight * mYfact);
	const int yDst = mLcdRect.y();          // y-Koord. immer gleich
	int       xDst, xSrc, ySrc, x, y, w, h;
	int       col, row;
	double    xx = mLcdRect.x();
	char      c;
	bool      isPointSucc;

	str += " "; // fuer Vergleich mit Nachfolger '.'
	for (unsigned int i = 0; i< str.size()-1; i++)
	{
		c = str[i];
		if (c == '.') continue;
		if (str[i+1] == '.') isPointSucc = true; else isPointSucc = false;
		getSourceCoords(c, isPointSucc, row, col);
		xSrc = col * wSrc;
		ySrc = row * hSrc;
		xDst = int(xx);
		bitBlt(pmP, xDst, yDst, &mScaledDigitsPixmap, xSrc, ySrc, wSrc, hSrc);
		xx += mDigitWidth * mXfact;
	}
	mLcdRect.coords(&x, &y, &w, &h);
	update(x, y, w, h);
}

//****************** ******************* ******************* *******************

/// Liefert zu einem Buchstaben die zugehoerige Zeile und Spalte in der Pixmap.
///
/// Die Pixmap hat folgenden Aufbau:
/// \code
/// Zeile |   0  1  2  3  4  5  6  7  8  9  10 Spalte
/// ------|----------------------------------------
///    0  |   0  1  2  3  4  5  6  7  8  9  U
///    1  |   0. 1. 2. 3. 4. 5. 6. 7. 8. 9. I
///    2  |      '  "  -  ~  =  E  r  o  s  _
///    3  |    . '. ". -. ~. =. A. r. b  J  ]
/// \endcode
/// \param c           gesuchter Character
/// \param isPointSucc gibt an, ob dem Zeichen ein Punkt folgt
/// \retval row        gefundene Zeilennummer in der Pixmap
/// \retval col        gefundene Spaltennummer in der Pixmap
//  fse, 01.11.01

void Ti30mainDlg::getSourceCoords(char c, bool isPointSucc, int& row, int& col)
{
	if (c >= '0' && c <= '9')
	{
		row = 0;
		col = c - '0';
	}
	else
	{
		row = 2;
		switch (c)
		{
			case '#' : col =  0; break;
			case '\'': col =  1; break;
			case '\"': col =  2; break;
			case '-' : col =  3; break;

			case '~' : col =  4; break;
			case '=' : col =  5; break;
			case 'E' : col =  6; break;
			case 'r' : col =  7; break;
			case 'o' : col =  8; break;

			case 's' : col =  9; break;
			case '_' : col = 10; break;
			case 'U' : col = 10; row = 0; break;
			case 'I' : col = 10; row = 1; break;
			case 'B' : col =  8; row = 1; break; // 8 mit Punkt, fuer mThinkStr
			case 'A' : col =  6; row = 3; break;
			case 'b' : col =  8; row = 3; break;
			case 'J' : col =  9; row = 3; break;
			case ']' : col = 10; row = 3; break;
			default: col = 0;
		}
	}
	if (isPointSucc) ++row;
}

//******************************************************************************
// Button-Slots:
//******************************************************************************

/// Slot-Funktion fuer die meisten Buttons.
//  fse, 23.02.02

void Ti30mainDlg::butAction()
{
	QPushButton* senderP = (QPushButton*)(sender());
	string sym = senderP->name();
	string dummy;
	static string lastSym = "";

	static string lastSymButOne = "";
	DisplayMode dm = mCalcP->getCfgsP()->mDisplayMode;

	if (isError() || mIsOff) return;
	sym = extractSym(senderP);
	butActionBase(sym, dummy);
	if (dm == MORE_REALISTIC)
	{
		if (lastSym == "pi" || lastSym == "rcl" || lastSym == "exc")
			lastSym = lastSymButOne; // noetig fuer z.B. 3^pi=
		showAnyThinkSequence(sym, lastSym);
	}
	lastSymButOne = lastSym;
	lastSym = sym;

	if (sym != "inv" && sym != "deg" && sym != "rad" && sym != "gra")
		setLastType(B_ACTION);
}

//************************************** ***************************************

/** Erzeugt (angenaehert) eine der Sequenzen von Zeichen, die der TI-30 ausgibt,
    wenn er ueber schwierige Aufgaben (Funktionen) "nachdenkt".
    \param sym     aktuell eingegebenes Symbol
    \param lastSym zuletzt eingegebenes Symbol
*/
//  fse, 04.11.01

void Ti30mainDlg::showAnyThinkSequence(string sym, string lastSym)
{
	if      (sym == "rcp"                  ) { mTimThinking.start(50); mThinkStr = "__]]"                 ; return; }
	else if (sym == "sqr"                  ) { mTimThinking.start(50); mThinkStr = "UAB"                  ; return; }
	else if (sym == "sqrt"                 ) { mTimThinking.start(50); mThinkStr = "_]_]_]"               ; return; }
	else if (sym == "sin" || sym == "asin" ) { mTimThinking.start(50); mThinkStr = "11IIJJUUAABBBBB_]_]_]"; return; }
	else if (sym == "cos" || sym == "acos" ) { mTimThinking.start(50); mThinkStr = "11IIJJUUAABBBBB_]_]_]"; return; }
	else if (sym == "tan" || sym == "atan" ) { mTimThinking.start(50); mThinkStr = "11IIJJUUAABB____"     ; return; }
	else if (sym == "log" || sym == "pow10") { mTimThinking.start(50); mThinkStr = "11IIJJUUAA88____"     ; return; }
	else if (sym == "ln"  || sym == "ex"   ) { mTimThinking.start(50); mThinkStr = "11IIJJUUAA88____"     ; return; }
	else if (sym == "="   || lastSym == "/" || sym == "*" || sym == "-" || sym == "+")
			{ showEquThinkSequence(lastSym); return; }
	mTimThinking.start(50); mThinkStr = "_";
}

//************************* ************************* **************************

/** Zeigt spezielle "Thinking"-Sequenz, wenn '=' gedrueckt wurde.
    \param lastSym Symbol, das vor dem '=' eingegeben wurde.
*/

void Ti30mainDlg::showEquThinkSequence(string lastSym)
{
	if (lastSym == "pow" || lastSym == "ipow") mThinkStr = "11IIJJUUAABBAUJAB";
	else if (lastSym == "/" || lastSym == "*" || lastSym == "-" || lastSym == "+") mThinkStr = "UAB";
	else mThinkStr = "_";
	mTimThinking.start(50);
}

//************************* ************************* **************************

/// Erzeugt eine konkrete Sequenz von Zeichen, die der TI-30 beim "Nachdenken"
/// ausgibt.
//  fse, 04.11.01

void Ti30mainDlg::timeoutThinking()

{
	static string ledStr(' ', 9);

	if (mThinkStr.size() == 0)
	{
		mTimThinking.stop();
		setDisplayFields();
	}
	else
	{
		ledStr[8] = mThinkStr[0];
		mThinkStr = mThinkStr.substr(1);
		setLed(ledStr);
	}
}

//******************************************************************************

/// Slot-Funktion fuer die Zifferntasten.

void Ti30mainDlg::butDigit()
{
	QPushButton* senderP = (QPushButton*)(sender());
	string sym;

	if (isError() || mIsOff) return;
	sym = extractSym(senderP);
	mCompP->setIsInv (false);
	butDigitBase(sym);
}

//******************************************************************************

/// Slot-Funktion fuer die Off-Taste.
/// Wenn Button "OFF" zum ersten Mal gedrckt wird, wird das Diplay geloescht.
/// Beim zweiten Mal wird der Taschenrechner beendet.
//  fse, 08.11.01

void Ti30mainDlg::butOff()
{
	ONL_LOG(gnl) << "button: Off" << NL;
	switchOff();
}

//************************************** ***************************************

/// Schaltet Rechner aus.
//  fse, 03.03.03: updtCfgFile() jetzt im Destruktor des Calculators. 

void Ti30mainDlg::switchOff()
{
	DisplayMode  dm = mCalcP->getCfgsP()->mDisplayMode;
	AcNrm nrm;

	if (dm == SIMPLE || mIsOff)
	{
		Cfgs* cfgsP = mCalcP->getCfgsP();
		QRect rect = geometry();
		cfgsP->mXpos   = rect.x();
		cfgsP->mYpos   = rect.y();
		cfgsP->mWidth  = rect.width();
		cfgsP->mHeight = rect.height();
		//cfgsP->updtCfgFile();
		done(0);
	}
	initCalcDlgBase();
	nrm.process();
	mIsOff = true;
	if (dm != SIMPLE && mIsPixLoaded)
	{
		mTim150ms.stop();
		mTim45s.stop();
		mTim8min.stop();
		setLed("         ");
	}
}

//******************************************************************************

/// Slot-Funktion fuer die On-Taste.
/// Wenn Button "ON" zum ersten Mal gedrueckt wird, wird das Display geloescht.
/// Beim zweiten Mal werden die Stacks initialisiert.
//  fse, 13.10.01

void Ti30mainDlg::butOn()
{
	ONL_LOG(gnl) << "button: On" << NL;
	if (mLastType == B_ON || isError())
	{
		mCalcP->process("initwomem");
	}
	else
	{
		mCalcP->process("nrm");
		mCalcP->process("0");
		mCalcP->getComputerP()->setConstOpP(NULL);
		mCalcP->getComputerP()->setConstVal(0);
	}
	mDsplP->setDspMode(NRM);
	copyBaseDisplayFields();

	mIsOff = false;
	mDisplayStr = "0.";
	setDisplayFields();
	setLastType(B_ON);
}

//******************************************************************************

/// Slot-Funktion fuer die INV-Taste.

void Ti30mainDlg::butInv()
{
	QPushButton* senderP = (QPushButton*)(sender());
	string sym;

	if (isError() || mIsOff) return;
	sym = extractSym(senderP);
	mCompP->setIsInv (false);
	butAnyModeBase(sym);
}

//******************************************************************************

/// Slot-Funktion fuer die  Exponentialtaste.
//  fse, 15.06.02
//  fse, 05.08.03: Aufruf updateDisplayAfterDigit(): Flag "e" ergaenzt.

void Ti30mainDlg::butEe()
{
	string::size_type ePos  = mDisplayStr.find('e');
	bool              isSci = ePos != string::npos;
	string            sym = "ee", dummy;
//	DisplayMode       dm = mCalcP->getCfgsP()->mDisplayMode;

	ONL_LOG(gnl) << "button: Ee" << NL;
	if (isError() || mIsOff) return;
	if (isSci)
	{
		//double val = atof(dsplP->getSeven().c_str());
		if (mLastType == B_DIGIT) return;
		butActionBase(sym, dummy); // Komma verschieben
		setLastType(B_EE);
		copyBaseDisplayFields();
		setDisplayFields();
	}
	else
	{
		if (mDsplP->getDspMode() == SCI)
		{ 
			if (mCompP->getIsInv())
				mCalcP->process("nrm"); // wenn gerade Ziffer in Anzeige => SCI wieder abschalten
			else
			{
				mDisplayStr += "e+000";
				updateDisplayAfterDigit("e"+mDisplayStr);
				setLastType(B_DIGIT);
			}
		}
		else
		{
			double val = fabs(atof(mDisplayStr.c_str()));
			if (/*!mCompP->getIsInv() &&*/ val < 100000) // fse, 05.08.03
			{
				// SCI-Modus einschalten:
				mCalcP->process(sym);
				mDisplayStr += "e+000";
				updateDisplayAfterDigit("e"+mDisplayStr);
			}
			//setLastType(B_DIGIT); // fse, 05.08.03
			setLastType(B_EE);      //
		}
	}
}

//******************************************************************************

/// Slot-Funktion fuer Winkelmodi DEG, RAD, GRA.
//  fse, 27.10.01

void Ti30mainDlg::butDrg()
{
	string sym;
	AngMode am = mCompP->getAngMode();

	ONL_LOG(gnl) << "button: Drg" << NL;
	if (isError() || mIsOff) return;
	switch (am)
	{
		case DEG: sym = "rad"; break;
		case RAD: sym = "gra"; break;
		case GRA: sym = "deg"; break;
		default:  sym = "deg";
	}
	mCalcP->process(sym);
	updateDisplayAfterDigit(mDisplayStr);
}

//******************************************************************************

/// Funktion erzeugt die Tabelle mit allen Shortcuts.
//  fse, 17.08.02

void Ti30mainDlg::makeShortCutTab()
{
	mShortCutTab['0' ] = m0P       ;
	mShortCutTab['1' ] = m1P       ;
	mShortCutTab['2' ] = m2P       ;
	mShortCutTab['3' ] = m3P       ;
	mShortCutTab['4' ] = m4P       ;
	mShortCutTab['5' ] = m5P       ;
	mShortCutTab['6' ] = m6P       ;
	mShortCutTab['7' ] = m7P       ;
	mShortCutTab['8' ] = m8P       ;
	mShortCutTab['9' ] = m9P       ;
	mShortCutTab['.' ] = mPointP   ;
	mShortCutTab[',' ] = mPointP   ;
	mShortCutTab['+' ] = mAddP     ;
	mShortCutTab['-' ] = mSubP     ;
	mShortCutTab['*' ] = mMulP     ;
	mShortCutTab['/' ] = mDivP     ;
	mShortCutTab['=' ] = mEquP     ;
	mShortCutTab['\r'] = mEquP     ;
	mShortCutTab['i' ] = mInvP     ;
	mShortCutTab['s' ] = mSinP     ;
	mShortCutTab['c' ] = mCosP     ;
	mShortCutTab['T' ] = mTanP     ;
	mShortCutTab['e' ] = mEeP      ;
	mShortCutTab['l' ] = mLogP     ;
	mShortCutTab['n' ] = mLnP      ;
	mShortCutTab['y' ] = mPowP     ;
	mShortCutTab['p' ] = mPiP      ;
	mShortCutTab['%' ] = mPrcP     ;
	mShortCutTab['(' ] = mBraP     ;
	mShortCutTab[')' ] = mKetP     ;
	mShortCutTab['S' ] = mStoP     ;
	mShortCutTab['R' ] = mRclP     ;
	mShortCutTab['U' ] = mSumP     ;
	mShortCutTab['E' ] = mExcP     ;
	mShortCutTab['k' ] = mConstP   ;
	mShortCutTab['x' ] = mRcpP     ;
	mShortCutTab['^' ] = mSqrP     ;
	mShortCutTab['r' ] = mSqrtP    ;
	mShortCutTab['w' ] = mSqrtP    ;
	mShortCutTab['O' ] = mOffP     ;
	mShortCutTab['a' ] = mOnP      ; // Esc
	mShortCutTab['o' ] = mOnP      ;
	mShortCutTab['C' ] = mOnP      ;
	mShortCutTab['d' ] = mDrgP     ; 
	mShortCutTab['_' ] = mSignP    ; 
	mShortCutTab['g' ] = mSignP    ; 
	mShortCutTab['t' ] = mCalcTypeP; 
}                       

//******************************************************************************
// Tastaturbedienung:
//******************************************************************************

/** Mapped alle Tastatureingaben auf Tastendruecke in der Dialogbox.
    \param evP Zeiger auf das Qt-Tasten-Event
*/
//  fse, 25.12.01

void Ti30mainDlg::keyPressEvent(QKeyEvent* evP)
{
	QMouseEvent me(QEvent::MouseButtonPress, QPoint(5,5), LeftButton, LeftButton);

	execKeyEvent(evP, me);
}

//******************************************************************************

/** Mapped alle Tastatureingaben auf Tastendruecke in der Dialogbox.
    \param evP Zeiger auf das Qt-Tasten-Event
*/
//  fse, 25.12.01

void Ti30mainDlg::keyReleaseEvent(QKeyEvent* evP)
{
	QMouseEvent me(QEvent::MouseButtonRelease, QPoint(5,5), LeftButton, NoButton);

	execKeyEvent(evP, me);
	return;
}

//******************************************************************************

//******************************************************************************
//************************************** ***************************************
//************************* ************************* **************************
//****************** ******************* ******************* *******************
//************** *************** *************** *************** ***************
//************ ************ ************ ************ ************ *************
